static char rcsid[] = "$Id: concordance.c 213948 2018-03-02 23:24:52Z twu $";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "concordance.h"

#include <stdlib.h>		/* For qsort */
#include <stdio.h>
#include <string.h>
#include <strings.h>

#include "assert.h"
#include "mem.h"
#include "univdiag.h"
#include "univdiagdef.h"
#include "transcript.h"
#include "stage3hrdef.h"


static int subopt_levels;
static bool novelsplicingp;

static Chrpos_T pairmax_transcriptome;
static Chrpos_T pairmax_linear;
static Chrpos_T pairmax_circular;

static Chrpos_T expected_pairlength;
static Chrpos_T pairlength_deviation;
static Chrpos_T adjacent_pairlength;

static bool *circularp;
static bool merge_samechr_p;


#define MAX_HITS 1000

#define T Stage3end_T

#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif

/* Details */
#ifdef DEBUG1
#define debug1(x) x
#else
#define debug1(x)
#endif

/* Filter paths */
#ifdef DEBUG2
#define debug2(x) x
#else
#define debug2(x)
#endif



#if 0
/* Previously called by stage1hr.c, but now we use Transcript_concordance */
/* Since we haven't called merge procedure yet, we can just look at the Intlist_head */
List_T
Stage3hr_filter_concordant_tr (List_T *disjoint, List_T hits, List_T mates) {
  List_T common = NULL, p, q;
  T hit, mate;
  int trnum;
  Chrpos_T start1, end2;
  bool concordantp;

  *disjoint = (List_T) NULL;
  for (p = hits; p != NULL; p = List_next(p)) {
    hit = (T) List_head(p);
    trnum = Intlist_head(hit->trnums);
    concordantp = false;

    for (q = mates; q != NULL; q = List_next(q)) {
      mate = (T) List_head(q);
      if (Intlist_head(mate->trnums) == trnum) {
	start1 = (Chrpos_T) Intlist_head(hit->trstarts);
	end2 = (Chrpos_T) Intlist_head(mate->trends);
	if (start1 < end2) {
	  if (end2 <= start1 + pairmax_transcriptome) {
	    concordantp = true;
	  }
	} else {
	  if (start1 <= end2 + pairmax_transcriptome) {
	    concordantp = true;
	  }
	}
      }
    }

    if (concordantp == true) {
      debug7(printf("trnum %d is concordant\n",trnum));
      common = List_push(common,(void *) hit);
    } else {
      debug7(printf("trnum %d is discordant\n",trnum));
      *disjoint = List_push(*disjoint,(void *) hit);
    }
  }

  debug7(printf("Returning %d common hits\n",List_length(common)));
  return common;
}
#endif


static char *
print_sense (int sense) {
  if (sense == SENSE_NULL) {
    return "sense:null";
  } else if (sense == SENSE_ANTI) {
    return "sense:anti";
  } else if (sense == SENSE_FORWARD) {
    return "sense:fwd";
  } else {
    abort();
  }
}


static int
do_transcriptome_plus (int *concordant_score, List_T *hitpairs,
		       T *hits5, int nhits5, T *hits3, int nhits3,
		       int genestrand) {
  int nconcordant = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  int min_insertlength;
  int score;
  Chrnum_T chrnum;

  if (nhits5 > 0 && nhits3 > 0) {
    i = j = 0;
    while (i < nhits5) {
      hit5 = hits5[i];
      chrnum = hit5->effective_chrnum;
      
#ifdef DEBUG
      printf("plus/plus: i=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
	     i,nhits5,hit5->effective_chrnum,hit5->genomicstart - hit5->chroffset,hit5->genomicend - hit5->chroffset,
	     print_sense(hit5->sensedir),Method_string(hit5->method),hit5->circularalias,hit5);
      if (j >= 0 && j < nhits3) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits3,hits3[j]->effective_chrnum,
	       hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset);
      }
      printf("\n");
#endif
      
      while (j >= 0 && hits3[j]->effective_chrnum == chrnum) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	j--;
      }
      j++;		/* Finish backup */
      
      /* No need for advance */
      
      while (j < nhits3 && hits3[j]->effective_chrnum == chrnum) {
	debug(printf("  samechr: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	hit3 = hits3[j];
	
	if (Transcript_intersect_p(&min_insertlength,hit5->transcripts,hit3->transcripts) == false) {
	  debug(printf(" => transcripts do not intersect:\n"));
	  debug(Transcript_print_nums(hit5->transcripts));
	  debug(Transcript_print_nums(hit5->transcripts));
	  debug(printf("\n"));
	  
	} else {
	  debug(printf(" => concordant transcriptome with min_insertlength %d\n",min_insertlength));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/true)) != NULL) {
	    stage3pair->insertlength = min_insertlength;
	    
	    /* No need to update found_score */
	    debug(printf(" Have new pair with scores %d + %d\n",stage3pair->hit5->score,stage3pair->hit3->score));
	    
	    if ((score = (stage3pair->hit5->querylength - stage3pair->hit5->nmatches) + 
		 (stage3pair->hit3->querylength - stage3pair->hit3->nmatches)) < *concordant_score) {
	      *concordant_score = score;
	      assert(stage3pair->hit5->nmatches <= stage3pair->hit5->querylength);
	      assert(stage3pair->hit3->nmatches <= stage3pair->hit3->querylength);
	      debug(printf(" => Updating concordant_score to be %d = %d + %d\n",
			   *concordant_score,stage3pair->hit5->querylength - stage3pair->hit5->nmatches,
			   stage3pair->hit3->querylength - stage3pair->hit3->nmatches));
	    }
	    *hitpairs = List_push(*hitpairs,(void *) stage3pair);
	    nconcordant++;
	  }
	}
	debug(printf("\n"));
	
	j++;
      }
      j--;		/* Finish advance */
      
      i++;
    }
  }

  return nconcordant;
}


static int
do_transcriptome_minus (int *concordant_score, List_T *hitpairs,
			T *hits5, int nhits5, T *hits3, int nhits3,
			int genestrand) {
  int nconcordant = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  int min_insertlength;
  int score;
  Chrnum_T chrnum;

  if (nhits3 > 0 && nhits5 > 0) {
    i = j = 0;
    while (i < nhits3) {
      hit3 = hits3[i];
      chrnum = hit3->effective_chrnum;
      
#ifdef DEBUG
      printf("minus/minus: i=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
	     i,nhits3,hit3->effective_chrnum,hit3->genomicstart - hit3->chroffset,hit3->genomicend - hit3->chroffset,
	     print_sense(hit3->sensedir),Method_string(hit3->method),hit3->circularalias,hit3);
      if (j >= 0 && j < nhits5) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits5,hits5[j]->effective_chrnum,
	       hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset);
      }
      printf("\n");
#endif
      
      while (j >= 0 && hits5[j]->effective_chrnum == chrnum) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	j--;
      }
      j++;			/* Finish backup */
      
      /* No need for advance */
      
      while (j < nhits5 && hits5[j]->effective_chrnum == chrnum) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	hit5 = hits5[j];
	
	if (Transcript_intersect_p(&min_insertlength,hit5->transcripts,hit3->transcripts) == false) {
	  debug(printf(" => transcripts do not intersect:\n"));
	  debug(Transcript_print_nums(hit5->transcripts));
	  debug(Transcript_print_nums(hit5->transcripts));
	  debug(printf("\n"));
	  
	} else {
	  debug(printf(" => concordant transcriptome with min_insertlength %d\n",min_insertlength));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/true)) != NULL) {
	    stage3pair->insertlength = min_insertlength;
	    
	    if ((score = (stage3pair->hit5->querylength - stage3pair->hit5->nmatches) + 
		 (stage3pair->hit3->querylength - stage3pair->hit3->nmatches)) < *concordant_score) {
	      *concordant_score = score;
	      assert(stage3pair->hit5->nmatches <= stage3pair->hit5->querylength);
	      assert(stage3pair->hit3->nmatches <= stage3pair->hit3->querylength);
	      debug(printf(" => Updating concordant_score to be %d = %d + %d\n",
			   *concordant_score,stage3pair->hit5->querylength - stage3pair->hit5->nmatches,
			   stage3pair->hit3->querylength - stage3pair->hit3->nmatches));
	    }
	    *hitpairs = List_push(*hitpairs,(void *) stage3pair);
	    nconcordant++;
	  }
	}
	debug(printf("\n"));
	
	j++;
      }
      j--;		/* Finish advance */
      
      i++;
    }
  }

  return nconcordant;
}



List_T
Concordance_pair_up_transcriptome (bool *abort_pairing_p, int *concordant_score, List_T hitpairs,

				   Ladder_T ladder5_plus, Ladder_T ladder5_minus,
				   Ladder_T ladder3_plus, Ladder_T ladder3_minus,

				   int maxpairedpaths, int genestrand) {
  int nconcordant = 0, n;  /* was List_length(hitpairs), but this hurts minus if there are too many plus hits */

  T *hits5, *hits3;
  int nhits5, nhits3;
  int max_frontier_score, frontier_score, cutoff_level, level, score5, score3;


  debug(printf("Starting Concordance_pair_up_transcriptome\n"));

  /* Initial value, which resets to frontier_score + subopt_levels upon our first hitpair */
  cutoff_level = Ladder_cutoff(ladder5_plus) + Ladder_cutoff(ladder3_plus);
  if ((level = Ladder_cutoff(ladder5_minus) + Ladder_cutoff(ladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  max_frontier_score = cutoff_level;

  frontier_score = 0;
  while (*abort_pairing_p == false && frontier_score <= cutoff_level) {
    debug1(printf("frontier_score = %d\n",frontier_score));
    for (score5 = 0; score5 <= frontier_score; score5++) {
      score3 = frontier_score - score5;
      debug1(printf("score5 = %d, score3 = %d\n",score5,score3));

      /* plus/plus: hits5_plus against hits3_plus (really on minus) */
      if (score5 <= Ladder_cutoff(ladder5_plus) && score3 <= Ladder_cutoff(ladder3_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_plus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_plus,score3);
	debug(printf("at score %d+%d, nhits5_plus = %d, nhits3_plus = %d\n",score5,score3,nhits5,nhits3));

	if ((n = do_transcriptome_plus(&(*concordant_score),&hitpairs,
				       hits5,nhits5,hits3,nhits3,genestrand)) > 0) {
	  if (nconcordant == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nconcordant += n) > maxpairedpaths) {
	    debug(printf(" -- %d concordant paths exceeds %d",nconcordant,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* minus/minus: hits3_minus (really on plus) against hits5_minus */
      if (score5 <= Ladder_cutoff(ladder5_minus) && score3 <= Ladder_cutoff(ladder3_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_minus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_minus,score3);
	debug(printf("at score %d+%d, nhits5_minus = %d, nhits3_minus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_transcriptome_minus(&(*concordant_score),&hitpairs,
					hits5,nhits5,hits3,nhits3,genestrand)) > 0) {
	  if (nconcordant == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nconcordant += n) > maxpairedpaths) {
	    debug(printf(" -- %d concordant paths exceeds %d",nconcordant,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }
    }

    frontier_score++;
  }

  debug(printf("Finished with Concordance_pair_up_transcriptome: %d concordant\n",List_length(hitpairs)));

  return hitpairs;
}


static int
do_genome_plus (int *adjacent_score, int *concordant_score, List_T *hitpairs,
		T *hits5, int nhits5, T *hits3, int nhits3,
		int querylength5, int querylength3, int genestrand) {
  int nconcordant = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  Univcoord_T insert_start;
  int score;
  int pairmax;

  if (nhits5 > 0 && nhits3 > 0) {
    i = j = 0;
    while (i < nhits5) {
      hit5 = hits5[i];
      if (circularp[hit5->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      
      insert_start = hit5->genomicend - querylength5;
#ifdef DEBUG
      printf("plus/plus: i=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
	     i,nhits5,hit5->effective_chrnum,hit5->genomicstart - hit5->chroffset,hit5->genomicend - hit5->chroffset,
	     print_sense(hit5->sensedir),Method_string(hit5->method),hit5->circularalias,hit5);
      if (j >= 0 && j < nhits3) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits3,hits3[j]->effective_chrnum,
	       hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset);
      }
      printf("\n");
#endif

      while (j >= 0 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	j--;
      }
      j++;		/* Finish backup */

      while (j < nhits3 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	j++;
      }
      
      while (j < nhits3 && hits3[j]->genomicstart + querylength3 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	hit3 = hits3[j];
	
	if (hit5->effective_chrnum != hit3->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,hit3->effective_chrnum));
	  
	} else if (SENSE_INCONSISTENT_P(hit5->sensedir,hit3->sensedir)) {
	  debug(printf(" => sense inconsistent: %d | %d = %d",hit5->sensedir,hit3->sensedir,hit5->sensedir|hit3->sensedir));
	  
	} else if (hit3->genomicend < hit5->genomicstart) {
	  debug(printf(" => scramble because end3 %llu < start5 %llu\n",
		       (unsigned long long) hit3->genomicend,(unsigned long long) hit5->genomicstart));
	  
	} else {
	  debug(printf(" => concordant effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit5->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/false)) != NULL) {
	    
	    if ((score = (stage3pair->hit5->querylength - stage3pair->hit5->nmatches) + 
		 (stage3pair->hit3->querylength - stage3pair->hit3->nmatches)) < *concordant_score) {
	      *concordant_score = score;
	      assert(stage3pair->hit5->nmatches <= stage3pair->hit5->querylength);
	      assert(stage3pair->hit3->nmatches <= stage3pair->hit3->querylength);
	      debug(printf(" => Updating concordant_score to be %d = %d + %d\n",
			   *concordant_score,stage3pair->hit5->querylength - stage3pair->hit5->nmatches,
			   stage3pair->hit3->querylength - stage3pair->hit3->nmatches));
	    }
	    if (stage3pair->insertlength <= adjacent_pairlength && score < *adjacent_score) {
	      *adjacent_score = score;
	    }
	    *hitpairs = List_push(*hitpairs,(void *) stage3pair);
	    nconcordant++;
	  }
	}
	debug(printf("\n"));
	
	j++;
      }
      j--;		/* Finish advance */
      
      i++;
    }
  }

  return nconcordant;
}



static int
do_genome_minus (int *adjacent_score, int *concordant_score, List_T *hitpairs,
		 T *hits5, int nhits5, T *hits3, int nhits3,
		 int querylength5, int querylength3, int genestrand) {
  int nconcordant = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  Univcoord_T insert_start;
  int score;
  int pairmax;


  if (nhits3 > 0 && nhits5 > 0) {
    i = j = 0;
    while (i < nhits3) {
      hit3 = hits3[i];
      if (circularp[hit3->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      insert_start = hit3->genomicstart - querylength3;
#ifdef DEBUG
      printf("minus/minus: i=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
	     i,nhits3,hit3->effective_chrnum,hit3->genomicstart - hit3->chroffset,hit3->genomicend - hit3->chroffset,
	     print_sense(hit3->sensedir),Method_string(hit3->method),hit3->circularalias,hit3);
      if (j >= 0 && j < nhits5) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits5,hits5[j]->effective_chrnum,
	       hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset);
      }
      printf("\n");
#endif

      while (j >= 0 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	j--;
      }
      j++;			/* Finish backup */
      
      while (j < nhits5 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	j++;
      }
      
      while (j < nhits5 && hits5[j]->genomicend + querylength5 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	hit5 = hits5[j];

	if (hit3->effective_chrnum != hit5->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,hit3->effective_chrnum));

	} else if (SENSE_INCONSISTENT_P(hit3->sensedir,hit5->sensedir)) {
	  debug(printf(" => sense inconsistent: %d | %d = %d",hit5->sensedir,hit3->sensedir,hit5->sensedir|hit3->sensedir));
		
	} else if (hit5->genomicstart < hit3->genomicend) {
	  debug(printf(" => scramble because start5 %llu < end3 %llu\n",
		       (unsigned long long) hit5->genomicstart,(unsigned long long) hit3->genomicend));
		
	} else {
	  debug(printf(" => concordant effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit3->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/false)) != NULL) {
		  
	    if ((score = (stage3pair->hit5->querylength - stage3pair->hit5->nmatches) + 
		 (stage3pair->hit3->querylength - stage3pair->hit3->nmatches)) < *concordant_score) {
	      *concordant_score = score;
	      assert(stage3pair->hit5->nmatches <= stage3pair->hit5->querylength);
	      assert(stage3pair->hit3->nmatches <= stage3pair->hit3->querylength);
	      debug(printf(" => Updating concordant_score to be %d = %d + %d\n",
			   *concordant_score,stage3pair->hit5->querylength - stage3pair->hit5->nmatches,
			   stage3pair->hit3->querylength - stage3pair->hit3->nmatches));
	    }
	    if (stage3pair->insertlength <= adjacent_pairlength && score < *adjacent_score) {
	      *adjacent_score = score;
	    }
	    *hitpairs = List_push(*hitpairs,(void *) stage3pair);
	    nconcordant++;
	  }
	}
	debug(printf("\n"));
	      
	j++;
      }
      j--;		/* Finish advance */
	    
      i++;
    }
  }

  return nconcordant;
}



/* Finds concordant pairs if nconcordant is 0 */
List_T
Concordance_pair_up_genome (bool *abort_pairing_p, int *adjacent_score, int *concordant_score, List_T hitpairs,

			    List_T hitlist5_gplus, List_T hitlist5_gminus,
			    List_T hitlist3_gplus, List_T hitlist3_gminus,
			    
			    Ladder_T ladder5_plus, Ladder_T ladder5_minus,
			    Ladder_T ladder3_plus, Ladder_T ladder3_minus,

			    int querylength5, int querylength3, int maxpairedpaths, int genestrand) {
  int nconcordant = 0, n; /* Also used as an indicator of when to reset cutoff_level */
  Ladder_T newladder5_plus, newladder5_minus, newladder3_plus, newladder3_minus;

  T *hits5, *hits3;
  int nhits5, nhits3;

  int max_frontier_score, frontier_score, score5, score3;
  int cutoff_level, level;
  

  debug(printf("Starting Concordance_pair_up_genome\n"));

#if 0
  /* Rely instead on Ladder_cutoff, in case there are a few good hits */
  if (List_length(hitlist5_gplus) > MAX_HITS) {
    Stage3end_gc(hitlist5_gplus);
    newladder5_plus = Ladder_new((List_T) NULL,/*end5p*/true);
  } else {
    newladder5_plus = Ladder_new(hitlist5_gplus,/*end5p*/true);
    Ladder_merge(ladder5_plus,newladder5_plus);
  }

  if (List_length(hitlist5_gminus) > MAX_HITS) {
    Stage3end_gc(hitlist5_gminus);
    newladder5_minus = Ladder_new((List_T) NULL,/*end5p*/true);
  } else {
    newladder5_minus = Ladder_new(hitlist5_gminus,/*end5p*/true);
    Ladder_merge(ladder5_minus,newladder5_minus);
  }

  if (List_length(hitlist3_gplus) > MAX_HITS) {
    Stage3end_gc(hitlist3_gplus);
    newladder3_plus = Ladder_new((List_T) NULL,/*end5p*/false);
  } else {
    newladder3_plus = Ladder_new(hitlist3_gplus,/*end5p*/false);
    Ladder_merge(ladder3_plus,newladder3_plus);
  }

  if (List_length(hitlist3_gminus) > MAX_HITS) {
    Stage3end_gc(hitlist3_gminus);
    newladder3_minus = Ladder_new((List_T) NULL,/*end5p*/false);
  } else {
    newladder3_minus = Ladder_new(hitlist3_gminus,/*end5p*/false);
    Ladder_merge(ladder3_minus,newladder3_minus);
  }

#else
  newladder5_plus = Ladder_new(hitlist5_gplus,/*end5p*/true);
  newladder5_minus = Ladder_new(hitlist5_gminus,/*end5p*/true);
  newladder3_plus = Ladder_new(hitlist3_gplus,/*end5p*/false);
  newladder3_minus = Ladder_new(hitlist3_gminus,/*end5p*/false);

  Ladder_merge(ladder5_plus,newladder5_plus);
  Ladder_merge(ladder5_minus,newladder5_minus);
  Ladder_merge(ladder3_plus,newladder3_plus);
  Ladder_merge(ladder3_minus,newladder3_minus);
#endif

  
  /* Initial value, which resets to frontier_score + subopt_levels upon our first hitpair */
  cutoff_level = Ladder_cutoff(newladder5_plus) + Ladder_cutoff(ladder3_plus);
  if ((level = Ladder_cutoff(newladder5_minus) + Ladder_cutoff(ladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  if ((level = Ladder_cutoff(ladder5_plus) + Ladder_cutoff(newladder3_plus)) > cutoff_level) {
    cutoff_level = level;
  }
  if ((level = Ladder_cutoff(ladder5_minus) + Ladder_cutoff(newladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  max_frontier_score = cutoff_level;

  frontier_score = 0;
  /* Eliminating check on frontier_score against *found_score, because too greedy and can miss correct result */
  while (*abort_pairing_p == false && frontier_score <= cutoff_level) {
    debug1(printf("frontier_score = %d\n",frontier_score));
    for (score5 = 0; score5 <= frontier_score; score5++) {
      score3 = frontier_score - score5;
      debug1(printf("score5 = %d, score3 = %d\n",score5,score3));

      /* New hits 5 vs all hits 3 */
      /* plus/plus: hits5_plus against hits3_plus (really on minus) */
      if (score5 <= Ladder_cutoff(newladder5_plus) && score3 <= Ladder_cutoff(ladder3_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,newladder5_plus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_plus,score3);
	debug(printf("at score %d+%d, nnewhits5_plus = %d, nhits3_plus = %d\n",score5,score3,nhits5,nhits3));

	if ((n = do_genome_plus(&(*adjacent_score),&(*concordant_score),&hitpairs,
				hits5,nhits5,hits3,nhits3,
				querylength5,querylength3,genestrand)) > 0) {
	  if (nconcordant == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nconcordant += n) > maxpairedpaths) {
	    debug(printf(" -- %d concordant paths exceeds %d",nconcordant,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* minus/minus: hits3_minus (really on plus) against newhits5_minus */
      if (score5 <= Ladder_cutoff(newladder5_minus) && score3 <= Ladder_cutoff(ladder3_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,newladder5_minus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_minus,score3);
	debug(printf("at score %d+%d, nnewhits5_minus = %d, nhits3_minus = %d\n",score5,score3,nhits5,nhits3));

	if ((n = do_genome_minus(&(*adjacent_score),&(*concordant_score),&hitpairs,
				 hits5,nhits5,hits3,nhits3,
				 querylength5,querylength3,genestrand)) > 0) {
	  if (nconcordant == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nconcordant += n) > maxpairedpaths) {
	    debug(printf(" -- %d concordant paths exceeds %d",nconcordant,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* All hits 5 vs new hits 3 */
      /* plus/plus: hits5_plus against newhits3_plus (really on minus) */
      if (score5 <= Ladder_cutoff(ladder5_plus) && score3 <= Ladder_cutoff(newladder3_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_plus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,newladder3_plus,score3);
	debug(printf("at score %d+%d, nhits5_plus = %d, nnewhits3_plus = %d\n",score5,score3,nhits5,nhits3));

	if ((n = do_genome_plus(&(*adjacent_score),&(*concordant_score),&hitpairs,
				hits5,nhits5,hits3,nhits3,
				querylength5,querylength3,genestrand)) > 0) {
	  if (nconcordant == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nconcordant += n) > maxpairedpaths) {
	    debug(printf(" -- %d concordant paths exceeds %d",nconcordant,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* minus/minus: newhits3_minus (really on plus) against hits5_minus */
      if (score5 <= Ladder_cutoff(ladder5_minus) && score3 <= Ladder_cutoff(newladder3_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_minus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,newladder3_minus,score3);

	debug(printf("at score %d+%d, nhits5_minus = %d, nnewhits3_minus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_genome_minus(&(*adjacent_score),&(*concordant_score),&hitpairs,
				 hits5,nhits5,hits3,nhits3,
				 querylength5,querylength3,genestrand)) > 0) {
	  if (nconcordant == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nconcordant += n) > maxpairedpaths) {
	    debug(printf(" -- %d concordant paths exceeds %d",nconcordant,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }
    }

    frontier_score++;
  }


  Ladder_free(&newladder5_plus);
  Ladder_free(&newladder5_minus);
  Ladder_free(&newladder3_plus);
  Ladder_free(&newladder3_minus);

  Ladder_gc_duplicates(ladder5_plus);
  Ladder_gc_duplicates(ladder5_minus);
  Ladder_gc_duplicates(ladder3_plus);
  Ladder_gc_duplicates(ladder3_minus);

  debug(printf("Finished with Concordance_pair_up_genome: %d concordant\n",List_length(hitpairs)));

  return hitpairs;
}


static int
do_distant_plus_plus (int *concordant_score, List_T *hitpairs,
		      List_T *conc_transloc, List_T *samechr,
		      T *hits5, int nhits5, T *hits3, int nhits3,
		      int querylength5, int querylength3, int genestrand) {
  int nfound = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  int score;
  Univcoord_T insert_start;
  Chrpos_T pairmax;

  if (nhits5 > 0 && nhits3 > 0) {
    i = j = 0;
    while (i < nhits5) {
      hit5 = hits5[i];
      if (circularp[hit5->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      
      insert_start = hit5->genomicend - querylength5;
#ifdef DEBUG
      printf("plus/plus: i=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
	     i,nhits5,hit5->effective_chrnum,hit5->genomicstart - hit5->chroffset,hit5->genomicend - hit5->chroffset,
	     print_sense(hit5->sensedir),Method_string(hit5->method),hit5->circularalias,hit5);
      if (j >= 0 && j < nhits3) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits3,hits3[j]->effective_chrnum,
	       hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset);
      }
      printf("\n");
#endif
      
      while (j >= 0 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	j--;
      }
      j++;		/* Finish backup */
      
      while (j < nhits3 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	j++;
      }
      
      while (j < nhits3 && hits3[j]->genomicstart + querylength3 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]->circularalias,hits3[j]));
	hit3 = hits3[j];
	
	if (hit5->effective_chrnum != hit3->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,hit3->effective_chrnum));
	  
	} else if (hit5->distant_splice_p == true && hit3->distant_splice_p == true /* && hit5->other_chrnum != hit3->other_chrnum */) {
	  /* Could potentially miss an alignment if the two ends overlap */
	  debug(printf(" => double splice translocations"));
	  
	} else if (hit5->distant_splice_p == true || hit3->distant_splice_p == true) {
	  debug(printf(" => conc_transloc effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit5->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT_TRANSLOCATIONS,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/false)) != NULL) {
	    *conc_transloc = List_push(*conc_transloc,(void *) stage3pair);
	    nfound++;
	  }
	  
	} else if (SENSE_INCONSISTENT_P(hit5->sensedir,hit3->sensedir)) {
	  debug(printf(" => sense inconsistent: %d | %d = %d",hit5->sensedir,hit3->sensedir,hit5->sensedir|hit3->sensedir));
	  
	} else if (hit3->genomicend < hit5->genomicstart) {
	  debug(printf(" => scramble because end3 %llu < start5 %llu\n",
		       (unsigned long long) hit3->genomicend,(unsigned long long) hit5->genomicstart));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/PAIRED_SCRAMBLE,
					   /*expect_concordant_p*/false,/*transcriptome_guided_p*/false)) != NULL) {
	    *samechr = List_push(*samechr,(void *) stage3pair);
	    nfound++;
	  }
	  
	} else {
	  /* Unexpected? */
	  debug(printf(" => concordant effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit5->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/false)) != NULL) {
	    
	    if ((score = (stage3pair->hit5->querylength - stage3pair->hit5->nmatches) + 
		 (stage3pair->hit3->querylength - stage3pair->hit3->nmatches)) < *concordant_score) {
	      *concordant_score = score;
	      assert(stage3pair->hit5->nmatches <= stage3pair->hit5->querylength);
	      assert(stage3pair->hit3->nmatches <= stage3pair->hit3->querylength);
	      debug(printf(" => Updating concordant_score to be %d = %d + %d\n",
			   *concordant_score,stage3pair->hit5->querylength - stage3pair->hit5->nmatches,
			   stage3pair->hit3->querylength - stage3pair->hit3->nmatches));
	    }
	    *hitpairs = List_push(*hitpairs,(void *) stage3pair);
	    nfound++;
	  }
	}
	debug(printf("\n"));
	
	j++;
      }
      j--;		/* Finish advance */
      
      i++;
    }
  }
  
  return nfound;
}


static int
do_distant_minus_minus (int *concordant_score, List_T *hitpairs,
			List_T *conc_transloc, List_T *samechr,
			T *hits5, int nhits5, T *hits3, int nhits3,
			int querylength5, int querylength3, int genestrand) {
  int nfound = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  int score;
  Univcoord_T insert_start;
  Chrpos_T pairmax;

  if (nhits3 > 0 && nhits5 > 0) {
    i = j = 0;
    while (i < nhits3) {
      hit3 = hits3[i];
      if (circularp[hit3->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      insert_start = hit3->genomicstart - querylength3;
#ifdef DEBUG
      printf("minus/minus: i=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
	     i,nhits3,hit3->effective_chrnum,hit3->genomicstart - hit3->chroffset,hit3->genomicend - hit3->chroffset,
	     print_sense(hit3->sensedir),Method_string(hit3->method),hit3->circularalias,hit3);
      if (j >= 0 && j < nhits5) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits5,hits5[j]->effective_chrnum,
	       hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset);
      }
      printf("\n");
#endif
      
      while (j >= 0 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	j--;
      }
      j++;			/* Finish backup */
      
      while (j < nhits5 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	j++;
      }
      
      while (j < nhits5 && hits5[j]->genomicend + querylength5 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s circularalias:%d %p",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]->circularalias,hits5[j]));
	hit5 = hits5[j];
	
	/* Do want to see pairs previously seen */
	if (hit3->effective_chrnum != hit5->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,hit3->effective_chrnum));
	  
	} else if (hit5->distant_splice_p == true && hit3->distant_splice_p == true /* && hit5->other_chrnum != hit3->other_chrnum */) {
	  /* Could potentially miss an alignment if the two ends overlap */
	  debug(printf(" => double splice translocations"));
	  
	} else if (hit5->distant_splice_p == true || hit3->distant_splice_p == true) {
	  debug(printf(" => conc_transloc effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit3->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT_TRANSLOCATIONS,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/false)) != NULL) {
	    *conc_transloc = List_push(*conc_transloc,(void *) stage3pair);
	    nfound++;
	  }
		
	} else if (SENSE_INCONSISTENT_P(hit3->sensedir,hit5->sensedir)) {
	  debug(printf(" => sense inconsistent: %d | %d = %d",hit5->sensedir,hit3->sensedir,hit5->sensedir|hit3->sensedir));
		
	} else if (hit5->genomicstart < hit3->genomicend) {
	  debug(printf(" => scramble because start5 %llu < end3 %llu\n",
		       (unsigned long long) hit5->genomicstart,(unsigned long long) hit3->genomicend));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/PAIRED_SCRAMBLE,
					   /*expect_concordant_p*/false,/*transcriptome_guided_p*/false)) != NULL) {
	    *samechr = List_push(*samechr,(void *) stage3pair);
	    nfound++;
	  }
		
	} else {
	  /* Unexpected? */
	  debug(printf(" => concordant effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit3->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/CONCORDANT,
					   /*expect_concordant_p*/true,/*transcriptome_guided_p*/false)) != NULL) {
		  
	    if ((score = (stage3pair->hit5->querylength - stage3pair->hit5->nmatches) + 
		 (stage3pair->hit3->querylength - stage3pair->hit3->nmatches)) < *concordant_score) {
	      *concordant_score = score;
	      assert(stage3pair->hit5->nmatches <= stage3pair->hit5->querylength);
	      assert(stage3pair->hit3->nmatches <= stage3pair->hit3->querylength);
	      debug(printf(" => Updating concordant_score to be %d = %d + %d\n",
			   *concordant_score,stage3pair->hit5->querylength - stage3pair->hit5->nmatches,
			   stage3pair->hit3->querylength - stage3pair->hit3->nmatches));
	    }
	    *hitpairs = List_push(*hitpairs,(void *) stage3pair);
	    nfound++;
	  }
	}
	debug(printf("\n"));
	      
	j++;
      }
      j--;		/* Finish advance */
	    
      i++;
    }
  }
  
  return nfound;
}


static int
do_distant_plus_minus (List_T *samechr,
		       T *hits5, int nhits5, T *hits3, int nhits3,
		       int querylength5, int querylength3, int genestrand) {
  int nfound = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  Univcoord_T insert_start;
  Chrpos_T pairmax;

  if (nhits5 > 0 && nhits3 > 0) {
    i = j = 0;
    while (i < nhits5) {
      hit5 = hits5[i];
      if (circularp[hit5->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      insert_start = hit5->genomicend - querylength5;
#ifdef DEBUG
      printf("plus/minus: i=%d/%d #%d:%u..%u %s %s %p",
	     i,nhits5,hit5->effective_chrnum,hit5->genomicstart - hit5->chroffset,hit5->genomicend - hit5->chroffset,
	     print_sense(hit5->sensedir),Method_string(hit5->method),hit5);
      if (j >= 0 && j < nhits3) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits3,hits3[j]->effective_chrnum,
	       hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset);
      }
      printf("\n");
#endif
	    
      while (j >= 0 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]));
	j--;
      }
      j++;		/* Finish backup */

      while (j < nhits3 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u %s %s %p\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]));
	j++;
      }

      while (j < nhits3 && hits3[j]->genomicstart + querylength3 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s %p",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset,
		     print_sense(hits3[j]->sensedir),Method_string(hits3[j]->method),hits3[j]));
	hit3 = hits3[j];
		
	if (hit5->effective_chrnum != hit3->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,hit3->effective_chrnum));
		
	} else if (hit5->distant_splice_p == true && hit3->distant_splice_p == true /* && hit5->other_chrnum != hit3->other_chrnum */) {
	  debug(printf(" => double splice translocations"));
		
	} else if (SENSE_INCONSISTENT_FOR_INVERSION_P(hit5->sensedir,hit3->sensedir)) {
	  debug(printf(" => sense inconsistent for inversion"));
#if 0
	} else if (hits3[j]->genomicstart + querylength3 <= insert_start) {
	  debug(printf(" => scramble"));
	  if (nsamechr <= maxpairedpaths &&
	      (stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/PAIRED_SCRAMBLE,
					   /*expect_concordant_p*/false,/*transcriptome_guided_p*/false)) != NULL) {
	    *samechr = List_push(*samechr,(void *) stage3pair);
	    nfound++;
	  }
#endif
	} else {
	  debug(printf(" => inversion effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit5->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/PAIRED_INVERSION,
					   /*expect_concordant_p*/false,/*transcriptome_guided_p*/false)) != NULL) {
	    *samechr = List_push(*samechr,(void *) stage3pair);
	    nfound++;
	  }
	}
	debug(printf("\n"));

	j++;
      }
      j--;		/* Finish advance */

      i++;
    }
  }

  return nfound;
}


static int
do_distant_minus_plus (List_T *samechr,
		       T *hits5, int nhits5, T *hits3, int nhits3,
		       int querylength5, int querylength3, int genestrand) {
  int nfound = 0;
  int i, j;
  Stage3pair_T stage3pair;
  T hit5, hit3;
  Univcoord_T insert_start;
  Chrpos_T pairmax;

  if (nhits3 > 0 && nhits5 > 0) {
    i = j = 0;
    while (i < nhits3) {
      hit3 = hits3[i];
      if (circularp[hit3->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      insert_start = hit3->genomicstart - querylength3;
#ifdef DEBUG
      printf("minus/plus: i=%d/%d #%d:%u..%u %s %s %p",
	     i,nhits3,hit3->effective_chrnum,hit3->genomicstart - hit3->chroffset,hit3->genomicend - hit3->chroffset,
	     print_sense(hit3->sensedir),Method_string(hit3->method),hit3);
      if (j >= 0 && j < nhits5) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits5,hits5[j]->effective_chrnum,
	       hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset);
      }
      printf("\n");
#endif

      while (j >= 0 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u %s %s %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]));
	j--;
      }
      j++;			/* Finish backup */

      while (j < nhits5 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u %s %s %p\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]));
	j++;
      }

      while (j < nhits5 && hits5[j]->genomicend + querylength5 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u %s %s %p",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset,
		     print_sense(hits5[j]->sensedir),Method_string(hits5[j]->method),hits5[j]));
	hit5 = hits5[j];

	if (hit3->effective_chrnum != hit5->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,hit3->effective_chrnum));
		
	} else if (hit5->distant_splice_p == true && hit3->distant_splice_p == true /* && hit5->other_chrnum != hit3->other_chrnum */) {
	  debug(printf(" => double splice translocations"));
		
	} else if (SENSE_INCONSISTENT_FOR_INVERSION_P(hit3->sensedir,hit5->sensedir)) {
	  debug(printf(" => sense inconsistent for inversion"));
#if 0
	} else if (hits5[j]->genomicend + querylength5 <= insert_start) {
	  debug(printf(" => scramble"));
	  if ((*nsamechr) <= maxpairedpaths &&
	      (stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/PAIRED_SCRAMBLE,
					   /*expect_concordant_p*/false,/*transcriptome_guided_p*/false)) != NULL) {
	    *samechr = List_push(*samechr,(void *) stage3pair);
	    nfound++;
	  }
#endif
	} else {
	  debug(printf(" => inversion effchr %d (chrnum5 %d, chrnum3 %d)",
		       hit3->effective_chrnum,hit5->chrnum,hit3->chrnum));
	  if ((stage3pair = Stage3pair_new(hit5,hit3,genestrand,/*pairtype*/PAIRED_INVERSION,
					   /*expect_concordant_p*/false,/*transcriptome_guided_p*/false)) != NULL) {
	    *samechr = List_push(*samechr,(void *) stage3pair);
	    nfound++;
	  }
	}
	debug(printf("\n"));
	      
	j++;
      }
      j--;		/* Finish advance */
	    
      i++;
    }
  }

  return nfound;
}



List_T
Concordance_pair_up_distant (bool *abort_pairing_p, int *concordant_score,
			     List_T *samechr, List_T *conc_transloc, List_T hitpairs, 

			     List_T hitlist5_gplus, List_T hitlist5_gminus,
			     List_T hitlist3_gplus, List_T hitlist3_gminus,

			     Ladder_T ladder5_plus, Ladder_T ladder5_minus,
			     Ladder_T ladder3_plus, Ladder_T ladder3_minus,

			     int querylength5, int querylength3, int maxpairedpaths, int genestrand) {
  int nfound = 0, n;
  int max_frontier_score, frontier_score, score5, score3;
  int cutoff_level, level;
  Ladder_T newladder5_plus, newladder5_minus, newladder3_plus, newladder3_minus;

  T *hits5, *hits3;
  int nhits5, nhits3;


  debug(printf("Starting Concordance_pair_up_distant\n"));

#if 0
  /* Rely instead on Ladder_cutoff, in case there are a few good hits */
  if (List_length(hitlist5_gplus) > MAX_HITS) {
    Stage3end_gc(hitlist5_gplus);
    newladder5_plus = Ladder_new((List_T) NULL,/*end5p*/true);
  } else {
    newladder5_plus = Ladder_new(hitlist5_gplus,/*end5p*/true);
    Ladder_merge(ladder5_plus,newladder5_plus);
  }

  if (List_length(hitlist5_gminus) > MAX_HITS) {
    Stage3end_gc(hitlist5_gminus);
    newladder5_minus = Ladder_new((List_T) NULL,/*end5p*/true);
  } else {
    newladder5_minus = Ladder_new(hitlist5_gminus,/*end5p*/true);
    Ladder_merge(ladder5_minus,newladder5_minus);
  }

  if (List_length(hitlist3_gplus) > MAX_HITS) {
    Stage3end_gc(hitlist3_gplus);
    newladder3_plus = Ladder_new((List_T) NULL,/*end5p*/false);
  } else {
    newladder3_plus = Ladder_new(hitlist3_gplus,/*end5p*/false);
    Ladder_merge(ladder3_plus,newladder3_plus);
  }

  if (List_length(hitlist3_gminus) > MAX_HITS) {
    Stage3end_gc(hitlist3_gminus);
    newladder3_minus = Ladder_new((List_T) NULL,/*end5p*/false);
  } else {
    newladder3_minus = Ladder_new(hitlist3_gminus,/*end5p*/false);
    Ladder_merge(ladder3_minus,newladder3_minus);
  }

#else
  newladder5_plus = Ladder_new(hitlist5_gplus,/*end5p*/true);
  newladder5_minus = Ladder_new(hitlist5_gminus,/*end5p*/true);
  newladder3_plus = Ladder_new(hitlist3_gplus,/*end5p*/false);
  newladder3_minus = Ladder_new(hitlist3_gminus,/*end5p*/false);

  Ladder_merge(ladder5_plus,newladder5_plus);
  Ladder_merge(ladder5_minus,newladder5_minus);
  Ladder_merge(ladder3_plus,newladder3_plus);
  Ladder_merge(ladder3_minus,newladder3_minus);
#endif


  cutoff_level = Ladder_cutoff(newladder5_plus) + Ladder_cutoff(ladder3_plus);
  if ((level = Ladder_cutoff(newladder5_minus) + Ladder_cutoff(ladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  if ((level = Ladder_cutoff(ladder5_plus) + Ladder_cutoff(newladder3_plus)) > cutoff_level) {
    cutoff_level = level;
  }
  if ((level = Ladder_cutoff(ladder5_minus) + Ladder_cutoff(newladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  if ((level = Ladder_cutoff(ladder5_plus) + Ladder_cutoff(ladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  if ((level = Ladder_cutoff(ladder5_minus) + Ladder_cutoff(ladder3_plus)) > cutoff_level) {
    cutoff_level = level;
  }
  max_frontier_score = cutoff_level;

  frontier_score = 0;
  while (frontier_score <= cutoff_level) {
    debug1(printf("frontier_score = %d\n",frontier_score));
    for (score5 = 0; score5 <= frontier_score; score5++) {
      score3 = frontier_score - score5;
      debug1(printf("score5 = %d, score3 = %d\n",score5,score3));

      /* New hits 5 vs all hits 3 */
      /* plus/plus: hits5_plus against hits3_plus (really on minus) */
      if (score5 <= Ladder_cutoff(newladder5_plus) && score3 <= Ladder_cutoff(ladder3_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,newladder5_plus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_plus,score3);
	debug(printf("at score %d+%d, nnewhits5_plus = %d, nhits3_plus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_distant_plus_plus(&(*concordant_score),&hitpairs,&(*conc_transloc),&(*samechr),
				      hits5,nhits5,hits3,nhits3,
				      querylength5,querylength3,genestrand)) > 0) {
	  if (nfound == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nfound += n) > maxpairedpaths) {
	    debug(printf(" -- %d found paths exceeds %d",nfound,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* minus/minus: hits3_minus (really on plus) against hits5_minus */
      if (score5 <= Ladder_cutoff(newladder5_minus) && score3 <= Ladder_cutoff(ladder3_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,newladder5_minus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_minus,score3);
	debug(printf("at score %d+%d, nnewhits5_minus = %d, nhits3_minus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_distant_minus_minus(&(*concordant_score),&hitpairs,&(*conc_transloc),&(*samechr),
					hits5,nhits5,hits3,nhits3,
					querylength5,querylength3,genestrand)) > 0) {
	  if (nfound == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nfound += n) > maxpairedpaths) {
	    debug(printf(" -- %d found paths exceeds %d",nfound,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }
	

      /* All hits 5 vs new hits 3 */
      /* plus/plus: hits5_plus against hits3_plus (really on minus) */
      if (score5 <= Ladder_cutoff(ladder5_plus) && score3 <= Ladder_cutoff(newladder3_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_plus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,newladder3_plus,score3);
	debug(printf("at score %d+%d, nnewhits5_plus = %d, nhits3_plus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_distant_plus_plus(&(*concordant_score),&hitpairs,&(*conc_transloc),&(*samechr),
				      hits5,nhits5,hits3,nhits3,
				      querylength5,querylength3,genestrand)) > 0) {
	  if (nfound == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nfound += n) > maxpairedpaths) {
	    debug(printf(" -- %d found paths exceeds %d",nfound,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* minus/minus: hits3_minus (really on plus) against hits5_minus */
      if (score5 <= Ladder_cutoff(ladder5_minus) && score3 <= Ladder_cutoff(newladder3_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_minus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,newladder3_minus,score3);
	debug(printf("at score %d+%d, nnewhits5_minus = %d, nhits3_minus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_distant_minus_minus(&(*concordant_score),&hitpairs,&(*conc_transloc),&(*samechr),
					hits5,nhits5,hits3,nhits3,
					querylength5,querylength3,genestrand)) > 0) {
	  if (nfound == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nfound += n) > maxpairedpaths) {
	    debug(printf(" -- %d found paths exceeds %d",nfound,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }


      /* All hits5 vs all hits3 */
      /* plus/minus (inversions): hits5_plus against hits3_minus */
      if (score5 <= Ladder_cutoff(ladder5_plus) && score3 <= Ladder_cutoff(ladder3_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_plus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_minus,score3);
	debug(printf("at score %d+%d, nhits5_plus = %d, nhits3_minus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_distant_plus_minus(&(*samechr),hits5,nhits5,hits3,nhits3,
				       querylength5,querylength3,genestrand)) > 0) {
	  if (nfound == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nfound += n) > maxpairedpaths) {
	    debug(printf(" -- %d found paths exceeds %d",nfound,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }

      /* minus/plus (inversions): hits3_plus against hits5_minus */
      if (score5 <= Ladder_cutoff(ladder5_minus) && score3 <= Ladder_cutoff(ladder3_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_minus,score5);
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_plus,score3);
	debug(printf("at score %d+%d, nhits5_minus = %d, nhits3_plus = %d\n",score5,score3,nhits5,nhits3));
	if ((n = do_distant_minus_plus(&(*samechr),hits5,nhits5,hits3,nhits3,
				       querylength5,querylength3,genestrand)) > 0) {
	  if (nfound == 0) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	  }
	  if ((nfound += n) > maxpairedpaths) {
	    debug(printf(" -- %d found paths exceeds %d",nfound,maxpairedpaths));
	    *abort_pairing_p = true;
	  }
	}
      }
    }

    frontier_score++;
  }


  Ladder_free(&newladder5_plus);
  Ladder_free(&newladder5_minus);
  Ladder_free(&newladder3_plus);
  Ladder_free(&newladder3_minus);

  Ladder_gc_duplicates(ladder5_plus);
  Ladder_gc_duplicates(ladder5_minus);
  Ladder_gc_duplicates(ladder3_plus);
  Ladder_gc_duplicates(ladder3_minus);

  debug(printf("Finished with Concordance_pair_up_distant: %d concordant, %d samechr, %d conc_transloc\n",
		List_length(hitpairs),List_length(*samechr),List_length(*conc_transloc)));

  return hitpairs;
}



static int
path_ascending_cmp (const void *a, const void *b) {
  List_T x = * (List_T *) a;
  List_T y = * (List_T *) b;
  Univdiag_T diagonalx, diagonaly;
  Univcoord_T path_start_x, path_start_y, path_end_x, path_end_y;

  diagonalx = List_head(x);
  diagonaly = List_head(y);
  path_start_x = diagonalx->univdiagonal;
  path_start_y = diagonaly->univdiagonal;

  if (path_start_x < path_start_y) {
    return -1;
  } else if (path_start_y < path_start_x) {
    return +1;
  } else {
    diagonalx = (Univdiag_T) List_last_value(x);
    diagonaly = (Univdiag_T) List_last_value(y);
    path_end_x = diagonalx->univdiagonal;
    path_end_y = diagonaly->univdiagonal;
    if (path_end_x < path_end_y) {
      return -1;
    } else if (path_end_y < path_end_x) {
      return +1;
    } else {
      return 0;
    }
  }
}


static int
hit_ascending_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->high < y->high) {
    return -1;
  } else if (y->high < x->high) {
    return +1;
  } else if (x->low < y->low) {
    return -1;
  } else if (y->low < x->low) {
    return +1;
  } else {
    return 0;
  }
}


/* Modeled after do_genome_plus */
static Intlist_T
do_records5_plus (Intlist_T anchors, Record_T *records5, int *candidates5, int ncandidates5,
		  T *hits3, int nhits3, int querylength5, int querylength3) {
  int ii, i, j;
  Record_T record5;
  T hit3;
  Univcoord_T insert_start;
  int pairmax;


  if (ncandidates5 > 0 && nhits3 > 0) {
    ii = j = 0;
    while (ii < ncandidates5) {
      i = candidates5[ii];
      record5 = records5[i];
      if (circularp[record5->chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
    
      insert_start = record5->genomicend - querylength5;
#ifdef DEBUG
      printf("plus/plus: ii=%d/%d #%d:%u..%u",
	     ii,ncandidates5,record5->chrnum,record5->genomicstart - record5->chroffset,record5->genomicend - record5->chroffset);
      if (j >= 0 && j < nhits3) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits3,hits3[j]->effective_chrnum,
	       hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset);
      }
      printf("\n");
#endif
    
      while (j >= 0 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset));
	j--;
      }
      j++;		/* Finish backup */
    
      while (j < nhits3 && 
	     hits3[j]->genomicstart + querylength3 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u\n",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset));
	j++;
      }
    
      while (j < nhits3 && hits3[j]->genomicstart + querylength3 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u",
		     j,nhits3,hits3[j]->effective_chrnum,hits3[j]->genomicstart - hits3[j]->chroffset,hits3[j]->genomicend - hits3[j]->chroffset));
	hit3 = hits3[j];
      
	if (record5->chrnum != hit3->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",record5->chrnum,hit3->effective_chrnum));
	
	} else if (hit3->genomicend < record5->genomicstart) {
	  debug(printf(" => scramble because end3 %llu < start5 %llu\n",
		       (unsigned long long) hit3->genomicend,(unsigned long long) record5->genomicstart));
	
	} else {
	  debug(printf(" => concordant"));
	  anchors = Intlist_push(anchors,i);
	}
	debug(printf("\n"));
      
	j++;
      }
      j--;		/* Finish advance */
    
      ii++;
    }
  }

  return anchors;
}

/* Modeled after do_genome_minus */
static Intlist_T
do_records5_minus (Intlist_T anchors, Record_T *records5, int *candidates5, int ncandidates5,
		   T *hits3, int nhits3, int querylength5, int querylength3) {
  int i, jj, j;
  Record_T record5;
  T hit3;
  Univcoord_T insert_start;
  int pairmax;

  if (ncandidates5 > 0 && nhits3 > 0) {
    i = jj = 0;
    while (i < nhits3) {
      hit3 = hits3[i];
      if (circularp[hit3->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }

      insert_start = hit3->genomicstart - querylength3;
#ifdef DEBUG
      printf("minus/minus: i=%d/%d #%d:%u..%u",
	     i,nhits3,hit3->effective_chrnum,hit3->genomicstart - hit3->chroffset,hit3->genomicend - hit3->chroffset);
      if (jj >= 0 && jj < ncandidates5) {
	j = candidates5[jj];
	record5 = records5[j];
	printf("    jj=%d/%d #%d:%u..%u",jj,ncandidates5,record5->chrnum,
	       record5->genomicstart - record5->chroffset,record5->genomicend - record5->chroffset);
      }
      printf("\n");
#endif

      while (jj >= 0 && 
	     records5[candidates5[jj]]->genomicend + querylength5 /* for scramble: */ + pairmax > insert_start) {
	debug(j = candidates5[jj]);
	debug(record5 = records5[j]);
	debug(printf("  backup: j=%d/%d #%d:%u..%u\n",
		     j,ncandidates5,record5->chrnum,record5->genomicstart - record5->chroffset,record5->genomicend - record5->chroffset));
	jj--;
      }
      jj++;			/* Finish backup */
      
      while (jj < ncandidates5 && 
	     records5[candidates5[jj]]->genomicend + querylength5 /* for scramble: */ + pairmax <= insert_start) {
	debug(j = candidates5[jj]);
	debug(record5 = records5[j]);
	debug(printf("  advance: jj=%d/%d #%d:%u..%u\n",
		     jj,ncandidates5,record5->chrnum,record5->genomicstart - record5->chroffset,record5->genomicend - record5->chroffset));
	jj++;
      }
      
      while (jj < ncandidates5 && records5[candidates5[jj]]->genomicend + querylength5 <= pairmax + insert_start) {
	j = candidates5[jj];
	record5 = records5[j];
	debug(printf("  overlap: jj=%d/%d #%d:%u..%u",
		     jj,ncandidates5,record5->chrnum,record5->genomicstart - record5->chroffset,record5->genomicend - record5->chroffset));

	if (hit3->effective_chrnum != record5->chrnum) {
	  debug(printf(" => diff chrs %d and %d",record5->chrnum,hit3->effective_chrnum));

	} else if (record5->genomicstart < hit3->genomicend) {
	  debug(printf(" => scramble because start5 %llu < end3 %llu\n",
		       (unsigned long long) record5->genomicstart,(unsigned long long) hit3->genomicend));
		
	} else {
	  debug(printf(" => concordant"));
	  anchors = Intlist_push(anchors,j);
	}
	debug(printf("\n"));
	      
	jj++;
      }
      jj--;		/* Finish advance */
	    
      i++;
    }
  }

  return anchors;
}


int
Concordance_filter_records5 (Intlist_T *plus_anchors, Intlist_T *minus_anchors, int score_start,
			     Record_T *records5_plus, int *candidates5_plus, int ncandidates5_plus,
			     Record_T *records5_minus, int *candidates5_minus, int ncandidates5_minus,
			     Ladder_T ladder3_plus, Ladder_T ladder3_minus,
			     int querylength5, int querylength3) {
  bool foundp = false;

  T *hits3;
  int nhits3;
  int max_frontier_score, frontier_score, cutoff_level, level, score3;

  *plus_anchors = *minus_anchors = (Intlist_T) NULL;

  cutoff_level = Ladder_cutoff(ladder3_plus);
  if ((level = Ladder_cutoff(ladder3_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  max_frontier_score = cutoff_level;

  frontier_score = score_start;
  while (frontier_score <= cutoff_level) {
    debug1(printf("frontier_score = %d\n",frontier_score));
    for (score3 = 0; score3 <= frontier_score; score3++) {
      debug1(printf("score3 = %d\n",score3));

      /* plus/plus */
      if (score3 <= Ladder_cutoff(ladder3_plus)) {
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_plus,score3);
	debug(printf("at score %d, nhits3_plus = %d\n",score3,nhits3));

	if ((*plus_anchors = do_records5_plus(*plus_anchors,
					      records5_plus,candidates5_plus,ncandidates5_plus,
					      hits3,nhits3,querylength5,querylength3)) != NULL) {
	  if (foundp == false) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	    foundp = true;
	  }
	}
      }

      /* minus/minus */
      if (score3 <= Ladder_cutoff(ladder3_minus)) {
	hits3 = Ladder_hits_for_score(&nhits3,ladder3_minus,score3);
	debug(printf("at score %d, nhits3_minus = %d\n",score3,nhits3));

	if ((*minus_anchors = do_records5_minus(*minus_anchors,
						records5_minus,candidates5_minus,ncandidates5_minus,
						hits3,nhits3,querylength5,querylength3)) != NULL) {
	  if (foundp == false) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	    foundp = true;
	  }
	}
      }
    }

    frontier_score++;
  }

  return frontier_score;
}



/* Modeled after do_genome_plus */
static Intlist_T
do_records3_plus (Intlist_T anchors, T *hits5, int nhits5,
		  Record_T *records3, int *candidates3, int ncandidates3,
		  int querylength5, int querylength3) {
  int i, jj, j;
  T hit5;
  Record_T record3;
  Univcoord_T insert_start;
  int pairmax;

  if (nhits5 > 0 && ncandidates3 > 0) {
    i = jj = 0;
    while (i < nhits5) {
      hit5 = hits5[i];
      if (circularp[hit5->effective_chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }
      
      insert_start = hit5->genomicend - querylength5;
#ifdef DEBUG
      printf("plus/plus: i=%d/%d #%d:%u..%u",
	     i,nhits5,hit5->effective_chrnum,hit5->genomicstart - hit5->chroffset,hit5->genomicend - hit5->chroffset);
      if (jj >= 0 && jj < ncandidates3) {
	j = candidates3[jj];
	record3 = records3[j];
	printf("    jj=%d/%d #%d:%u..%u",jj,ncandidates3,record3->chrnum,
	       record3->genomicstart - record3->chroffset,record3->genomicend - record3->chroffset);
      }
      printf("\n");
#endif

      while (jj >= 0 && 
	     records3[candidates3[jj]]->genomicstart + querylength3 /* for scramble: */ + pairmax > insert_start) {
	debug(j = candidates3[jj]);
	debug(record3 = records3[j]);
	debug(printf("  backup: jj=%d/%d #%d:%u..%u\n",
		     jj,ncandidates3,record3->chrnum,record3->genomicstart - record3->chroffset,record3->genomicend - record3->chroffset));
	jj--;
      }
      jj++;		/* Finish backup */

      while (jj < ncandidates3 && 
	     records3[candidates3[jj]]->genomicstart + querylength3 /* for scramble: */ + pairmax <= insert_start) {
	debug(j = candidates3[jj]);
	debug(record3 = records3[j]);
	debug(printf("  advance: jj=%d/%d #%d:%u..%u\n",
		     jj,ncandidates3,record3->chrnum,record3->genomicstart - record3->chroffset,record3->genomicend - record3->chroffset));
	jj++;
      }
      
      while (jj < ncandidates3 && records3[candidates3[jj]]->genomicstart + querylength3 <= pairmax + insert_start) {
	j = candidates3[jj];
	record3 = records3[j];
	debug(printf("  overlap: jj=%d/%d #%d:%u..%u",
		     jj,ncandidates3,record3->chrnum,record3->genomicstart - record3->chroffset,record3->genomicend - record3->chroffset));
	
	if (hit5->effective_chrnum != record3->chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,record3->chrnum));
	  
	} else if (record3->genomicend < hit5->genomicstart) {
	  debug(printf(" => scramble because end3 %llu < start5 %llu\n",
		       (unsigned long long) record3->genomicend,(unsigned long long) hit5->genomicstart));
	  
	} else {
	  debug(printf(" => concordant"));
	  anchors = Intlist_push(anchors,j);
	}
	debug(printf("\n"));
	
	jj++;
      }
      jj--;		/* Finish advance */
      
      i++;
    }
  }

  return anchors;
}


/* Modeled after do_genome_minus */
static Intlist_T
do_records3_minus (Intlist_T anchors, T *hits5, int nhits5,
		   Record_T *records3, int *candidates3, int ncandidates3,
		   int querylength5, int querylength3) {
  int ii, i, j;
  T hit5;
  Record_T record3;
  Univcoord_T insert_start;
  int pairmax;

  if (nhits5 > 0 && ncandidates3 > 0) {
    ii = j = 0;
    while (ii < ncandidates3) {
      i = candidates3[ii];
      record3 = records3[i];
      if (circularp[record3->chrnum] == true) {
	pairmax = pairmax_circular;
      } else {
	pairmax = pairmax_linear;
      }

      insert_start = record3->genomicstart - querylength3;
#ifdef DEBUG
      printf("minus/minus: ii=%d/%d #%d:%u..%u",
	     ii,ncandidates3,record3->chrnum,record3->genomicstart - record3->chroffset,record3->genomicend - record3->chroffset);
      if (j >= 0 && j < nhits5) {
	printf("    j=%d/%d #%d:%u..%u",j,nhits5,hits5[j]->effective_chrnum,
	       hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset);
      }
      printf("\n");
#endif

      while (j >= 0 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax > insert_start) {
	debug(printf("  backup: j=%d/%d #%d:%u..%u\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset));
	j--;
      }
      j++;			/* Finish backup */
      
      while (j < nhits5 && 
	     hits5[j]->genomicend + querylength5 /* for scramble: */ + pairmax <= insert_start) {
	debug(printf("  advance: j=%d/%d #%d:%u..%u\n",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset));
	j++;
      }
      
      while (j < nhits5 && hits5[j]->genomicend + querylength5 <= pairmax + insert_start) {
	debug(printf("  overlap: j=%d/%d #%d:%u..%u",
		     j,nhits5,hits5[j]->effective_chrnum,hits5[j]->genomicstart - hits5[j]->chroffset,hits5[j]->genomicend - hits5[j]->chroffset));
	hit5 = hits5[j];

	if (record3->chrnum != hit5->effective_chrnum) {
	  debug(printf(" => diff chrs %d and %d",hit5->effective_chrnum,record3->chrnum));

	} else if (hit5->genomicstart < record3->genomicend) {
	  debug(printf(" => scramble because start5 %llu < end3 %llu\n",
		       (unsigned long long) hit5->genomicstart,(unsigned long long) record3->genomicend));
		
	} else {
	  debug(printf(" => concordant"));
	  anchors = Intlist_push(anchors,i);
	}
	debug(printf("\n"));
	      
	j++;
      }
      j--;		/* Finish advance */
	    
      ii++;
    }
  }

  return anchors;
}

int
Concordance_filter_records3 (Intlist_T *plus_anchors, Intlist_T *minus_anchors, int score_start,
			     Record_T *records3_plus, int *candidates3_plus, int ncandidates3_plus,
			     Record_T *records3_minus, int *candidates3_minus, int ncandidates3_minus,
			     Ladder_T ladder5_plus, Ladder_T ladder5_minus,
			     int querylength5, int querylength3) {
  bool foundp = false;

  T *hits5;
  int nhits5;
  int max_frontier_score, frontier_score, cutoff_level, level, score5;

  *plus_anchors = *minus_anchors = (Intlist_T) NULL;

  cutoff_level = Ladder_cutoff(ladder5_plus);
  if ((level = Ladder_cutoff(ladder5_minus)) > cutoff_level) {
    cutoff_level = level;
  }
  max_frontier_score = cutoff_level;

  frontier_score = score_start;
  while (frontier_score <= cutoff_level) {
    debug1(printf("frontier_score = %d\n",frontier_score));
    for (score5 = 0; score5 <= frontier_score; score5++) {
      debug1(printf("score5 = %d\n",score5));

      /* plus/plus */
      if (score5 <= Ladder_cutoff(ladder5_plus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_plus,score5);
	debug(printf("at score %d, nhits5_plus = %d\n",score5,nhits5));

	if ((*plus_anchors = do_records3_plus(*plus_anchors,hits5,nhits5,
					      records3_plus,candidates3_plus,ncandidates3_plus,
					      querylength5,querylength3)) != NULL) {
	  if (foundp == false) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	    foundp = true;
	  }
	}
      }

      /* minus/minus */
      if (score5 <= Ladder_cutoff(ladder5_minus)) {
	hits5 = Ladder_hits_for_score(&nhits5,ladder5_minus,score5);
	debug(printf("at score %d, nhits5_minus = %d\n",score5,nhits5));

	if ((*minus_anchors = do_records3_minus(*minus_anchors,hits5,nhits5,
						records3_minus,candidates3_minus,ncandidates3_minus,
						querylength5,querylength3)) != NULL) {
	  if (foundp == false) {
	    if ((cutoff_level = frontier_score + subopt_levels) > max_frontier_score) {
	      cutoff_level = max_frontier_score;
	    }
	    foundp = true;
	  }
	}
      }
    }

    frontier_score++;
  }

  return frontier_score;
}




/* Returns paths (list of Univdiag_T objects) that are concordant with
   hits (list of Stage3end_T objects) on the other side */
/* Paths have lower coordinates than the hits */
/* TODO: Consider a binary search */
List_T
Concordance_filter_paths_low (List_T *remainder, List_T pathlist, List_T hitlist, int querylength_paths, int querylength_hits) {
  List_T filtered_paths = NULL;

  List_T *paths, path;
  T *hits;
  int npaths, nhits, i, j;

  Univdiag_T diagonal;
  Univcoord_T path_start, path_end;
  int pairmax_adj;
  bool overlapp;

  debug2(printf("Entered Concordance_filter_paths_low with %d paths and %d hits\n",
		 List_length(pathlist),List_length(hitlist)));

  *remainder = (List_T) NULL;
  if (pathlist == NULL || hitlist == NULL) {
    return (List_T) NULL;

  } else {
    pairmax_adj = pairmax_linear + querylength_paths;
    debug2(printf("pairmax_adj %d\n",pairmax_adj));

    /* Sort paths and hits */
    paths = (List_T *) List_to_array_n(&npaths,pathlist);
    hits = (T *) List_to_array_n(&nhits,hitlist);
    qsort(paths,npaths,sizeof(List_T),path_ascending_cmp);
    qsort(hits,nhits,sizeof(T),hit_ascending_cmp);

    i = j = 0;
    while (i < npaths) {
      path = paths[i];
      diagonal = (Univdiag_T) List_head(path);
      path_start = diagonal->univdiagonal;
    
      diagonal = (Univdiag_T) List_last_value(path);
      path_end = diagonal->univdiagonal + querylength_paths;

#ifdef DEBUG2
      printf("i=%d/%d %u..%u",i,npaths,path_start,path_end);
      if (j >= 0 && j < nhits) {
	printf("    j=%d/%d %u..%u",j,nhits,hits[j]->low,hits[j]->high);
      }
      printf("\n");
#endif

      while (j >= 0 && hits[j]->high < path_start) {
	debug2(printf("  backup: j=%d/%d %u..%u\n",j,nhits,hits[j]->low,hits[j]->high));
	j--;
      }
      j++;		/* Finish backup */
      debug2(printf("Backed up to %d\n",j));
      /* j is the first hit where path is close to hit, unless we backed up too much */

      /* path_start <= hits[j]->high */
      while (j+1 < nhits && hits[j+1]->high >= path_start) {
	debug2(printf("  advance: j=%d/%d %u..%u\n",j+1,nhits,hits[j+1]->low,hits[j+1]->high));
	j++;
      }
      /* j is the first hit where path is close to hit */

      overlapp = false;
      /* (hits[j]->low + querylength_hits) - (path_end - querylength_paths) <= pairmax */
      /* hits[j]->low + querylength_hits <= path_end + (pairmax + querylength_paths) */
      while (j < nhits && hits[j]->low + querylength_hits <= path_end + pairmax_adj) {
	debug2(printf("  overlap: j=%d/%d %u..%u\n",j,nhits,hits[j]->low,hits[j]->high));
	overlapp = true;
	j++;
      }
      j--;		/* Finish advance */
	    
      if (overlapp == true) {
	filtered_paths = List_push(filtered_paths,(void *) path);
      } else {
	*remainder = List_push(*remainder,(void *) path);
      }

      i++;
    }

    FREE(hits);
    FREE(paths);
    return filtered_paths;
  }
}


/* Paths have higher coordinates than the hits */
/* TODO: Consider a binary search */
List_T
Concordance_filter_paths_high (List_T *remainder, List_T pathlist, List_T hitlist, int querylength_paths, int querylength_hits) {
  List_T filtered_paths = NULL;

  List_T *paths, path;
  T *hits;
  int npaths, nhits, i, j;

  Univdiag_T diagonal;
  Univcoord_T path_start, path_end;
  int pairmax;			/* pairmax_linear */
  int querylength_both;
  bool overlapp;

  debug2(printf("Entered Concordance_filter_paths_high with %d paths and %d hits\n",
		 List_length(pathlist),List_length(hitlist)));

  *remainder = (List_T) NULL;
  if (pathlist == NULL || hitlist == NULL) {
    return (List_T) NULL;

  } else {
    pairmax = pairmax_linear;
    querylength_both = querylength_paths + querylength_hits;
    debug2(printf("pairmax %d, querylength_both %d\n",pairmax,querylength_both));

    /* Sort paths and hits */
    paths = (List_T *) List_to_array_n(&npaths,pathlist);
    hits = (T *) List_to_array_n(&nhits,hitlist);
    qsort(paths,npaths,sizeof(List_T),path_ascending_cmp);
    qsort(hits,nhits,sizeof(T),hit_ascending_cmp);

    i = j = 0;
    while (i < npaths) {
      path = paths[i];
      diagonal = (Univdiag_T) List_head(path);
      path_start = diagonal->univdiagonal;
    
      diagonal = (Univdiag_T) List_last_value(path);
      path_end = diagonal->univdiagonal + querylength_paths;

#ifdef DEBUG2
      printf("i=%d/%d %u..%u",i,npaths,path_start,path_end);
      if (j >= 0 && j < nhits) {
	printf("    j=%d/%d %u..%u",j,nhits,hits[j]->low,hits[j]->high);
      }
      printf("\n");
#endif

      while (j >= 0 && hits[j]->high + pairmax > path_start + querylength_both) {
	debug2(printf("  backup: j=%d/%d %u..%u\n",j,nhits,hits[j]->low,hits[j]->high));
	j--;
      }
      j++;		/* Finish backup */
      debug2(printf("Backed up to %d\n",j));
      /* j is the first hit where path is close to hit, unless we backed up too much */

      /* (path_start + querylength_paths) - (hits[j]->high - querylength_hits) <= pairmax */
      /* path_start + (querylength_paths + querylength_hits) <= hits[j]->high + pairmax */
      while (j+1 < nhits && hits[j+1]->high + pairmax <= path_start + querylength_both) {
	debug2(printf("  advance: j=%d/%d %u..%u\n",j+1,nhits,hits[j+1]->low,hits[j+1]->high));
	j++;
      }
      /* j is the first hit where path is close to hit */

      /* path_end >= hits[j]->low */
      overlapp = false;
      while (j < nhits && hits[j]->low <= path_end) {
	debug2(printf("  overlap: j=%d/%d %u..%u\n",j,nhits,hits[j]->low,hits[j]->high));
	overlapp = true;
	j++;
      }
      j--;		/* Finish advance */
	    
      if (overlapp == true) {
	filtered_paths = List_push(filtered_paths,(void *) path);
      } else {
	*remainder = List_push(*remainder,(void *) path);
      }

      i++;
    }
    
    FREE(hits);
    FREE(paths);
    return filtered_paths;
  }
}


void
Concordance_setup (int subopt_levels_in, bool novelsplicingp_in,
		   Chrpos_T pairmax_transcriptome_in,
		   Chrpos_T pairmax_linear_in, Chrpos_T pairmax_circular_in,
		   Chrpos_T expected_pairlength_in, Chrpos_T pairlength_deviation_in, 
		   bool *circularp_in, bool merge_samechr_p_in) {
  subopt_levels = subopt_levels_in;
  novelsplicingp = novelsplicingp_in;

  pairmax_transcriptome = pairmax_transcriptome_in;
  pairmax_linear = pairmax_linear_in;
  pairmax_circular = pairmax_circular_in;

  expected_pairlength = expected_pairlength_in;
  pairlength_deviation = pairlength_deviation_in;
  adjacent_pairlength = expected_pairlength + pairlength_deviation;

  circularp = circularp_in;
  merge_samechr_p = merge_samechr_p_in;

  return;
}
