static char rcsid[] = "$Id: extension-search.c 214796 2018-04-21 00:59:35Z twu $";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifndef HAVE_MEMCPY
#define memcpy(d,s,n) bcopy((s),(d),(n))
#endif

#include "extension-search.h"

#ifdef WORDS_BIGENDIAN
#define CONVERT(x) Bigendian_convert_uint(x)
#include "bigendian.h"
#else
#define CONVERT(x) (x)
#include "littleendian.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>		/* For munmap */

#include "mem.h"
#include "bool.h"
#include "assert.h"
#include "access.h"
#include "types.h"
#include "oligo.h"

#include "list.h"
#include "maxent_hr.h"

#include "genome128_hr.h"
#include "genome128_consec.h"

#include "univdiagdef.h"
#include "univdiag.h"
#include "path-solve.h"

#if defined(HAVE_SSE2)
#include <emmintrin.h>
#endif
#ifdef HAVE_SSE4_1
#include <smmintrin.h>
#endif
#ifdef HAVE_AVX2
#include <immintrin.h>
#endif
#ifdef HAVE_AVX512
#include <immintrin.h>
#endif


/* conservative behaves better for high-quality sequences and is faster */
/* #define LIBERAL 1 */
#define LIBERAL_KMER 8

/* #define LOCALDB_EARLY 1 */

/* Some limit is needed to prevent GSNAP from running very slowly */
#define MAX_HITS_FOR_BEST_ELT 1000


/* #define CHECK_OLIGOS 1 */

#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif

/* binary_search */
#ifdef DEBUG10
#define debug10(x) x
#else
#define debug10(x)
#endif

/* Sorting of diagonals */
#ifdef DEBUG12
#define debug12(x) x
#else
#define debug12(x)
#endif

/* filter_elts */
#ifdef DEBUG13
#define debug13(x) x
#else
#define debug13(x)
#endif


static Mode_T mode;

static Genome_T genomebits;
static Genome_T genomebits_alt;

static Univ_IIT_T chromosome_iit;
static int circular_typeint;
static bool *circularp;

static Genome_T genomebits;
static Genome_T genomebits_alt;
static Indexdb_T indexdb;
static Indexdb_T indexdb2;


static char conversion_fwd[128];
static char conversion_rev[128];

static Chrpos_T overall_max_distance_genome;
static Chrpos_T overall_end_distance_genome;
static Chrpos_T shortsplicedist;
static Chrpos_T min_intronlength;
static Chrpos_T max_deletionlen;
static Chrpos_T max_insertionlen_default;
static int max_end_deletions;
static int max_middle_insertions_default;

static int index1part;
static int local1part;
static int leftreadshift;
static Oligospace_T oligobase_mask;


#ifdef LARGE_GENOMES
#define GETPOS(high,low) (((Univcoord_T) high << 32) + low)
#endif


void
Extension_search_setup (Mode_T mode_in, Univ_IIT_T chromosome_iit_in, int circular_typeint_in, bool *circularp_in,
			Genome_T genomebits_in, Genome_T genomebits_alt_in, Indexdb_T indexdb_in, Indexdb_T indexdb2_in,
			Chrpos_T shortsplicedist_in, Chrpos_T shortsplicedist_novelend,
			int min_intronlength_in, int max_deletionlength, int max_end_deletions_in,
			int max_middle_insertions_in, int max_end_insertions,
			int index1part_in, int local1part_in) {
  int i;

  mode = mode_in;

  genomebits = genomebits_in;
  genomebits_alt = genomebits_alt_in;
  indexdb = indexdb_in;
  indexdb2 = indexdb2_in;

  genomebits = genomebits_in;
  genomebits_alt = genomebits_alt_in;
  circularp = circularp_in;

  for (i = 0; i < 128; i++) {
    conversion_fwd[i] = i;
    conversion_rev[i] = i;
  }
  if (mode == STANDARD) {
    /* Don't change conversion */
  } else if (mode == CMET_STRANDED || mode == CMET_NONSTRANDED) {
    conversion_fwd['C'] = 'T';	/* CT */
    conversion_rev['G'] = 'A';	/* GA */
  } else if (mode == ATOI_STRANDED || mode == ATOI_NONSTRANDED) {
    conversion_fwd['A'] = 'G';	/* AG */
    conversion_rev['T'] = 'C';	/* TC */
  } else if (mode == TTOC_STRANDED || mode == TTOC_NONSTRANDED) {
    conversion_fwd['T'] = 'C';	/* TC */
    conversion_rev['A'] = 'G';	/* AG */
  }

  chromosome_iit = chromosome_iit_in;
  circular_typeint = circular_typeint_in;
  shortsplicedist = shortsplicedist_in;

  min_intronlength = min_intronlength_in;
  max_deletionlen = max_deletionlength;
  max_end_deletions = max_end_deletions_in;
  max_middle_insertions_default = max_middle_insertions_in;
  if (max_middle_insertions_in > max_end_insertions) {
    max_insertionlen_default = max_middle_insertions_in;
  } else {
    max_insertionlen_default = max_end_insertions;
  }

  if (shortsplicedist > max_deletionlen) {
    overall_max_distance_genome = shortsplicedist;
  } else {
    overall_max_distance_genome = max_deletionlen;
  }

  if (shortsplicedist_novelend > max_deletionlen) {
    overall_end_distance_genome = shortsplicedist_novelend;
  } else {
    overall_end_distance_genome = max_deletionlen;
  }

  index1part = index1part_in;
  local1part = local1part_in;

#ifdef HAVE_64_BIT
  leftreadshift = 64 - index1part - index1part;
  oligobase_mask = ~(~ (Oligospace_T) 0 << 2*index1part);
#else
  leftreadshift = 32 - index1part - index1part;
  oligobase_mask = ~(~ (Oligospace_T) 0 << 2*index1part);
#endif

  return;
}


/* Simplified version of Spanningelt_T */
#define T Elt_T


#if 0
/* Works only for queryfwd.  Use Elt_read functions instead */
static T
Elt_new (int querypos, int nmatches, Univcoord_T *all_diagonals, int n_all_diagonals) {
  T new = (T) MALLOC(sizeof(*new));
#ifdef DEBUG
  int i;
#endif

  new->querystart = querypos;
  new->queryend = querypos + nmatches - 1;
  new->nmatches = nmatches;
  debug(printf("Making an Elt with querystart %d, nmatches %d => queryend %d\n",
	       new->querystart,new->nmatches,new->queryend));

  new->all_diagonals = all_diagonals;
  new->n_all_diagonals = n_all_diagonals;

  new->diagonals = &(new->all_diagonals[0]);
  new->ndiagonals = n_all_diagonals;

#ifdef DEBUG
  printf("Diagonals:");
  for (i = 0; i < n_all_diagonals; i++) {
    printf(" %u",all_diagonals[i]);
  }
  printf("\n");
#endif

  new->lowi = 0;
  new->highi = n_all_diagonals;

  return new;
}
#endif


static void
Elt_free (T *old) {

  FREE((*old)->all_diagonals);
  FREE(*old);
  return;
}

void
Elt_gc (List_T *set) {
  List_T p;
  T elt;

  for (p = *set; p != NULL; p = List_next(p)) {
    elt = (T) List_head(p);
    Elt_free(&elt);
  }
  List_free(&(*set));
  return;
}



#if 0
static int
nt_querylength (char *query, int querylength) {
  int i;
  char c;

  i = 0;
  while (i < querylength && ((c = query[i]) == 'A' || c == 'C' || c == 'G' || c == 'T')) {
    i++;
  }

  return i;
}
#endif

#ifdef CHECK_OLIGOS
static Oligospace_T
nt_oligo (char *query, int indexsize) {
  Oligospace_T oligo = 0U;
  int i;

  for (i = 0; i < indexsize; i++) {
    oligo *= 4;
    
    printf("%c",query[i]);
    switch (query[i]) {
    case 'A': break;
    case 'C': oligo += 1; break;
    case 'G': oligo += 2; break;
    case 'T': oligo += 3; break;
    default:
      fprintf(stderr,"Saw N in nt_oligo\n");
      abort();
    }
  }
  printf("\n");

  return oligo;
}

#define LOW_TWO_BITS 0x3


static char *
oligo_nt (UINT4 oligo, int oligosize) {
  char *nt = MALLOC((oligosize+1)*sizeof(char));
  int i, j;
  UINT4 lowbits;

  j = oligosize-1;
  for (i = 0; i < oligosize; i++) {
    lowbits = oligo & LOW_TWO_BITS;
    switch (lowbits) {
    case RIGHT_A: nt[j] = 'A'; break;
    case RIGHT_C: nt[j] = 'C'; break;
    case RIGHT_G: nt[j] = 'G'; break;
    case RIGHT_T: nt[j] = 'T'; break;
    }
    oligo >>= 2;
    j--;
  }
  nt[oligosize] = '\0';

  return nt;
}

#endif


#if !defined(HAVE_SSE4_2)
#define count_leading_zeroes_32(diff) ((diff >> 16) ? clz_table[diff >> 16] : 16 + clz_table[diff])
#elif defined(HAVE_LZCNT)
#define count_leading_zeroes_32(diff) _lzcnt_u32(diff)
#elif defined(HAVE_BUILTIN_CLZ)
#define count_leading_zeroes_32(diff) __builtin_clz(diff)
#else
#define count_leading_zeroes_32(diff) ((diff >> 16) ? clz_table[diff >> 16] : 16 + clz_table[diff])
#endif

#if !defined(HAVE_SSE4_2)
#define count_trailing_zeroes_32(diff) mod_37_bit_position[(-diff & diff) % 37]
#elif defined(HAVE_TZCNT)
#define count_trailing_zeroes_32(diff) _tzcnt_u32(diff)
#elif defined(HAVE_BUILTIN_CTZ)
#define count_trailing_zeroes_32(diff) __builtin_ctz(diff)
#else
/* lowbit = -diff & diff */
#define count_trailing_zeroes_32(diff) mod_37_bit_position[(-diff & diff) % 37]
#endif


/* query is a substring of the original, starting with queryoffset */
/* Performs the same function as Sarray_lookup, which returns the diagonals */
T
Elt_read_queryfwd (
#ifdef LARGE_GENOMES
		   unsigned char *positions_high,
#endif
		   UINT4 *positions, int n, int diagterm, int querylength, int querystart,
		   Compress_T query_compress, int index1part, bool plusp, int genestrand) {
  T new;
  int max_nmatches;
  Univcoord_T *best_diagonals, *out, diagonal;
  int *nmatches;
  int i;
#ifdef LIBERAL
  int queryend;
  int nmismatches3;
#endif
  

  debug(printf("Got %d positions\n",n));
  if (n == 0) {
    return (T) NULL;
  } else if (querystart >= querylength - index1part) {
    return (T) NULL;
  } else {
    max_nmatches = 0;
    nmatches = (int *) MALLOC(n*sizeof(int));
    for (i = 0; i < n; i++) {
#ifdef LARGE_GENOMES
      assert(GETPOS(positions_high[i],positions[i]) >= (Univcoord_T) -diagterm);
      diagonal = GETPOS(positions_high[i],positions[i]) + diagterm;
#else
      assert(positions[i] >= (Univcoord_T) -diagterm);
      diagonal = positions[i] + diagterm;
#endif
      debug(printf("plusp %d, diagonal is %u, querystart is %d\n",plusp,diagonal,querystart));

#ifdef LIBERAL	
      /* Uses Genome_first_kmer_right, which is liberal */
      queryend = Genome_first_kmer_right(&nmismatches3,genomebits,query_compress,/*left*/diagonal,
					 /*pos5*/querystart,/*pos3*/querylength,
					 plusp,genestrand,/*query_unk_mismatch_local_p*/false,
					 /*kmer*/LIBERAL_KMER);
      if ((nmatches[i] = queryend - querystart) > max_nmatches) {
	max_nmatches = nmatches[i];
      }
#else
      /* Uses Genome_consecutive_matches_rightward, which is conservative */
      nmatches[i] = index1part +
	Genome_consecutive_matches_rightward(genomebits,query_compress,/*left*/diagonal,
					     /*pos5*/querystart+index1part,/*pos3*/querylength,
					     plusp,genestrand);
      if (nmatches[i] > max_nmatches) {
	max_nmatches = nmatches[i];
      }
#endif
    }

    out = best_diagonals = MALLOC(n*sizeof(Univcoord_T));
    for (i = 0; i < n; i++) {
      if (nmatches[i] >= max_nmatches - /*index1interval*/ 3 + 1) {
#ifdef LARGE_GENOMES
	*out++ = GETPOS(positions_high[i],positions[i]) + diagterm;
#else
	*out++ = positions[i] + diagterm;
#endif
      }
    }
    FREE(nmatches);
   

    new = (T) MALLOC(sizeof(*new));
    new->querystart = querystart;
    new->queryend = querystart + max_nmatches;
    new->nmatches = max_nmatches;

    assert(new->queryend <= (int) querylength);
    debug(printf("Making an Elt with querystart %d, nmatches %d => queryend %d\n",
		 new->querystart,new->nmatches,new->queryend));

    new->all_diagonals = new->diagonals = best_diagonals;
    new->n_all_diagonals = new->ndiagonals = out - best_diagonals;

#ifdef DEBUG
    printf("Diagonals:");
    for (i = 0; i < new->ndiagonals; i++) {
      printf(" %u",new->diagonals[i]);
    }
    printf("\n");
#endif

    new->lowi = 0;
    new->highi = new->n_all_diagonals;

    return new;
  }
}


T
Elt_read_queryrev (
#ifdef LARGE_GENOMES
		   unsigned char *positions_high,
#endif
		   UINT4 *positions, int n, int diagterm, int querylength, int queryend,
		   Compress_T query_compress, int index1part, bool plusp, int genestrand) {
  T new;
  int max_nmatches;
  Univcoord_T *best_diagonals, *out, diagonal;
  int *nmatches;
  int i;
#ifdef LIBERAL
  int querystart;
  int nmismatches5;
#endif
  

  debug(printf("Got %d diagonals\n",n));
  if (n == 0) {
    return (T) NULL;
  } else if (queryend < index1part) {
    return (T) NULL;
  } else {
    max_nmatches = 0;
    nmatches = (int *) MALLOC(n*sizeof(int));
    for (i = 0; i < n; i++) {
#ifdef LARGE_GENOMES
      assert(GETPOS(positions_high[i],positions[i]) >= (Univcoord_T) -diagterm);
      diagonal = GETPOS(positions_high[i],positions[i]) + diagterm;
#else
      assert(positions[i] >= (Univcoord_T) -diagterm);
      diagonal = positions[i] + diagterm;
#endif
      debug(printf("plusp %d, diagonal is %u, queryend is %d\n",plusp,diagonal,queryend));

#ifdef LIBERAL	
      /* Uses Genome_first_kmer_left, which is liberal */
      querystart = Genome_first_kmer_left(&nmismatches5,genomebits,query_compress,/*left*/diagonal,
					  /*pos5*/0,/*pos3*/queryend,
					  plusp,genestrand,/*query_unk_mismatch_local_p*/false,
					  /*kmer*/LIBERAL_KMER);
      if ((nmatches[i] = queryend - querystart) > max_nmatches) {
	max_nmatches = nmatches[i];
      }
#else
      /* Uses Genome_consecutive_matches_rightward, which is conservative */
      nmatches[i] = index1part +
	Genome_consecutive_matches_leftward(genomebits,query_compress,/*left*/diagonal,
					    /*pos5*/0,/*pos3*/queryend - index1part,plusp,genestrand);
      if (nmatches[i] > max_nmatches) {
	max_nmatches = nmatches[i];
      }
#endif
    }

    out = best_diagonals = MALLOC(n*sizeof(Univcoord_T));
    for (i = 0; i < n; i++) {
      if (nmatches[i] >= max_nmatches - /*index1interval*/ 3 + 1) {
#ifdef LARGE_GENOMES
	*out++ = GETPOS(positions_high[i],positions[i]) + diagterm;
#else
	*out++ = positions[i] + diagterm;
#endif
      }
    }
    FREE(nmatches);
   

    new = (T) MALLOC(sizeof(*new));
    new->queryend = queryend;
    new->querystart = queryend - max_nmatches;
    new->nmatches = max_nmatches;

    assert(new->querystart >= 0);
    debug(printf("Making an Elt with queryend %d, nmatches %d => querystart %d\n",
		 new->queryend,new->nmatches,new->querystart));

    new->all_diagonals = new->diagonals = best_diagonals;
    new->n_all_diagonals = new->ndiagonals = out - best_diagonals;

#ifdef DEBUG
    printf("Diagonals:");
    for (i = 0; i < new->ndiagonals; i++) {
      printf(" %u",new->diagonals[i]);
    }
    printf("\n");
#endif

    new->lowi = 0;
    new->highi = new->n_all_diagonals;

    return new;
  }
}



static int
binary_search_univcoord (int lowi, int highi, Univcoord_T *diagonals, Univcoord_T goal) {
  int middlei;

  debug10(printf("entered binary search with lowi=%d, highi=%d, goal=%u\n",lowi,highi,goal));

  while (lowi < highi) {
    middlei = lowi + ((highi - lowi) / 2);
    debug10(printf("  binary: %d:%u %d:%u %d:%u   vs. %u\n",
		   lowi,diagonals[lowi],middlei,diagonals[middlei],
		   highi-1,diagonals[highi-1],goal));
    if (goal < diagonals[middlei]) {
      highi = middlei;
    } else if (goal > diagonals[middlei]) {
      lowi = middlei + 1;
    } else {
      debug10(printf("binary search returns %d\n",middlei));
      return middlei;
    }
  }

  debug10(printf("binary search returns %d\n",highi));
  return highi;
}


static void
Elt_filter_diagonals (T this, Univcoord_T low, Univcoord_T high) {
  int lowi, highi;

  debug(printf("Entered Elt_filter_diagonals on %d..%d with %d diagonals with low %u and high %u, nmatches %d\n",
	       this->querystart,this->queryend,this->n_all_diagonals,low,high,this->nmatches));

  /* low_adj and high_adj are inclusive */
  lowi = binary_search_univcoord(/*lowi*/0,/*highi*/this->n_all_diagonals,this->all_diagonals,/*goal*/low);
  highi = binary_search_univcoord(lowi,/*highi*/this->n_all_diagonals,this->all_diagonals,/*goal*/high + 1) - 1;
  if ((this->ndiagonals = highi - lowi + 1) == 0) {
    this->diagonals = (Univcoord_T *) NULL;

  } else {
    this->diagonals = &(this->all_diagonals[lowi]);
  }

  return;
}


#ifdef DEBUG
static void
Elt_dump (T elt) {
  int k;

  printf("Elt %d..%d with %d diagonals:\n",elt->querystart,elt->queryend,elt->ndiagonals);
  for (k = 0; k < elt->ndiagonals; k++) {
    printf("  %u\n",elt->diagonals[k]);
  }
  printf("\n");

  return;
}
#endif


#define add_bounded(x,plusterm,highbound) ((x + (plusterm) >= highbound) ? (highbound - 1) : x + (plusterm))
#define subtract_bounded(x,minusterm,lowbound) ((x < lowbound + (minusterm)) ? lowbound : x - (minusterm))


static void
filter_elts (T *elt_array, int best_i, int nelts) {
  T elt, middle;
  int i;

  middle = elt_array[best_i];

  debug13(printf("Best elt is %d..%d with %d diagonals\n",middle->querystart,middle->queryend,middle->ndiagonals));

  /* Filter elts and diagonals for right side */
  debug13(printf("Filtering elts on right side\n"));
  for (i = best_i + 1; i < nelts; i++) {
    elt = elt_array[i];
    debug13(printf("Handling right elt %d: %p\n",i,elt));
    if (elt->querystart >= middle->querystart && elt->queryend <= middle->queryend) {
      debug13(printf("Not allowing right elt that is subsumed by middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    } else if (elt->querystart <= middle->querystart) {
      debug13(printf("Not allowing right elt that extends left of middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    } else if ((middle->queryend - elt->querystart) > (middle->queryend - middle->querystart) / 2) {
      debug13(printf("Not allowing right elt that mainly overlaps middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    } else if ((middle->queryend - elt->querystart) > (elt->queryend - elt->querystart) / 2) {
      debug13(printf("Not allowing right elt that mainly overlaps middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    }
  }

  /* Filter elts and diagonals for left side */
  debug13(printf("Filtering elts on left side\n"));
  for (i = best_i - 1; i >= 0; --i) {
    elt = elt_array[i];
    debug13(printf("Handling left elt %d: %p\n",i,elt));
    if (elt->querystart >= middle->querystart && elt->queryend <= middle->queryend) {
      debug13(printf("Not allowing left elt that is subsumed by middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    } else if (elt->queryend >= middle->queryend) {
      debug13(printf("Not allowing left elt that extends right of middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    } else if ((elt->queryend - middle->querystart) > (middle->queryend - middle->querystart) / 2) {
      debug13(printf("Not allowing left elt that mainly overlaps middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    } else if ((elt->queryend - middle->querystart) > (elt->queryend - elt->querystart) / 2) {
      debug13(printf("Not allowing left elt that mainly overlaps middle elt: query %d..%d\n",
		     elt->querystart,elt->queryend));
      /* Elt_free(&elt); */
      elt_array[i] = (Elt_T) NULL;
    }
  }

  return;
}


/* Currently returning a score, but this is not used by the caller */
static int
get_diagonals (Univdiag_T *middle_diagonal, List_T *best_right_diagonals, List_T *best_left_diagonals, 
	       List_T *all_right_diagonals, List_T *all_left_diagonals, int querylength,
	       Univcoord_T chroffset, Univcoord_T chrhigh, Univcoord_T goal,
	       T *elt_array, int best_i, int nelts) {
  int best_score_right, best_score_left, best_score, score;
  T elt;
  Univcoord_T low, high, left_position, right_position;

  int min_querystart, max_queryend; /* Checked using localdb */
  Univcoord_T *diagonals;
  int ndiagonals;

  int i, j, k;
  List_T p;

  Univdiag_T *diagonal_array, diagonal, prev_diagonal;
  int querypos;
  List_T left_diagonals, right_diagonals;

  int max_insertionlen;

  if (max_middle_insertions_default >= 0) {
    max_insertionlen = max_insertionlen_default;
  } else {
    max_insertionlen = querylength;
  }

  debug13(printf("\n***Entered get_diagonals with goal %u, chroffset %u, chrhigh %u\n",goal,chroffset,chrhigh));

  /* Create diagonals.  We give a bonus of +1 for being on the same
     diagonal.  This means that we should count consecutive regions
     within each diagonal as 2 points.  Then an indel or gap will
     give only 1 point, or a relative penalty. */

  elt = elt_array[best_i];
  left_position = goal + elt->querystart;
  right_position = goal + elt->queryend;
  min_querystart = elt->querystart;
  max_queryend = elt->queryend;

  *middle_diagonal = Univdiag_new(elt->querystart,elt->queryend,/*univdiagonal*/goal);
  (*middle_diagonal)->intscore = 2*(elt->queryend - elt->querystart + 1);
  debug13(printf("Creating middle diagonal: query %d..%d, diagonal %u = goal %u - chroffset %u\n",
		 elt->querystart,elt->queryend,goal - chroffset,goal,chroffset));


  /* Filter diagonals for right side */
  right_diagonals = (List_T) NULL;

  debug13(printf("Filtering elts and diagonals for right side.  Subtracting %d and adding %d from goal %u\n",
		 max_insertionlen,overall_end_distance_genome,goal));
  low = subtract_bounded(goal,/*minusterm*/max_insertionlen,chroffset);
  high = add_bounded(goal,/*plusterm*/overall_end_distance_genome,chrhigh);

  for (i = best_i + 1; i < nelts; i++) {
    if ((elt = elt_array[i]) != NULL) {
      debug13(printf("Handling right elt %d: %p\n",i,elt));
      Elt_filter_diagonals(elt,low,high);
      if (elt->ndiagonals > 0) {
	diagonals = elt->diagonals;
	ndiagonals = elt->ndiagonals;
      } else {
	debug13(printf("Elt %d..%d has no diagonals\n",elt->querystart,elt->queryend));
	/* ? TODO: Restrict so we search only short elts */
#ifdef LOCALDB_EARLY
	diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,/*pos5*/elt->querystart,/*pos3*/elt->queryend,
					  low,high);
	if (elt->queryend > max_queryend) {
	  /* To prevent us from trying again */
	  max_queryend = elt->queryend;
	}
#else
	ndiagonals = 0;
#endif
      }

      /* successp = false; */
      for (j = 0; j < ndiagonals; j++) {
	debug13(printf("position %d is %u\n",j,diagonals[j]));
	if (diagonals[j] + elt->queryend <= right_position) {
	  debug13(printf("Not creating right diagonal that does not advance past right_position %u: query %d..%d, diagonal %u\n",
			 right_position - chroffset,elt->querystart,elt->queryend,diagonals[j] - chroffset));
	} else {
	  debug13(printf("Creating right diagonal: query %d..%d, diagonal %u\n",
			 elt->querystart,elt->queryend,diagonals[j] - chroffset));
	  right_diagonals = List_push(right_diagonals,
				      (void *) Univdiag_new(elt->querystart,elt->queryend,/*univdiagonal*/diagonals[j]));
	  /* In resetting right_position, we are assuming that this diagonal will be used */
	  /* right_position = diagonals[j] + elt->queryend; */
	  if (elt->queryend > max_queryend) {
	    /* In case we succeeded with the original elt diagonals */
	    max_queryend = elt->queryend;
	  }
	  /* successp = true; */
	}
      }

#if 0
      /* In resetting low and high, we are assuming that this diagonal will be used */
      if (successp == true) {
	/* Success: Update low and high for next search */
	low = subtract_bounded(diagonals[0],/*minusterm*/max_insertionlen,chroffset);
	high = add_bounded(diagonals[ndiagonals-1],/*plusterm*/overall_end_distance_genome,chrhigh);
	debug13(printf("Updated low and high based on %u and %u\n",diagonals[0],diagonals[ndiagonals-1]));
      }
#endif

#ifdef LOCALDB_EARLY
      if (elt->ndiagonals == 0) {
	FREE(diagonals);
      }
#endif
    }
  }

  debug13(printf("max_queryend = %d\n",max_queryend));
#ifdef LOCALDB_EARLY
  if (querylength - max_queryend >= local1part) {
    /* Try to complete the right end */
    diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,/*pos5*/max_queryend,/*pos3*/querylength,
				      low,high);
    for (j = 0; j < ndiagonals; j++) {
      debug13(printf("position %d is %u\n",j,diagonals[j]));
      if (diagonals[j] + /*elt->queryend*/querylength <= right_position) {
	debug13(printf("Not creating right diagonal that does not advance past right_position %u: query %d..%d, diagonal %u\n",
		       right_position - chroffset,max_queryend,querylength,diagonals[j] - chroffset));
      } else {
	right_diagonals = List_push(right_diagonals,
				    (void *) Univdiag_new(max_queryend,querylength,/*univdiagonal*/diagonals[j]));
      }
    }
    FREE(diagonals);
  }
#endif

  right_diagonals = List_reverse(right_diagonals);



  /* Filter diagonals for left side */
  left_diagonals = (List_T) NULL;

  debug13(printf("Filtering elts and diagonals for left side.  Subtracting %d and adding %d from goal %u\n",
		 overall_end_distance_genome,max_insertionlen,goal));
  low = subtract_bounded(goal,/*minusterm*/overall_end_distance_genome,chroffset);
  high = add_bounded(goal,/*plusterm*/max_insertionlen,chrhigh);

  for (i = best_i - 1; i >= 0; --i) {
    if ((elt = elt_array[i]) != NULL) {
      debug13(printf("Handling left elt %d: %p\n",i,elt));
      Elt_filter_diagonals(elt,low,high);
      if (elt->ndiagonals > 0) {
	diagonals = elt->diagonals;
	ndiagonals = elt->ndiagonals;
      } else {
	debug13(printf("Elt %d..%d has no diagonals\n",elt->querystart,elt->queryend));
	/* ? TODO: Restrict so we search only short elts */
#ifdef LOCALDB_EARLY
	diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,/*pos5*/elt->querystart,/*pos3*/elt->queryend,
					  low,high);
	if (elt->querystart < min_querystart) {
	  /* To prevent us from trying again */
	  min_querystart = elt->querystart;
	}
#else
	ndiagonals = 0;
#endif
      }

      /* successp = false; */
      for (j = ndiagonals - 1; j >= 0; --j) {
	debug13(printf("position %d is %u\n",j,diagonals[j]));
	if (diagonals[j] + elt->querystart >= left_position) {
	  debug13(printf("Not creating left diagonal that does not advance past left_position %u: query %d..%d, diagonal %u\n",
			 left_position - chroffset,elt->querystart,elt->queryend,diagonals[j] - chroffset));
	} else {
	  debug13(printf("Creating left diagonal: query %d..%d, diagonal %u\n",
			 elt->querystart,elt->queryend,diagonals[j] - chroffset));
	  left_diagonals = List_push(left_diagonals,
				     (void *) Univdiag_new(elt->querystart,elt->queryend,/*univdiagonal*/diagonals[j]));
	  /* In resetting left_position, we are assuming that this diagonal will be used */
	  /* left_position = diagonals[j] + elt->querystart; */
	  if (elt->querystart < min_querystart) {
	    /* In case we succeeded with the original elt diagonals */
	    min_querystart = elt->querystart;
	  }
	  /* successp = true; */
	}
      }

#if 0
      /* In resetting right_position, we are assuming that this diagonal will be used */
      if (successp == true) {
	/* Success: Update low and high for next search */
	low = subtract_bounded(diagonals[0],/*minusterm*/overall_end_distance_genome,chroffset);
	high = add_bounded(diagonals[ndiagonals-1],/*plusterm*/max_insertionlen,chrhigh);
	debug13(printf("Updated low and high based on %u and %u\n",diagonals[0],diagonals[ndiagonals-1]));
      }
#endif

#ifdef LOCALDB_EARLY
      if (elt->ndiagonals == 0) {
	FREE(diagonals);
      }
#endif
    }
  }

  debug13(printf("min_querystart = %d\n",min_querystart));
#ifdef LOCALDB_EARLY
  if (min_querystart >= local1part) {
    /* Try to complete the left end */
    diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,/*pos5*/0,/*pos3*/min_querystart,
				      low,high);
    for (j = ndiagonals - 1; j >= 0; --j) {
      debug13(printf("position %d is %u\n",j,diagonals[j]));
      if (diagonals[j] /* + querystart (0) */ >= left_position) {
	debug13(printf("Not creating left diagonal that does not advance past left_position %u: query %d..%d, diagonal %u\n",
		       left_position - chroffset,0,min_querystart,diagonals[j] - chroffset));
      } else {
	left_diagonals = List_push(left_diagonals,
				   (void *) Univdiag_new(0,min_querystart,/*univdiagonal*/diagonals[j]));
      }
    }
    FREE(diagonals);
  }
#endif

  left_diagonals = List_reverse(left_diagonals);



  /* A.  Compute right diagonals */
  /* A1.  Scoring for dynamic programming */
  diagonal_array = (Univdiag_T *) List_to_array_n(&ndiagonals,right_diagonals);
  List_free(&right_diagonals);
#ifdef DEBUG12
  printf("Right side before consolidating\n");
  for (i = 0; i < ndiagonals; i++) {
    diagonal = diagonal_array[i];
    printf("%d..%d at %u\n",diagonal->querystart,diagonal->queryend,diagonal->univdiagonal);
  }
#endif

  *all_right_diagonals = (List_T) NULL;
  qsort(diagonal_array,ndiagonals,sizeof(Univdiag_T),Univdiag_diagonal_cmp);
  i = 0;
  while (i < ndiagonals) {
    j = i;
    while (j < ndiagonals && diagonal_array[j]->univdiagonal == diagonal_array[i]->univdiagonal) {
      j++;
    }
    if (j == i) {
      *all_right_diagonals = List_push(*all_right_diagonals,(void *) diagonal_array[i]);
    } else {
      *all_right_diagonals = List_push(*all_right_diagonals,
				       (void *) Univdiag_new(diagonal_array[i]->querystart,
							     diagonal_array[j-1]->queryend,
							     diagonal_array[i]->univdiagonal));
      for (k = i; k < j; k++) {
	Univdiag_free(&(diagonal_array[k]));
      }
    }
    i = j;
  }
  FREE(diagonal_array);

  /* TODO: May be able to skip this sorting step */
  diagonal_array = (Univdiag_T *) List_to_array_n(&ndiagonals,*all_right_diagonals);
  qsort(diagonal_array,ndiagonals,sizeof(Univdiag_T),Univdiag_ascending_cmp);
#ifdef DEBUG12
  printf("Right side after consolidating and sorting\n");
  for (i = 0; i < ndiagonals; i++) {
    diagonal = diagonal_array[i];
    printf("%d..%d at %u\n",diagonal->querystart,diagonal->queryend,diagonal->univdiagonal);
  }
#endif


  debug13(printf("%d diagonals:\n",ndiagonals));
  for (i = 0; i < ndiagonals; i++) {
    diagonal = diagonal_array[i];
    debug13(printf("%d: %d..%d at %u\n",i,diagonal->querystart,diagonal->queryend,diagonal->univdiagonal));

    low = subtract_bounded(diagonal->univdiagonal,overall_max_distance_genome,chroffset);
    high = add_bounded(diagonal->univdiagonal,max_insertionlen,chrhigh);
    querypos = diagonal->querystart;
    best_score = 0;

    for (j = i - 1; j >= 0; --j) {
      prev_diagonal = diagonal_array[j];
      debug13(printf("  %d: %d..%d at %u  ",j,prev_diagonal->querystart,prev_diagonal->queryend,prev_diagonal->univdiagonal));

      if (prev_diagonal->queryend >= querypos) {
	debug13(printf("Skipping because queryend %d >= querypos %d\n",prev_diagonal->queryend,querypos));
      } else if (prev_diagonal->univdiagonal < low) {
	debug13(printf("Skipping because diagonal %u < low_chrpos %u\n",prev_diagonal->univdiagonal,low));
      } else if (prev_diagonal->univdiagonal > high) {
	debug13(printf("Skipping because diagonal %u > high_chrpos %u\n",prev_diagonal->univdiagonal,high));
      } else {
	score = prev_diagonal->intscore;
	if (prev_diagonal->univdiagonal == diagonal->univdiagonal) {
	  score += 1;
	}
	if (score <= best_score) {
	  debug13(printf("Skipping because score %d <= best_score %d\n",score,best_score));
	} else {
	  best_score = score;
	  diagonal->prev = prev_diagonal;
	  debug13(printf("Updating best score to be %d.  Prev diagonal is %d..%d at %u\n",
			 best_score,prev_diagonal->querystart,prev_diagonal->queryend,prev_diagonal->univdiagonal));
	}
      }
    }

    /* Handle links to middle diagonal */
    prev_diagonal = *middle_diagonal;
    debug13(printf("  Middle: %d..%d at %u  ",prev_diagonal->querystart,prev_diagonal->queryend,prev_diagonal->univdiagonal));
    if (prev_diagonal->queryend >= querypos) {
      debug13(printf("Skipping because queryend %d >= querypos %d\n",prev_diagonal->queryend,querypos));
    } else if (prev_diagonal->univdiagonal < low) {
      debug13(printf("Skipping because diagonal %u < low_chrpos %u\n",prev_diagonal->univdiagonal,low));
    } else if (prev_diagonal->univdiagonal > high) {
      debug13(printf("Skipping because diagonal %u > high_chrpos %u\n",prev_diagonal->univdiagonal,high));
    } else {
      score = prev_diagonal->intscore;
      if (prev_diagonal->univdiagonal == diagonal->univdiagonal) {
	score += 1;		/* This bonus means we should double count contiguous region within each segment */
      }
      if (score <= best_score) {
	debug13(printf("Skipping because score %d <= best_score %d\n",score,best_score));
      } else {
	best_score = score;
	/* diagonal->prev = (Univdiag_T) NULL; */
	debug13(printf("Updating best score (for link to middle diagonal) to be %d\n",best_score));
      }
    }

    diagonal->intscore = best_score + 2*diagonal->nconsecutive;
    debug13(printf("Right diagonal %d..%d at %u gets score %d\n",
		   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,diagonal->intscore));
  }
  FREE(diagonal_array);


  /* A2.  Optimizing for dynamic programming */
  best_score_right = 0;
  *best_right_diagonals = (List_T) NULL;
  for (p = *all_right_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    debug13(printf("Considering right diagonal %d..%d at %u (vs middle diagonal %u)\n",
		   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,(*middle_diagonal)->univdiagonal));
    if (diagonal->intscore > best_score_right) {
      best_score_right = diagonal->intscore;
      List_free(&(*best_right_diagonals));
      *best_right_diagonals = List_push(NULL,(void *) diagonal);
    } else if (diagonal->intscore == best_score_right) {
      *best_right_diagonals = List_push(*best_right_diagonals,(void *) diagonal);
    }
  }


  /* C.  Compute left diagonals */
  /* C1.  Scoring for dynamic programming */
  diagonal_array = (Univdiag_T *) List_to_array_n(&ndiagonals,left_diagonals);
  List_free(&left_diagonals);
#ifdef DEBUG12
  printf("Left side before consolidating\n");
  for (i = 0; i < ndiagonals; i++) {
    diagonal = diagonal_array[i];
    printf("%d..%d at %u\n",diagonal->querystart,diagonal->queryend,diagonal->univdiagonal);
  }
#endif

  *all_left_diagonals = (List_T) NULL;
  qsort(diagonal_array,ndiagonals,sizeof(Univdiag_T),Univdiag_diagonal_cmp);
  i = 0;
  while (i < ndiagonals) {
    j = i;
    while (j < ndiagonals && diagonal_array[j]->univdiagonal == diagonal_array[i]->univdiagonal) {
      j++;
    }
    if (j == i) {
      *all_left_diagonals = List_push(*all_left_diagonals,(void *) diagonal_array[i]);
    } else {
      *all_left_diagonals = List_push(*all_left_diagonals,
				      (void *) Univdiag_new(diagonal_array[i]->querystart,
							    diagonal_array[j-1]->queryend,
							    diagonal_array[i]->univdiagonal));
      for (k = i; k < j; k++) {
	Univdiag_free(&(diagonal_array[k]));
      }
    }
    i = j;
  }
  FREE(diagonal_array);

  /* TODO: May be able to skip this sorting step */
  diagonal_array = (Univdiag_T *) List_to_array_n(&ndiagonals,*all_left_diagonals);
  qsort(diagonal_array,ndiagonals,sizeof(Univdiag_T),Univdiag_descending_cmp);
#ifdef DEBUG12
  printf("Left side after consolidating and sorting\n");
  for (i = 0; i < ndiagonals; i++) {
    diagonal = diagonal_array[i];
    printf("%d..%d at %u\n",diagonal->querystart,diagonal->queryend,diagonal->univdiagonal);
  }
#endif


  debug13(printf("%d diagonals:\n",ndiagonals));
  for (i = 0; i < ndiagonals; i++) {
    diagonal = diagonal_array[i];
    debug13(printf("%d: %d..%d at %u\n",i,diagonal->querystart,diagonal->queryend,diagonal->univdiagonal));

    low = subtract_bounded(diagonal->univdiagonal,max_insertionlen,chroffset);
    high = add_bounded(diagonal->univdiagonal,overall_max_distance_genome,chrhigh);
    querypos = diagonal->queryend;
    best_score = 0;

    for (j = i - 1; j >= 0; --j) {
      prev_diagonal = diagonal_array[j];
      debug13(printf("  %d: %d..%d at %u  ",j,prev_diagonal->querystart,prev_diagonal->queryend,prev_diagonal->univdiagonal));

      if (prev_diagonal->querystart <= querypos) {
	debug13(printf("Skipping because querystart %d <= querypos %d\n",prev_diagonal->querystart,querypos));
      } else if (prev_diagonal->univdiagonal < low) {
	debug13(printf("Skipping because diagonal %u < low %u\n",prev_diagonal->univdiagonal,low));
      } else if (prev_diagonal->univdiagonal > high) {
	debug13(printf("Skipping because diagonal %u > high %u\n",prev_diagonal->univdiagonal,high));
      } else {
	score = prev_diagonal->intscore;
	if (prev_diagonal->univdiagonal == diagonal->univdiagonal) {
	  score += 1;
	}
	if (score <= best_score) {
	  debug13(printf("Skipping because score %d <= best_score %d\n",score,best_score));
	} else {
	  best_score = score;
	  diagonal->prev = prev_diagonal;
	  debug13(printf("Updating best score to be %d.  Prev diagonal is %d..%d at %u\n",
			 best_score,prev_diagonal->querystart,prev_diagonal->queryend,prev_diagonal->univdiagonal));
	}
      }
    }

    /* Handle links to middle diagonal */
    prev_diagonal = *middle_diagonal;
    debug13(printf("  Middle: %d..%d at %u  ",prev_diagonal->querystart,prev_diagonal->queryend,prev_diagonal->univdiagonal));
    if (prev_diagonal->querystart <= querypos) {
      debug13(printf("Skipping because querystart %d <= querypos %d\n",prev_diagonal->querystart,querypos));
    } else if (prev_diagonal->univdiagonal < low) {
      debug13(printf("Skipping because diagonal %u < low_chrpos %u\n",prev_diagonal->univdiagonal,low));
    } else if (prev_diagonal->univdiagonal > high) {
      debug13(printf("Skipping because diagonal %u > high_chrpos %u\n",prev_diagonal->univdiagonal,high));
    } else {
      score = prev_diagonal->intscore;
      if (prev_diagonal->univdiagonal == diagonal->univdiagonal) {
	score += 1;		/* This bonus means we should double count contiguous region within each segment */
      }
      if (score <= best_score) {
	debug13(printf("Skipping because score %d <= best_score %d\n",prev_diagonal->intscore,best_score));
      } else {
	best_score = score;
	/* diagonal->prev = (Univdiag_T) NULL; */
	debug13(printf("Updating best score (for link to middle diagonal) to be %d\n",best_score));
      }
    }

    diagonal->intscore = best_score + 2*diagonal->nconsecutive;
    debug13(printf("Left diagonal %d..%d at %u gets score %d\n",
		   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,diagonal->intscore));
  }
  FREE(diagonal_array);


  /* C2.  Optimizing for dynamic programming */
  best_score_left = 0;
  *best_left_diagonals = (List_T) NULL;
  for (p = *all_left_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (diagonal->intscore > best_score_left) {
      best_score_left = diagonal->intscore;
      List_free(&(*best_left_diagonals));
      *best_left_diagonals = List_push(NULL,(void *) diagonal);
    } else if (diagonal->intscore == best_score_left) {
      *best_left_diagonals = List_push(*best_left_diagonals,(void *) diagonal);
    }
  }

#if 0
  printf("Best on the left\n");
  for (p = *best_left_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("Score %d: %d..%d at %u\n",diagonal->intscore,diagonal->querystart,diagonal->queryend,diagonal->diagonal);
  }
#endif


  if (best_score_left == 0 && best_score_right == 0) {
    return (*middle_diagonal)->intscore;
  } else if (best_score_left == 0) {
    return best_score_right;
  } else if (best_score_right == 0) {
    return best_score_left;
  } else {
    /* middle_diagonal score is double counted */
    return best_score_left + best_score_right - (*middle_diagonal)->intscore;
  }
}


#if 0
static void
diagonals_dump (Univcoord_T *diagonals, int ndiagonals) {
  int i;

  for (i = 0; i < ndiagonals; i++) {
    printf(" %u",diagonals[i]);
  }
  printf("\n");

  return;
}
#endif


static bool
Extension_search_queryfwd (int *found_score, List_T *hits_gplus, List_T *hits_gminus,

			   /* Used by Distant_rna_solve */
			   List_T *plus_set, List_T *minus_set,

			   List_T *plus_complete_paths, List_T *minus_complete_paths,
			   Stage1_T stage1, char *queryuc_ptr, char *queryrc, int querylength,
			   Compress_T query_compress_fwd, Compress_T query_compress_rev, 
			   int nmisses_allowed, int genestrand,
			   bool paired_end_p, int level) {
  bool successp = false;
  List_T complete_path;

  T best_plus_elt, best_minus_elt, elt, *plus_elt_array, *minus_elt_array;
  int best_plus_nmatches, best_minus_nmatches;
  bool foundp;
  int max_plus_querypos, max_minus_querypos, plus_querypos[3], minus_querypos[3];
  int i;
  Chrnum_T chrnum;
  Univcoord_T chroffset, chrhigh, left;
  Chrpos_T chrlength;

  int nseeds_plus, nseeds_minus;
  int mod;
  int niter_plus, niter_minus, best_plus_i, best_minus_i, nplus, nminus;
  Univdiag_T *middle_diagonals_plus = NULL, *middle_diagonals_minus = NULL;
  List_T *best_right_diagonals_plus = NULL, *best_left_diagonals_plus = NULL,
    *all_right_diagonals_plus = NULL, *all_left_diagonals_plus = NULL,
    *best_right_diagonals_minus = NULL, *best_left_diagonals_minus = NULL,
    *all_right_diagonals_minus = NULL, *all_left_diagonals_minus = NULL;

  Univdiag_T diagonal;
#ifdef LARGE_GENOMES
  unsigned char *positions_high;
#endif
  UINT4 *positions;
  int npositions;

  int query_lastpos = querylength - index1part;
  int queryoffset;
#if 0
  Reader_T reader_plus, reader_minus;
  Oligostate_T last_state_plus, last_state_minus;
  Oligospace_T oligo, oligo2;
  Oligospace_T forward_plus, revcomp_plus, forward_minus, revcomp_minus, forward_oligo, revcomp_oligo;
#endif
#ifdef DEBUG
  List_T p;
#endif


  debug(printf("\nStarting Extension_search_queryfwd with querylength %d and nmisses_allowed %d, genestrand %d\n",
	       querylength,nmisses_allowed,genestrand));

#if 0
  /* Allow calls to Extension_search_queryfwd in addition to other methods */
  *hits_gplus = *hits_gminus = (List_T) NULL;
#endif

  if (nmisses_allowed < 0) {
    nmisses_allowed = 0;
#if 0
  } else {
    /* It is possible that this makes GSNAP too slow */
    nmisses_allowed = querylength;
#endif
  }

#if 0
  if (genestrand == +2) {
    plus_conversion = conversion_rev;
    minus_conversion = conversion_fwd;
  } else {
    plus_conversion = conversion_fwd;
    minus_conversion = conversion_rev;
  }
#endif

#if 0
  reader_plus = Reader_new(queryuc_ptr,/*querystart*/0,/*queryend*/querylength);
  last_state_plus = INIT;
  forward_plus = revcomp_plus = 0;

  reader_minus = Reader_new(queryrc,/*querystart*/0,/*queryend*/querylength);
  last_state_minus = INIT;
  forward_minus = revcomp_minus = 0;
#endif

  /* I.  Race from plus and minus start to end */
  *plus_set = *minus_set = (List_T) NULL;
  best_plus_nmatches = best_minus_nmatches = 0;
  best_plus_elt = best_minus_elt = (T) NULL;
  max_plus_querypos = max_minus_querypos = 0;
  plus_querypos[0] = minus_querypos[0] = 0;
  plus_querypos[1] = minus_querypos[1] = 1;
  plus_querypos[2] = minus_querypos[2] = 2;
  niter_plus = niter_minus = 0;

  while (niter_plus <= nmisses_allowed && niter_minus <= nmisses_allowed &&
	 max_plus_querypos < query_lastpos && max_minus_querypos < query_lastpos) {
    for (mod = 0; mod < 3; mod++) {
      if ((queryoffset = plus_querypos[mod]) >= query_lastpos) {
	/* Skip */
	plus_querypos[mod] += 3;
      } else if (stage1->plus_validp[queryoffset] == false) {
	/* Skip */
	plus_querypos[mod] += 3;
      } else {
	if (stage1->plus_retrievedp[queryoffset] == true) {
#ifdef LARGE_GENOMES
	  positions_high = stage1->plus_positions_high[queryoffset];
#endif
	  positions = stage1->plus_positions[queryoffset];
	  npositions = stage1->plus_npositions[queryoffset];
	  debug(printf("Already have %d plus positions at %d\n",npositions,queryoffset));
	} else {
	  /* These should be lefts */
	  assert(stage1->plus_positions[queryoffset] == NULL);
	  stage1->plus_retrievedp[queryoffset] = true;
#ifdef LARGE_GENOMES
	  npositions = stage1->plus_npositions[queryoffset] =
	    Indexdb_largeptr_with_diagterm(&stage1->plus_positions_high[queryoffset],&stage1->plus_positions[queryoffset],
					   indexdb,stage1->plus_oligos[queryoffset],/*diagterm*/-queryoffset);
	  positions_high = stage1->plus_positions_high[queryoffset];
#else
	  npositions = stage1->plus_npositions[queryoffset] =
	    Indexdb_ptr_with_diagterm(&stage1->plus_positions[queryoffset],indexdb,
				      stage1->plus_oligos[queryoffset],/*diagterm*/-queryoffset);
#endif
	  positions = stage1->plus_positions[queryoffset];
	}

	if ((elt = Elt_read_queryfwd(
#ifdef LARGE_GENOMES
				     positions_high,
#endif
				     positions,npositions,/*diagterm*/-queryoffset,
				     querylength,/*querystart*/queryoffset,
				     query_compress_fwd,index1part,/*plusp*/true,
				     genestrand)) == NULL) {
	  plus_querypos[mod] += 3;
	} else {
	  if (elt->nmatches > best_plus_nmatches && elt->n_all_diagonals <= MAX_HITS_FOR_BEST_ELT) {
	    best_plus_elt = elt;
	    best_plus_nmatches = elt->nmatches;
	    best_plus_i = niter_plus;
	  }
	  *plus_set = List_push(*plus_set,elt);
	  plus_querypos[mod] += elt->nmatches;
	  niter_plus++;
	}
      }

      if (plus_querypos[mod] > max_plus_querypos) {
	max_plus_querypos = plus_querypos[mod];
      }


      /* querypos_rc uses standard Stage1_T convention, but we have switched to sarray convention */
      /* querypos_rc = query_lastpos - queryoffset; */
      if ((queryoffset = minus_querypos[mod]) >= query_lastpos) {
	/* Skip */
	minus_querypos[mod] += 3;
      } else if (stage1->minus_validp[queryoffset] == false) {
	/* Skip */
	minus_querypos[mod] += 3;
      } else {
	if (stage1->minus_retrievedp[queryoffset] == true) {
#ifdef LARGE_GENOMES
	  positions_high = stage1->minus_positions_high[queryoffset];
#endif
	  positions = stage1->minus_positions[queryoffset];
	  npositions = stage1->minus_npositions[queryoffset];
	  debug(printf("Already have %d minus positions at %d\n",npositions,queryoffset));
	} else {
	  /* These should be lefts */
	  assert(stage1->minus_positions[queryoffset] == NULL);
	  stage1->minus_retrievedp[queryoffset] = true;
	  /* In standard stage1 convention, diagterm would be
	     -querylength + querypos + index1part = -(query_lastpos - querypos) = -queryoffset */
#ifdef LARGE_GENOMES
	  npositions = stage1->minus_npositions[queryoffset] =
	    Indexdb_largeptr_with_diagterm(&stage1->minus_positions_high[queryoffset],&stage1->minus_positions[queryoffset],
					   indexdb2,stage1->minus_oligos[queryoffset],/*diagterm*/-queryoffset);
	  positions_high = stage1->minus_positions_high[queryoffset];
#else
	  npositions = stage1->minus_npositions[queryoffset] =
	    Indexdb_ptr_with_diagterm(&stage1->minus_positions[queryoffset],indexdb2,
				      stage1->minus_oligos[queryoffset],/*diagterm*/-queryoffset);
#endif
	  positions = stage1->minus_positions[queryoffset];
	}
	
	if ((elt = Elt_read_queryfwd(
#ifdef LARGE_GENOMES
				     positions_high,
#endif
				     positions,npositions,/*diagterm*/-queryoffset,
				     querylength,/*querystart*/queryoffset,
				     query_compress_rev,index1part,/*plusp*/false,
				     genestrand)) == NULL) {
	  minus_querypos[mod] += 3;
	} else {
	  if (elt->nmatches > best_minus_nmatches && elt->n_all_diagonals <= MAX_HITS_FOR_BEST_ELT) {
	    best_minus_elt = elt;
	    best_minus_nmatches = elt->nmatches;
	    best_minus_i = niter_minus;
	  }
	  *minus_set = List_push(*minus_set,elt);
	  minus_querypos[mod] += elt->nmatches;
	  niter_minus++;
	}
      }
	
      if (minus_querypos[mod] > max_minus_querypos) {
	max_minus_querypos = minus_querypos[mod];
      }
    }

#ifdef DEBUG
    printf("\n");
    for (mod = 0; mod < 3; mod++) {
      printf("mod %d, plus_querypos %d\n",mod,plus_querypos[mod]);
      printf("mod %d, minus_querypos %d\n",mod,minus_querypos[mod]);
    }
    printf("max_plus_querypos %d\n",max_plus_querypos);
    printf("max_maxus_querypos %d\n",max_minus_querypos);
    printf("\n");
#endif

#ifndef LIBERAL
    /* Skip the presumed mismatch */
    max_plus_querypos += 1;
    max_minus_querypos += 1;
#endif

    plus_querypos[0] = max_plus_querypos;
    plus_querypos[1] = max_plus_querypos + 1;
    plus_querypos[2] = max_plus_querypos + 2;

    minus_querypos[0] = max_minus_querypos;
    minus_querypos[1] = max_minus_querypos + 1;
    minus_querypos[2] = max_minus_querypos + 2;
  }


#ifdef DEBUG
  printf("plus set:\n");
  for (p = *plus_set; p != NULL; p = List_next(p)) {
    Elt_dump((T) List_head(p));
  }
  printf("\n");
  printf("minus set:\n");
  for (p = *minus_set; p != NULL; p = List_next(p)) {
    Elt_dump((T) List_head(p));
  }
#endif


  if (max_minus_querypos >= query_lastpos) {
    /* This branch is critical for high speed (about 4x faster),
       although we could miss some hits that need to be solved by
       another method */
    /* minus side won, so skip plus side */
    nseeds_plus = 0;

  } else if (max_plus_querypos < query_lastpos || best_plus_elt == NULL) {
    debug(printf("Plus: Still could not find large pieces: plus_querypos %d < query_lastpos %d, best_plus_elt %p\n",
		 max_plus_querypos,query_lastpos,best_plus_elt));
    nseeds_plus = 0;

  } else {
    /* Elt_fill_diagonals_all(best_plus_elt,plus_sarray); */
    if (best_plus_elt->ndiagonals == 0) {
      /* Could happen if there are too many diagonals */
      debug(printf("Plus: Best elt has no diagonals\n"));
      nseeds_plus = 0;

    } else {
      *plus_set = List_reverse(*plus_set);
      plus_elt_array = (T *) List_to_array_n(&nplus,*plus_set);

#ifdef DEBUG
      printf("LEFT\n");
      for (i = 0; i < best_plus_i; i++) {
	Elt_dump(plus_elt_array[i]);
      }
      printf("MIDDLE\n");
      Elt_dump(plus_elt_array[best_plus_i]);
      printf("RIGHT\n");
      for (i = best_plus_i + 1; i < nplus; i++) {
	Elt_dump(plus_elt_array[i]);
      }
#endif

      filter_elts(plus_elt_array,best_plus_i,nplus);

      nseeds_plus = best_plus_elt->ndiagonals;
      /* Assigned only if score is high */

      middle_diagonals_plus = (Univdiag_T *) MALLOC(nseeds_plus*sizeof(Univdiag_T));
      best_right_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));
      best_left_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));
      all_right_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));
      all_left_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));

      chrnum = 1;
      Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
      for (i = 0; i < nseeds_plus; i++) {
	left = best_plus_elt->diagonals[i];
	debug(printf("Comparing left %u with chrhigh %u\n",left,chrhigh));
	if (left >= chrhigh) {
	  chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	  /* *chrhigh += 1U; */
	}
	/* May not want to solve for best_right_diagonals and best_left_diagonals.  Use oligoindex instead. */
	get_diagonals(&(middle_diagonals_plus[i]),
		      &(best_right_diagonals_plus[i]),&(best_left_diagonals_plus[i]),
		      &(all_right_diagonals_plus[i]),&(all_left_diagonals_plus[i]),
		      querylength,chroffset,chrhigh,/*goal*/left,
		      plus_elt_array,best_plus_i,nplus);
      }

      FREE(plus_elt_array);
    }
  }

  if (max_plus_querypos >= query_lastpos) {
    /* This branch is critical for high speed (about 4x faster),
       although we could miss some hits that need to be solved by
       another method */
    /* plus side won, so skip minus side */
    nseeds_minus = 0;

  } else if (max_minus_querypos < query_lastpos || best_minus_elt == NULL) {
    debug(printf("Minus: Still could not find large pieces: minus_querypos %d < query_lastpos %d, best_minus_elt %p\n",
		 max_minus_querypos,query_lastpos,best_minus_elt));
    nseeds_minus = 0;

  } else {
    /* Elt_fill_diagonals_all(best_minus_elt,minus_sarray); */
    if (best_minus_elt->ndiagonals == 0) {
      /* Could happen if there are too many diagonals */
      debug(printf("Minus: Best elt has no diagonals\n"));
      nseeds_minus = 0;

    } else {
      *minus_set = List_reverse(*minus_set);
      minus_elt_array = (T *) List_to_array_n(&nminus,*minus_set);

#ifdef DEBUG
      printf("LEFT\n");
      for (i = 0; i < best_minus_i; i++) {
	Elt_dump(minus_elt_array[i]);
      }
      printf("MIDDLE\n");
      Elt_dump(minus_elt_array[best_minus_i]);
      printf("RIGHT\n");
      for (i = best_minus_i + 1; i < nminus; i++) {
	Elt_dump(minus_elt_array[i]);
      }
#endif

      filter_elts(minus_elt_array,best_minus_i,nminus);

      nseeds_minus = best_minus_elt->ndiagonals;
      /* Assigned only if score is high */

      middle_diagonals_minus = (Univdiag_T *) MALLOC(nseeds_minus*sizeof(Univdiag_T));
      best_right_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));
      best_left_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));
      all_right_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));
      all_left_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));

      chrnum = 1;
      Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
      for (i = 0; i < nseeds_minus; i++) {
	left = best_minus_elt->diagonals[i];
	debug(printf("Comparing left %u with chrhigh %u\n",left,chrhigh));
	if (left >= chrhigh) {
	  chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	  /* *chrhigh += 1U; */
	}
	/* May not want to solve for best_right_diagonals and best_left_diagonals.  Use oligoindex instead. */
	get_diagonals(&(middle_diagonals_minus[i]),
		      &(best_right_diagonals_minus[i]),&(best_left_diagonals_minus[i]),
		      &(all_right_diagonals_minus[i]),&(all_left_diagonals_minus[i]),
		      querylength,chroffset,chrhigh,/*goal*/left,
		      minus_elt_array,best_minus_i,nminus);
      }

      FREE(minus_elt_array);
    }
  }

#if 0
  /* Because we don't always left-extend, we cannot trust best_score */
  best_score = 0;
  for (i = 0; i < nseeds_plus; i++) {
    if (scores_plus[i] > best_score) {
      best_score = scores_plus[i];
    }
  }
  for (i = 0; i < nseeds_minus; i++) {
    if (scores_minus[i] > best_score) {
      best_score = scores_minus[i];
    }
  }
#endif

  debug(printf("Have %d nseeds_plus and %d nseeds_minus\n",nseeds_plus,nseeds_minus));


  chrnum = 1;
  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
  for (i = 0; i < nseeds_plus; i++) {
    if (1 /*|| scores_plus[i] > best_score - 20*/) {
      diagonal = middle_diagonals_plus[i];
      left = diagonal->univdiagonal;
      if (left >= chrhigh) {
	chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	/* *chrhigh += 1U; */
      }
      debug13(printf("chrnum %d, diagonal %u, chrlength %d, chrpos %d\n",chrnum,left,chrlength,left-chroffset));
      debug13(printf("Entering find_best_path_genome with diagonal %d..%d, %d best right, and %d best left\n",
		     diagonal->querystart,diagonal->queryend,List_length(best_right_diagonals_plus[i]),
		     List_length(best_left_diagonals_plus[i])));

      *hits_gplus = Path_solve_from_diagonals(&foundp,&(*found_score),&complete_path,*hits_gplus,
					      /*middle_diagonal*/diagonal,best_right_diagonals_plus[i],
					      best_left_diagonals_plus[i],
					      /*queryptr*/queryuc_ptr,querylength,query_compress_fwd,
					      chrnum,chroffset,chrhigh,chrlength,/*plusp*/true,genestrand,
					      /*nmismatches_allowed*/nmisses_allowed,paired_end_p,
					      /*METHOD*/EXT,level);

      if (foundp == false) {
	*plus_complete_paths = List_push(*plus_complete_paths,(void *) complete_path);
      } else {
	successp = true;
      }
    }
  }

  chrnum = 1;
  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
  for (i = 0; i < nseeds_minus; i++) {
    if (1 /*|| scores_minus[i] > best_score - 20*/) {
      diagonal = middle_diagonals_minus[i];
      left = diagonal->univdiagonal;
      if (left >= chrhigh) {
	chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	/* *chrhigh += 1U; */
      }

      debug13(printf("chrnum %d, diagonal %u, chrlength %d, chrpos %d\n",chrnum,left,chrlength,left-chroffset));
      debug13(printf("Entering find_best_path_genome with diagonal %d..%d, %d best right, and %d best left\n",
		     diagonal->querystart,diagonal->queryend,List_length(best_right_diagonals_minus[i]),
		     List_length(best_left_diagonals_minus[i])));

      *hits_gminus = Path_solve_from_diagonals(&foundp,&(*found_score),&complete_path,*hits_gminus,
					       /*middle_diagonal*/diagonal,best_right_diagonals_minus[i],
					       best_left_diagonals_minus[i],
					       /*queryptr*/queryrc,querylength,query_compress_rev,
					       chrnum,chroffset,chrhigh,chrlength,/*plusp*/false,genestrand,
					       /*nmismatches_allowed*/nmisses_allowed,paired_end_p,
					       /*method*/EXT,level);

      if (foundp == false) {
	*minus_complete_paths = List_push(*minus_complete_paths,(void *) complete_path);
      } else {
	successp = true;
      }
    }
  }


  if (nseeds_minus > 0) {
    for (i = 0; i < nseeds_minus; i++) {
      Univdiag_free(&(middle_diagonals_minus[i]));
      List_free(&(best_right_diagonals_minus[i]));
      List_free(&(best_left_diagonals_minus[i]));
      Univdiag_gc(&(all_right_diagonals_minus[i]));
      Univdiag_gc(&(all_left_diagonals_minus[i]));
    }
    FREE(middle_diagonals_minus);
    FREE(best_right_diagonals_minus);
    FREE(best_left_diagonals_minus);
    FREE(all_right_diagonals_minus);
    FREE(all_left_diagonals_minus);
  }

  if (nseeds_plus > 0) {
    for (i = 0; i < nseeds_plus; i++) {
      Univdiag_free(&(middle_diagonals_plus[i]));
      List_free(&(best_right_diagonals_plus[i]));
      List_free(&(best_left_diagonals_plus[i]));
      Univdiag_gc(&(all_right_diagonals_plus[i]));
      Univdiag_gc(&(all_left_diagonals_plus[i]));
    }
    FREE(middle_diagonals_plus);
    FREE(best_right_diagonals_plus);
    FREE(best_left_diagonals_plus);
    FREE(all_right_diagonals_plus);
    FREE(all_left_diagonals_plus);
  }

  debug(printf("Found %d plus and %d minus hits\n",List_length(*hits_gplus),List_length(*hits_gminus)));

#if 0
  Reader_free(&reader_minus);
  Reader_free(&reader_plus);
#endif

  return successp;
}


/* Note: A second try, starting at queryrev, may solve only a few more
   cases presented to this method (i.e., cases that cannot be solved
   by kmer-end search).  It therefore may be best to push these cases
   to the next method. */

static bool
Extension_search_queryrev (int *found_score, List_T *hits_gplus, List_T *hits_gminus,

			   /* Used by Distant_rna_solve */
			   List_T *plus_set, List_T *minus_set,

			   List_T *plus_complete_paths, List_T *minus_complete_paths,
			   Stage1_T stage1, char *queryuc_ptr, char *queryrc, int querylength,
			   Compress_T query_compress_fwd, Compress_T query_compress_rev, 
			   int nmisses_allowed, int genestrand,
			   bool paired_end_p, int level) {
  bool successp = false;
  List_T complete_path;

  T best_plus_elt, best_minus_elt, elt, *plus_elt_array, *minus_elt_array;
  int best_plus_nmatches, best_minus_nmatches;
  bool foundp;
  int min_plus_querypos, min_minus_querypos, plus_querypos[3], minus_querypos[3];
  int i;
  Chrnum_T chrnum;
  Univcoord_T chroffset, chrhigh, left;
  Chrpos_T chrlength;

  int nseeds_plus, nseeds_minus;
  int mod;
  int niter_plus, niter_minus, best_plus_i, best_minus_i, nplus, nminus;
  Univdiag_T *middle_diagonals_plus = NULL, *middle_diagonals_minus = NULL;
  List_T *best_right_diagonals_plus = NULL, *best_left_diagonals_plus = NULL,
    *all_right_diagonals_plus = NULL, *all_left_diagonals_plus = NULL,
    *best_right_diagonals_minus = NULL, *best_left_diagonals_minus = NULL,
    *all_right_diagonals_minus = NULL, *all_left_diagonals_minus = NULL;

  Univdiag_T diagonal;
#ifdef LARGE_GENOMES
  unsigned char *positions_high;
#endif
  UINT4 *positions;
  int npositions;

  int query_lastpos = querylength - index1part;
  int queryoffset;
#if 0
  Reader_T reader_plus, reader_minus;
  Oligostate_T last_state_plus, last_state_minus;
  Oligospace_T oligo, oligo2;
  Oligospace_T forward_plus, revcomp_plus, forward_minus, revcomp_minus, forward_oligo, revcomp_oligo;
#endif
#ifdef DEBUG
  List_T p;
#endif

  debug(printf("\nStarting Extension_search_queryrev with querylength %d and nmisses_allowed %d, genestrand %d\n",
	       querylength,nmisses_allowed,genestrand));

#if 0
  /* Allow calls to Extension_search_queryrev in addition to other methods */
  *hits_gplus = *hits_gminus = (List_T) NULL;
#endif

  if (nmisses_allowed < 0) {
    nmisses_allowed = 0;
#if 0
  } else {
    /* It is possible that this makes GSNAP too slow */
    nmisses_allowed = querylength;
#endif
  }

#if 0
  if (genestrand == +2) {
    plus_conversion = conversion_rev;
    minus_conversion = conversion_fwd;
  } else {
    plus_conversion = conversion_fwd;
    minus_conversion = conversion_rev;
  }
#endif

#if 0
  reader_plus = Reader_new(queryuc_ptr,/*querystart*/0,/*queryend*/querylength);
  last_state_plus = INIT;
  forward_plus = revcomp_plus = 0;

  reader_minus = Reader_new(queryrc,/*querystart*/0,/*queryend*/querylength);
  last_state_minus = INIT;
  forward_minus = revcomp_minus = 0;
#endif

  /* I.  Race from plus and minus end to start */
  *plus_set = *minus_set = (List_T) NULL;
  best_plus_nmatches = best_minus_nmatches = 0;
  best_plus_elt = best_minus_elt = (T) NULL;
  min_plus_querypos = min_minus_querypos = query_lastpos;
  plus_querypos[0] = minus_querypos[0] = query_lastpos;
  plus_querypos[1] = minus_querypos[1] = query_lastpos-1;
  plus_querypos[2] = minus_querypos[2] = query_lastpos-2;
  niter_plus = niter_minus = 0;

  while (niter_plus <= nmisses_allowed && niter_minus <= nmisses_allowed &&
	 min_plus_querypos > 0 && min_minus_querypos > 0) {
    for (mod = 0; mod < 3; mod++) {
      if ((queryoffset = plus_querypos[mod]) < 0) {
	/* Skip */
	plus_querypos[mod] -= 3;
      } else if (stage1->plus_validp[queryoffset] == false) {
	/* Skip */
	plus_querypos[mod] -= 3;
      } else {
	if (stage1->plus_retrievedp[queryoffset] == true) {
#ifdef LARGE_GENOMES
	  positions_high = stage1->plus_positions_high[queryoffset];
#endif
	  positions = stage1->plus_positions[queryoffset];
	  npositions = stage1->plus_npositions[queryoffset];
	} else {
	  /* These should be lefts */
	  assert(stage1->plus_positions[queryoffset] == NULL);
	  stage1->plus_retrievedp[queryoffset] = true;
#ifdef LARGE_GENOMES
	  npositions = stage1->plus_npositions[queryoffset] =
	    Indexdb_largeptr_with_diagterm(&stage1->plus_positions_high[queryoffset],&stage1->plus_positions[queryoffset],
					   indexdb,stage1->plus_oligos[queryoffset],/*diagterm*/-queryoffset);
	  positions_high = stage1->plus_positions_high[queryoffset];
#else
	  npositions = stage1->plus_npositions[queryoffset] =
	    Indexdb_ptr_with_diagterm(&stage1->plus_positions[queryoffset],indexdb,
				      stage1->plus_oligos[queryoffset],/*diagterm*/-queryoffset);
#endif
	  positions = stage1->plus_positions[queryoffset];
	}
	
	if ((elt = Elt_read_queryrev(
#ifdef LARGE_GENOMES
				     positions_high,
#endif
				     positions,npositions,/*diagterm*/-queryoffset,
				     querylength,/*queryend*/queryoffset + index1part,
				     query_compress_fwd,index1part,/*plusp*/true,
				     genestrand)) == NULL) {
	  plus_querypos[mod] -= 3;
	} else {
	  if (elt->nmatches > best_plus_nmatches && elt->n_all_diagonals <= MAX_HITS_FOR_BEST_ELT) {
	    best_plus_elt = elt;
	    best_plus_nmatches = elt->nmatches;
	    best_plus_i = niter_plus;
	  }
	  *plus_set = List_push(*plus_set,elt);
	  plus_querypos[mod] -= elt->nmatches;
	  niter_plus++;
	}
      }

      if (plus_querypos[mod] < min_plus_querypos) {
	min_plus_querypos = plus_querypos[mod];
      }


      /* querypos_rc uses standard Stage1_T convention, but we have switched to sarray convention */
      /* querypos_rc = query_lastpos - queryoffset; */
      if ((queryoffset = minus_querypos[mod]) < 0) {
	/* Skip */
	minus_querypos[mod] -= 3;
      } else if (stage1->minus_validp[queryoffset] == false) {
	/* Skip */
	minus_querypos[mod] -= 3;
      } else {
	if (stage1->minus_retrievedp[queryoffset] == true) {
#ifdef LARGE_GENOMES
	  positions_high = stage1->minus_positions_high[queryoffset];
#endif
	  positions = stage1->minus_positions[queryoffset];
	  npositions = stage1->minus_npositions[queryoffset];
	} else {
	  /* These should be lefts */
	  assert(stage1->minus_positions[queryoffset] == NULL);
	  stage1->minus_retrievedp[queryoffset] = true;
	  /* In standard stage1 convention, diagterm would be
	     -querylength + querypos + index1part = -(query_lastpos - querypos) = -queryoffset */
#ifdef LARGE_GENOMES
	  npositions = stage1->minus_npositions[queryoffset] =
	    Indexdb_largeptr_with_diagterm(&stage1->minus_positions_high[queryoffset],&stage1->minus_positions[queryoffset],
					   indexdb2,stage1->minus_oligos[queryoffset],/*diagterm*/-queryoffset);
	  positions_high = stage1->minus_positions_high[queryoffset];
#else
	  npositions = stage1->minus_npositions[queryoffset] =
	    Indexdb_ptr_with_diagterm(&stage1->minus_positions[queryoffset],indexdb2,
				      stage1->minus_oligos[queryoffset],/*diagterm*/-queryoffset);
#endif
	  positions = stage1->minus_positions[queryoffset];
	}
	
	if ((elt = Elt_read_queryrev(
#ifdef LARGE_GENOMES
				     positions_high,
#endif
				     positions,npositions,/*diagterm*/-queryoffset,
				     querylength,/*queryend*/queryoffset + index1part,
				     query_compress_rev,index1part,/*plusp*/false,
				     genestrand)) == NULL) {
	  minus_querypos[mod] -= 3;
	} else {
	  if (elt->nmatches > best_minus_nmatches && elt->n_all_diagonals <= MAX_HITS_FOR_BEST_ELT) {
	    best_minus_elt = elt;
	    best_minus_nmatches = elt->nmatches;
	    best_minus_i = niter_minus;
	  }
	  *minus_set = List_push(*minus_set,elt);
	  minus_querypos[mod] -= elt->nmatches;
	  niter_minus++;
	}
      }

      if (minus_querypos[mod] < min_minus_querypos) {
	min_minus_querypos = minus_querypos[mod];
      }
    }

#ifdef DEBUG
    printf("\n");
    for (mod = 0; mod < 3; mod++) {
      printf("mod %d, plus_querypos %d\n",mod,plus_querypos[mod]);
      printf("mod %d, minus_querypos %d\n",mod,minus_querypos[mod]);
    }
    printf("min_plus_querypos %d\n",min_plus_querypos);
    printf("min_minus_querypos %d\n",min_minus_querypos);
    printf("\n");
#endif

#ifndef LIBERAL
    /* Skip the presumed mismatch */
    min_plus_querypos -= 1;
    min_minus_querypos -= 1;
#endif

    plus_querypos[0] = min_plus_querypos;
    plus_querypos[1] = min_plus_querypos - 1;
    plus_querypos[2] = min_plus_querypos - 2;

    minus_querypos[0] = min_minus_querypos;
    minus_querypos[1] = min_minus_querypos - 1;
    minus_querypos[2] = min_minus_querypos - 2;
  }


#ifdef DEBUG
  printf("plus set:\n");
  for (p = *plus_set; p != NULL; p = List_next(p)) {
    Elt_dump((T) List_head(p));
  }
  printf("\n");
  printf("minus set:\n");
  for (p = *minus_set; p != NULL; p = List_next(p)) {
    Elt_dump((T) List_head(p));
  }
#endif

  if (min_minus_querypos <= 0) {
    /* This branch is critical for high speed (about 4x faster),
       although we could miss some hits that need to be solved by
       another method */
    /* minus side won, so skip plus side */
    nseeds_plus = 0;

  } else if (min_plus_querypos > 0 || best_plus_elt == NULL) {
    debug(printf("Plus: Still could not find large pieces: plus_querypos %d > 0, best_plus_elt %p\n",
		 min_plus_querypos,best_plus_elt));
    nseeds_plus = 0;

  } else {
    /* Elt_fill_diagonals_all(best_plus_elt,plus_sarray); */
    if (best_plus_elt->ndiagonals == 0) {
      /* Could happen if there are too many diagonals */
      debug(printf("Plus: Best elt has no diagonals\n"));
      nseeds_plus = 0;

    } else {
      /* *plus_set = List_reverse(*plus_set);  -- not needed for queryrev */
      plus_elt_array = (T *) List_to_array_n(&nplus,*plus_set);
      best_plus_i = (nplus - 1) - best_plus_i; /* Critical for making best_i correct without a List_reverse */
      assert(plus_elt_array[best_plus_i] == best_plus_elt);

#ifdef DEBUG
      printf("LEFT\n");
      for (i = 0; i < best_plus_i; i++) {
	Elt_dump(plus_elt_array[i]);
      }
      printf("MIDDLE\n");
      Elt_dump(plus_elt_array[best_plus_i]);
      printf("RIGHT\n");
      for (i = best_plus_i + 1; i < nplus; i++) {
	Elt_dump(plus_elt_array[i]);
      }
#endif

      filter_elts(plus_elt_array,best_plus_i,nplus);

      nseeds_plus = best_plus_elt->ndiagonals;
      /* Assigned only if score is high */

      middle_diagonals_plus = (Univdiag_T *) MALLOC(nseeds_plus*sizeof(Univdiag_T));
      best_right_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));
      best_left_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));
      all_right_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));
      all_left_diagonals_plus = (List_T *) MALLOC(nseeds_plus*sizeof(List_T));

      chrnum = 1;
      Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
      for (i = 0; i < nseeds_plus; i++) {
	left = best_plus_elt->diagonals[i];
	debug(printf("Comparing left %u with chrhigh %u\n",left,chrhigh));
	if (left >= chrhigh) {
	  chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	  /* *chrhigh += 1U; */
	}
	/* May not want to solve for best_right_diagonals and best_left_diagonals.  Use oligoindex instead. */
	get_diagonals(&(middle_diagonals_plus[i]),
		      &(best_right_diagonals_plus[i]),&(best_left_diagonals_plus[i]),
		      &(all_right_diagonals_plus[i]),&(all_left_diagonals_plus[i]),
		      querylength,chroffset,chrhigh,/*goal*/left,
		      plus_elt_array,best_plus_i,nplus);
      }

      FREE(plus_elt_array);
    }
  }

  if (min_plus_querypos <= 0) {
    /* This branch is critical for high speed (about 4x faster),
       although we could miss some hits that need to be solved by
       another method */
    /* plus side won, so skip minus side */
    nseeds_minus = 0;

  } else if (min_minus_querypos > 0 || best_minus_elt == NULL) {
    debug(printf("Minus: Still could not find large pieces: minus_querypos %d > 0, best_minus_elt %p\n",
		 min_minus_querypos,best_minus_elt));
    nseeds_minus = 0;

  } else {
    /* Elt_fill_diagonals_all(best_minus_elt,minus_sarray); */
    if (best_minus_elt->ndiagonals == 0) {
      /* Could happen if there are too many diagonals */
      debug(printf("Minus: Best elt has no diagonals\n"));
      nseeds_minus = 0;

    } else {
      /* *minus_set = List_reverse(*minus_set); -- not needed for queryrev */
      minus_elt_array = (T *) List_to_array_n(&nminus,*minus_set);
      best_minus_i = (nminus - 1) - best_minus_i; /* Critical for making best_i correct without a List_reverse */
      assert(minus_elt_array[best_minus_i] == best_minus_elt);

#ifdef DEBUG
      printf("LEFT\n");
      for (i = 0; i < best_minus_i; i++) {
	Elt_dump(minus_elt_array[i]);
      }
      printf("MIDDLE\n");
      Elt_dump(minus_elt_array[best_minus_i]);
      printf("RIGHT\n");
      for (i = best_minus_i + 1; i < nminus; i++) {
	Elt_dump(minus_elt_array[i]);
      }
#endif

      filter_elts(minus_elt_array,best_minus_i,nminus);

      nseeds_minus = best_minus_elt->ndiagonals;
      /* Assigned only if score is high */

      middle_diagonals_minus = (Univdiag_T *) MALLOC(nseeds_minus*sizeof(Univdiag_T));
      best_right_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));
      best_left_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));
      all_right_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));
      all_left_diagonals_minus = (List_T *) MALLOC(nseeds_minus*sizeof(List_T));

      chrnum = 1;
      Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
      for (i = 0; i < nseeds_minus; i++) {
	left = best_minus_elt->diagonals[i];
	debug(printf("Comparing left %u with chrhigh %u\n",left,chrhigh));
	if (left >= chrhigh) {
	  chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	  /* *chrhigh += 1U; */
	}
	/* May not want to solve for best_right_diagonals and best_left_diagonals.  Use oligoindex instead. */
	get_diagonals(&(middle_diagonals_minus[i]),
		      &(best_right_diagonals_minus[i]),&(best_left_diagonals_minus[i]),
		      &(all_right_diagonals_minus[i]),&(all_left_diagonals_minus[i]),
		      querylength,chroffset,chrhigh,/*goal*/left,
		      minus_elt_array,best_minus_i,nminus);
      }

      FREE(minus_elt_array);
    }
  }

#if 0
  /* Because we don't always left-extend, we cannot trust best_score */
  best_score = 0;
  for (i = 0; i < nseeds_plus; i++) {
    if (scores_plus[i] > best_score) {
      best_score = scores_plus[i];
    }
  }
  for (i = 0; i < nseeds_minus; i++) {
    if (scores_minus[i] > best_score) {
      best_score = scores_minus[i];
    }
  }
#endif

  debug(printf("Have %d nseeds_plus and %d nseeds_minus\n",nseeds_plus,nseeds_minus));


  chrnum = 1;
  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
  for (i = 0; i < nseeds_plus; i++) {
    if (1 /*|| scores_plus[i] > best_score - 20*/) {
      diagonal = middle_diagonals_plus[i];
      left = diagonal->univdiagonal;
      if (left >= chrhigh) {
	chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	/* *chrhigh += 1U; */
      }
      debug13(printf("chrnum %d, diagonal %u, chrlength %d, chrpos %d\n",chrnum,left,chrlength,left-chroffset));
      debug13(printf("Entering find_best_path_genome with diagonal %d..%d, %d best right, and %d best left\n",
		     diagonal->querystart,diagonal->queryend,List_length(best_right_diagonals_plus[i]),
		     List_length(best_left_diagonals_plus[i])));

      *hits_gplus = Path_solve_from_diagonals(&foundp,&(*found_score),&complete_path,*hits_gplus,
					      /*middle_diagonal*/diagonal,best_right_diagonals_plus[i],
					      best_left_diagonals_plus[i],
					      /*queryptr*/queryuc_ptr,querylength,query_compress_fwd,
					      chrnum,chroffset,chrhigh,chrlength,/*plusp*/true,genestrand,
					      /*nmismatches_allowed*/nmisses_allowed,paired_end_p,
					      /*method*/EXT,level);

      if (foundp == false) {
	*plus_complete_paths = List_push(*plus_complete_paths,(void *) complete_path);
      } else {
	successp = true;
      }
    }
  }

  chrnum = 1;
  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,/*chrnum*/1,circular_typeint);
  for (i = 0; i < nseeds_minus; i++) {
    if (1 /*|| scores_minus[i] > best_score - 20*/) {
      diagonal = middle_diagonals_minus[i];
      left = diagonal->univdiagonal;
      if (left >= chrhigh) {
	chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
	Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);
	/* *chrhigh += 1U; */
      }

      debug13(printf("chrnum %d, diagonal %u, chrlength %d, chrpos %d\n",chrnum,left,chrlength,left-chroffset));
      debug13(printf("Entering find_best_path_genome with diagonal %d..%d, %d best right, and %d best left\n",
		     diagonal->querystart,diagonal->queryend,List_length(best_right_diagonals_minus[i]),
		     List_length(best_left_diagonals_minus[i])));

      *hits_gminus = Path_solve_from_diagonals(&foundp,&(*found_score),&complete_path,*hits_gminus,
					       /*middle_diagonal*/diagonal,best_right_diagonals_minus[i],
					       best_left_diagonals_minus[i],
					       /*queryptr*/queryrc,querylength,query_compress_rev,
					       chrnum,chroffset,chrhigh,chrlength,/*plusp*/false,genestrand,
					       /*nmismatches_allowed*/nmisses_allowed,paired_end_p,
					       /*method*/EXT,level);

      if (foundp == false) {
	*minus_complete_paths = List_push(*minus_complete_paths,(void *) complete_path);
      } else {
	successp = true;
      }
    }
  }


  if (nseeds_minus > 0) {
    for (i = 0; i < nseeds_minus; i++) {
      Univdiag_free(&(middle_diagonals_minus[i]));
      List_free(&(best_right_diagonals_minus[i]));
      List_free(&(best_left_diagonals_minus[i]));
      Univdiag_gc(&(all_right_diagonals_minus[i]));
      Univdiag_gc(&(all_left_diagonals_minus[i]));
    }
    FREE(middle_diagonals_minus);
    FREE(best_right_diagonals_minus);
    FREE(best_left_diagonals_minus);
    FREE(all_right_diagonals_minus);
    FREE(all_left_diagonals_minus);
  }

  if (nseeds_plus > 0) {
    for (i = 0; i < nseeds_plus; i++) {
      Univdiag_free(&(middle_diagonals_plus[i]));
      List_free(&(best_right_diagonals_plus[i]));
      List_free(&(best_left_diagonals_plus[i]));
      Univdiag_gc(&(all_right_diagonals_plus[i]));
      Univdiag_gc(&(all_left_diagonals_plus[i]));
    }
    FREE(middle_diagonals_plus);
    FREE(best_right_diagonals_plus);
    FREE(best_left_diagonals_plus);
    FREE(all_right_diagonals_plus);
    FREE(all_left_diagonals_plus);
  }

  /* Caller needs to call Elt_gc(&plus_set) and Elt_gc(&minus_set) */

  debug(printf("Found %d plus and %d minus hits\n",List_length(*hits_gplus),List_length(*hits_gminus)));

#if 0
  Reader_free(&reader_minus);
  Reader_free(&reader_plus);
#endif

  return successp;
}


void
Extension_search (int *found_score, List_T *hits_gplus, List_T *hits_gminus,

		  /* Used by Distant_rna_solve */
		  List_T *queryfwd_plus_set, List_T *queryfwd_minus_set,
		  List_T *queryrev_plus_set, List_T *queryrev_minus_set,

		  Stage1_T stage1, char *queryuc_ptr, char *queryrc, int querylength,
		  Compress_T query_compress_fwd, Compress_T query_compress_rev, 

		  Pairpool_T pairpool, Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR,
		  Oligoindex_array_T oligoindices_minor, Diagpool_T diagpool, Cellpool_T cellpool,

		  int nmismatches_allowed, int genestrand, bool paired_end_p, int level) {
  bool success_fwd_p, success_rev_p;
  List_T plus_complete_paths = NULL, minus_complete_paths = NULL;

  success_fwd_p = Extension_search_queryfwd(&(*found_score),&(*hits_gplus),&(*hits_gminus),
					    &(*queryfwd_plus_set),&(*queryfwd_minus_set),
					    &plus_complete_paths,&minus_complete_paths,
					    stage1,queryuc_ptr,queryrc,querylength,
					    query_compress_fwd,query_compress_rev,
					    nmismatches_allowed,genestrand,paired_end_p,level);
  
  success_rev_p = Extension_search_queryrev(&(*found_score),&(*hits_gplus),&(*hits_gminus),
					    &(*queryrev_plus_set),&(*queryrev_minus_set),
					    &plus_complete_paths,&minus_complete_paths,
					    stage1,queryuc_ptr,queryrc,querylength,
					    query_compress_fwd,query_compress_rev,
					    nmismatches_allowed,genestrand,paired_end_p,level);

  if (success_fwd_p == false && success_rev_p == false) {
    Path_solve_via_gmap(&(*found_score),&(*hits_gplus),&(*hits_gminus),
			plus_complete_paths,minus_complete_paths,genestrand,
			queryuc_ptr,querylength,query_compress_fwd,query_compress_rev,
			pairpool,dynprogL,dynprogM,dynprogR,oligoindices_minor,diagpool,cellpool,
			paired_end_p,/*method*/EXT_GMAP,level);
  }

  Univdiag_list_gc(&plus_complete_paths);
  Univdiag_list_gc(&minus_complete_paths);

  return;
}

