static char rcsid[] = "$Id: merge-simd-records.c 214107 2018-03-10 09:01:49Z twu $";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "merge-simd-records.h"
#include "assert.h"
#include "mem.h"
#include "bool.h"
#include "types.h"
#include "popcount.h"		/* For clz_table */
#include "merge-uint4.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>		/* For memcpy */


/* #define PYRAMID_SIZE 4 */
/* #define KEY_MASK (~0U << 2) */

#define PYRAMID_SIZE 32
#define KEY_MASK (~0U << 5)


/* Results of Merge_records */
#ifdef DEBUG0
#define debug0(x) x
#else
#define debug0(x)
#endif

#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif

#ifdef DEBUG2
#define debug2(x) x
#else
#define debug2(x)
#endif



#define PARENT(i) (i >> 1)
#define LEFT(i) (i << 1)
#define RIGHT(i) ((i << 1) | 1)



/* Assumes heapi < base put into LEFT(heapi) */
static int
pyramid_merge_prealloc (unsigned int **heap, unsigned int *curr_storage, unsigned int *prev_storage,
			int *nelts, int pyramid_start, int pyramid_end) {
  unsigned int *temp;
  int nodei;
  int nalloc;
#ifdef HAVE_SSE4_1
  int n;
#endif
#ifdef DEBUG2
  int i;
#endif

  while (pyramid_end > pyramid_start) {
    debug2(printf("Merging level: %d..%d\n",pyramid_start,pyramid_end));
    nalloc = 0;

    nodei = pyramid_end;
    while (nodei >= pyramid_start) {
      debug2(printf("Merging nodes %d (%d elts) and %d (%d elts) => %d\n",
		   nodei-1,nelts[nodei-1],nodei,nelts[nodei],PARENT(nodei)));
      heap[PARENT(nodei)] = Merge_uint4(/*dest*/&(curr_storage[nalloc]),heap[nodei-1],heap[nodei],nelts[nodei-1],nelts[nodei]);
      CHECK_ALIGN(heap[PARENT(nodei)]);
      /* Have to align start of each entry curr_storage[nalloc], regardless of end padding */
#ifdef HAVE_SSE4_1
      n = nelts[PARENT(nodei)] = nelts[nodei-1] + nelts[nodei];
      nalloc += PAD_UINT4(n);
#else
      nalloc += (nelts[PARENT(nodei)] = nelts[nodei-1] + nelts[nodei]);
#endif
      debug2(printf("Created list %p of length %d at node %d\n",
		    heap[PARENT(nodei)],nelts[PARENT(nodei)],PARENT(nodei)));

#ifdef DEBUG2
      for (i = 0; i < nelts[PARENT(nodei)]; i++) {
	printf("%u\n",heap[PARENT(nodei)][i]);
      }
#endif

      /* Freeing memory one row at a time, so don't do it here */
      nodei -= 2;
    }

    /* Swap memory spaces */
    debug(printf("Swapping storage spaces\n"));
    temp = prev_storage;
    prev_storage = curr_storage;
    curr_storage = temp;

    /* Go up a level */
    pyramid_end = PARENT(pyramid_end);
    pyramid_start = PARENT(pyramid_start);
  }

  debug(printf("Returning ancestor %d\n\n",pyramid_start));
  return pyramid_start;
}



static Record_T **
make_record_heap (int **nelts, List_T stream_list, Intlist_T streamsize_list, 
		  Intlist_T querypos_list, Intlist_T diagterm_list, int nstreams, int base,
		  struct Record_T *all_records) {
  Record_T **record_heap;
  UINT4 *diagonals, diagterm;
  int heapsize, null_pyramid_start, heapi, basei;
  int querypos;
  int i, k;

  heapsize = 2*nstreams - 1;
  null_pyramid_start = (heapsize + PYRAMID_SIZE - 1)/PYRAMID_SIZE * PYRAMID_SIZE; /* full or partial pyramid for entries below this */

  /* Add PYRAMID_SIZE to handle partial pyramid */
  record_heap = (Record_T **) CALLOC(heapsize + PYRAMID_SIZE,sizeof(Record_T *));
  *nelts = (int *) CALLOC(heapsize + PYRAMID_SIZE,sizeof(int));

  /* Process as (base - 1) downto nstreams, then heapsize downto base,
     because stream_list is in reverse order of elts */
  k = 0;
  for (heapi = base - 1; heapi >= PARENT(null_pyramid_start); heapi--) {
    /* Put all information into penultimate row */
    diagonals = (UINT4 *) List_head(stream_list); /* already padded */
    stream_list = List_next(stream_list);

    (*nelts)[heapi] = Intlist_head(streamsize_list);
    streamsize_list = Intlist_next(streamsize_list);

    querypos = Intlist_head(querypos_list);
    querypos_list = Intlist_next(querypos_list);

    diagterm = (UINT4) Intlist_head(diagterm_list);
    diagterm_list = Intlist_next(diagterm_list);

    record_heap[heapi] = (Record_T *) MALLOC(((*nelts)[heapi]) * sizeof(Record_T));
    debug2(printf("NULL: Assigning node %d with %d elts (%p)",heapi,(*nelts)[heapi],record_heap[heapi]));

    for (i = 0; i < (*nelts)[heapi]; i++) {
      /* Process in forward order to keep records in order */
      assert(diagonals[i] >= (Univcoord_T) -diagterm);
      all_records[k].diagonal = diagonals[i] + diagterm;
      all_records[k].querypos = querypos;
      record_heap[heapi][i] = &(all_records[k]);
      debug2(printf(" %u+%d",diagonals[i],querypos));
      k++;
    }
    debug2(printf("\n"));
  }
    
  for ( ; heapi >= nstreams; heapi--) {
    /* Move all information down to left child */
    basei = LEFT(heapi);

    diagonals = (UINT4 *) List_head(stream_list); /* already padded */
    stream_list = List_next(stream_list);

    (*nelts)[basei] = Intlist_head(streamsize_list);
    streamsize_list = Intlist_next(streamsize_list);

    querypos = Intlist_head(querypos_list);
    querypos_list = Intlist_next(querypos_list);

    diagterm = (UINT4) Intlist_head(diagterm_list);
    diagterm_list = Intlist_next(diagterm_list);

    record_heap[basei] = (Record_T *) MALLOC(((*nelts)[basei]) * sizeof(Record_T));
    debug2(printf("PART: Assigning node %d => %d with %d elts (%p)",heapi,basei,(*nelts)[basei],record_heap[basei]));

    for (i = 0; i < (*nelts)[basei]; i++) {
      /* Process in forward order to keep records in order */
      assert(diagonals[i] >= (Univcoord_T) -diagterm);
      all_records[k].diagonal = diagonals[i] + diagterm;
      all_records[k].querypos = querypos;
      record_heap[basei][i] = &(all_records[k]);
      debug2(printf(" %u+%d",diagonals[i],querypos));
      k++;
    }
    debug2(printf("\n"));
  }

  for (heapi = heapsize; heapi >= base; heapi--) {
    /* Put all information into base row */
    diagonals = (UINT4 *) List_head(stream_list); /* already padded */
    stream_list = List_next(stream_list);

    (*nelts)[heapi] = Intlist_head(streamsize_list);
    streamsize_list = Intlist_next(streamsize_list);

    querypos = Intlist_head(querypos_list);
    querypos_list = Intlist_next(querypos_list);

    diagterm = (UINT4) Intlist_head(diagterm_list);
    diagterm_list = Intlist_next(diagterm_list);

    record_heap[heapi] = (Record_T *) MALLOC(((*nelts)[heapi]) * sizeof(Record_T));
    debug2(printf("FULL: Assigning node %d with %d elts (%p)",heapi,(*nelts)[heapi],record_heap[heapi]));

    for (i = 0; i < (*nelts)[heapi]; i++) {
      /* Process in forward order to keep records in order */
      assert(diagonals[i] >= (Univcoord_T) -diagterm);
      all_records[k].diagonal = diagonals[i] + diagterm;
      all_records[k].querypos = querypos;
      record_heap[heapi][i] = &(all_records[k]);
      debug2(printf(" %u+%d",diagonals[i],querypos));
      k++;
    }
    debug2(printf("\n"));
  }

  return record_heap;
}


/* For initializing heap, there are three categories:
   base..(heapsize % PYRAMID_SIZE) + PYRAMID_SIZE: Fill bottom row
   straddling heapsize: Pull down some nodes to bottom row
   heapsize..(2*base - 1): Fill penultimate row */
/* Candidate is any record with more than a single k-mer matching */
Record_T *
Merge_records (int *nelts1, int **candidates, int *ncandidates,
	       List_T stream_list, Intlist_T streamsize_list,
	       Intlist_T querypos_list, Intlist_T diagterm_list,
	       struct Record_T *all_records, int index1part) {
  Intlist_T candidate_list;
  Record_T *result, **record_heap, curr;
  UINT4 **key_heap, *prev_storage, *curr_storage;
  int ptrs[PYRAMID_SIZE];
  int *nelts, nalloc;
  int nstreams, heapsize, base, ancestori, pyramid_start, pyramid_end,
    node_start, node_end;
  int bits;
  int heapi, streami, i, j, k;
  UINT4 diagonal;

  debug(printf("Entered Merge_records\n"));

  if ((nstreams = List_length(stream_list)) == 0) {
    *candidates = (int *) NULL;
    *ncandidates = 0;
    *nelts1 = 0;
    return (Record_T *) NULL;

  } else {
    heapsize = 2*nstreams - 1;	/* also index of last node */
#ifdef HAVE_BUILTIN_CLZ
    bits = 31 - __builtin_clz(heapsize);
#elif defined(HAVE_ASM_BSR)
    asm("bsr %1,%0" : "=r"(bits) : "r"(heapsize));
#else
    bits = 31 - ((heapsize >> 16) ? clz_table[heapsize >> 16] : 16 + clz_table[heapsize]); 
#endif
    base = (1 << bits);
    debug(printf("nstreams %d, heapsize %d, base %d\n",nstreams,heapsize,base));
    record_heap = make_record_heap(&nelts,stream_list,streamsize_list,querypos_list,diagterm_list,
				   nstreams,base,all_records);
  }

  if (nstreams == 1) {
    *nelts1 = nelts[1];	/* Temporary value needed for collapsing of results */
    result = record_heap[1];

  } else {
    key_heap = (UINT4 **) CALLOC(heapsize + PYRAMID_SIZE,sizeof(UINT4 *));

    /* Middle pyramids */
    while (base > PYRAMID_SIZE) {
      for (pyramid_start = 2*base - PYRAMID_SIZE, pyramid_end = 2*base - 1; pyramid_start >= base;
	   pyramid_start -= PYRAMID_SIZE, pyramid_end -= PYRAMID_SIZE) {
	debug2(printf("records: pyramid_start %d, pyramid_end %d, nstreams %d",pyramid_start,pyramid_end,nstreams));
	
	if (pyramid_start > heapsize) {
	  node_start = PARENT(pyramid_start);
	  node_end = PARENT(pyramid_end);
	  debug2(printf(" => node_start %d, node_end %d\n",node_start,node_end));
	} else {
	  node_start = pyramid_start;
	  node_end = pyramid_end;
	}
	debug2(printf("\n"));
	
	/* Allocate memory for the pyramid */
	nalloc = 0;
	/* Have to align start of each entry prev_storage[nalloc] and curr_storage[nalloc], regardless of end padding */
#ifdef HAVE_SSE4_1
	for (heapi = node_start; heapi <= node_end; heapi++) {
	  nalloc += PAD_UINT4(nelts[heapi]);
	}
	prev_storage = (UINT4 *) MALLOC_ALIGN(nalloc * sizeof(UINT4));
	curr_storage = (UINT4 *) MALLOC_ALIGN(nalloc * sizeof(UINT4));
#else
	for (heapi = node_start; heapi <= node_end; heapi++) {
	  nalloc += nelts[heapi];
	}
	prev_storage = (UINT4 *) MALLOC(nalloc * sizeof(UINT4));
	curr_storage = (UINT4 *) MALLOC(nalloc * sizeof(UINT4));
#endif
	
	/* Convert structures to integers (key_heap) */
	nalloc = 0;
	for (heapi = node_start, streami = 0; heapi <= node_end; heapi++, streami++) {
	  debug2(printf("Creating key node %d from %p\n",heapi,record_heap[heapi]));
	  /* key_heap[heapi] = (UINT4 *) MALLOC((npadded + 1) * sizeof(UINT4)); */
	  key_heap[heapi] = &(prev_storage[nalloc]);
	  for (i = 0; i < nelts[heapi]; i++) {
	    key_heap[heapi][i] = (record_heap[heapi][i]->diagonal & KEY_MASK) + streami;
	  }
	  /* Had to align start of each entry prev_storage[nalloc], regardless of end padding */
#ifdef HAVE_SSE4_1
	  nalloc += PAD_UINT4(nelts[heapi]);
#else
	  nalloc += nelts[heapi];
#endif
	}
	
	ancestori = pyramid_merge_prealloc(key_heap,curr_storage,prev_storage,nelts,
					   node_start,node_end);
	
	/* Convert integers to structures */
	record_heap[ancestori] = (Record_T *) MALLOC(nelts[ancestori] * sizeof(Record_T));
	memset(ptrs,0,PYRAMID_SIZE*sizeof(int));
	k = 0;
	for (i = 0; i < nelts[ancestori]; i++) {
	  streami = key_heap[ancestori][i] & ~KEY_MASK;
	  record_heap[ancestori][k++] = record_heap[node_start + streami][ptrs[streami]++];
	}
	
	/* Free base heaps */
	for (heapi = node_start; heapi <= node_end; heapi++) {
	  FREE(record_heap[heapi]);
	}
	
	/* Free key_heap storage */
	FREE_ALIGN(prev_storage);
	FREE_ALIGN(curr_storage);
	
      }
      base = ancestori;
    }
    
    /* Last pyramid */
    pyramid_start = base;
    pyramid_end = 2*base - 1;
    debug2(printf("records: pyramid_start %d, pyramid_end %d, nstreams %d\n",pyramid_start,pyramid_end,nstreams));
    
    /* Allocate memory for the pyramid */
    nalloc = 0;
    
    /* Have to align start of each entry prev_storage[nalloc] and curr_storage[nalloc], regardless of end padding */
#ifdef HAVE_SSE4_1
    for (heapi = pyramid_start; heapi <= pyramid_end; heapi++) {
      nalloc += PAD_UINT4(nelts[heapi]);
    }
    prev_storage = (UINT4 *) MALLOC_ALIGN(nalloc * sizeof(UINT4));
    curr_storage = (UINT4 *) MALLOC_ALIGN(nalloc * sizeof(UINT4));
#else
    for (heapi = pyramid_start; heapi <= pyramid_end; heapi++) {
      nalloc += nelts[heapi];
    }
    prev_storage = (UINT4 *) MALLOC(nalloc * sizeof(UINT4));
    curr_storage = (UINT4 *) MALLOC(nalloc * sizeof(UINT4));
#endif
    
    /* Convert structures to integers (key_heap) */
    nalloc = 0;
    for (heapi = pyramid_start, streami = 0; heapi <= pyramid_end; heapi++, streami++) {
      /* key_heap[heapi] = (UINT4 *) MALLOC((npadded + 1) * sizeof(UINT4)); */
      key_heap[heapi] = &(prev_storage[nalloc]);
      for (i = 0; i < nelts[heapi]; i++) {
	key_heap[heapi][i] = (record_heap[heapi][i]->diagonal & KEY_MASK) + streami;
      }
      /* Had to align start each entry prev_storage[nalloc], regardless of end padding */
#ifdef HAVE_SSE4_1
      nalloc += PAD_UINT4(nelts[heapi]);
#else
      nalloc += nelts[heapi];
#endif
    }
    
    ancestori = pyramid_merge_prealloc(key_heap,curr_storage,prev_storage,
				       nelts,pyramid_start,pyramid_end);
    /* ancestori should be 1 */
    
    /* Convert integers to structures */
    record_heap[ancestori] = (Record_T *) MALLOC(nelts[ancestori] * sizeof(Record_T));
    memset(ptrs,0,PYRAMID_SIZE*sizeof(int));
    k = 0;
    for (i = 0; i < nelts[ancestori]; i++) {
      streami = key_heap[ancestori][i] & ~KEY_MASK;
      record_heap[ancestori][k++] = record_heap[pyramid_start + streami][ptrs[streami]++];
    }
    
    /* Free base heaps (unless pyramid_start == 1, implying that base == 1) */
    for (heapi = pyramid_start, streami = 0; heapi <= pyramid_end; heapi++, streami++) {
      FREE(record_heap[pyramid_start + streami]);
    }
    
    /* Free key_heap storage */
    FREE_ALIGN(prev_storage);
    FREE_ALIGN(curr_storage);
    
    
    *nelts1 = nelts[1];	/* Temporary value, needed for final insertion sort and collapsing of results*/
    result = record_heap[1];
    
    /* Final insertion sort to correct for truncation of keys */
    for (j = 1; j < *nelts1; j++) {
      curr = result[j];
      i = j - 1;
      /* For a stable merge sort, is the second condition possible?  Yes */
      while (i >= 0 && (result[i]->diagonal > curr->diagonal ||
			(result[i]->diagonal == curr->diagonal &&
			 result[i]->querypos > curr->querypos))) {
	assert(result[i]->diagonal >= curr->diagonal);
	result[i+1] = result[i];
	i--;
      }
      result[i+1] = curr;
    }

    FREE(key_heap);
  }

  FREE(nelts);
  FREE(record_heap);


#ifdef DEBUG0
  printf("Merge_records returning result of length %d\n",*nelts1);
  for (i = 0; i < *nelts1; i++) {
    printf("%u %d\n",result[i]->diagonal,result[i]->querypos);
  }
#endif

  /* Collapse records */
  k = 0;
  i = 0;
  candidate_list = (Intlist_T) NULL;
  while (i < *nelts1) {
    diagonal = result[i]->diagonal;
    j = i + 1;
    while (j < *nelts1 && result[j]->diagonal == diagonal) {
      j++;
    }

    result[k]->diagonal = diagonal;
    result[k]->querypos = result[i]->querypos;
    result[k]->queryend = result[j-1]->querypos + index1part;
    if (j - 1 > i) {
      candidate_list = Intlist_push(candidate_list,k);
    }
    k++;

    i = j;
  }

  *nelts1 = k;

#ifdef DEBUG0
  printf("After collapsing, Merge_records returning result of length %d\n",*nelts1);
  for (i = 0; i < *nelts1; i++) {
    printf("%u %d..%d\n",result[i]->diagonal,result[i]->querypos,result[i]->queryend);
  }
#endif

  candidate_list = Intlist_reverse(candidate_list);
  *candidates = Intlist_to_array(&(*ncandidates),candidate_list);
  Intlist_free(&candidate_list);

  return result;
}

