/* $Id: segment-search.h 214025 2018-03-05 07:06:43Z twu $ */
#ifndef SEGMENT_SEARCH_INCLUDED
#define SEGMENT_SEARCH_INCLUDED
#ifdef HAVE_CONFIG_H
#include <config.h>		/* For HAVE_64_BIT */
#endif

#include "bool.h"
#include "types.h"
#include "mode.h"
#include "genomicpos.h"

#include "shortread.h"
#include "iit-read-univ.h"
#include "genome.h"

#include "stage1hr.h"
#include "resulthr.h"		/* For Pairtype_T */

#ifdef PMAP
#include "oligoindex_pmap.h"
#else
#include "oligoindex_hr.h"
#endif
#include "pairpool.h"
#include "diagpool.h"
#include "cellpool.h"
#include "dynprog.h"
#ifdef LARGE_GENOMES
#include "merge-heap-records.h"
#else
#include "merge-simd-records.h"
#endif


extern Record_T *
Segment_identify (int *nsegments, struct Record_T **all_records,
		  int **candidates, int *ncandidates,
#ifdef LARGE_GENOMES
		  unsigned char **positions_high,
#endif
		  UINT4 **positions, int *npositions, bool *validp,
		  int querylength, int sizelimit, bool plusp);

extern void
Segment_search_filtered (int *found_score, List_T *plus_hits, List_T *minus_hits,
			 List_T *plus_complete_paths, List_T *minus_complete_paths,

			 Record_T *plus_records, int plus_nrecords, Intlist_T plus_anchors,
			 Record_T *minus_records, int minus_nrecords, Intlist_T minus_anchors,

			 char *queryuc_ptr, char *queryrc, int querylength,
			 Compress_T query_compress_fwd, Compress_T query_compress_rev,
			 int genestrand, bool paired_end_p, int level);

extern void
Segment_search_all (int *found_score, List_T *plus_hits, List_T *minus_hits,
		    List_T *plus_complete_paths, List_T *minus_complete_paths,

		    Record_T *plus_records, int plus_nrecords, 
		    int *plus_candidates, int plus_ncandidates,
		    Record_T *minus_records, int minus_nrecords,
		    int *minus_candidates, int minus_ncandidates,

		    char *queryuc_ptr, char *queryrc, int querylength,
		    Compress_T query_compress_fwd, Compress_T query_compress_rev,
		    int genestrand, bool paired_end_p, int level);

extern void
Segment_search_setup (int index1part_in, int index1interval_in,
		      int max_anchors_in, Univ_IIT_T chromosome_iit_in, int nchromosomes_in,
		      int circular_typeint_in, Mode_T mode_in,
		      Univcoord_T *splicesites_in, Splicetype_T *splicetypes_in,
		      Chrpos_T *splicedists_in, int nsplicesites_in,
		      int max_middle_deletions, Chrpos_T shortsplicedist_in,
		      int maxpeelback_in);

extern void
Segment_search_cleanup ();

#endif

