static char rcsid[] = "$I$";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "univdiag.h"
#include "univdiagdef.h"

#include <stdio.h>
#include <stdlib.h>
#include "mem.h"
#include "assert.h"


#define T Univdiag_T


T
Univdiag_new (int querystart, int queryend, Univcoord_T univdiagonal) {
  T new = (T) MALLOC(sizeof(*new));

  assert(queryend > querystart);

  new->univdiagonal = univdiagonal;
  new->querystart = querystart;
  new->queryend = queryend;
  new->nconsecutive = queryend - querystart + 1;
  new->nmismatches_known_p = false;

  new->intscore = -1;		/* Unknown number of matches */
  new->nlinked = 0;
  new->prev = (Univdiag_T) NULL;

  return new;
}


T
Univdiag_copy (T this) {
  return Univdiag_new(this->querystart,this->queryend,this->univdiagonal);
}


List_T
Univdiag_copy_list (List_T list) {
  List_T new = NULL, p;
  T this;

  for (p = list; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    new = List_push(new,(void *) Univdiag_new(this->querystart,this->queryend,this->univdiagonal));
  }
  
  return List_reverse(new);
}
  


#if 0
T
Univdiag_new_fillin (int querystart, int queryend, int indexsize, Univcoord_T univdiagonal) {
  T new = (T) MALLOC(sizeof(*new));

  new->univdiagonal = univdiagonal;
  new->querystart = querystart;
  new->queryend = queryend + indexsize - 1;
  new->nconsecutive = new->queryend - querystart + 1;
  new->nmismatches_known_p = false;

  new->intscore = -1;
  new->nlinked = 0;
  new->prev = (Univdiag_T) NULL;

  return new;
}
#endif


void
Univdiag_free (T *old) {
  FREE(*old);
  return;
}

void
Univdiag_gc (List_T *list) {
  T univdiagonal;
  List_T p;

  for (p = *list; p != NULL; p = List_next(p)) {
    univdiagonal = (T) List_head(p);
    FREE(univdiagonal);
  }
  List_free(&(*list));
  return;
}

void
Univdiag_list_gc (List_T *paths) {
  List_T path, p;

  for (p = *paths; p != NULL; p = List_next(p)) {
    path = (List_T) List_head(p);
    Univdiag_gc(&path);
  }
  List_free(&(*paths));
  return;
}


int
Univdiag_list_length (List_T path) {
  int length = 0;
  T this;
  List_T p;

  for (p = path; p != NULL; p = List_next(p)) {
    this = (T) List_head(p);
    length += this->queryend - this->querystart + 1;
  }
  return length;
}


int
Univdiag_ascending_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->querystart < y->querystart) {
    return -1;
  } else if (y->querystart < x->querystart) {
    return +1;
  } else if (x->queryend < y->queryend) {
    return -1;
  } else if (y->queryend < x->queryend) {
    return +1;
  } else if (x->univdiagonal < y->univdiagonal) {
    return -1;
  } else if (y->univdiagonal < x->univdiagonal) {
    return +1;
  } else {
    return 0;
  }
}


int
Univdiag_descending_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->querystart > y->querystart) {
    return -1;
  } else if (y->querystart > x->querystart) {
    return +1;
  } else if (x->queryend > y->queryend) {
    return -1;
  } else if (y->queryend > x->queryend) {
    return +1;
  } else if (x->univdiagonal > y->univdiagonal) {
    return -1;
  } else if (y->univdiagonal > x->univdiagonal) {
    return +1;
  } else {
    return 0;
  }
}



int
Univdiag_diagonal_cmp (const void *a, const void *b) {
  T x = * (T *) a;
  T y = * (T *) b;

  if (x->univdiagonal < y->univdiagonal) {
    return -1;
  } else if (y->univdiagonal < x->univdiagonal) {
    return +1;
  } else if (x->querystart < y->querystart) {
    return -1;
  } else if (y->querystart < x->querystart) {
    return +1;
  } else if (x->queryend < y->queryend) {
    return -1;
  } else if (y->queryend < x->queryend) {
    return +1;
  } else {
    return 0;
  }
}
