static char rcsid[] = "$Id: path-solve.c 215952 2018-07-06 05:58:20Z twu $";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifndef HAVE_MEMCPY
#define memcpy(d,s,n) bcopy((s),(d),(n))
#endif

#include "path-solve.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>		/* For munmap */

#include "mem.h"
#include "assert.h"
#include "access.h"

#include "dynprog.h"		/* For Dynprog_consistent_p */
#include "comp.h"
#include "splice.h"
#include "indel.h"
#include "intron.h"
#include "maxent_hr.h"

#include "univdiagdef.h"
#include "substring.h"
#include "junction.h"

#include "stage2.h"
#include "stage3.h"
#include "stage3hr.h"
#include "sedgesort.h"


#define MIN_INTRONLEN 9
#define USE_LOCALDB 1
#define MAX_GMAP_PATHS 10


/* If we turn on EXTEND_TO_ENDS, then substrings will be judged as poor, and we will not get any terminal alignments */
/* But if we turn off EXTEND_TO_ENDS, then fusion alignments will be judged as acceptable, and we will not get any distant alignments */
/* The fix is to turn off EXTEND_TO_ENDS and make sure found_score is not adjusted for splice site ends */
/* #define EXTEND_TO_ENDS 1 */

/* known splicing */
#ifdef DEBUG4S
#define debug4s(x) x
#else
#define debug4s(x)
#endif

/* binary_search */
#ifdef DEBUG10
#define debug10(x) x
#else
#define debug10(x)
#endif

/* best_path_genome */
#ifdef DEBUG13
#define debug13(x) x
#else
#define debug13(x)
#endif

/* via gmap */
#ifdef DEBUG14
#define debug14(x) x
#else
#define debug14(x)
#endif

  
static Genome_T genomebits;
static Genome_T genomebits_alt;

static bool *circularp;

static int index1part;
static int index1interval;
static int local1part;
static Localdb_T localdb;
static Localdb_T localdb2;

static int endtrim_allowed = 8;

static Chrpos_T min_intronlength;
static Chrpos_T max_deletionlen;

/* Splicing */
static bool novelsplicingp;
static Univcoord_T *splicesites;
static Splicetype_T *splicetypes;
static Chrpos_T *splicedists;
static int nsplicesites;

/* For localdb */
static int max_insertionlen;
static Chrpos_T overall_end_distance_genome;

static Univ_IIT_T chromosome_iit;
static int circular_typeint;
static int gmap_mode;
static int maxpeelback;


static int
binary_search_univcoord (int lowi, int highi, Univcoord_T *positions, Univcoord_T goal) {
  int middlei;

  debug10(printf("entered binary search with lowi=%d, highi=%d, goal=%u\n",lowi,highi,goal));

  while (lowi < highi) {
    middlei = lowi + ((highi - lowi) / 2);
    debug10(printf("  binary: %d:%u %d:%u %d:%u   vs. %u\n",
		   lowi,positions[lowi],middlei,positions[middlei],
		   highi-1,positions[highi-1],goal));
    if (goal < positions[middlei]) {
      highi = middlei;
    } else if (goal > positions[middlei]) {
      lowi = middlei + 1;
    } else {
      debug10(printf("binary search returns %d\n",middlei));
      return middlei;
    }
  }

  debug10(printf("binary search returns %d\n",highi));
  return highi;
}


#define add_bounded(x,plusterm,highbound) ((x + (plusterm) >= highbound) ? (highbound - 1) : x + (plusterm))
#define subtract_bounded(x,minusterm,lowbound) ((x < lowbound + (minusterm)) ? lowbound : x - (minusterm))


#if 0
static List_T
resolve_end_diagonals (Univcoord_T *right_univdiagonal_sense, int *right_splicepos_sense,
		       int *right_nmismatches_sense, Junction_T *right_junction_sense,
		       Substring_T *right_alts_sense,
		       Univcoord_T *right_univdiagonal_antisense, int *right_splicepos_antisense,
		       int *right_nmismatches_antisense, Junction_T *right_junction_antisense,
		       Substring_T *right_alts_antisense,

		       Univcoord_T *left_univdiagonal_sense, int *left_splicepos_sense,
		       int *left_nmismatches_sense, Junction_T *left_junction_sense,
		       Substring_T *left_alts_sense,
		       Univcoord_T *left_univdiagonal_antisense, int *left_splicepos_antisense,
		       int *left_nmismatches_antisense, Junction_T *left_junction_antisense,
		       Substring_T *left_alts_antisense,

		       Univdiag_T middle_diagonal, List_T right_diagonals, List_T left_diagonals,

		       int querylength, Compress_T query_compress,
		       Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
		       bool plusp, int genestrand, int max_mismatches_allowed) {
  List_T complete_path = NULL, right_paths, left_paths;
  List_T p;

  List_T right_samechr_diagonals, left_samechr_diagonals;
  List_T diagonal_path, alts_path;
  Univdiag_T diagonal, common_diagonal, prev_diagonal, right_indel_diagonal = NULL, left_indel_diagonal = NULL;
  int ndiagonals;

  Univcoord_T left, prev_left, mindiff, diff;

  Chrpos_T splice_distance;
  int querystart, queryend, splice_pos;
  int best_knowni_i, best_knowni_j, best_nmismatches_i, best_nmismatches_j;
  double best_prob_i, best_prob_j, donor_prob, acceptor_prob;
  bool splice_found_p;

  Chrpos_T first_dist_sense, second_dist_sense, first_dist_antisense, second_dist_antisense;
  double first_prob_sense, second_prob_sense, first_prob_antisense, second_prob_antisense;
  int firsti_sense, secondi_sense, firsti_antisense, secondi_antisense;
  int sensei, antisensei;

  Intlist_T right_endpoints_sense = NULL, right_endpoints_antisense = NULL, 
    right_queryends_sense = NULL, right_queryends_antisense = NULL;
  Univcoordlist_T right_alts_coords_sense = NULL, right_alts_coords_antisense = NULL;
  Intlist_T right_alts_knowni_sense = NULL, right_alts_knowni_antisense = NULL;
  Intlist_T right_alts_nmismatchesi_sense = NULL, right_alts_nmismatchesi_antisense = NULL,
    right_alts_nmismatchesj_sense = NULL, right_alts_nmismatchesj_antisense = NULL;
  Doublelist_T right_alts_probsi_sense = NULL, right_alts_probsi_antisense = NULL,
    right_alts_probsj_sense = NULL, right_alts_probsj_antisense = NULL;

  Intlist_T left_endpoints_sense = NULL, left_endpoints_antisense = NULL, 
    left_querystarts_sense = NULL, left_querystarts_antisense = NULL;
  Univcoordlist_T left_alts_coords_sense = NULL, left_alts_coords_antisense = NULL;
  Intlist_T left_alts_knowni_sense = NULL, left_alts_knowni_antisense = NULL;
  Intlist_T left_alts_nmismatchesi_sense = NULL, left_alts_nmismatchesi_antisense = NULL,
    left_alts_nmismatchesj_sense = NULL, left_alts_nmismatchesj_antisense = NULL;
  Doublelist_T left_alts_probsi_sense = NULL, left_alts_probsi_antisense = NULL,
    left_alts_probsj_sense = NULL, left_alts_probsj_antisense = NULL;

  Spliceinfo_T spliceinfo;
  int j;

  spliceinfo = Spliceinfo_new(querylength);

  debug13(printf("***Entered resolve_end_diagonals\n"));


  *right_univdiagonal_sense = *right_univdiagonal_antisense = 
    *left_univdiagonal_sense = *left_univdiagonal_antisense = (Univcoord_T) 0;
  *right_splicepos_sense = *right_splicepos_antisense = 
    *left_splicepos_sense = *left_splicepos_antisense = 0;
  *right_nmismatches_sense = *right_nmismatches_antisense = 
    *left_nmismatches_sense = *left_nmismatches_antisense = -1; /* Need to recalculate */
  *right_alts_sense = *right_alts_antisense = 
    *left_alts_sense = *left_alts_antisense = (Substring_T) NULL;


  right_samechr_diagonals = (List_T) NULL;
  for (p = right_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (diagonal->univdiagonal >= chrhigh) {
      /* Skip.  Diagonal is in next chromosome */
    } else {
      right_samechr_diagonals = List_push(right_samechr_diagonals,(void *) diagonal);
    }
  }

  left_samechr_diagonals = (List_T) NULL;
  for (p = left_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (diagonal->univdiagonal + querylength <= chroffset) {
      /* Skip.  Diagonal is in previous chromosome */
    } else {
      left_samechr_diagonals = List_push(left_samechr_diagonals,(void *) diagonal);
    }
  }


  /* A3.  Traceback for dynamic programming */
  right_paths = (List_T) NULL;
  debug13(printf("On right, have %d right_diagonals\n",List_length(right_samechr_diagonals)));
  if ((ndiagonals = List_length(right_samechr_diagonals)) == 0) {
    common_diagonal = (Univdiag_T) NULL;

    /* querystart = middle_diagonal->queryend + 1; */
    left = middle_diagonal->univdiagonal;

  } else if (ndiagonals == 1) {
    common_diagonal = (Univdiag_T) List_head(right_samechr_diagonals);

    /* querystart = common_diagonal->queryend + 1; */
    left = common_diagonal->univdiagonal;

  } else {
    debug13(printf("3 Multiple (%d) right diagonals\n",ndiagonals));

    /* Distinguish between common and divergent diagonals */
    for (p = right_samechr_diagonals; p != NULL; p = List_next(p)) {
      diagonal = (Univdiag_T) List_head(p);
      debug13(printf("Path:\n"));
      while (diagonal != NULL) {
	debug13(printf("Diagonal: query %d..%d, diagonal %u\n",
		       diagonal->querystart,diagonal->queryend,diagonal->univdiagonal));
	diagonal->nlinked += 1;
	diagonal = diagonal->prev;
      }
    }

    /* Handle divergent diagonals */
    /* Now that we are running oligoindex, we may need to obtain only the last common_diagonal */
    for (p = right_samechr_diagonals; p != NULL; p = List_next(p)) {
      alts_path = (List_T) NULL;
      diagonal = (Univdiag_T) List_head(p);
      while (diagonal != NULL && diagonal->nlinked < ndiagonals) {
	alts_path = List_push(alts_path,(void *) diagonal);
	diagonal = diagonal->prev;
      }
      right_paths = List_push(right_paths,(void *) alts_path);

      common_diagonal = diagonal; /* Last elt on prev path.  Save for later */
      /* printf("Common diagonal is now %p\n",common_diagonal); */
    }
#ifdef DEBUG13
    if (common_diagonal != NULL) {
      printf("Common diagonal is: query %d..%d, diagonal %u\n",
	     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal);
    }
#endif


    if (common_diagonal == NULL) {
      /* All paths connect directly to the middle diagonal, so there is no common diagonal */
      prev_diagonal = middle_diagonal;
      /* querystart = middle_diagonal_3->queryend + 1; */
      prev_left = middle_diagonal->univdiagonal;
    } else {
      prev_diagonal = common_diagonal;
      /* querystart = common_diagonal->queryend + 1; */
      prev_left = common_diagonal->univdiagonal;
    }

    /* Distinguish right paths by looking for smallest indel (which wins) or splicing */
    debug13(printf("Have %d right_paths.  Distinguish by looking for indels\n",List_length(right_paths)));
    for (p = right_paths; p != NULL; p = List_next(p)) {
      alts_path = (List_T) List_head(p);
      diagonal = (Univdiag_T) List_head(alts_path);
      left = diagonal->univdiagonal;
      debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));

      if (left < prev_left) {
	/* Insertion */
	debug13(printf("Found insertion\n"));
	if (right_indel_diagonal == NULL) {
	  right_indel_diagonal = diagonal;
	  mindiff = prev_left - left;
	} else if ((diff = prev_left - left) < mindiff) {
	  right_indel_diagonal = diagonal;
	  mindiff = diff;
	}

      } else if (left - prev_left < MIN_INTRONLEN) {
	/* Deletion */
	debug13(printf("Found deletion\n"));
	if (right_indel_diagonal == NULL) {
	  right_indel_diagonal = diagonal;
	  mindiff = left - prev_left;
	} else if ((diff = left - prev_left) < mindiff) {
	  right_indel_diagonal = diagonal;
	  mindiff = diff;
	}
      }
    }

    if (right_indel_diagonal != NULL) {
      /* Push onto middle path later */
      /* querystart = right_indel_diagonal->queryend + 1; */
      left = right_indel_diagonal->univdiagonal;

    } else {
      debug13(printf("Still have %d right_paths.  Distinguish by looking for best splice\n",List_length(right_paths)));
      first_dist_sense = second_dist_sense = 0;
      first_prob_sense = second_prob_sense = 0.0;
      firsti_sense = secondi_sense = -1;
      first_dist_antisense = second_dist_antisense = 0;
      first_prob_antisense = second_prob_antisense = 0.0;
      firsti_antisense = secondi_antisense = -1;
      sensei = antisensei = 0;

      splice_found_p = false;
      for (p = right_paths; p != NULL; p = List_next(p)) {
	alts_path = (List_T) List_head(p);
	diagonal = (Univdiag_T) List_head(alts_path);
	left = diagonal->univdiagonal;

	spliceinfo->segmenti_donor_nknown = spliceinfo->segmenti_antiacceptor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(prev_left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,prev_left);
	  while (j < nsplicesites && splicesites[j] < prev_left + querylength) {
	    if (splicetypes[j] == DONOR) {
	      debug4s(printf("Setting known donor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_donor_knowni[spliceinfo->segmenti_donor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIACCEPTOR) {
	      debug4s(printf("Setting known antiacceptor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_antiacceptor_knowni[spliceinfo->segmenti_antiacceptor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = querylength + 100;
	spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = querylength + 100;

	spliceinfo->segmentj_acceptor_nknown = spliceinfo->segmentj_antidonor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,left);
	  while (j < nsplicesites && splicesites[j] < left + querylength) {
	    if (splicetypes[j] == ACCEPTOR) {
	      debug4s(printf("Setting known acceptor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_acceptor_knowni[spliceinfo->segmentj_acceptor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIDONOR) {
	      debug4s(printf("Setting known antidonor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_antidonor_knowni[spliceinfo->segmentj_antidonor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = querylength + 100;
	spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = querylength + 100;

	/* splice_distance = left - prev_left; */
#if 0
	max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
	debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
	if (prev_diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
	if (diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
#endif

	if ((splice_pos = Splice_resolve_sense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
					       &best_prob_i,&best_prob_j,
					       /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
					       prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
					       spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found sense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  splice_found_p = true;
	  if (best_prob_i + best_prob_j > first_prob_sense) {
	    second_dist_sense = first_dist_sense;
	    second_prob_sense = first_prob_sense;
	    secondi_sense = firsti_sense;
	    first_dist_sense = left - prev_left;
	    first_prob_sense = best_prob_i + best_prob_j;
	    firsti_sense = sensei;
	  } else if (best_prob_i + best_prob_j > second_prob_sense) {
	    second_dist_sense = left - prev_left;
	    second_prob_sense = best_prob_i + best_prob_j;
	    secondi_sense = sensei;
	  }

	  right_endpoints_sense = Intlist_push(right_endpoints_sense,splice_pos);
	  right_queryends_sense = Intlist_push(right_queryends_sense,diagonal->queryend/*+ 1*/);
	  right_alts_coords_sense = Univcoordlist_push(right_alts_coords_sense,left + splice_pos);
	  right_alts_knowni_sense = Intlist_push(right_alts_knowni_sense,best_knowni_j);
	  right_alts_nmismatchesi_sense = Intlist_push(right_alts_nmismatchesi_sense,best_nmismatches_i);
	  right_alts_nmismatchesj_sense = Intlist_push(right_alts_nmismatchesj_sense,best_nmismatches_j);
	  right_alts_probsi_sense = Doublelist_push(right_alts_probsi_sense,best_prob_i);
	  right_alts_probsj_sense = Doublelist_push(right_alts_probsj_sense,best_prob_j);
	  sensei++;
	}

	if ((splice_pos = Splice_resolve_antisense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
						   &best_prob_i,&best_prob_j,
						   /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
						   prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
						   spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found antisense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  splice_found_p = true;
	  if (best_prob_i + best_prob_j > first_prob_antisense) {
	    second_dist_antisense = first_dist_antisense;
	    second_prob_antisense = first_prob_antisense;
	    secondi_antisense = firsti_antisense;
	    first_dist_antisense = left - prev_left;
	    first_prob_antisense = best_prob_i + best_prob_j;
	    firsti_antisense = antisensei;
	  } else if (best_prob_i + best_prob_j > second_prob_antisense) {
	    second_dist_antisense = left - prev_left;
	    second_prob_antisense = best_prob_i + best_prob_j;
	    secondi_antisense = antisensei;
	  }

	  right_endpoints_antisense = Intlist_push(right_endpoints_antisense,splice_pos);
	  right_queryends_antisense = Intlist_push(right_queryends_antisense,diagonal->queryend/*+ 1*/);
	  right_alts_coords_antisense = Univcoordlist_push(right_alts_coords_antisense,left + splice_pos);
	  right_alts_knowni_antisense = Intlist_push(right_alts_knowni_antisense,best_knowni_j);
	  right_alts_nmismatchesi_antisense = Intlist_push(right_alts_nmismatchesi_antisense,best_nmismatches_i);
	  right_alts_nmismatchesj_antisense = Intlist_push(right_alts_nmismatchesj_antisense,best_nmismatches_j);
	  right_alts_probsi_antisense = Doublelist_push(right_alts_probsi_antisense,best_prob_i);
	  right_alts_probsj_antisense = Doublelist_push(right_alts_probsj_antisense,best_prob_j);
	  antisensei++;
	}
      }


      if (Intlist_length(right_endpoints_sense) > 1) {
	if (first_dist_sense < second_dist_sense/2) {
	  debug13(printf("first dist sense %u is significantly shorter than second dist sense %u.  Keeping %d from end\n",
			 first_dist_sense,second_dist_sense,firsti_sense));
	  firsti_sense = Intlist_length(right_endpoints_sense) - 1 - firsti_sense; /* Because we don't reverse lists */
	  right_endpoints_sense = Intlist_keep_one(right_endpoints_sense,firsti_sense);
	  right_queryends_sense = Intlist_keep_one(right_queryends_sense,firsti_sense);
	  right_alts_coords_sense = Univcoordlist_keep_one(right_alts_coords_sense,firsti_sense);
	  right_alts_knowni_sense = Intlist_keep_one(right_alts_knowni_sense,firsti_sense);
	  right_alts_nmismatchesi_sense = Intlist_keep_one(right_alts_nmismatchesi_sense,firsti_sense);
	  right_alts_nmismatchesj_sense = Intlist_keep_one(right_alts_nmismatchesj_sense,firsti_sense);
	  right_alts_probsi_sense = Doublelist_keep_one(right_alts_probsi_sense,firsti_sense);
	  right_alts_probsj_sense = Doublelist_keep_one(right_alts_probsj_sense,firsti_sense);
	}
      }

      if (Intlist_length(right_endpoints_antisense) > 1) {
	if (first_dist_antisense < second_dist_antisense/2) {
	  debug13(printf("first dist antisense %u is significantly shorter than second dist antisense %u.  Keeping %d from end\n",
			 first_dist_antisense,second_dist_antisense,firsti_antisense));
	  firsti_antisense = Intlist_length(right_endpoints_antisense) - 1 - firsti_antisense; /* Because we don't reverse lists */
	  right_endpoints_antisense = Intlist_keep_one(right_endpoints_antisense,firsti_antisense);
	  right_queryends_antisense = Intlist_keep_one(right_queryends_antisense,firsti_antisense);
	  right_alts_coords_antisense = Univcoordlist_keep_one(right_alts_coords_antisense,firsti_antisense);
	  right_alts_knowni_antisense = Intlist_keep_one(right_alts_knowni_antisense,firsti_antisense);
	  right_alts_nmismatchesi_antisense = Intlist_keep_one(right_alts_nmismatchesi_antisense,firsti_antisense);
	  right_alts_nmismatchesj_antisense = Intlist_keep_one(right_alts_nmismatchesj_antisense,firsti_antisense);
	  right_alts_probsi_antisense = Doublelist_keep_one(right_alts_probsi_antisense,firsti_antisense);
	  right_alts_probsj_antisense = Doublelist_keep_one(right_alts_probsj_antisense,firsti_antisense);
	}
      } 

      if (splice_found_p == false) {
	debug13(printf("Could not find a splice on right.  Probably indel+splice.  Taking the closest right diagonal instead\n"));
	for (p = right_paths; p != NULL; p = List_next(p)) {
	  alts_path = (List_T) List_head(p);
	  diagonal = (Univdiag_T) List_head(alts_path);
	  left = diagonal->univdiagonal;
	  debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));
	  if (right_indel_diagonal == NULL) {
	    right_indel_diagonal = diagonal;
	    mindiff = left - prev_left;
	  } else if ((diff = left - prev_left) < mindiff) {
	    right_indel_diagonal = diagonal;
	    mindiff = diff;
	  }
	}

      } else {
	
	/* SENSE */
	if (right_endpoints_sense == NULL) {
	  /* Skip.  Antisense splice found */

	} else if (Intlist_length(right_endpoints_sense) == 1) {
	  /* Only one sense splice on right */
	  debug13(printf("Only one sense splice on right, which should have %d and %d mismatches\n",
			 Intlist_head(right_alts_nmismatchesi_sense),Intlist_head(right_alts_nmismatchesj_sense)));
	  splice_pos = Intlist_head(right_endpoints_sense);
	  left = Univcoordlist_head(right_alts_coords_sense) - splice_pos;

	  *right_univdiagonal_sense = left;
	  *right_splicepos_sense = splice_pos;
	  *right_nmismatches_sense = Intlist_head(right_alts_nmismatchesj_sense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = left - prev_left;
	  if (plusp == true) {
	    donor_prob = Doublelist_head(right_alts_probsi_sense);
	    acceptor_prob = Doublelist_head(right_alts_probsj_sense);
	  } else {
	    acceptor_prob = Doublelist_head(right_alts_probsi_sense);
	    donor_prob = Doublelist_head(right_alts_probsj_sense);
	  }
	  debug13(printf("Creating a right_junction_sense with splice distance %u (left %u - prev_left %u)\n",
			 splice_distance,left,prev_left));
	  *right_junction_sense = Junction_new_splice(splice_distance,SENSE_FORWARD,donor_prob,acceptor_prob);
#endif
	  
	} else if (Intlist_vary(right_endpoints_sense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */
	  
	} else {
	  /* Alternative substring on right */
	  debug13(printf("Alternative substring on right\n"));
	  splice_pos = Intlist_head(right_endpoints_sense);
	  queryend = Intlist_head(right_queryends_sense); /* Should all be the same */

	  if (plusp == true) {
	    *right_alts_sense = Substring_new_alts_A(/*querystart*/splice_pos,queryend,splice_pos,querylength,
						     plusp,genestrand,query_compress,
						     chrnum,chroffset,chrhigh,chrlength,
						     right_alts_coords_sense,right_alts_knowni_sense,
						     right_alts_nmismatchesj_sense,right_alts_probsj_sense,
						     /*alts_common_prob*/Doublelist_head(right_alts_probsi_sense),
						     /*substring1p*/false);
	  } else {
	    *right_alts_sense = Substring_new_alts_D(/*querystart*/querylength - queryend,
						       /*queryend*/querylength - splice_pos,splice_pos,querylength,
						     plusp,genestrand,query_compress,
						     chrnum,chroffset,chrhigh,chrlength,
						     right_alts_coords_sense,right_alts_knowni_sense,
						     right_alts_nmismatchesj_sense,right_alts_probsj_sense,
						     /*alts_common_prob*/Doublelist_head(right_alts_probsi_sense),
						     /*substring1p*/true);
	  }
	}


	/* ANTISENSE */
	if (right_endpoints_antisense == NULL) {
	  /* Skip.  Sense splice found */

        } else if (Intlist_length(right_endpoints_antisense) == 1) {
	  /* Only one antisense splice on right */
	  debug13(printf("Only one antisense splice on right, which should have %d and %d mismatches\n",
			 Intlist_head(right_alts_nmismatchesi_antisense),Intlist_head(right_alts_nmismatchesj_antisense)));
	  splice_pos = Intlist_head(right_endpoints_antisense);
	  left = Univcoordlist_head(right_alts_coords_antisense) - splice_pos;

	  *right_univdiagonal_antisense = left;
	  *right_splicepos_antisense = splice_pos;
	  *right_nmismatches_antisense = Intlist_head(right_alts_nmismatchesj_antisense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = left - prev_left;
	  if (plusp == true) {
	    acceptor_prob = Doublelist_head(right_alts_probsi_antisense);
	    donor_prob = Doublelist_head(right_alts_probsj_antisense);
	  } else {
	    donor_prob = Doublelist_head(right_alts_probsi_antisense);
	    acceptor_prob = Doublelist_head(right_alts_probsj_antisense);
	  }
	  debug13(printf("Creating a right_junction_antisense with splice distance %u (left %u - prev_left %u)\n",
			 splice_distance,left,prev_left));
	  *right_junction_antisense = Junction_new_splice(splice_distance,SENSE_ANTI,donor_prob,acceptor_prob);
#endif

	} else if (Intlist_vary(right_endpoints_antisense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */

	} else {
	  /* Alternative substring on right */
	  debug13(printf("Alternative substring on right\n"));
	  splice_pos = Intlist_head(right_endpoints_antisense);
	  queryend = Intlist_head(right_queryends_antisense); /* Should all be the same */
	  
	  if (plusp == true) {
	    *right_alts_antisense = Substring_new_alts_D(/*querystart*/splice_pos,queryend,splice_pos,querylength,
							 plusp,genestrand,query_compress,
							 chrnum,chroffset,chrhigh,chrlength,
							 right_alts_coords_antisense,right_alts_knowni_antisense,
							 right_alts_nmismatchesj_antisense,right_alts_probsj_antisense,
							 /*alts_common_prob*/Doublelist_head(right_alts_probsi_antisense),
							 /*substring1p*/false);
	  } else {
	    *right_alts_antisense = Substring_new_alts_A(/*querystart*/querylength - queryend,
							 /*queryend*/querylength - splice_pos,splice_pos,querylength,
							 plusp,genestrand,query_compress,
							 chrnum,chroffset,chrhigh,chrlength,
							 right_alts_coords_antisense,right_alts_knowni_antisense,
							 right_alts_nmismatchesj_antisense,right_alts_probsj_antisense,
							 /*alts_common_prob*/Doublelist_head(right_alts_probsi_antisense),
							 /*substring1p*/true);
	  }
	}
      }
    }
  }


  complete_path = (List_T) NULL;

  if (right_indel_diagonal != NULL) {
    debug13(printf("Pushing right indel diagonal onto middle: query %d..%d, diagonal %u\n",
		   right_indel_diagonal->querystart,right_indel_diagonal->queryend,right_indel_diagonal->univdiagonal - chroffset));
    complete_path = List_push(complete_path,(void *) Univdiag_copy(right_indel_diagonal));
  }


  /* A5. Process common diagonal from right */
  while (common_diagonal != NULL) {
    if (middle_diagonal->univdiagonal < chroffset + chrlength && common_diagonal->univdiagonal > chroffset + chrlength) {
      debug13(printf("Cannot handle common diagonal across circular origin: query %d..%d, diagonal %u\n",
		     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal - chroffset));
    } else {
      complete_path = List_push(complete_path,(void *) Univdiag_copy(common_diagonal));
      debug13(printf("Pushing common diagonal onto middle: query %d..%d, diagonal %u\n",
		     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal - chroffset));
    }
    common_diagonal = common_diagonal->prev;
  }

  /* B. Process original middle path */
  complete_path = List_push(complete_path,(void *) Univdiag_copy(middle_diagonal));


  /* C3.  Traceback for dynamic programming */
  left_paths = (List_T) NULL;
  debug13(printf("On left, have %d left_samechr_diagonals (resolve_end_diagonals)\n",List_length(left_samechr_diagonals)));
  if ((ndiagonals = List_length(left_samechr_diagonals)) == 0) {
    common_diagonal = (Univdiag_T) NULL;

    /* queryend = middle_diagonal->querystart; */
    left = middle_diagonal->univdiagonal;

  } else if (ndiagonals == 1) {
    common_diagonal = (Univdiag_T) List_head(left_samechr_diagonals);

    /* queryend = common_diagonal->querystart; */
    left = common_diagonal->univdiagonal;

  } else {
    debug13(printf("4 Multiple (%d) left diagonals\n",ndiagonals));

    /* Distinguish between common and divergent diagonals */
    for (p = left_samechr_diagonals; p != NULL; p = List_next(p)) {
      diagonal = (Univdiag_T) List_head(p);
      debug13(printf("Path:\n"));
      while (diagonal != NULL) {
	debug13(printf("Diagonal: query %d..%d, diagonal %u\n",
		       diagonal->querystart,diagonal->queryend,diagonal->univdiagonal));
	diagonal->nlinked += 1;
	diagonal = diagonal->prev;
      }
    }

    /* Handle divergent diagonals */
    /* Now that we are running oligoindex, we may need to obtain only the last common_diagonal */
    for (p = left_samechr_diagonals; p != NULL; p = List_next(p)) {
      alts_path = (List_T) NULL;
      diagonal = (Univdiag_T) List_head(p);
      while (diagonal != NULL && diagonal->nlinked < ndiagonals) {
	alts_path = List_push(alts_path,(void *) diagonal);
	diagonal = diagonal->prev;
      }
      left_paths = List_push(left_paths,(void *) alts_path);

      common_diagonal = diagonal; /* Last elt on prev path.  Save for later */
      /* printf("Common diagonal is now %p\n",common_diagonal); */
    }
#ifdef DEBUG13
    if (common_diagonal != NULL) {
      printf("Common diagonal is: query %d..%d, diagonal %u\n",
	     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal);
    }
#endif


    if (common_diagonal == NULL) {
      /* All paths connect directly to the middle diagonal, so there is no common diagonal */
      diagonal = middle_diagonal;
      /* queryend = middle_diagonal->querystart; */
      left = middle_diagonal->univdiagonal;
    } else {
      diagonal = common_diagonal;
      /* queryend = common_diagonal->querystart; */
      left = common_diagonal->univdiagonal;
    }

    /* Distinguish left paths by looking for smallest indel (which wins) or splicing */
    debug13(printf("Have %d left_paths.  Distinguish by looking for indel\n",List_length(left_paths)));
    for (p = left_paths; p != NULL; p = List_next(p)) {
      alts_path = (List_T) List_head(p);
      prev_diagonal = (Univdiag_T) List_head(alts_path);
      prev_left = prev_diagonal->univdiagonal;
      debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));

      if (left < prev_left) {
	/* Insertion */
	debug13(printf("Found insertion\n"));
	if (left_indel_diagonal == NULL) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = prev_left - left;
	} else if ((diff = prev_left - left) < mindiff) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = diff;
	}

      } else if (left - prev_left < MIN_INTRONLEN) {
	/* Deletion */
	debug13(printf("Found deletion\n"));
	if (left_indel_diagonal == NULL) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = left - prev_left;
	} else if ((diff = left - prev_left) < mindiff) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = diff;
	}
      }
    }

    if (left_indel_diagonal != NULL) {
      /* Push onto middle path later */
      left = left_indel_diagonal->univdiagonal;
      /* queryend = left_indel_diagonal->querystart; */

    } else {
      debug13(printf("Still have %d left_paths.  Distinguish by looking for best splice\n",List_length(left_paths)));
      first_dist_sense = second_dist_sense = 0;
      first_prob_sense = second_prob_sense = 0.0;
      firsti_sense = secondi_sense = -1;
      first_dist_antisense = second_dist_antisense = 0;
      first_prob_antisense = second_prob_antisense = 0.0;
      firsti_antisense = secondi_antisense = -1;
      sensei = antisensei = 0;

      splice_found_p = false;
      for (p = left_paths; p != NULL; p = List_next(p)) {
	alts_path = (List_T) List_head(p);
	prev_diagonal = (Univdiag_T) List_head(alts_path);
	prev_left = prev_diagonal->univdiagonal;

	spliceinfo->segmenti_donor_nknown = spliceinfo->segmenti_antiacceptor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(prev_left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,prev_left);
	  while (j < nsplicesites && splicesites[j] < prev_left + querylength) {
	    if (splicetypes[j] == DONOR) {
	      debug4s(printf("Setting known donor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_donor_knowni[spliceinfo->segmenti_donor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIACCEPTOR) {
	      debug4s(printf("Setting known antiacceptor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_antiacceptor_knowni[spliceinfo->segmenti_antiacceptor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = querylength + 100;
	spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = querylength + 100;

	spliceinfo->segmentj_acceptor_nknown = spliceinfo->segmentj_antidonor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,left);
	  while (j < nsplicesites && splicesites[j] < left + querylength) {
	    if (splicetypes[j] == ACCEPTOR) {
	      debug4s(printf("Setting known acceptor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_acceptor_knowni[spliceinfo->segmentj_acceptor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIDONOR) {
	      debug4s(printf("Setting known antidonor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_antidonor_knowni[spliceinfo->segmentj_antidonor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = querylength + 100;
	spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = querylength + 100;

	/* splice_distance = left - prev_left; */
#if 0
	max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
	debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
	if (prev_diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
	if (diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
#endif

	if ((splice_pos = Splice_resolve_sense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
					       &best_prob_i,&best_prob_j,
					       /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
					       prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
					       spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found sense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  splice_found_p = true;
	  if (best_prob_i + best_prob_j > first_prob_sense) {
	    second_dist_sense = first_dist_sense;
	    second_prob_sense = first_prob_sense;
	    secondi_sense = firsti_sense;
	    first_dist_sense = left - prev_left;
	    first_prob_sense = best_prob_i + best_prob_j;
	    firsti_sense = sensei;
	  } else if (best_prob_i + best_prob_j > second_prob_sense) {
	    second_dist_sense = left - prev_left;
	    second_prob_sense = best_prob_i + best_prob_j;
	    secondi_sense = sensei;
	  }

	  left_endpoints_sense = Intlist_push(left_endpoints_sense,splice_pos);
	  left_querystarts_sense = Intlist_push(left_querystarts_sense,prev_diagonal->querystart);
	  left_alts_coords_sense = Univcoordlist_push(left_alts_coords_sense,prev_left + splice_pos);
	  left_alts_knowni_sense = Intlist_push(left_alts_knowni_sense,best_knowni_i);
	  left_alts_nmismatchesi_sense = Intlist_push(left_alts_nmismatchesi_sense,best_nmismatches_i);
	  left_alts_nmismatchesj_sense = Intlist_push(left_alts_nmismatchesj_sense,best_nmismatches_j);
	  left_alts_probsi_sense = Doublelist_push(left_alts_probsi_sense,best_prob_i);
	  left_alts_probsj_sense = Doublelist_push(left_alts_probsj_sense,best_prob_j);
	  sensei++;
	}

	if ((splice_pos = Splice_resolve_antisense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
						   &best_prob_i,&best_prob_j,
						   /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
						   prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
						   spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found antisense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  splice_found_p = true;
	  if (best_prob_i + best_prob_j > first_prob_antisense) {
	    second_dist_antisense = first_dist_antisense;
	    second_prob_antisense = first_prob_antisense;
	    secondi_antisense = firsti_antisense;
	    first_dist_antisense = left - prev_left;
	    first_prob_antisense = best_prob_i + best_prob_j;
	    firsti_antisense = antisensei;
	  } else if (best_prob_i + best_prob_j > second_prob_antisense) {
	    second_dist_antisense = left - prev_left;
	    second_prob_antisense = best_prob_i + best_prob_j;
	    secondi_antisense = antisensei;
	  }

	  left_endpoints_antisense = Intlist_push(left_endpoints_antisense,splice_pos);
	  left_querystarts_antisense = Intlist_push(left_querystarts_antisense,prev_diagonal->querystart);
	  left_alts_coords_antisense = Univcoordlist_push(left_alts_coords_antisense,prev_left + splice_pos);
	  left_alts_knowni_antisense = Intlist_push(left_alts_knowni_antisense,best_knowni_i);
	  left_alts_nmismatchesi_antisense = Intlist_push(left_alts_nmismatchesi_antisense,best_nmismatches_i);
	  left_alts_nmismatchesj_antisense = Intlist_push(left_alts_nmismatchesj_antisense,best_nmismatches_j);
	  left_alts_probsi_antisense = Doublelist_push(left_alts_probsi_antisense,best_prob_i);
	  left_alts_probsj_antisense = Doublelist_push(left_alts_probsj_antisense,best_prob_j);
	  antisensei++;
	}
      }


      if (Intlist_length(left_endpoints_sense) > 1) {
	if (first_dist_sense < second_dist_sense/2) {
	  debug13(printf("first dist sense %u is significantly shorter than second dist sense %u.  Keeping %d from end\n",
			 first_dist_sense,second_dist_sense,firsti_sense));
	  firsti_sense = Intlist_length(left_endpoints_sense) - 1 - firsti_sense; /* Because we don't reverse lists */
	  left_endpoints_sense = Intlist_keep_one(left_endpoints_sense,firsti_sense);
	  left_querystarts_sense = Intlist_keep_one(left_querystarts_sense,firsti_sense);
	  left_alts_coords_sense = Univcoordlist_keep_one(left_alts_coords_sense,firsti_sense);
	  left_alts_knowni_sense = Intlist_keep_one(left_alts_knowni_sense,firsti_sense);
	  left_alts_nmismatchesi_sense = Intlist_keep_one(left_alts_nmismatchesi_sense,firsti_sense);
	  left_alts_nmismatchesj_sense = Intlist_keep_one(left_alts_nmismatchesj_sense,firsti_sense);
	  left_alts_probsi_sense = Doublelist_keep_one(left_alts_probsi_sense,firsti_sense);
	  left_alts_probsj_sense = Doublelist_keep_one(left_alts_probsj_sense,firsti_sense);
	}
      }
      
      if (Intlist_length(left_endpoints_antisense) > 1) {
	if (first_dist_antisense < second_dist_antisense/2) {
	  debug13(printf("first dist antisense %u is significantly shorter than second dist antisense %u.  Keeping %d from end\n",
			 first_dist_antisense,second_dist_antisense,firsti_antisense));
	  firsti_antisense = Intlist_length(left_endpoints_antisense) - 1 - firsti_antisense; /* Because we don't reverse lists */
	  left_endpoints_antisense = Intlist_keep_one(left_endpoints_antisense,firsti_antisense);
	  left_querystarts_antisense = Intlist_keep_one(left_querystarts_antisense,firsti_antisense);
	  left_alts_coords_antisense = Univcoordlist_keep_one(left_alts_coords_antisense,firsti_antisense);
	  left_alts_knowni_antisense = Intlist_keep_one(left_alts_knowni_antisense,firsti_antisense);
	  left_alts_nmismatchesi_antisense = Intlist_keep_one(left_alts_nmismatchesi_antisense,firsti_antisense);
	  left_alts_nmismatchesj_antisense = Intlist_keep_one(left_alts_nmismatchesj_antisense,firsti_antisense);
	  left_alts_probsi_antisense = Doublelist_keep_one(left_alts_probsi_antisense,firsti_antisense);
	  left_alts_probsj_antisense = Doublelist_keep_one(left_alts_probsj_antisense,firsti_antisense);
	}
      }

      if (splice_found_p == false) {  /* was left_endpoints_sense == NULL && left_endpoints_antisense == NULL */
	debug13(printf("Could not find a splice on left.  Probably indel+splice.  Taking the closest left diagonal instead\n"));
	for (p = left_paths; p != NULL; p = List_next(p)) {
	  alts_path = (List_T) List_head(p);
	  prev_diagonal = (Univdiag_T) List_head(alts_path);
	  prev_left = prev_diagonal->univdiagonal;
	  debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));
	  if (left_indel_diagonal == NULL) {
	    left_indel_diagonal = prev_diagonal;
	    mindiff = left - prev_left;
	  } else if ((diff = left - prev_left) < mindiff) {
	    left_indel_diagonal = prev_diagonal;
	    mindiff = diff;
	  }
	}

      } else {

	/* SENSE */
	if (left_endpoints_sense == NULL) {
	  /* Skip.  Antisense splice found */
	  /* Previously extended endpoints leftward */

	} else if (Intlist_length(left_endpoints_sense) == 1) {
	  /* Only one sense splice on left */
	  debug13(printf("Only one sense splice on left, which should have %d and %d mismatches, plusp %d\n",
			 Intlist_head(left_alts_nmismatchesi_sense),Intlist_head(left_alts_nmismatchesj_sense),plusp));
	  splice_pos = Intlist_head(left_endpoints_sense);
	  left = Univcoordlist_head(left_alts_coords_sense) - splice_pos;

	  *left_univdiagonal_sense = left;
	  *left_splicepos_sense = splice_pos;
	  *left_nmismatches_sense = Intlist_head(left_alts_nmismatchesi_sense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = prev_left - left;
	  if (plusp == true) {
	    donor_prob = Doublelist_head(left_alts_probsi_sense);
	    acceptor_prob = Doublelist_head(left_alts_probsj_sense);
	  } else {
	    acceptor_prob = Doublelist_head(left_alts_probsi_sense);
	    donor_prob = Doublelist_head(left_alts_probsj_sense);
	  }
	  debug13(printf("Creating a left_junction_sense with splice distance %u\n",splice_distance));
	  *left_junction_sense = Junction_new_splice(splice_distance,SENSE_FORWARD,donor_prob,acceptor_prob);
#endif

	} else if (Intlist_vary(left_endpoints_sense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */
	  /* Previously extended endpoints leftward */

	} else {
	  /* Alternative substring on left */
	  debug13(printf("Alternative substring on left\n"));
	  splice_pos = Intlist_head(left_endpoints_sense);
	  querystart = Intlist_head(left_querystarts_sense); /* Should all be the same */

	  if (plusp == true) {
	    *left_alts_sense = Substring_new_alts_D(querystart,/*queryend*/splice_pos,splice_pos,querylength,
						    plusp,genestrand,query_compress,
						    chrnum,chroffset,chrhigh,chrlength,
						    left_alts_coords_sense,left_alts_knowni_sense,
						    left_alts_nmismatchesi_sense,left_alts_probsi_sense,
						    /*alts_common_prob*/Doublelist_head(left_alts_probsj_sense),
						    /*substring1p*/true);
	  } else {
	    *left_alts_sense = Substring_new_alts_A(/*querystart*/querylength - splice_pos,
						    /*queryend*/querylength - querystart,splice_pos,querylength,
						    plusp,genestrand,query_compress,
						    chrnum,chroffset,chrhigh,chrlength,
						    left_alts_coords_sense,left_alts_knowni_sense,
						    left_alts_nmismatchesi_sense,left_alts_probsi_sense,
						    /*alts_common_prob*/Doublelist_head(left_alts_probsj_sense),
						    /*substring1p*/false);
	  }
	}


	/* ANTISENSE */
	if (left_endpoints_antisense == NULL) {
	  /* Skip.  Sense splice found */
	  /* Previously extended endpoints leftward */

	} else if (Intlist_length(left_endpoints_antisense) == 1) {
	  /* Only one antisense splice on left */
	  debug13(printf("Only one antisense splice on left, which should have %d and %d mismatches\n",
			 Intlist_head(left_alts_nmismatchesi_antisense),Intlist_head(left_alts_nmismatchesj_antisense)));
	  splice_pos = Intlist_head(left_endpoints_antisense);
	  left = Univcoordlist_head(left_alts_coords_antisense) - splice_pos;

	  *left_univdiagonal_antisense = left;
	  *left_splicepos_antisense = splice_pos;
	  *left_nmismatches_antisense = Intlist_head(left_alts_nmismatchesi_antisense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = prev_left - left;
	  if (plusp == true) {
	    acceptor_prob = Doublelist_head(left_alts_probsi_antisense);
	    donor_prob = Doublelist_head(left_alts_probsj_antisense);
	  } else {
	    donor_prob = Doublelist_head(left_alts_probsi_antisense);
	    acceptor_prob = Doublelist_head(left_alts_probsj_antisense);
	  }
	  debug13(printf("Creating a left_junction_antisense with splice distance %u\n",splice_distance));
	  *left_junction_antisense = Junction_new_splice(splice_distance,SENSE_ANTI,donor_prob,acceptor_prob);
#endif

	} else if (Intlist_vary(left_endpoints_antisense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */
	  /* Previously extended endpoints leftward */

	} else {
	  /* Alternative substring on left */
	  debug13(printf("Alternative substring on left\n"));
	  splice_pos = Intlist_head(left_endpoints_antisense);
	  querystart = Intlist_head(left_querystarts_antisense); /* Should all be the same */

	  if (plusp == true) {
	    *left_alts_antisense = Substring_new_alts_A(querystart,/*queryend*/splice_pos,splice_pos,querylength,
							plusp,genestrand,query_compress,
							chrnum,chroffset,chrhigh,chrlength,
							left_alts_coords_antisense,left_alts_knowni_antisense,
							left_alts_nmismatchesi_antisense,left_alts_probsi_antisense,
							/*alts_common_prob*/Doublelist_head(left_alts_probsj_antisense),
							/*substring1p*/true);
	  } else {
	    *left_alts_antisense = Substring_new_alts_D(/*querystart*/querylength - splice_pos,
							/*queryend*/querylength - querystart,splice_pos,querylength,
							plusp,genestrand,query_compress,
							chrnum,chroffset,chrhigh,chrlength,
							left_alts_coords_antisense,left_alts_knowni_antisense,
							left_alts_nmismatchesi_antisense,left_alts_probsi_antisense,
							/*alts_common_prob*/Doublelist_head(left_alts_probsj_antisense),
							/*substring1p*/false);
	  }
	}
      }
    }
  }


  /* C5. Process left diagonals in reverse */
  diagonal_path = (List_T) NULL;

  while (common_diagonal != NULL) {
    if (middle_diagonal->univdiagonal > chroffset + chrlength && common_diagonal->univdiagonal < chroffset + chrlength) {
      debug13(printf("Cannot handle common diagonal across circular origin: query %d..%d, diagonal %u\n",
		     diagonal->querystart,diagonal->queryend,diagonal->univdiagonal - chroffset));
    } else {
      diagonal_path = List_push(diagonal_path,(void *) common_diagonal);
      debug13(printf("Pushing common diagonal onto middle: query %d..%d, diagonal %u\n",
		     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal - chroffset));
    }
    common_diagonal = common_diagonal->prev;
  }

  /* Pops off in reverse */
  debug13(printf("Putting common diagonals (length %d) onto middle\n",List_length(diagonal_path)));
  for (p = diagonal_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    complete_path = List_push(complete_path,(void *) Univdiag_copy(diagonal));
  }
  List_free(&diagonal_path);

  if (left_indel_diagonal != NULL) {
    debug13(printf("Pushing left indel diagonal onto middle: query %d..%d, diagonal %u\n",
		   left_indel_diagonal->querystart,left_indel_diagonal->queryend,left_indel_diagonal->univdiagonal - chroffset));
    complete_path = List_push(complete_path,(void *) Univdiag_copy(left_indel_diagonal));
  }


  debug13(printf("***Exiting resolve_end_diagonals\n"));

  Spliceinfo_free(&spliceinfo);

  for (p = right_paths; p != NULL; p = List_next(p)) {
    alts_path = (List_T) List_head(p);
    List_free(&alts_path);
  }
  List_free(&right_paths);

  for (p = left_paths; p != NULL; p = List_next(p)) {
    alts_path = (List_T) List_head(p);
    List_free(&alts_path);
  }
  List_free(&left_paths);

#if 0
  /* Make endpoints consistent between adjacent univdiags */
  p = complete_path;
  q = List_next(complete_path);
  univdiag2 = List_head(p);
  while (q != NULL) {
    univdiag1 = univdiag2;
    univdiag2 = List_head(q);
    if (univdiag1->queryend >= univdiag2->querystart) {
      overlap = univdiag1->queryend - univdiag2->querystart + 1;
      univdiag1->queryend -= (overlap + 1)/2;
      univdiag2->querystart += (overlap + 1)/2;
    }
    p = q;
    q = List_next(q);
  }
#endif

  Intlist_free(&right_endpoints_sense); Intlist_free(&right_endpoints_antisense);
  Intlist_free(&right_queryends_sense); Intlist_free(&right_queryends_antisense);
  Univcoordlist_free(&right_alts_coords_sense); Univcoordlist_free(&right_alts_coords_antisense);
  Intlist_free(&right_alts_knowni_sense); Intlist_free(&right_alts_knowni_antisense);
  Intlist_free(&right_alts_nmismatchesi_sense); Intlist_free(&right_alts_nmismatchesi_antisense);
  Intlist_free(&right_alts_nmismatchesj_sense); Intlist_free(&right_alts_nmismatchesj_antisense);
  Doublelist_free(&right_alts_probsi_sense); Doublelist_free(&right_alts_probsi_antisense);
  Doublelist_free(&right_alts_probsj_sense); Doublelist_free(&right_alts_probsj_antisense);

  Intlist_free(&left_endpoints_sense); Intlist_free(&left_endpoints_antisense);
  Intlist_free(&left_querystarts_sense); Intlist_free(&left_querystarts_antisense);
  Univcoordlist_free(&left_alts_coords_sense); Univcoordlist_free(&left_alts_coords_antisense);
  Intlist_free(&left_alts_knowni_sense); Intlist_free(&left_alts_knowni_antisense);
  Intlist_free(&left_alts_nmismatchesi_sense); Intlist_free(&left_alts_nmismatchesi_antisense);
  Intlist_free(&left_alts_nmismatchesj_sense); Intlist_free(&left_alts_nmismatchesj_antisense);
  Doublelist_free(&left_alts_probsi_sense); Doublelist_free(&left_alts_probsi_antisense);
  Doublelist_free(&left_alts_probsj_sense); Doublelist_free(&left_alts_probsj_antisense);

  List_free(&left_samechr_diagonals);
  List_free(&right_samechr_diagonals);

  return complete_path;
}
#endif


/* Head of list is distal, tail is middle, so we can keep adding to the list */
static List_T
resolve_right_diagonals (Univcoord_T *right_univdiagonal_sense, int *right_splicepos_sense,
			 int *right_nmismatches_sense,
#ifdef PRECOMPUTE_JUNCTIONS
			 Junction_T *right_junction_sense,
#endif
			 Substring_T *right_alts_sense,
			 Univcoord_T *right_univdiagonal_antisense, int *right_splicepos_antisense,
			 int *right_nmismatches_antisense,
#ifdef PRECOMPUTE_JUNCTIONS
			 Junction_T *right_junction_antisense,
#endif
			 Substring_T *right_alts_antisense,

			 List_T best_right_path, Univdiag_T middle_diagonal, List_T right_diagonals,

			 int querylength, Compress_T query_compress,
			 Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
			 bool plusp, int genestrand, int max_mismatches_allowed) {
  List_T right_paths, p;
  Univdiag_T right_indel_diagonal = NULL;

  List_T right_samechr_diagonals;
  List_T diagonal_path, alts_path;
  Univdiag_T diagonal, common_diagonal, prev_diagonal;
  int ndiagonals;

  Univcoord_T left, prev_left, mindiff, diff;

  Chrpos_T splice_distance;
  int querystart, queryend, splice_pos;
  int best_knowni_i, best_knowni_j, best_nmismatches_i, best_nmismatches_j;
  double best_prob_i, best_prob_j, donor_prob, acceptor_prob;

  Chrpos_T first_dist_sense, second_dist_sense, first_dist_antisense, second_dist_antisense;
  double first_prob_sense, second_prob_sense, first_prob_antisense, second_prob_antisense;
  int firsti_sense, secondi_sense, firsti_antisense, secondi_antisense;
  int sensei, antisensei;

  Intlist_T right_endpoints_sense = NULL, right_endpoints_antisense = NULL, 
    right_queryends_sense = NULL, right_queryends_antisense = NULL;
  Univcoordlist_T right_alts_coords_sense = NULL, right_alts_coords_antisense = NULL;
  Intlist_T right_alts_knowni_sense = NULL, right_alts_knowni_antisense = NULL;
  Intlist_T right_alts_nmismatchesi_sense = NULL, right_alts_nmismatchesi_antisense = NULL,
    right_alts_nmismatchesj_sense = NULL, right_alts_nmismatchesj_antisense = NULL;
  Doublelist_T right_alts_probsi_sense = NULL, right_alts_probsi_antisense = NULL,
    right_alts_probsj_sense = NULL, right_alts_probsj_antisense = NULL;

  Spliceinfo_T spliceinfo;
  int j;

  spliceinfo = Spliceinfo_new(querylength);

  debug13(printf("***Entered resolve_right_diagonals\n"));


  *right_univdiagonal_sense = *right_univdiagonal_antisense = (Univcoord_T) 0;
  *right_splicepos_sense = *right_splicepos_antisense = 0;
  *right_nmismatches_sense = *right_nmismatches_antisense = -1; /* Need to recalculate */
  *right_alts_sense = *right_alts_antisense = (Substring_T) NULL;

  right_samechr_diagonals = (List_T) NULL;
  for (p = right_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (diagonal->univdiagonal >= chrhigh) {
      /* Skip.  Diagonal is in next chromosome */
    } else {
      right_samechr_diagonals = List_push(right_samechr_diagonals,(void *) diagonal);
    }
  }


  /* A3.  Traceback for dynamic programming */
  right_paths = (List_T) NULL;
  debug13(printf("On right, have %d right_diagonals\n",List_length(right_samechr_diagonals)));
  if ((ndiagonals = List_length(right_samechr_diagonals)) == 0) {
    common_diagonal = (Univdiag_T) NULL;

    /* querystart = middle_diagonal->queryend + 1; */
    left = middle_diagonal->univdiagonal;

  } else if (ndiagonals == 1) {
    common_diagonal = (Univdiag_T) List_head(right_samechr_diagonals);

    /* querystart = common_diagonal->queryend + 1; */
    left = common_diagonal->univdiagonal;

  } else {
    debug13(printf("3 Multiple (%d) right diagonals\n",ndiagonals));

    /* Distinguish between common and divergent diagonals */
    for (p = right_samechr_diagonals; p != NULL; p = List_next(p)) {
      diagonal = (Univdiag_T) List_head(p);
      debug13(printf("Path:\n"));
      while (diagonal != NULL) {
	debug13(printf("Diagonal: query %d..%d, diagonal %u\n",
		       diagonal->querystart,diagonal->queryend,diagonal->univdiagonal));
	diagonal->nlinked += 1;
	diagonal = diagonal->prev;
      }
    }

    /* Handle divergent diagonals */
    /* Now that we are running oligoindex, we may need to obtain only the last common_diagonal */
    for (p = right_samechr_diagonals; p != NULL; p = List_next(p)) {
      alts_path = (List_T) NULL;
      diagonal = (Univdiag_T) List_head(p);
      while (diagonal != NULL && diagonal->nlinked < ndiagonals) {
	alts_path = List_push(alts_path,(void *) diagonal);
	diagonal = diagonal->prev;
      }
      right_paths = List_push(right_paths,(void *) alts_path);

      common_diagonal = diagonal; /* Last elt on prev path.  Save for later */
      /* printf("Common diagonal is now %p\n",common_diagonal); */
    }
#ifdef DEBUG13
    if (common_diagonal != NULL) {
      printf("Common diagonal is: query %d..%d, diagonal %u\n",
	     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal);
    }
#endif


    if (common_diagonal == NULL) {
      /* All paths connect directly to the middle diagonal, so there is no common diagonal */
      prev_diagonal = middle_diagonal;
      /* querystart = middle_diagonal_3->queryend + 1; */
      prev_left = middle_diagonal->univdiagonal;
    } else {
      prev_diagonal = common_diagonal;
      /* querystart = common_diagonal->queryend + 1; */
      prev_left = common_diagonal->univdiagonal;
    }

    /* Distinguish right paths by looking for smallest indel (which wins) or splicing */
    debug13(printf("Have %d right_paths.  Distinguish by looking for indels\n",List_length(right_paths)));
    for (p = right_paths; p != NULL; p = List_next(p)) {
      alts_path = (List_T) List_head(p);
      diagonal = (Univdiag_T) List_head(alts_path);
      left = diagonal->univdiagonal;
      debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));

      if (left < prev_left) {
	/* Insertion */
	debug13(printf("Found insertion\n"));
	if (right_indel_diagonal == NULL) {
	  right_indel_diagonal = diagonal;
	  mindiff = prev_left - left;
	} else if ((diff = prev_left - left) < mindiff) {
	  right_indel_diagonal = diagonal;
	  mindiff = diff;
	}

      } else if (left - prev_left < MIN_INTRONLEN) {
	/* Deletion */
	debug13(printf("Found deletion\n"));
	if (right_indel_diagonal == NULL) {
	  right_indel_diagonal = diagonal;
	  mindiff = left - prev_left;
	} else if ((diff = left - prev_left) < mindiff) {
	  right_indel_diagonal = diagonal;
	  mindiff = diff;
	}
      }
    }

    if (right_indel_diagonal != NULL) {
      /* Push onto middle path later */
      /* querystart = right_indel_diagonal->queryend + 1; */
      left = right_indel_diagonal->univdiagonal;

    } else {
      debug13(printf("Still have %d right_paths.  Distinguish by looking for best splice\n",List_length(right_paths)));
      first_dist_sense = second_dist_sense = 0;
      first_prob_sense = second_prob_sense = 0.0;
      firsti_sense = secondi_sense = -1;
      first_dist_antisense = second_dist_antisense = 0;
      first_prob_antisense = second_prob_antisense = 0.0;
      firsti_antisense = secondi_antisense = -1;
      sensei = antisensei = 0;

      for (p = right_paths; p != NULL; p = List_next(p)) {
	alts_path = (List_T) List_head(p);
	diagonal = (Univdiag_T) List_head(alts_path);
	left = diagonal->univdiagonal;

	spliceinfo->segmenti_donor_nknown = spliceinfo->segmenti_antiacceptor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(prev_left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,prev_left);
	  while (j < nsplicesites && splicesites[j] < prev_left + querylength) {
	    if (splicetypes[j] == DONOR) {
	      debug4s(printf("Setting known donor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_donor_knowni[spliceinfo->segmenti_donor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIACCEPTOR) {
	      debug4s(printf("Setting known antiacceptor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_antiacceptor_knowni[spliceinfo->segmenti_antiacceptor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = querylength + 100;
	spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = querylength + 100;

	spliceinfo->segmentj_acceptor_nknown = spliceinfo->segmentj_antidonor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,left);
	  while (j < nsplicesites && splicesites[j] < left + querylength) {
	    if (splicetypes[j] == ACCEPTOR) {
	      debug4s(printf("Setting known acceptor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_acceptor_knowni[spliceinfo->segmentj_acceptor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIDONOR) {
	      debug4s(printf("Setting known antidonor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_antidonor_knowni[spliceinfo->segmentj_antidonor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = querylength + 100;
	spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = querylength + 100;

	/* splice_distance = left - prev_left; */
#if 0
	max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
	debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
	if (prev_diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
	if (diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
#endif

	if ((splice_pos = Splice_resolve_sense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
					       &best_prob_i,&best_prob_j,
					       /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
					       prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
					       spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found sense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  if (best_prob_i + best_prob_j > first_prob_sense) {
	    second_dist_sense = first_dist_sense;
	    second_prob_sense = first_prob_sense;
	    secondi_sense = firsti_sense;
	    first_dist_sense = left - prev_left;
	    first_prob_sense = best_prob_i + best_prob_j;
	    firsti_sense = sensei;
	  } else if (best_prob_i + best_prob_j > second_prob_sense) {
	    second_dist_sense = left - prev_left;
	    second_prob_sense = best_prob_i + best_prob_j;
	    secondi_sense = sensei;
	  }

	  right_endpoints_sense = Intlist_push(right_endpoints_sense,splice_pos);
	  right_queryends_sense = Intlist_push(right_queryends_sense,diagonal->queryend/*+ 1*/);
	  right_alts_coords_sense = Univcoordlist_push(right_alts_coords_sense,left + splice_pos);
	  right_alts_knowni_sense = Intlist_push(right_alts_knowni_sense,best_knowni_j);
	  right_alts_nmismatchesi_sense = Intlist_push(right_alts_nmismatchesi_sense,best_nmismatches_i);
	  right_alts_nmismatchesj_sense = Intlist_push(right_alts_nmismatchesj_sense,best_nmismatches_j);
	  right_alts_probsi_sense = Doublelist_push(right_alts_probsi_sense,best_prob_i);
	  right_alts_probsj_sense = Doublelist_push(right_alts_probsj_sense,best_prob_j);
	  sensei++;
	}

	if ((splice_pos = Splice_resolve_antisense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
						   &best_prob_i,&best_prob_j,
						   /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
						   prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
						   spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found antisense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  if (best_prob_i + best_prob_j > first_prob_antisense) {
	    second_dist_antisense = first_dist_antisense;
	    second_prob_antisense = first_prob_antisense;
	    secondi_antisense = firsti_antisense;
	    first_dist_antisense = left - prev_left;
	    first_prob_antisense = best_prob_i + best_prob_j;
	    firsti_antisense = antisensei;
	  } else if (best_prob_i + best_prob_j > second_prob_antisense) {
	    second_dist_antisense = left - prev_left;
	    second_prob_antisense = best_prob_i + best_prob_j;
	    secondi_antisense = antisensei;
	  }

	  right_endpoints_antisense = Intlist_push(right_endpoints_antisense,splice_pos);
	  right_queryends_antisense = Intlist_push(right_queryends_antisense,diagonal->queryend/*+ 1*/);
	  right_alts_coords_antisense = Univcoordlist_push(right_alts_coords_antisense,left + splice_pos);
	  right_alts_knowni_antisense = Intlist_push(right_alts_knowni_antisense,best_knowni_j);
	  right_alts_nmismatchesi_antisense = Intlist_push(right_alts_nmismatchesi_antisense,best_nmismatches_i);
	  right_alts_nmismatchesj_antisense = Intlist_push(right_alts_nmismatchesj_antisense,best_nmismatches_j);
	  right_alts_probsi_antisense = Doublelist_push(right_alts_probsi_antisense,best_prob_i);
	  right_alts_probsj_antisense = Doublelist_push(right_alts_probsj_antisense,best_prob_j);
	  antisensei++;
	}
      }


      if (Intlist_length(right_endpoints_sense) > 1) {
	if (first_dist_sense < second_dist_sense/2) {
	  debug13(printf("first dist sense %u is significantly shorter than second dist sense %u.  Keeping %d from end\n",
			 first_dist_sense,second_dist_sense,firsti_sense));
	  firsti_sense = Intlist_length(right_endpoints_sense) - 1 - firsti_sense; /* Because we don't reverse lists */
	  right_endpoints_sense = Intlist_keep_one(right_endpoints_sense,firsti_sense);
	  right_queryends_sense = Intlist_keep_one(right_queryends_sense,firsti_sense);
	  right_alts_coords_sense = Univcoordlist_keep_one(right_alts_coords_sense,firsti_sense);
	  right_alts_knowni_sense = Intlist_keep_one(right_alts_knowni_sense,firsti_sense);
	  right_alts_nmismatchesi_sense = Intlist_keep_one(right_alts_nmismatchesi_sense,firsti_sense);
	  right_alts_nmismatchesj_sense = Intlist_keep_one(right_alts_nmismatchesj_sense,firsti_sense);
	  right_alts_probsi_sense = Doublelist_keep_one(right_alts_probsi_sense,firsti_sense);
	  right_alts_probsj_sense = Doublelist_keep_one(right_alts_probsj_sense,firsti_sense);
	}
      }

      if (Intlist_length(right_endpoints_antisense) > 1) {
	if (first_dist_antisense < second_dist_antisense/2) {
	  debug13(printf("first dist antisense %u is significantly shorter than second dist antisense %u.  Keeping %d from end\n",
			 first_dist_antisense,second_dist_antisense,firsti_antisense));
	  firsti_antisense = Intlist_length(right_endpoints_antisense) - 1 - firsti_antisense; /* Because we don't reverse lists */
	  right_endpoints_antisense = Intlist_keep_one(right_endpoints_antisense,firsti_antisense);
	  right_queryends_antisense = Intlist_keep_one(right_queryends_antisense,firsti_antisense);
	  right_alts_coords_antisense = Univcoordlist_keep_one(right_alts_coords_antisense,firsti_antisense);
	  right_alts_knowni_antisense = Intlist_keep_one(right_alts_knowni_antisense,firsti_antisense);
	  right_alts_nmismatchesi_antisense = Intlist_keep_one(right_alts_nmismatchesi_antisense,firsti_antisense);
	  right_alts_nmismatchesj_antisense = Intlist_keep_one(right_alts_nmismatchesj_antisense,firsti_antisense);
	  right_alts_probsi_antisense = Doublelist_keep_one(right_alts_probsi_antisense,firsti_antisense);
	  right_alts_probsj_antisense = Doublelist_keep_one(right_alts_probsj_antisense,firsti_antisense);
	}
      } 

      if (right_endpoints_sense == NULL && right_endpoints_antisense == NULL) {
	debug13(printf("Could not find a splice on right.  Probably indel+splice.  Taking the closest right diagonal instead\n"));
	for (p = right_paths; p != NULL; p = List_next(p)) {
	  alts_path = (List_T) List_head(p);
	  diagonal = (Univdiag_T) List_head(alts_path);
	  left = diagonal->univdiagonal;
	  debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));
	  if (right_indel_diagonal == NULL) {
	    right_indel_diagonal = diagonal;
	    mindiff = left - prev_left;
	  } else if ((diff = left - prev_left) < mindiff) {
	    right_indel_diagonal = diagonal;
	    mindiff = diff;
	  }
	}

      } else {
	
	/* SENSE */
	if (right_endpoints_sense == NULL) {
	  /* Skip.  Antisense splice found */

	} else if (Intlist_length(right_endpoints_sense) == 1) {
	  /* Only one sense splice on right */
	  debug13(printf("Only one sense splice on right, which should have %d and %d mismatches\n",
			 Intlist_head(right_alts_nmismatchesi_sense),Intlist_head(right_alts_nmismatchesj_sense)));
	  splice_pos = Intlist_head(right_endpoints_sense);
	  left = Univcoordlist_head(right_alts_coords_sense) - splice_pos;

	  *right_univdiagonal_sense = left;
	  *right_splicepos_sense = splice_pos;
	  *right_nmismatches_sense = Intlist_head(right_alts_nmismatchesj_sense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = left - prev_left;
	  if (plusp == true) {
	    donor_prob = Doublelist_head(right_alts_probsi_sense);
	    acceptor_prob = Doublelist_head(right_alts_probsj_sense);
	  } else {
	    acceptor_prob = Doublelist_head(right_alts_probsi_sense);
	    donor_prob = Doublelist_head(right_alts_probsj_sense);
	  }
	  debug13(printf("Creating a right_junction_sense with splice distance %u\n",splice_distance));
	  *right_junction_sense = Junction_new_splice(splice_distance,SENSE_FORWARD,donor_prob,acceptor_prob);
#endif
	  
	} else if (Intlist_vary(right_endpoints_sense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */
	  
	} else {
	  /* Alternative substring on right */
	  debug13(printf("Alternative substring on right\n"));
	  splice_pos = Intlist_head(right_endpoints_sense);
	  queryend = Intlist_head(right_queryends_sense); /* Should all be the same */

	  if (plusp == true) {
	    *right_alts_sense = Substring_new_alts_A(/*querystart*/splice_pos,queryend,splice_pos,querylength,
						     plusp,genestrand,query_compress,
						     chrnum,chroffset,chrhigh,chrlength,
						     right_alts_coords_sense,right_alts_knowni_sense,
						     right_alts_nmismatchesj_sense,right_alts_probsj_sense,
						     /*alts_common_prob*/Doublelist_head(right_alts_probsi_sense),
						     /*substring1p*/false);
	  } else {
	    *right_alts_sense = Substring_new_alts_D(/*querystart*/querylength - queryend,
						     /*queryend*/querylength - splice_pos,splice_pos,querylength,
						     plusp,genestrand,query_compress,
						     chrnum,chroffset,chrhigh,chrlength,
						     right_alts_coords_sense,right_alts_knowni_sense,
						     right_alts_nmismatchesj_sense,right_alts_probsj_sense,
						     /*alts_common_prob*/Doublelist_head(right_alts_probsi_sense),
						     /*substring1p*/true);
	  }
	}


	/* ANTISENSE */
	if (right_endpoints_antisense == NULL) {
	  /* Skip.  Sense splice found */

        } else if (Intlist_length(right_endpoints_antisense) == 1) {
	  /* Only one antisense splice on right */
	  debug13(printf("Only one antisense splice on right, which should have %d and %d mismatches\n",
			 Intlist_head(right_alts_nmismatchesi_antisense),Intlist_head(right_alts_nmismatchesj_antisense)));
	  splice_pos = Intlist_head(right_endpoints_antisense);
	  left = Univcoordlist_head(right_alts_coords_antisense) - splice_pos;

	  *right_univdiagonal_antisense = left;
	  *right_splicepos_antisense = splice_pos;
	  *right_nmismatches_antisense = Intlist_head(right_alts_nmismatchesj_antisense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = left - prev_left;
	  if (plusp == true) {
	    acceptor_prob = Doublelist_head(right_alts_probsi_antisense);
	    donor_prob = Doublelist_head(right_alts_probsj_antisense);
	  } else {
	    donor_prob = Doublelist_head(right_alts_probsi_antisense);
	    acceptor_prob = Doublelist_head(right_alts_probsj_antisense);
	  }
	  debug13(printf("Creating a right_junction_antisense with splice distance %u\n",splice_distance));
	  *right_junction_antisense = Junction_new_splice(splice_distance,SENSE_ANTI,donor_prob,acceptor_prob);
#endif

	} else if (Intlist_vary(right_endpoints_antisense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */

	} else {
	  /* Alternative substring on right */
	  debug13(printf("Alternative substring on right\n"));
	  splice_pos = Intlist_head(right_endpoints_antisense);
	  queryend = Intlist_head(right_queryends_antisense); /* Should all be the same */
	  
	  if (plusp == true) {
	    *right_alts_antisense = Substring_new_alts_D(/*querystart*/splice_pos,queryend,splice_pos,querylength,
							 plusp,genestrand,query_compress,
							 chrnum,chroffset,chrhigh,chrlength,
							 right_alts_coords_antisense,right_alts_knowni_antisense,
							 right_alts_nmismatchesj_antisense,right_alts_probsj_antisense,
							 /*alts_common_prob*/Doublelist_head(right_alts_probsi_antisense),
							 /*substring1p*/false);
	  } else {
	    *right_alts_antisense = Substring_new_alts_A(/*querystart*/querylength - queryend,
							 /*queryend*/querylength - splice_pos,splice_pos,querylength,
							 plusp,genestrand,query_compress,
							 chrnum,chroffset,chrhigh,chrlength,
							 right_alts_coords_antisense,right_alts_knowni_antisense,
							 right_alts_nmismatchesj_antisense,right_alts_probsj_antisense,
							 /*alts_common_prob*/Doublelist_head(right_alts_probsi_antisense),
							 /*substring1p*/true);
	  }
	}
      }
    }
  }

  /* A5. Process right common diagonals in reverse */
  diagonal_path = (List_T) NULL;

  while (common_diagonal != NULL) {
    if (middle_diagonal->univdiagonal < chroffset + chrlength && common_diagonal->univdiagonal > chroffset + chrlength) {
      debug13(printf("Cannot handle common diagonal across circular origin: query %d..%d, diagonal %u\n",
		     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal - chroffset));
    } else {
      diagonal_path = List_push(diagonal_path,(void *) common_diagonal);
      debug13(printf("Pushing common diagonal onto middle: query %d..%d, diagonal %u\n",
		     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal - chroffset));
    }
    common_diagonal = common_diagonal->prev;
  }

  /* Pop off in reverse */
  debug13(printf("Putting common diagonals (length %d) onto middle\n",List_length(diagonal_path)));
  for (p = diagonal_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    best_right_path = List_push(best_right_path,(void *) Univdiag_copy(diagonal));
  }
  List_free(&diagonal_path);

  if (right_indel_diagonal != NULL) {
    debug13(printf("Pushing right indel diagonal: query %d..%d, diagonal %u\n",
		   right_indel_diagonal->querystart,right_indel_diagonal->queryend,right_indel_diagonal->univdiagonal - chroffset));
    best_right_path = List_push(best_right_path,(void *) Univdiag_copy(right_indel_diagonal));
  }


#ifdef DEBUG13
  printf("***Exiting resolve_right_diagonals\n");
  printf("right_sense: diagonal %u with splice pos %d and %d mismatches vs right_alts_sense %p\n",
	 *right_univdiagonal_sense,*right_splicepos_sense,*right_nmismatches_sense,*right_alts_sense);
  printf("right_antisense: diagonal %u with splice pos %d and %d mismatches vs right_alts_antisense %p\n",
	 *right_univdiagonal_antisense,*right_splicepos_antisense,*right_nmismatches_antisense,*right_alts_antisense);
#endif

  Spliceinfo_free(&spliceinfo);

  for (p = right_paths; p != NULL; p = List_next(p)) {
    alts_path = (List_T) List_head(p);
    List_free(&alts_path);
  }
  List_free(&right_paths);

  Intlist_free(&right_endpoints_sense); Intlist_free(&right_endpoints_antisense);
  Intlist_free(&right_queryends_sense); Intlist_free(&right_queryends_antisense);
  Univcoordlist_free(&right_alts_coords_sense); Univcoordlist_free(&right_alts_coords_antisense);
  Intlist_free(&right_alts_knowni_sense); Intlist_free(&right_alts_knowni_antisense);
  Intlist_free(&right_alts_nmismatchesi_sense); Intlist_free(&right_alts_nmismatchesi_antisense);
  Intlist_free(&right_alts_nmismatchesj_sense); Intlist_free(&right_alts_nmismatchesj_antisense);
  Doublelist_free(&right_alts_probsi_sense); Doublelist_free(&right_alts_probsi_antisense);
  Doublelist_free(&right_alts_probsj_sense); Doublelist_free(&right_alts_probsj_antisense);

  List_free(&right_samechr_diagonals);

  return best_right_path;
}


/* Head of list is distal, tail is middle, so we can keep adding to the list */
static List_T
resolve_left_diagonals (Univcoord_T *left_univdiagonal_sense, int *left_splicepos_sense,
			int *left_nmismatches_sense,
#ifdef PRECOMPUTE_JUNCTIONS
			Junction_T *left_junction_sense,
#endif
			Substring_T *left_alts_sense,
			Univcoord_T *left_univdiagonal_antisense, int *left_splicepos_antisense,
			int *left_nmismatches_antisense,
#ifdef PRECOMPUTE_JUNCTIONS
			Junction_T *left_junction_antisense,
#endif
			Substring_T *left_alts_antisense,

			List_T best_left_path, Univdiag_T middle_diagonal, List_T left_diagonals,

			int querylength, Compress_T query_compress,
			Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
			bool plusp, int genestrand, int max_mismatches_allowed) {
  List_T left_paths, p;
  Univdiag_T left_indel_diagonal = NULL;

  List_T left_samechr_diagonals;
  List_T diagonal_path, alts_path;
  Univdiag_T diagonal, common_diagonal, prev_diagonal;
  int ndiagonals;

  Univcoord_T left, prev_left, mindiff, diff;

  Chrpos_T splice_distance;
  int querystart, queryend, splice_pos;
  int best_knowni_i, best_knowni_j, best_nmismatches_i, best_nmismatches_j;
  double best_prob_i, best_prob_j, donor_prob, acceptor_prob;

  Chrpos_T first_dist_sense, second_dist_sense, first_dist_antisense, second_dist_antisense;
  double first_prob_sense, second_prob_sense, first_prob_antisense, second_prob_antisense;
  int firsti_sense, secondi_sense, firsti_antisense, secondi_antisense;
  int sensei, antisensei;

  Intlist_T left_endpoints_sense = NULL, left_endpoints_antisense = NULL, 
    left_querystarts_sense = NULL, left_querystarts_antisense = NULL;
  Univcoordlist_T left_alts_coords_sense = NULL, left_alts_coords_antisense = NULL;
  Intlist_T left_alts_knowni_sense = NULL, left_alts_knowni_antisense = NULL;
  Intlist_T left_alts_nmismatchesi_sense = NULL, left_alts_nmismatchesi_antisense = NULL,
    left_alts_nmismatchesj_sense = NULL, left_alts_nmismatchesj_antisense = NULL;
  Doublelist_T left_alts_probsi_sense = NULL, left_alts_probsi_antisense = NULL,
    left_alts_probsj_sense = NULL, left_alts_probsj_antisense = NULL;

  Spliceinfo_T spliceinfo;
  int j;

  spliceinfo = Spliceinfo_new(querylength);

  debug13(printf("***Entered resolve_left_diagonals\n"));

  *left_univdiagonal_sense = *left_univdiagonal_antisense = (Univcoord_T) 0;
  *left_splicepos_sense = *left_splicepos_antisense = 0;
  *left_nmismatches_sense = *left_nmismatches_antisense = -1; /* Need to recalculate */
  *left_alts_sense = *left_alts_antisense = (Substring_T) NULL;

  left_samechr_diagonals = (List_T) NULL;
  for (p = left_diagonals; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (diagonal->univdiagonal + querylength <= chroffset) {
      /* Skip.  Diagonal is in previous chromosome */
    } else {
      left_samechr_diagonals = List_push(left_samechr_diagonals,(void *) diagonal);
    }
  }


  /* C3.  Traceback for dynamic programming */
  left_paths = (List_T) NULL;
  debug13(printf("On left, have %d left_diagonals (resolve_left_diagonals)\n",List_length(left_samechr_diagonals)));
  if ((ndiagonals = List_length(left_samechr_diagonals)) == 0) {
    common_diagonal = (Univdiag_T) NULL;

    /* queryend = middle_diagonal->querystart; */
    left = middle_diagonal->univdiagonal;

  } else if (ndiagonals == 1) {
    common_diagonal = (Univdiag_T) List_head(left_samechr_diagonals);

    /* queryend = common_diagonal->querystart; */
    left = common_diagonal->univdiagonal;

  } else {
    debug13(printf("4 Multiple (%d) left diagonals\n",ndiagonals));

    /* Distinguish between common and divergent diagonals */
    for (p = left_samechr_diagonals; p != NULL; p = List_next(p)) {
      diagonal = (Univdiag_T) List_head(p);
      debug13(printf("Path:\n"));
      while (diagonal != NULL) {
	debug13(printf("Diagonal: query %d..%d, diagonal %u\n",
		       diagonal->querystart,diagonal->queryend,diagonal->univdiagonal));
	diagonal->nlinked += 1;
	diagonal = diagonal->prev;
      }
    }

    /* Handle divergent diagonals */
    /* Now that we are running oligoindex, we may need to obtain only the last common_diagonal */
    for (p = left_samechr_diagonals; p != NULL; p = List_next(p)) {
      alts_path = (List_T) NULL;
      diagonal = (Univdiag_T) List_head(p);
      while (diagonal != NULL && diagonal->nlinked < ndiagonals) {
	alts_path = List_push(alts_path,(void *) diagonal);
	diagonal = diagonal->prev;
      }
      left_paths = List_push(left_paths,(void *) alts_path);

      common_diagonal = diagonal; /* Last elt on prev path.  Save for later */
      /* printf("Common diagonal is now %p\n",common_diagonal); */
    }
#ifdef DEBUG13
    if (common_diagonal != NULL) {
      printf("Common diagonal is: query %d..%d, diagonal %u\n",
	     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal);
    }
#endif


    if (common_diagonal == NULL) {
      /* All paths connect directly to the middle diagonal, so there is no common diagonal */
      diagonal = middle_diagonal;
      /* queryend = middle_diagonal->querystart; */
      left = middle_diagonal->univdiagonal;
    } else {
      diagonal = common_diagonal;
      /* queryend = common_diagonal->querystart; */
      left = common_diagonal->univdiagonal;
    }

    /* Distinguish left paths by looking for smallest indel (which wins) or splicing */
    debug13(printf("Have %d left_paths.  Distinguish by looking for indel\n",List_length(left_paths)));
    for (p = left_paths; p != NULL; p = List_next(p)) {
      alts_path = (List_T) List_head(p);
      prev_diagonal = (Univdiag_T) List_head(alts_path);
      prev_left = prev_diagonal->univdiagonal;
      debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));

      if (left < prev_left) {
	/* Insertion */
	debug13(printf("Found insertion\n"));
	if (left_indel_diagonal == NULL) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = prev_left - left;
	} else if ((diff = prev_left - left) < mindiff) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = diff;
	}

      } else if (left - prev_left < MIN_INTRONLEN) {
	/* Deletion */
	debug13(printf("Found deletion\n"));
	if (left_indel_diagonal == NULL) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = left - prev_left;
	} else if ((diff = left - prev_left) < mindiff) {
	  left_indel_diagonal = prev_diagonal;
	  mindiff = diff;
	}
      }
    }

    if (left_indel_diagonal != NULL) {
      /* Push onto middle path later */
      left = left_indel_diagonal->univdiagonal;
      /* queryend = left_indel_diagonal->querystart; */

    } else {
      debug13(printf("Still have %d left_paths.  Distinguish by looking for best splice\n",List_length(left_paths)));
      first_dist_sense = second_dist_sense = 0;
      first_prob_sense = second_prob_sense = 0.0;
      firsti_sense = secondi_sense = -1;
      first_dist_antisense = second_dist_antisense = 0;
      first_prob_antisense = second_prob_antisense = 0.0;
      firsti_antisense = secondi_antisense = -1;
      sensei = antisensei = 0;

      for (p = left_paths; p != NULL; p = List_next(p)) {
	alts_path = (List_T) List_head(p);
	prev_diagonal = (Univdiag_T) List_head(alts_path);
	prev_left = prev_diagonal->univdiagonal;

	spliceinfo->segmenti_donor_nknown = spliceinfo->segmenti_antiacceptor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(prev_left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,prev_left);
	  while (j < nsplicesites && splicesites[j] < prev_left + querylength) {
	    if (splicetypes[j] == DONOR) {
	      debug4s(printf("Setting known donor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_donor_knowni[spliceinfo->segmenti_donor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIACCEPTOR) {
	      debug4s(printf("Setting known antiacceptor %d for segmenti at %u\n",j,splicesites[j]));
	      spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = splicesites[j] - prev_left;
	      spliceinfo->segmenti_antiacceptor_knowni[spliceinfo->segmenti_antiacceptor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = querylength + 100;
	spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = querylength + 100;

	spliceinfo->segmentj_acceptor_nknown = spliceinfo->segmentj_antidonor_nknown = 0;
	if (nsplicesites > 0 &&
	    Splicetrie_splicesite_p(left,/*pos5*/1,/*pos3*/querylength) == true) {
	  j = binary_search_univcoord(0,nsplicesites,splicesites,left);
	  while (j < nsplicesites && splicesites[j] < left + querylength) {
	    if (splicetypes[j] == ACCEPTOR) {
	      debug4s(printf("Setting known acceptor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_acceptor_knowni[spliceinfo->segmentj_acceptor_nknown++] = j;
	    } else if (splicetypes[j] == ANTIDONOR) {
	      debug4s(printf("Setting known antidonor %d for segmentj at %u\n",j,splicesites[j]));
	      spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = splicesites[j] - left;
	      spliceinfo->segmentj_antidonor_knowni[spliceinfo->segmentj_antidonor_nknown++] = j;
	    }
	    j++;
	  }
	}
	spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = querylength + 100;
	spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = querylength + 100;

	/* splice_distance = left - prev_left; */
#if 0
	max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
	debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
	if (prev_diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
	if (diagonal->intscore > 0) {
	  max_mismatches_allowed += 1;
	}
#endif

	if ((splice_pos = Splice_resolve_sense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
					       &best_prob_i,&best_prob_j,
					       /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
					       prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
					       spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found sense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  if (best_prob_i + best_prob_j > first_prob_sense) {
	    second_dist_sense = first_dist_sense;
	    second_prob_sense = first_prob_sense;
	    secondi_sense = firsti_sense;
	    first_dist_sense = left - prev_left;
	    first_prob_sense = best_prob_i + best_prob_j;
	    firsti_sense = sensei;
	  } else if (best_prob_i + best_prob_j > second_prob_sense) {
	    second_dist_sense = left - prev_left;
	    second_prob_sense = best_prob_i + best_prob_j;
	    secondi_sense = sensei;
	  }

	  left_endpoints_sense = Intlist_push(left_endpoints_sense,splice_pos);
	  left_querystarts_sense = Intlist_push(left_querystarts_sense,prev_diagonal->querystart);
	  left_alts_coords_sense = Univcoordlist_push(left_alts_coords_sense,prev_left + splice_pos);
	  left_alts_knowni_sense = Intlist_push(left_alts_knowni_sense,best_knowni_i);
	  left_alts_nmismatchesi_sense = Intlist_push(left_alts_nmismatchesi_sense,best_nmismatches_i);
	  left_alts_nmismatchesj_sense = Intlist_push(left_alts_nmismatchesj_sense,best_nmismatches_j);
	  left_alts_probsi_sense = Doublelist_push(left_alts_probsi_sense,best_prob_i);
	  left_alts_probsj_sense = Doublelist_push(left_alts_probsj_sense,best_prob_j);
	  sensei++;
	}

	if ((splice_pos = Splice_resolve_antisense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
						   &best_prob_i,&best_prob_j,
						   /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
						   prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
						   spliceinfo,max_mismatches_allowed,plusp,genestrand)) > 0) {
	  debug13(printf("Found antisense splice_pos %d with nmismatches %d and %d, probs %f and %f\n",
			 splice_pos,best_nmismatches_i,best_nmismatches_j,best_prob_i,best_prob_j));
	  if (best_prob_i + best_prob_j > first_prob_antisense) {
	    second_dist_antisense = first_dist_antisense;
	    second_prob_antisense = first_prob_antisense;
	    secondi_antisense = firsti_antisense;
	    first_dist_antisense = left - prev_left;
	    first_prob_antisense = best_prob_i + best_prob_j;
	    firsti_antisense = antisensei;
	  } else if (best_prob_i + best_prob_j > second_prob_antisense) {
	    second_dist_antisense = left - prev_left;
	    second_prob_antisense = best_prob_i + best_prob_j;
	    secondi_antisense = antisensei;
	  }

	  left_endpoints_antisense = Intlist_push(left_endpoints_antisense,splice_pos);
	  left_querystarts_antisense = Intlist_push(left_querystarts_antisense,prev_diagonal->querystart);
	  left_alts_coords_antisense = Univcoordlist_push(left_alts_coords_antisense,prev_left + splice_pos);
	  left_alts_knowni_antisense = Intlist_push(left_alts_knowni_antisense,best_knowni_i);
	  left_alts_nmismatchesi_antisense = Intlist_push(left_alts_nmismatchesi_antisense,best_nmismatches_i);
	  left_alts_nmismatchesj_antisense = Intlist_push(left_alts_nmismatchesj_antisense,best_nmismatches_j);
	  left_alts_probsi_antisense = Doublelist_push(left_alts_probsi_antisense,best_prob_i);
	  left_alts_probsj_antisense = Doublelist_push(left_alts_probsj_antisense,best_prob_j);
	  antisensei++;
	}
      }


      if (Intlist_length(left_endpoints_sense) > 1) {
	if (first_dist_sense < second_dist_sense/2) {
	  debug13(printf("first dist sense %u is significantly shorter than second dist sense %u.  Keeping %d from end\n",
			 first_dist_sense,second_dist_sense,firsti_sense));
	  firsti_sense = Intlist_length(left_endpoints_sense) - 1 - firsti_sense; /* Because we don't reverse lists */
	  left_endpoints_sense = Intlist_keep_one(left_endpoints_sense,firsti_sense);
	  left_querystarts_sense = Intlist_keep_one(left_querystarts_sense,firsti_sense);
	  left_alts_coords_sense = Univcoordlist_keep_one(left_alts_coords_sense,firsti_sense);
	  left_alts_knowni_sense = Intlist_keep_one(left_alts_knowni_sense,firsti_sense);
	  left_alts_nmismatchesi_sense = Intlist_keep_one(left_alts_nmismatchesi_sense,firsti_sense);
	  left_alts_nmismatchesj_sense = Intlist_keep_one(left_alts_nmismatchesj_sense,firsti_sense);
	  left_alts_probsi_sense = Doublelist_keep_one(left_alts_probsi_sense,firsti_sense);
	  left_alts_probsj_sense = Doublelist_keep_one(left_alts_probsj_sense,firsti_sense);
	}
      }

      if (Intlist_length(left_endpoints_antisense) > 1) {
	if (first_dist_antisense < second_dist_antisense/2) {
	  debug13(printf("first dist antisense %u is significantly shorter than second dist antisense %u.  Keeping %d from end\n",
			 first_dist_antisense,second_dist_antisense,firsti_antisense));
	  firsti_antisense = Intlist_length(left_endpoints_antisense) - 1 - firsti_antisense; /* Because we don't reverse lists */
	  left_endpoints_antisense = Intlist_keep_one(left_endpoints_antisense,firsti_antisense);
	  left_querystarts_antisense = Intlist_keep_one(left_querystarts_antisense,firsti_antisense);
	  left_alts_coords_antisense = Univcoordlist_keep_one(left_alts_coords_antisense,firsti_antisense);
	  left_alts_knowni_antisense = Intlist_keep_one(left_alts_knowni_antisense,firsti_antisense);
	  left_alts_nmismatchesi_antisense = Intlist_keep_one(left_alts_nmismatchesi_antisense,firsti_antisense);
	  left_alts_nmismatchesj_antisense = Intlist_keep_one(left_alts_nmismatchesj_antisense,firsti_antisense);
	  left_alts_probsi_antisense = Doublelist_keep_one(left_alts_probsi_antisense,firsti_antisense);
	  left_alts_probsj_antisense = Doublelist_keep_one(left_alts_probsj_antisense,firsti_antisense);
	}
      }

      if (left_endpoints_sense == NULL && left_endpoints_antisense == NULL) {
	debug13(printf("Could not find a splice on left.  Probably indel+splice.  Taking the closest left diagonal instead\n"));
	for (p = left_paths; p != NULL; p = List_next(p)) {
	  alts_path = (List_T) List_head(p);
	  prev_diagonal = (Univdiag_T) List_head(alts_path);
	  prev_left = prev_diagonal->univdiagonal;
	  debug13(printf("left %u, prev_left %u, difference %d\n",left,prev_left,(int) left - prev_left));
	  if (left_indel_diagonal == NULL) {
	    left_indel_diagonal = prev_diagonal;
	    mindiff = left - prev_left;
	  } else if ((diff = left - prev_left) < mindiff) {
	    left_indel_diagonal = prev_diagonal;
	    mindiff = diff;
	  }
	}

      } else {

	/* SENSE */
	if (left_endpoints_sense == NULL) {
	  /* Skip.  Antisense splice found */
	  /* Previously extended endpoints leftward */

	} else if (Intlist_length(left_endpoints_sense) == 1) {
	  /* Only one sense splice on left */
	  debug13(printf("Only one sense splice on left, which should have %d and %d mismatches, plusp %d\n",
			 Intlist_head(left_alts_nmismatchesi_sense),Intlist_head(left_alts_nmismatchesj_sense),plusp));
	  splice_pos = Intlist_head(left_endpoints_sense);
	  left = Univcoordlist_head(left_alts_coords_sense) - splice_pos;

	  *left_univdiagonal_sense = left;
	  *left_splicepos_sense = splice_pos;
	  *left_nmismatches_sense = Intlist_head(left_alts_nmismatchesi_sense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = prev_left - left;
	  if (plusp == true) {
	    donor_prob = Doublelist_head(left_alts_probsi_sense);
	    acceptor_prob = Doublelist_head(left_alts_probsj_sense);
	  } else {
	    acceptor_prob = Doublelist_head(left_alts_probsi_sense);
	    donor_prob = Doublelist_head(left_alts_probsj_sense);
	  }
	  debug13(printf("Creating a left_junction_sense with splice distance %u (prev_left %u - left %u)\n",
			 splice_distance,prev_left,left));
	  *left_junction_sense = Junction_new_splice(splice_distance,SENSE_FORWARD,donor_prob,acceptor_prob);
#endif

	} else if (Intlist_vary(left_endpoints_sense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */
	  /* Previously extended endpoints leftward */

	} else {
	  /* Alternative substring on left */
	  debug13(printf("Alternative substring on left\n"));
	  splice_pos = Intlist_head(left_endpoints_sense);
	  querystart = Intlist_head(left_querystarts_sense); /* Should all be the same */

	  if (plusp == true) {
	    *left_alts_sense = Substring_new_alts_D(querystart,/*queryend*/splice_pos,splice_pos,querylength,
						    plusp,genestrand,query_compress,
						    chrnum,chroffset,chrhigh,chrlength,
						    left_alts_coords_sense,left_alts_knowni_sense,
						    left_alts_nmismatchesi_sense,left_alts_probsi_sense,
						    /*alts_common_prob*/Doublelist_head(left_alts_probsj_sense),
						    /*substring1p*/true);
	  } else {
	    *left_alts_sense = Substring_new_alts_A(/*querystart*/querylength - splice_pos,
						    /*queryend*/querylength - querystart,splice_pos,querylength,
						    plusp,genestrand,query_compress,
						    chrnum,chroffset,chrhigh,chrlength,
						    left_alts_coords_sense,left_alts_knowni_sense,
						    left_alts_nmismatchesi_sense,left_alts_probsi_sense,
						    /*alts_common_prob*/Doublelist_head(left_alts_probsj_sense),
						    /*substring1p*/false);
	  }
	}


	/* ANTISENSE */
	if (left_endpoints_antisense == NULL) {
	  /* Skip.  Sense splice found */
	  /* Previously extended endpoints leftward */

	} else if (Intlist_length(left_endpoints_antisense) == 1) {
	  /* Only one antisense splice on left */
	  debug13(printf("Only one antisense splice on left, which should have %d and %d mismatches\n",
			 Intlist_head(left_alts_nmismatchesi_antisense),Intlist_head(left_alts_nmismatchesj_antisense)));
	  splice_pos = Intlist_head(left_endpoints_antisense);
	  left = Univcoordlist_head(left_alts_coords_antisense) - splice_pos;

	  *left_univdiagonal_antisense = left;
	  *left_splicepos_antisense = splice_pos;
	  *left_nmismatches_antisense = Intlist_head(left_alts_nmismatchesi_antisense);

#ifdef PRECOMPUTE_JUNCTIONS
	  if (common_diagonal == NULL) {
	    /* All paths connect directly to the middle diagonal, so there is no common diagonal */
	    prev_left = middle_diagonal->univdiagonal;
	  } else {
	    prev_left = common_diagonal->univdiagonal;
	  }

	  splice_distance = prev_left - left;
	  if (plusp == true) {
	    acceptor_prob = Doublelist_head(left_alts_probsi_antisense);
	    donor_prob = Doublelist_head(left_alts_probsj_antisense);
	  } else {
	    donor_prob = Doublelist_head(left_alts_probsi_antisense);
	    acceptor_prob = Doublelist_head(left_alts_probsj_antisense);
	  }
	  debug13(printf("Creating a left_junction_antisense with splice distance %u (prev_left %u - left %u)\n",
			 splice_distance,prev_left,left));
	  *left_junction_antisense = Junction_new_splice(splice_distance,SENSE_ANTI,donor_prob,acceptor_prob);
#endif

	} else if (Intlist_vary(left_endpoints_antisense) == true) {
	  /* Skip.  Different splices cannot be combined into an alternative one */
	  /* Previously extended endpoints leftward */

	} else {
	  /* Alternative substring on left */
	  debug13(printf("Alternative substring on left\n"));
	  splice_pos = Intlist_head(left_endpoints_antisense);
	  querystart = Intlist_head(left_querystarts_antisense); /* Should all be the same */

	  if (plusp == true) {
	    *left_alts_antisense = Substring_new_alts_A(querystart,/*queryend*/splice_pos,splice_pos,querylength,
							plusp,genestrand,query_compress,
							chrnum,chroffset,chrhigh,chrlength,
							left_alts_coords_antisense,left_alts_knowni_antisense,
							left_alts_nmismatchesi_antisense,left_alts_probsi_antisense,
							/*alts_common_prob*/Doublelist_head(left_alts_probsj_antisense),
							/*substring1p*/true);
	  } else {
	    *left_alts_antisense = Substring_new_alts_D(/*querystart*/querylength - splice_pos,
							/*queryend*/querylength - querystart,splice_pos,querylength,
							plusp,genestrand,query_compress,
							chrnum,chroffset,chrhigh,chrlength,
							left_alts_coords_antisense,left_alts_knowni_antisense,
							left_alts_nmismatchesi_antisense,left_alts_probsi_antisense,
							/*alts_common_prob*/Doublelist_head(left_alts_probsj_antisense),
							/*substring1p*/false);
	  }
	}
      }
    }
  }


  /* C5. Process left common diagonals in reverse */
  diagonal_path = (List_T) NULL;

  while (common_diagonal != NULL) {
    if (middle_diagonal->univdiagonal > chroffset + chrlength && common_diagonal->univdiagonal < chroffset + chrlength) {
      debug13(printf("Cannot handle common diagonal across circular origin: query %d..%d, diagonal %u\n",
		     diagonal->querystart,diagonal->queryend,diagonal->univdiagonal - chroffset));
    } else {
      diagonal_path = List_push(diagonal_path, (void *) common_diagonal);
      debug13(printf("Pushing common diagonal onto middle: query %d..%d, diagonal %u\n",
		     common_diagonal->querystart,common_diagonal->queryend,common_diagonal->univdiagonal - chroffset));
    }
    common_diagonal = common_diagonal->prev;
  }

  /* Pop off in reverse */
  debug13(printf("Putting common diagonals (length %d) onto middle\n",List_length(diagonal_path)));
  for (p = diagonal_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    best_left_path = List_push(best_left_path,(void *) Univdiag_copy(diagonal));
  }
  List_free(&diagonal_path);

  if (left_indel_diagonal != NULL) {
    debug13(printf("Pushing left indel diagonal onto best: query %d..%d, diagonal %u\n",
		   left_indel_diagonal->querystart,left_indel_diagonal->queryend,left_indel_diagonal->univdiagonal - chroffset));
    best_left_path = List_push(best_left_path,(void *) Univdiag_copy(left_indel_diagonal));
  }


#ifdef DEBUG13
  printf("***Exiting resolve_left_diagonals\n");
  printf("left_sense: diagonal %u with splice pos %d and %d mismatches vs left_alts_sense %p\n",
	 *left_univdiagonal_sense,*left_splicepos_sense,*left_nmismatches_sense,*left_alts_sense);
  printf("left_antisense: diagonal %u with splice pos %d and %d mismatches vs left_alts_antisense %p\n",
	 *left_univdiagonal_antisense,*left_splicepos_antisense,*left_nmismatches_antisense,*left_alts_antisense);
#endif

  Spliceinfo_free(&spliceinfo);

  for (p = left_paths; p != NULL; p = List_next(p)) {
    alts_path = (List_T) List_head(p);
    List_free(&alts_path);
  }
  List_free(&left_paths);

  Intlist_free(&left_endpoints_sense); Intlist_free(&left_endpoints_antisense);
  Intlist_free(&left_querystarts_sense); Intlist_free(&left_querystarts_antisense);
  Univcoordlist_free(&left_alts_coords_sense); Univcoordlist_free(&left_alts_coords_antisense);
  Intlist_free(&left_alts_knowni_sense); Intlist_free(&left_alts_knowni_antisense);
  Intlist_free(&left_alts_nmismatchesi_sense); Intlist_free(&left_alts_nmismatchesi_antisense);
  Intlist_free(&left_alts_nmismatchesj_sense); Intlist_free(&left_alts_nmismatchesj_antisense);
  Doublelist_free(&left_alts_probsi_sense); Doublelist_free(&left_alts_probsi_antisense);
  Doublelist_free(&left_alts_probsj_sense); Doublelist_free(&left_alts_probsj_antisense);

  List_free(&left_samechr_diagonals);

  return best_left_path;
}




static bool
endpoints_acceptable_p (bool *intronp, List_T junctions, Intlist_T endpoints) {
  bool acceptable_p = true;
  Junction_T junction;
  List_T p;
  Intlist_T a;
  int last_endpoint;

  debug13(printf("Evaluating endpoints for acceptability: %s\n",Intlist_to_string(endpoints)));

  last_endpoint = -1;
  for (a = endpoints; a != NULL; a = Intlist_next(a)) {
    if (Intlist_head(a) <= last_endpoint) {
      debug13(printf("Endpoint %d <= %d, so unacceptable\n",Intlist_head(a),last_endpoint));
      acceptable_p = false;
    }
    last_endpoint = Intlist_head(a);
  }

  *intronp = false;
  for (p = junctions; p != NULL; p = List_next(p)) {
    junction = (Junction_T) List_head(p);
    if (junction == NULL) {
      debug13(printf("Junction is NULL, so unacceptable\n"));
      acceptable_p = false;
    } else if (Junction_type(junction) == SPLICE_JUNCTION) {
      *intronp = true;
    }
  }

  return acceptable_p;
}


static void
fix_query_ends (List_T path) {
  Univdiag_T prev_diagonal, diagonal;
  List_T p;
  int midpoint;

#if 0
  for (p = path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d %u\n",diagonal->querystart,diagonal->queryend,diagonal->univdiagonal);
  }
  printf("\n");
#endif

  prev_diagonal = (Univdiag_T) List_head(path);
  for (p = List_next(path); p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (prev_diagonal->queryend > diagonal->querystart) {
      midpoint = (prev_diagonal->queryend + diagonal->querystart)/2;
      prev_diagonal->queryend = midpoint - 1;
      diagonal->querystart = midpoint;
    }
    prev_diagonal = diagonal;
  }

  for (p = path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if (diagonal->queryend < diagonal->querystart) {
      abort();
    }
  }

  return;
}


/* Always solves against plus strand of genome.  Just provide either
   queryuc/query_compress_fwd (coords measured from beginning of
   sequence) or queryrc/query_compress_rev (coords measured from end
   of sequence).  All coordinates measured from low end.
   Sense/antisense is with respect to the plus strand.  But to
   interface with Stage3end_new_substring command, need to flip
   coordinates for case where queryrc aligns to plus strand. */

/* chrnum is fixed from middle_diagonal */
static List_T
solve_via_segments_genome (bool *foundp, int *found_score, List_T hits, List_T complete_path,

			   Univcoord_T right_univdiagonal_sense, int right_splicepos_sense,
			   int right_nmismatches_sense, Junction_T right_junction_sense,
			   Substring_T right_alts_sense,
			   Univcoord_T right_univdiagonal_antisense, int right_splicepos_antisense,
			   int right_nmismatches_antisense, Junction_T right_junction_antisense,
			   Substring_T right_alts_antisense,

			   Univcoord_T left_univdiagonal_sense, int left_splicepos_sense,
			   int left_nmismatches_sense, Junction_T left_junction_sense,
			   Substring_T left_alts_sense,
			   Univcoord_T left_univdiagonal_antisense, int left_splicepos_antisense,
			   int left_nmismatches_antisense, Junction_T left_junction_antisense,
			   Substring_T left_alts_antisense,

			   Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
			   int querylength, Compress_T query_compress, bool plusp, int genestrand, int max_mismatches_allowed,
			   bool paired_end_p, Method_T method, int level) {
  List_T super_path;
  Stage3end_T hit, hit1, hit2, best_hit;
  Junction_T new_junction, junction;
  int best_nmatches, nmatches;
  int sensedir, sense_sensedir, antisense_sensedir;

  List_T p;
  Univdiag_T diagonal, prev_diagonal, new_diagonal;
  Chrpos_T splice_distance;
  int querystart_for_merge;
  int nmismatches, prev_nmismatches;
  bool fillin_p;
#if 0
  bool abort_sense_p = false, abort_antisense_p = false;
#endif

  int indel_pos;
  int nindels;
  Univcoord_T deletionpos;

  int splice_pos;
  double donor_prob, acceptor_prob;
  int introntype;

  bool sense_acceptable_p, antisense_acceptable_p, sense_intronp, antisense_intronp;
  Univcoord_T left, prev_left;
  Univcoordlist_T sense_lefts = NULL, antisense_lefts = NULL;
  Intlist_T sense_nmismatches = NULL, antisense_nmismatches = NULL;
  Intlist_T sense_endpoints = NULL, antisense_endpoints = NULL;
  List_T sense_junctions = NULL, antisense_junctions = NULL;

  int int_ignore;
  Univcoord_T univcoord_ignore;

  int best_knowni_i, best_knowni_j, best_nmismatches_i, best_nmismatches_j;
  double best_prob_i, best_prob_j;

  Spliceinfo_T spliceinfo;
  int j;

  spliceinfo = Spliceinfo_new(querylength);

#ifdef DEBUG13
  printf("\n");
  printf("Entered solve_via_segments_genome\n");
  printf("Original diagonals:\n");
  for (p = complete_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d at %u [%u (%u - %u)]\n",
	   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,
	   diagonal->univdiagonal - chroffset,diagonal->univdiagonal,chroffset);
  }
  printf("\n");

  printf("right_sense: diagonal %u with splice pos %d and %d mismatches vs right_alts_sense %p\n",
	 right_univdiagonal_sense,right_splicepos_sense,right_nmismatches_sense,right_alts_sense);
  printf("right_antisense: diagonal %u with splice pos %d and %d mismatches vs right_alts_antisense %p\n",
	 right_univdiagonal_antisense,right_splicepos_antisense,right_nmismatches_antisense,right_alts_antisense);

  printf("left_sense: diagonal %u with splice pos %d and %d mismatches vs left_alts_sense %p\n",
	 left_univdiagonal_sense,left_splicepos_sense,left_nmismatches_sense,left_alts_sense);
  printf("left_antisense: diagonal %u with splice pos %d and %d mismatches vs left_alts_antisense %p\n",
	 left_univdiagonal_antisense,left_splicepos_antisense,left_nmismatches_antisense,left_alts_antisense);
  printf("\n");
#endif

  /*  Step 1:  Handle mismatches */
  *foundp = false;
  super_path = (List_T) NULL;

  p = complete_path;
  prev_diagonal = (Univdiag_T) List_head(p);
  querystart_for_merge = prev_diagonal->querystart;
  prev_left = prev_diagonal->univdiagonal;
  nmismatches = 0;
  fillin_p = false;

  for (p = List_next(p); p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    if ((left = diagonal->univdiagonal) == prev_left) {
      /* Mismatch.  Assuming that queryend is exclusive */
#if 0
      nmismatches += (diagonal->querystart - prev_diagonal->queryend); /* This could be an overestimate */
#endif
      debug13(printf("We have an unknown number of mismatches between %d..%d and %d..%d\n",
		     prev_diagonal->querystart,prev_diagonal->queryend,diagonal->querystart,diagonal->queryend));
      if (diagonal->nmismatches_known_p == false) {
	fillin_p = true;
      }
      diagonal->nmismatches_known_p = false;
#if 0
      diagonal->intscore = 100;		/* Always recompute mismatches for a super-diagonal */
      nmismatches = -1;	/* Always recompute mismatches for a super-diagonal */
#endif

    } else {
      /* Indel or splice */

      /* Handle previous segment (for prev_left) */
      new_diagonal = Univdiag_new(querystart_for_merge,prev_diagonal->queryend,prev_diagonal->univdiagonal);
      if (fillin_p == true || prev_diagonal->nmismatches_known_p == false) {
	new_diagonal->intscore = 100; /* Positive score allows for many mismatches in indel/splice routines */
      } else {
	new_diagonal->intscore = nmismatches;
      }
      super_path = List_push(super_path,(void *) new_diagonal);

      prev_left = left;
      querystart_for_merge = diagonal->querystart;
      nmismatches = 0;
      fillin_p = false;
    }

    prev_diagonal = diagonal;
  }

  new_diagonal = Univdiag_new(querystart_for_merge,prev_diagonal->queryend,prev_diagonal->univdiagonal);
  if (fillin_p == true || prev_diagonal->nmismatches_known_p == false) {
    new_diagonal->intscore = 100; /* Positive score allows for many mismatches in indel/splice routines */
  } else {
    new_diagonal->intscore = nmismatches;
  }
  super_path = List_push(super_path,(void *) new_diagonal);

  super_path = List_reverse(super_path);

#ifdef DEBUG13
  printf("\n");
  printf("Super diagonals on chrnum %d:\n",chrnum);
  for (p = super_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d at %u [%u] with %d mismatches\n",
	   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,
	   diagonal->univdiagonal - chroffset,diagonal->intscore);
  }
  printf("\n");
#endif


  /*  Step 1b:  Fix query ends */
  fix_query_ends(super_path);

#ifdef DEBUG13
  printf("\n");
  printf("After fixing query ends:\n");
  for (p = super_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d at %u [%u] with %d mismatches\n",
	   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,
	   diagonal->univdiagonal - chroffset,diagonal->intscore);
  }
  printf("\n");
#endif


  /*  Step 2:  Handle indels and splices */

  p = super_path;
  prev_diagonal = (Univdiag_T) List_head(p);
  prev_left = prev_diagonal->univdiagonal;

  debug13(printf("Diagonal %d..%d for left %u [%u]\n",
		 prev_diagonal->querystart,prev_diagonal->queryend,prev_left,prev_left - chroffset));

  sense_endpoints = Intlist_push(NULL,prev_diagonal->querystart);
  antisense_endpoints = Intlist_push(NULL,prev_diagonal->querystart);

  /* Previously pushed prev_diagonal->intscore, but that is not
     correct (unless there are no mismatches).  Pushing -1 indicates
     that we need to compute the value */
  if (prev_diagonal->intscore == 0) {
    sense_nmismatches = Intlist_push(NULL,0);
    antisense_nmismatches = Intlist_push(NULL,0);
  } else {
    sense_nmismatches = Intlist_push(NULL,-1);
    antisense_nmismatches = Intlist_push(NULL,-1);
  }

  for (p = List_next(p); p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    left = diagonal->univdiagonal;
    assert(left != prev_left); /* Because we already handled mismatches above */

    debug13(printf("Diagonal %d..%d at leftpos %u [%u], diff %d\n",
		   diagonal->querystart,diagonal->queryend,left,left - chroffset,left - prev_left));

    if (left < prev_left) {
      /* Insertion */
      nindels = prev_left - left;
#if 0
      max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
      debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
      if (prev_diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
      if (diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
#endif
      if ((indel_pos = Indel_resolve_middle_insertion(&best_nmismatches_i,&best_nmismatches_j,
						      /*left*/prev_left,/*indels*/+nindels,
						      /*mismatch_positions_left*/NULL,/*nmismatches_left*/0,
						      /*mismatch_positions_right*/NULL,/*nmismatches_right*/0,
						      /*ome*/genomebits,/*ome_alt*/genomebits_alt,query_compress,
						      prev_diagonal->querystart,diagonal->queryend,querylength,
						      max_mismatches_allowed,/*plusp:true*/true,genestrand,
						      /*want_lowest_coordinate_p*/true)) <= 0) {
	sense_junctions = List_push(sense_junctions,NULL);
	antisense_junctions = List_push(antisense_junctions,NULL);
      } else {
	sense_junctions = List_push(sense_junctions,Junction_new_insertion(nindels));
	antisense_junctions = List_push(antisense_junctions,Junction_new_insertion(nindels));
	diagonal->querystart += nindels; /* Needed for subsequent indel computation */
      }

#if 0
      if (indel_pos <= Intlist_head(sense_endpoints) + 1) {
	abort_sense_p = true;
      }
      if (indel_pos <= Intlist_head(antisense_endpoints) + 1) {
	abort_antisense_p = true;
      }
#endif

      if ((prev_nmismatches = Intlist_head(sense_nmismatches)) < 0) {
	/* Still need to compute */
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(sense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(sense_nmismatches,-1);
	sense_nmismatches = Intlist_push(sense_nmismatches,-1);
      } else {
	Intlist_head_set(sense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      }

      if ((prev_nmismatches = Intlist_head(antisense_nmismatches)) < 0) {
	/* Still need to compute */
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(antisense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(antisense_nmismatches,-1);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,-1);
      } else {
	Intlist_head_set(antisense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      }

      sense_lefts = Univcoordlist_push(sense_lefts,prev_left);
      antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);

      sense_endpoints = Intlist_push(sense_endpoints,indel_pos);
      antisense_endpoints = Intlist_push(antisense_endpoints,indel_pos);
      debug13(printf("insertion pos in range %d..%d is %d with nmismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,indel_pos,best_nmismatches_i,best_nmismatches_j));

    } else if (left <= prev_left + max_deletionlen) {
      /* Deletion (or short intron) */
      nindels = left - prev_left;
#if 0
      max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
      debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
      if (prev_diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
      if (diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
#endif
      if ((indel_pos = Indel_resolve_middle_deletion_or_splice(&introntype,&best_nmismatches_i,&best_nmismatches_j,
							       /*left*/prev_left,/*indels*/-nindels,
							       /*ome*/genomebits,/*ome_alt*/genomebits_alt,query_compress,
							       prev_diagonal->querystart,diagonal->queryend,querylength,
							       max_mismatches_allowed,/*plusp:true*/true,genestrand,
							       min_intronlength,/*want_lowest_coordinate_p*/true)) <= 0) {
	sense_junctions = List_push(sense_junctions,NULL);
	antisense_junctions = List_push(antisense_junctions,NULL);

      } else if ((Chrpos_T) nindels < min_intronlength ||
		 (novelsplicingp == false && splicesites == NULL) || circularp[chrnum] == true) {
	/* Cannot be an intron, so must be a deletion */
	deletionpos = prev_left + indel_pos;
	sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));
	antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

#if 0
	if (indel_pos <= Intlist_head(sense_endpoints) + 1) {
	  abort_sense_p = true;
	}
	if (indel_pos <= Intlist_head(antisense_endpoints) + 1) {
	  abort_antisense_p = true;
	}
#endif

      } else if ((sensedir = Intron_canonical_sensedir(introntype)) == SENSE_NULL) {
	/* No intron dinucleotids found, so must be a deletion */
	deletionpos = prev_left + indel_pos;
	sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));
	antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

#if 0
	if (indel_pos <= Intlist_head(sense_endpoints) + 1) {
	  abort_sense_p = true;
	}
	if (indel_pos <= Intlist_head(antisense_endpoints) + 1) {
	  abort_antisense_p = true;
	}
#endif

      } else {
	deletionpos = prev_left + indel_pos;
	if (plusp == true) {
	  if (sensedir == SENSE_FORWARD) {
	    donor_prob = Maxent_hr_donor_prob(deletionpos,chroffset);
	    acceptor_prob = Maxent_hr_acceptor_prob(deletionpos+nindels,chroffset);
	    sense_junctions = List_push(sense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_FORWARD,
									    donor_prob,acceptor_prob));
	    antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

	  } else {
	    donor_prob = Maxent_hr_antidonor_prob(deletionpos+nindels,chroffset);
	    acceptor_prob = Maxent_hr_antiacceptor_prob(deletionpos,chroffset);
	    antisense_junctions = List_push(antisense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_ANTI,
										    donor_prob,acceptor_prob));
	    sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));

	  }

	} else {
	  sensedir = (sensedir == SENSE_FORWARD) ? SENSE_ANTI : SENSE_FORWARD;
	  if (sensedir == SENSE_FORWARD) {
	    /* check */
	    donor_prob = Maxent_hr_antidonor_prob(deletionpos+nindels,chroffset);
	    acceptor_prob = Maxent_hr_antiacceptor_prob(deletionpos,chroffset);
	    sense_junctions = List_push(sense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_FORWARD,
									    donor_prob,acceptor_prob));
	    antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

	  } else {
	    /* check */
	    donor_prob = Maxent_hr_donor_prob(deletionpos,chroffset);
	    acceptor_prob = Maxent_hr_acceptor_prob(deletionpos+nindels,chroffset);
	    antisense_junctions = List_push(antisense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_ANTI,
										    donor_prob,acceptor_prob));
	    sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));
	  }
	}
      }

      if ((prev_nmismatches = Intlist_head(sense_nmismatches)) < 0) {
	/* Still need to compute */
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(sense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(sense_nmismatches,-1);
	sense_nmismatches = Intlist_push(sense_nmismatches,-1);
      } else {
	Intlist_head_set(sense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      }

      if ((prev_nmismatches = Intlist_head(antisense_nmismatches)) < 0) {
	/* Still need to compute */
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(antisense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(antisense_nmismatches,-1);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,-1);
      } else {
	Intlist_head_set(antisense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      }

      sense_lefts = Univcoordlist_push(sense_lefts,prev_left);
      antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);

      sense_endpoints = Intlist_push(sense_endpoints,indel_pos);
      antisense_endpoints = Intlist_push(antisense_endpoints,indel_pos);
      debug13(printf("deletion pos in range %d..%d is %d with nmismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,indel_pos,best_nmismatches_i,best_nmismatches_j));
      
    } else {
      /* Splice */
      spliceinfo->segmenti_donor_nknown = spliceinfo->segmenti_antiacceptor_nknown = 0;
      if (nsplicesites > 0 &&
	  Splicetrie_splicesite_p(prev_left,/*pos5*/1,/*pos3*/querylength) == true) {
	j = binary_search_univcoord(0,nsplicesites,splicesites,prev_left);
	while (j < nsplicesites && splicesites[j] < prev_left + querylength) {
	  if (splicetypes[j] == DONOR) {
	    debug4s(printf("Setting known donor %d for segmenti at %u\n",j,splicesites[j]));
	    spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = splicesites[j] - prev_left;
	    spliceinfo->segmenti_donor_knowni[spliceinfo->segmenti_donor_nknown++] = j;
	  } else if (splicetypes[j] == ANTIACCEPTOR) {
	    debug4s(printf("Setting known antiacceptor %d for segmenti at %u\n",j,splicesites[j]));
	    spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = splicesites[j] - prev_left;
	    spliceinfo->segmenti_antiacceptor_knowni[spliceinfo->segmenti_antiacceptor_nknown++] = j;
	  }
	  j++;
	}
      }
      spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = querylength + 100;
      spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = querylength + 100;
	  
      spliceinfo->segmentj_acceptor_nknown = spliceinfo->segmentj_antidonor_nknown = 0;
      if (nsplicesites > 0 &&
	  Splicetrie_splicesite_p(left,/*pos5*/1,/*pos3*/querylength) == true) {
	j = binary_search_univcoord(0,nsplicesites,splicesites,left);
	while (j < nsplicesites && splicesites[j] < left + querylength) {
	  if (splicetypes[j] == ACCEPTOR) {
	    debug4s(printf("Setting known acceptor %d for segmentj at %u\n",j,splicesites[j]));
	    spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = splicesites[j] - left;
	    spliceinfo->segmentj_acceptor_knowni[spliceinfo->segmentj_acceptor_nknown++] = j;
	  } else if (splicetypes[j] == ANTIDONOR) {
	    debug4s(printf("Setting known antidonor %d for segmentj at %u\n",j,splicesites[j]));
	    spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = splicesites[j] - left;
	    spliceinfo->segmentj_antidonor_knowni[spliceinfo->segmentj_antidonor_nknown++] = j;
	  }
	  j++;
	}
      }
      spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = querylength + 100;
      spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = querylength + 100;
      
      splice_distance = left - prev_left;
#if 0
      max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
      debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
      if (prev_diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
      if (diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
#endif

      if ((splice_pos = Splice_resolve_sense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
					     &best_prob_i,&best_prob_j,
					     /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
					     prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
					     spliceinfo,max_mismatches_allowed,plusp,genestrand)) <= 0) {
	sense_endpoints = Intlist_push(sense_endpoints,-1); /* Mark as invalid */
	sense_junctions = List_push(sense_junctions,NULL);
      } else if (plusp == true) {
	sense_endpoints = Intlist_push(sense_endpoints,splice_pos);
	sense_junctions = List_push(sense_junctions,Junction_new_splice(splice_distance,SENSE_FORWARD,
									/*donor_prob*/best_prob_i,/*acceptor_prob*/best_prob_j));
      } else {
	sense_endpoints = Intlist_push(sense_endpoints,splice_pos);
	sense_junctions = List_push(sense_junctions,Junction_new_splice(splice_distance,SENSE_FORWARD,
									/*donor_prob*/best_prob_j,/*acceptor_prob*/best_prob_i));
      }
      debug13(printf("sense splice_pos in range %d..%d is %d with mismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,splice_pos,best_nmismatches_i,best_nmismatches_j));
      if ((prev_nmismatches = Intlist_head(sense_nmismatches)) < 0) {
	/* Still need to compute */
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(sense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(sense_nmismatches,-1);
	sense_nmismatches = Intlist_push(sense_nmismatches,-1);
      } else {
	Intlist_head_set(sense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      }
      sense_lefts = Univcoordlist_push(sense_lefts,prev_left);

      if ((splice_pos = Splice_resolve_antisense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
						 &best_prob_i,&best_prob_j,
						 /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
						 prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
						 spliceinfo,max_mismatches_allowed,plusp,genestrand)) <= 0) {
	antisense_endpoints = Intlist_push(antisense_endpoints,-1); /* Mark as invalid */
	antisense_junctions = List_push(antisense_junctions,NULL);
      } else if (plusp == true) {
	antisense_endpoints = Intlist_push(antisense_endpoints,splice_pos);
	antisense_junctions = List_push(antisense_junctions,Junction_new_splice(splice_distance,SENSE_ANTI,
										/*donor_prob*/best_prob_j,/*acceptor_prob*/best_prob_i));
      } else {
	antisense_endpoints = Intlist_push(antisense_endpoints,splice_pos);
	antisense_junctions = List_push(antisense_junctions,Junction_new_splice(splice_distance,SENSE_ANTI,
										/*donor_prob*/best_prob_i,/*acceptor_prob*/best_prob_j));
      }
      debug13(printf("antisense splice_pos in range %d..%d is %d with nmismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,splice_pos,best_nmismatches_i,best_nmismatches_j));
      if ((prev_nmismatches = Intlist_head(antisense_nmismatches)) < 0) {
	/* Still need to compute */
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(antisense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(antisense_nmismatches,-1);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,-1);
      } else {
	Intlist_head_set(antisense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      }
      antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);
    }

    /* Handle previous segment (for prev_left) */
    prev_left = left;
    prev_diagonal = diagonal;
  }

  /* Finish up lists */
  sense_lefts = Univcoordlist_push(sense_lefts,prev_left);
  antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);
  sense_endpoints = Intlist_push(sense_endpoints,prev_diagonal->queryend/*+ 1*/);
  antisense_endpoints = Intlist_push(antisense_endpoints,prev_diagonal->queryend/*+ 1*/);


  debug13(printf("After step 2 (indels and splices)\n"));
  debug13(printf("sense (wrt plus): %s\n",Intlist_to_string(sense_endpoints)));
  debug13(printf("antisense (wrt plus): %s\n",Intlist_to_string(antisense_endpoints)));
  debug13(printf("sense lefts: %s\n",Univcoordlist_to_string(sense_lefts)));
  debug13(printf("antisense lefts: %s\n",Univcoordlist_to_string(antisense_lefts)));
  debug13(printf("sense nmismatches: %s\n",Intlist_to_string(sense_nmismatches)));
  debug13(printf("antisense nmismatches: %s\n",Intlist_to_string(antisense_nmismatches)));


  /*  Step 3:  Handle single or alternative sense/antisense ends on right */
  if (right_univdiagonal_sense > 0) {
    debug13(printf("Normal sense substring on right with %d mismatches\n",right_nmismatches_sense));
    debug13(printf("Ignoring right_junction_sense\n"));
#ifdef PRECOMPUTE_JUNCTIONS
    debug13(Junction_print(right_junction_sense));
#endif
    new_junction = Junction_new_generic(Univcoordlist_head(sense_lefts),right_univdiagonal_sense,
					/*querypos1*/right_splicepos_sense,
					/*querypos2*/right_splicepos_sense,
					chroffset,plusp,SENSE_FORWARD);
    debug13(printf("Using new junction\n"));
    debug13(Junction_print(new_junction));

    sense_lefts = Univcoordlist_push(sense_lefts,right_univdiagonal_sense);
    if (Intlist_head(sense_endpoints) != right_splicepos_sense) {
      Intlist_head_set(sense_endpoints,right_splicepos_sense); /* Use splice_pos instead of kmer-based endpoint */
      Intlist_head_set(sense_nmismatches,-1); /* Need to recompute nmismatches */
    }
    sense_endpoints = Intlist_push(sense_endpoints,querylength);
    sense_nmismatches = Intlist_push(sense_nmismatches,right_nmismatches_sense);
#if 0
    /* Previously pushed the given junction, but it can be wrong */
    sense_junctions = List_push(sense_junctions,right_junction_sense);
#else
    sense_junctions = List_push(sense_junctions,(void *) new_junction);
#endif


  } else if (right_alts_sense != NULL) {
    debug13(printf("Alternative substring on right\n"));
    splice_pos = Substring_amb_splice_pos(right_alts_sense);
#if 0
    if (plusp == false) {
      splice_pos = querylength - splice_pos;
    }
#endif
    if (Intlist_head(sense_endpoints) != splice_pos) {
      Intlist_head_set(sense_nmismatches,-1);
      Intlist_head_set(sense_endpoints,splice_pos);
    }
    /* sense_endpoints = Intlist_push(sense_endpoints,queryend); */

  } else if (Intlist_head(sense_endpoints) == querylength) {
    /* Last substring already goes to the end */
  } else {
#ifdef EXTEND_TO_ENDS
    Intlist_head_set(sense_endpoints,querylength);
    Intlist_head_set(sense_nmismatches,-1); /* Recalculate */
#endif
  }

  if (right_univdiagonal_antisense > 0) {
    debug13(printf("Normal antisense substring on right with %d mismatches\n",right_nmismatches_antisense));
    debug13(printf("Ignoring right_junction_antisense\n"));
#ifdef PRECOMPUTE_JUNCTIONS
    debug13(Junction_print(right_junction_antisense));
#endif
    new_junction = Junction_new_generic(Univcoordlist_head(antisense_lefts),right_univdiagonal_antisense,
					/*querypos1*/right_splicepos_antisense,
					/*querypos2*/right_splicepos_antisense,
                                        chroffset,plusp,SENSE_ANTI);
    debug13(printf("Using new junction\n"));
    debug13(Junction_print(new_junction));

    antisense_lefts = Univcoordlist_push(antisense_lefts,right_univdiagonal_antisense);
    if (Intlist_head(antisense_endpoints) != right_splicepos_antisense) {
      Intlist_head_set(antisense_endpoints,right_splicepos_antisense);  /* Use splice_pos instead of kmer-based endpoint */
      Intlist_head_set(antisense_nmismatches,-1); /* Need to recompute nmismatches */
    }
    antisense_endpoints = Intlist_push(antisense_endpoints,querylength);
    antisense_nmismatches = Intlist_push(antisense_nmismatches,right_nmismatches_antisense);
#if 0
    /* Previously pushed the given junction, but it can be wrong */
    antisense_junctions = List_push(antisense_junctions,right_junction_antisense);
#else
    antisense_junctions = List_push(antisense_junctions,(void *) new_junction);
#endif

  } else if (right_alts_antisense != NULL) {
    debug13(printf("Alternative substring on right\n"));
    splice_pos = Substring_amb_splice_pos(right_alts_antisense);
#if 0
    if (plusp == false) {
      splice_pos = querylength - splice_pos;
    }
#endif
    if (Intlist_head(antisense_endpoints) != splice_pos) {
      Intlist_head_set(antisense_nmismatches,-1);
      Intlist_head_set(antisense_endpoints,splice_pos);
    }
    /* antisense_endpoints = Intlist_push(antisense_endpoints,queryend); */

  } else if (Intlist_head(antisense_endpoints) == querylength) {
    /* Last substring already goes to the end */
  } else {
#ifdef EXTEND_TO_ENDS
    Intlist_head_set(antisense_endpoints,querylength);
    Intlist_head_set(antisense_nmismatches,-1); /* Recalculate */
#endif
  }


  debug13(printf("After step 3 (alternative ends on right)\n"));
  debug13(printf("sense (wrt plus): %s\n",Intlist_to_string(sense_endpoints)));
  debug13(printf("antisense (wrt plus): %s\n",Intlist_to_string(antisense_endpoints)));
  debug13(printf("sense lefts: %s\n",Univcoordlist_to_string(sense_lefts)));
  debug13(printf("antisense lefts: %s\n",Univcoordlist_to_string(antisense_lefts)));
  debug13(printf("sense nmismatches: %s\n",Intlist_to_string(sense_nmismatches)));
  debug13(printf("antisense nmismatches: %s\n",Intlist_to_string(antisense_nmismatches)));

  /*  Step 4:  Reverse sense and antisense alignments */
  sense_nmismatches = Intlist_reverse(sense_nmismatches);
  antisense_nmismatches = Intlist_reverse(antisense_nmismatches);
  sense_lefts = Univcoordlist_reverse(sense_lefts);
  antisense_lefts = Univcoordlist_reverse(antisense_lefts);
  sense_endpoints = Intlist_reverse(sense_endpoints);
  antisense_endpoints = Intlist_reverse(antisense_endpoints);
  sense_junctions = List_reverse(sense_junctions);
  antisense_junctions = List_reverse(antisense_junctions);

  debug13(printf("After step 4 (reverse alignments)\n"));
  debug13(printf("sense (wrt plus): %s\n",Intlist_to_string(sense_endpoints)));
  debug13(printf("antisense (wrt plus): %s\n",Intlist_to_string(antisense_endpoints)));
  debug13(printf("sense lefts: %s\n",Univcoordlist_to_string(sense_lefts)));
  debug13(printf("antisense lefts: %s\n",Univcoordlist_to_string(antisense_lefts)));
  debug13(printf("sense nmismatches: %s\n",Intlist_to_string(sense_nmismatches)));
  debug13(printf("antisense nmismatches: %s\n",Intlist_to_string(antisense_nmismatches)));


  /*  Step 5:  Handle single or alternative sense/antisense ends on left */
  if (left_univdiagonal_sense > 0) {
    debug13(printf("Normal sense substring on left with %d mismatches and left %u\n",
		   left_nmismatches_sense,left_univdiagonal_sense));
    debug13(printf("Ignoring left_junction_sense\n"));
#ifdef PRECOMPUTE_JUNCTIONS
    debug13(Junction_print(left_junction_sense));
#endif
    new_junction = Junction_new_generic(left_univdiagonal_sense,Univcoordlist_head(sense_lefts),
					/*querypos1*/left_splicepos_sense,
					/*querypos2*/left_splicepos_sense,
					chroffset,plusp,SENSE_FORWARD);
    debug13(printf("Using new junction\n"));
    debug13(Junction_print(new_junction));

    sense_lefts = Univcoordlist_push(sense_lefts,left_univdiagonal_sense);
    if (Intlist_head(sense_endpoints) != left_splicepos_sense) {
      Intlist_head_set(sense_endpoints,left_splicepos_sense); /* Use splice_pos instead of kmer-based endpoint */
      Intlist_head_set(sense_nmismatches,-1);		      /* Need to recompute nmismatches */
    }
    sense_endpoints = Intlist_push(sense_endpoints,0);
    sense_nmismatches = Intlist_push(sense_nmismatches,left_nmismatches_sense);
#if 0
    /* Previously pushed the given junction, but it can be wrong */
    sense_junctions = List_push(sense_junctions,left_junction_sense);
#else
    sense_junctions = List_push(sense_junctions,(void *) new_junction);
#endif


  } else if (left_alts_sense != NULL) {
    debug13(printf("Alternative substring on left\n"));
    splice_pos = Substring_amb_splice_pos(left_alts_sense);
#if 0
    if (plusp == false) {
      splice_pos = querylength - splice_pos;
    }
#endif
    if (Intlist_head(sense_endpoints) != splice_pos) {
      Intlist_head_set(sense_nmismatches,-1);
      Intlist_head_set(sense_endpoints,splice_pos);
    }
    /* sense_endpoints = Intlist_push(sense_endpoints,querystart); */

  } else if (Intlist_head(sense_endpoints) == 0) {
    /* First substring already goes to the beginning */
    debug13(printf("Sense: First substring already goes to the beginning\n"));
  } else {
#ifdef EXTEND_TO_ENDS
    Intlist_head_set(sense_endpoints,0);
    Intlist_head_set(sense_nmismatches,-1); /* Recalculate */
#endif
  }


  if (left_univdiagonal_antisense > 0) {
    debug13(printf("Normal antisense substring on left with %d mismatches ane left %u\n",
		   left_nmismatches_antisense,left_univdiagonal_antisense));
    debug13(printf("Ignoring left_junction_antisense\n"));
#ifdef PRECOMPUTE_JUNCTIONS
    debug13(Junction_print(left_junction_antisense));
#endif
    new_junction = Junction_new_generic(left_univdiagonal_antisense,Univcoordlist_head(antisense_lefts),
					/*querypos1*/left_splicepos_antisense,
					/*querypos2*/left_splicepos_antisense,
					chroffset,plusp,SENSE_ANTI);
    debug13(printf("Using new junction\n"));
    debug13(Junction_print(new_junction));

    antisense_lefts = Univcoordlist_push(antisense_lefts,left_univdiagonal_antisense);
    if (Intlist_head(antisense_endpoints) != left_splicepos_antisense) {
      Intlist_head_set(antisense_endpoints,left_splicepos_antisense); /* Use splice_pos instead of kmer-based endpoint */
      Intlist_head_set(antisense_nmismatches,-1);		      /* Need to recompute nmismatches */
    }
    antisense_endpoints = Intlist_push(antisense_endpoints,0);
    antisense_nmismatches = Intlist_push(antisense_nmismatches,left_nmismatches_antisense);
#if 0
    /* Previously pushed the given junction, but it can be wrong */
    antisense_junctions = List_push(antisense_junctions,left_junction_antisense);
#else
    antisense_junctions = List_push(antisense_junctions,(void *) new_junction);
#endif

  } else if (left_alts_antisense != NULL) {
    debug13(printf("Alternative substring on left\n"));
    splice_pos = Substring_amb_splice_pos(left_alts_antisense);
#if 0
    if (plusp == false) {
      splice_pos = querylength - splice_pos;
    }
#endif
    if (Intlist_head(antisense_endpoints) != splice_pos) {
      Intlist_head_set(antisense_nmismatches,-1);
      Intlist_head_set(antisense_endpoints,splice_pos);
    }
    /* antisense_endpoints = Intlist_push(antisense_endpoints,querystart); */

  } else if (Intlist_head(antisense_endpoints) == 0) {
    /* First substring already goes to the beginning */
    debug13(printf("Antisense: First substring already goes to the beginning\n"));
  } else {
#ifdef EXTEND_TO_ENDS
    Intlist_head_set(antisense_endpoints,0);
    Intlist_head_set(antisense_nmismatches,-1); /* Recalculate */
#endif
  }


  debug13(printf("After step 5 (alternative ends on left)\n"));
  debug13(printf("sense (wrt plus): %s\n",Intlist_to_string(sense_endpoints)));
  debug13(printf("antisense (wrt plus): %s\n",Intlist_to_string(antisense_endpoints)));
  debug13(printf("sense lefts: %s\n",Univcoordlist_to_string(sense_lefts)));
  debug13(printf("antisense lefts: %s\n",Univcoordlist_to_string(antisense_lefts)));
  debug13(printf("sense nmismatches: %s\n",Intlist_to_string(sense_nmismatches)));
  debug13(printf("antisense nmismatches: %s\n",Intlist_to_string(antisense_nmismatches)));

#ifdef DEBUG13
  printf("Sense junctions\n");
  for (p = sense_junctions; p != NULL; p = List_next(p)) {
    Junction_print(List_head(p));
  }
  printf("\n");
  printf("Antisense junctions\n");
  for (p = antisense_junctions; p != NULL; p = List_next(p)) {
    Junction_print(List_head(p));
  }
  printf("\n");

  printf("Sense lefts: %s\n",Univcoordlist_to_string(sense_lefts));
  printf("Antisense lefts: %s\n",Univcoordlist_to_string(antisense_lefts));
#endif


  /* Need to rely on probability filtering in splice.c to get correct
     results for sense and antisense */
  sense_acceptable_p = endpoints_acceptable_p(&sense_intronp,sense_junctions,sense_endpoints);
  antisense_acceptable_p = endpoints_acceptable_p(&antisense_intronp,antisense_junctions,
						  antisense_endpoints);
  debug13(printf("sense_acceptable_p %d, antisense_acceptable_p %d\n",
		 sense_acceptable_p,antisense_acceptable_p));
  debug13(printf("sense_intronp %d, antisense_intronp %d\n",sense_intronp,antisense_intronp));

  if (sense_acceptable_p == true && antisense_acceptable_p == true) {
    if (sense_intronp == true || right_alts_sense != NULL || left_alts_sense != NULL) {
      sense_sensedir = SENSE_FORWARD;
    } else {
      sense_sensedir = SENSE_NULL;
    }
    if (antisense_intronp == true || right_alts_antisense != NULL || left_alts_antisense != NULL) {
      antisense_sensedir = SENSE_ANTI;
    } else {
      antisense_sensedir = SENSE_NULL;
    }

    if (novelsplicingp == false ||
	(paired_end_p == false && sense_sensedir == SENSE_NULL && antisense_sensedir == SENSE_NULL)) {
      /* Create just one hit */
      if ((hit = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					  sense_nmismatches,sense_junctions,
					  querylength,query_compress,
					  /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_NULL,
					  chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					  /*extend_ends_p*/false,method,level)) == NULL) {
	debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
	Substring_free(&right_alts_sense);
	Substring_free(&left_alts_sense);
	Substring_free(&right_alts_antisense);
	Substring_free(&left_alts_antisense);

      } else {
	*foundp = true;

	if (Intlist_length(sense_endpoints) <= 2) {
	  hits = List_push(hits,(void *) hit);

	} else {
	  best_hit = hit;
	  best_nmatches = Stage3end_nmatches(hit);
	  debug13(printf("hit has %d matches\n",best_nmatches));

	  /* Try extending without the initial endpoint */
	  if ((hit1 = Stage3end_new_substrings(&(*found_score),Intlist_next(sense_endpoints),Univcoordlist_next(sense_lefts),
					       Intlist_next(sense_nmismatches),List_next(sense_junctions),
					       querylength,query_compress,
					       /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_NULL,
					       chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					       /*extend_ends_p*/true,method,level)) != NULL) {
	    if ((nmatches = Stage3end_nmatches(hit1)) > best_nmatches) {
	      best_hit = hit1;
	      best_nmatches = nmatches;
	    }
	    debug13(printf("hit1 has %d matches\n",nmatches));
	  }

	  /* Try extending without the final endpoint */
	  sense_endpoints = Intlist_reverse(Intlist_pop(Intlist_reverse(sense_endpoints),&int_ignore));
	  sense_lefts = Univcoordlist_reverse(Univcoordlist_pop(Univcoordlist_reverse(sense_lefts),&univcoord_ignore));
	  sense_nmismatches = Intlist_reverse(Intlist_pop(Intlist_reverse(sense_nmismatches),&int_ignore));
	  sense_junctions = List_reverse(List_pop(List_reverse(sense_junctions),(void **) &junction));
	  Junction_free(&junction);
	  if ((hit2 = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					       sense_nmismatches,sense_junctions,
					       querylength,query_compress,
					       /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_NULL,
					       chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					       /*extend_ends_p*/true,method,level)) != NULL) {
	    if ((nmatches = Stage3end_nmatches(hit2)) > best_nmatches) {
	      best_hit = hit2;
	      best_nmatches = nmatches;
	    }
	    debug13(printf("hit2 has %d matches\n",nmatches));
	  }

	  /* Take the best hit */
	  if (hit == best_hit) {
	    hits = List_push(hits,(void *) hit);
	    Stage3end_free(&hit1);
	    Stage3end_free(&hit2);
	  } else if (hit1 == best_hit) {
	    hits = List_push(hits,(void *) hit1);
	    Stage3end_free(&hit);
	    Stage3end_free(&hit2);
	  } else {
	    hits = List_push(hits,(void *) hit2);
	    Stage3end_free(&hit);
	    Stage3end_free(&hit1);
	  }
	}
      }

    } else {
      /* Create both sense and antisense hits */
      debug13(printf("Creating both sense and antisense hits\n"));
      
      if ((hit = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					  sense_nmismatches,sense_junctions,
					  querylength,query_compress,
					  right_alts_sense,left_alts_sense,plusp,genestrand,/*sensedir*/SENSE_FORWARD,
					  chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					  /*extend_ends_p*/false,method,level)) == NULL) {
	debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
	Substring_free(&right_alts_sense);
	Substring_free(&left_alts_sense);

      } else {
	*foundp = true;

	if (Intlist_length(sense_endpoints) <= 2) {
	  hits = List_push(hits,(void *) hit);

	} else {
	  best_hit = hit;
	  best_nmatches = Stage3end_nmatches(hit);
	  debug13(printf("hit has %d matches\n",best_nmatches));

	  /* Try extending without the initial endpoint */
	  if ((hit1 = Stage3end_new_substrings(&(*found_score),Intlist_next(sense_endpoints),Univcoordlist_next(sense_lefts),
					       Intlist_next(sense_nmismatches),List_next(sense_junctions),
					       querylength,query_compress,
					       /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_FORWARD,
					       chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					       /*extend_ends_p*/true,method,level)) != NULL) {
	    if ((nmatches = Stage3end_nmatches(hit1)) > best_nmatches) {
	      best_hit = hit1;
	      best_nmatches = nmatches;
	    }
	    debug13(printf("hit1 has %d matches\n",nmatches));
	  }

	  /* Try extending without the final endpoint */
	  sense_endpoints = Intlist_reverse(Intlist_pop(Intlist_reverse(sense_endpoints),&int_ignore));
	  sense_lefts = Univcoordlist_reverse(Univcoordlist_pop(Univcoordlist_reverse(sense_lefts),&univcoord_ignore));
	  sense_nmismatches = Intlist_reverse(Intlist_pop(Intlist_reverse(sense_nmismatches),&int_ignore));
	  sense_junctions = List_reverse(List_pop(List_reverse(sense_junctions),(void **) &junction));
	  Junction_free(&junction);
	  if ((hit2 = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					       sense_nmismatches,sense_junctions,
					       querylength,query_compress,
					       /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_FORWARD,
					       chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					       /*extend_ends_p*/true,method,level)) != NULL) {
	    if ((nmatches = Stage3end_nmatches(hit2)) > best_nmatches) {
	      best_hit = hit2;
	      best_nmatches = nmatches;
	    }
	    debug13(printf("hit2 has %d matches\n",nmatches));
	  }

	  /* Take the best hit */
	  if (hit == best_hit) {
	    hits = List_push(hits,(void *) hit);
	    Stage3end_free(&hit1);
	    Stage3end_free(&hit2);
	  } else if (hit1 == best_hit) {
	    hits = List_push(hits,(void *) hit1);
	    Stage3end_free(&hit);
	    Stage3end_free(&hit2);
	  } else {
	    hits = List_push(hits,(void *) hit2);
	    Stage3end_free(&hit);
	    Stage3end_free(&hit1);
	  }
	}
      }

      if ((hit = Stage3end_new_substrings(&(*found_score),antisense_endpoints,antisense_lefts,
					  antisense_nmismatches,antisense_junctions,
					  querylength,query_compress,
					  right_alts_antisense,left_alts_antisense,plusp,genestrand,/*sensedir*/SENSE_ANTI,
					  chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					  /*extend_ends_p*/false,method,level)) == NULL) {
	debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
	Substring_free(&right_alts_antisense);
	Substring_free(&left_alts_antisense);

      } else {
	*foundp = true;

	if (Intlist_length(antisense_endpoints) <= 2) {
	  hits = List_push(hits,(void *) hit);

	} else {
	  best_hit = hit;
	  best_nmatches = Stage3end_nmatches(hit);
	  debug13(printf("hit has %d matches\n",best_nmatches));

	  /* Try extending without the initial endpoint */
	  if ((hit1 = Stage3end_new_substrings(&(*found_score),Intlist_next(antisense_endpoints),Univcoordlist_next(antisense_lefts),
					       Intlist_next(antisense_nmismatches),List_next(antisense_junctions),
					       querylength,query_compress,
					       /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_ANTI,
					       chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					       /*extend_ends_p*/true,method,level)) != NULL) {
	    if ((nmatches = Stage3end_nmatches(hit1)) > best_nmatches) {
	      best_hit = hit1;
	      best_nmatches = nmatches;
	    }
	    debug13(printf("hit1 has %d matches\n",nmatches));
	  }

	  /* Try extending without the final endpoint */
	  antisense_endpoints = Intlist_reverse(Intlist_pop(Intlist_reverse(antisense_endpoints),&int_ignore));
	  antisense_lefts = Univcoordlist_reverse(Univcoordlist_pop(Univcoordlist_reverse(antisense_lefts),&univcoord_ignore));
	  antisense_nmismatches = Intlist_reverse(Intlist_pop(Intlist_reverse(antisense_nmismatches),&int_ignore));
	  antisense_junctions = List_reverse(List_pop(List_reverse(antisense_junctions),(void **) &junction));
	  Junction_free(&junction);
	  if ((hit2 = Stage3end_new_substrings(&(*found_score),antisense_endpoints,antisense_lefts,
					       antisense_nmismatches,antisense_junctions,
					       querylength,query_compress,
					       /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_ANTI,
					       chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					       /*extend_ends_p*/true,method,level)) != NULL) {
	    if ((nmatches = Stage3end_nmatches(hit2)) > best_nmatches) {
	      best_hit = hit2;
	      best_nmatches = nmatches;
	    }
	    debug13(printf("hit2 has %d matches\n",nmatches));
	  }

	  /* Take the best hit */
	  if (hit == best_hit) {
	    hits = List_push(hits,(void *) hit);
	    Stage3end_free(&hit1);
	    Stage3end_free(&hit2);
	  } else if (hit1 == best_hit) {
	    hits = List_push(hits,(void *) hit1);
	    Stage3end_free(&hit);
	    Stage3end_free(&hit2);
	  } else {
	    hits = List_push(hits,(void *) hit2);
	    Stage3end_free(&hit);
	    Stage3end_free(&hit1);
	  }
	}
      }
    }
    
  } else if (sense_acceptable_p == true) {
    if (sense_intronp == true || right_alts_sense != NULL || left_alts_sense != NULL) {
      sensedir = SENSE_FORWARD;
    } else {
      sensedir = SENSE_NULL;
    }
    if ((hit = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					sense_nmismatches,sense_junctions,
					querylength,query_compress,
					right_alts_sense,left_alts_sense,plusp,genestrand,sensedir,
					chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					/*extend_ends_p*/false,method,level)) == NULL) {
      debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
      Substring_free(&right_alts_sense);
      Substring_free(&left_alts_sense);

    } else {
      *foundp = true;

      if (Intlist_length(sense_endpoints) <= 2) {
	hits = List_push(hits,(void *) hit);

      } else {
	best_hit = hit;
	best_nmatches = Stage3end_nmatches(hit);
	debug13(printf("hit has %d matches\n",best_nmatches));

	/* Try extending without the initial endpoint */
	if ((hit1 = Stage3end_new_substrings(&(*found_score),Intlist_next(sense_endpoints),Univcoordlist_next(sense_lefts),
					     Intlist_next(sense_nmismatches),List_next(sense_junctions),
					     querylength,query_compress,
					     /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sensedir,
					     chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					     /*extend_ends_p*/true,method,level)) != NULL) {
	  if ((nmatches = Stage3end_nmatches(hit1)) > best_nmatches) {
	    best_hit = hit1;
	    best_nmatches = nmatches;
	  }
	  debug13(printf("hit1 has %d matches\n",nmatches));
	}
	
	/* Try extending without the final endpoint */
	sense_endpoints = Intlist_reverse(Intlist_pop(Intlist_reverse(sense_endpoints),&int_ignore));
	sense_lefts = Univcoordlist_reverse(Univcoordlist_pop(Univcoordlist_reverse(sense_lefts),&univcoord_ignore));
	sense_nmismatches = Intlist_reverse(Intlist_pop(Intlist_reverse(sense_nmismatches),&int_ignore));
	sense_junctions = List_reverse(List_pop(List_reverse(sense_junctions),(void **) &junction));
	Junction_free(&junction);
	if ((hit2 = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					     sense_nmismatches,sense_junctions,
					     querylength,query_compress,
					     /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sensedir,
					     chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					     /*extend_ends_p*/true,method,level)) != NULL) {
	  if ((nmatches = Stage3end_nmatches(hit2)) > best_nmatches) {
	    best_hit = hit2;
	    best_nmatches = nmatches;
	  }
	  debug13(printf("hit2 has %d matches\n",nmatches));
	}

	/* Take the best hit */
	if (hit == best_hit) {
	  hits = List_push(hits,(void *) hit);
	  Stage3end_free(&hit1);
	  Stage3end_free(&hit2);
	} else if (hit1 == best_hit) {
	  hits = List_push(hits,(void *) hit1);
	  Stage3end_free(&hit);
	  Stage3end_free(&hit2);
	} else {
	  hits = List_push(hits,(void *) hit2);
	  Stage3end_free(&hit);
	  Stage3end_free(&hit1);
	}
      }
    }

    Substring_free(&right_alts_antisense);
    Substring_free(&left_alts_antisense);

  } else if (antisense_acceptable_p == true) {
    if (antisense_intronp == true || right_alts_antisense != NULL || left_alts_antisense != NULL) {
      sensedir = SENSE_ANTI;
    } else {
      sensedir = SENSE_NULL;
    }
    if ((hit = Stage3end_new_substrings(&(*found_score),antisense_endpoints,antisense_lefts,
					antisense_nmismatches,antisense_junctions,
					querylength,query_compress,
					right_alts_antisense,left_alts_antisense,plusp,genestrand,sensedir,
					chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					/*extend_ends_p*/false,method,level)) == NULL) {
      debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
      Substring_free(&right_alts_antisense);
      Substring_free(&left_alts_antisense);

    } else {
      *foundp = true;

      if (Intlist_length(antisense_endpoints) <= 2) {
	hits = List_push(hits,(void *) hit);

      } else {
	best_hit = hit;
	best_nmatches = Stage3end_nmatches(hit);
	debug13(printf("hit has %d matches\n",best_nmatches));

	/* Try extending without the initial endpoint */
	if ((hit1 = Stage3end_new_substrings(&(*found_score),Intlist_next(antisense_endpoints),Univcoordlist_next(antisense_lefts),
					     Intlist_next(antisense_nmismatches),List_next(antisense_junctions),
					     querylength,query_compress,
					     /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sensedir,
					     chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					     /*extend_ends_p*/true,method,level)) != NULL) {
	  if ((nmatches = Stage3end_nmatches(hit1)) > best_nmatches) {
	    best_hit = hit1;
	    best_nmatches = nmatches;
	  }
	  debug13(printf("hit1 has %d matches\n",nmatches));
	}
	
	/* Try extending without the final endpoint */
	antisense_endpoints = Intlist_reverse(Intlist_pop(Intlist_reverse(antisense_endpoints),&int_ignore));
	antisense_lefts = Univcoordlist_reverse(Univcoordlist_pop(Univcoordlist_reverse(antisense_lefts),&univcoord_ignore));
	antisense_nmismatches = Intlist_reverse(Intlist_pop(Intlist_reverse(antisense_nmismatches),&int_ignore));
	antisense_junctions = List_reverse(List_pop(List_reverse(antisense_junctions),(void **) &junction));
	Junction_free(&junction);
	if ((hit2 = Stage3end_new_substrings(&(*found_score),antisense_endpoints,antisense_lefts,
					     antisense_nmismatches,antisense_junctions,
					     querylength,query_compress,
					     /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sensedir,
					     chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					     /*extend_ends_p*/true,method,level)) != NULL) {
	  if ((nmatches = Stage3end_nmatches(hit2)) > best_nmatches) {
	    best_hit = hit2;
	    best_nmatches = nmatches;
	  }
	  debug13(printf("hit2 has %d matches\n",nmatches));
	}

	/* Take the best hit */
	if (hit == best_hit) {
	  hits = List_push(hits,(void *) hit);
	  Stage3end_free(&hit1);
	  Stage3end_free(&hit2);
	} else if (hit1 == best_hit) {
	  hits = List_push(hits,(void *) hit1);
	  Stage3end_free(&hit);
	  Stage3end_free(&hit2);
	} else {
	  hits = List_push(hits,(void *) hit2);
	  Stage3end_free(&hit);
	  Stage3end_free(&hit1);
	}
      }
    }

    Substring_free(&right_alts_sense);
    Substring_free(&left_alts_sense);

  } else {
    /* Neither set of junctions/endpoints works */
    debug13(printf("ABORTING BECAUSE ENDPOINTS ARE UNACCEPTABLE (HAVE BREAKS)\n"));
    Substring_free(&right_alts_sense);
    Substring_free(&left_alts_sense);
    Substring_free(&right_alts_antisense);
    Substring_free(&left_alts_antisense);
  }


  Intlist_free(&sense_nmismatches);
  Intlist_free(&antisense_nmismatches);
  Univcoordlist_free(&sense_lefts);
  Univcoordlist_free(&antisense_lefts);
  Intlist_free(&sense_endpoints);
  Intlist_free(&antisense_endpoints);
  Junction_gc(&sense_junctions);
  Junction_gc(&antisense_junctions);
  
  for (p = super_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    Univdiag_free(&diagonal);
  }
  List_free(&super_path);

  Spliceinfo_free(&spliceinfo);

  return hits;
}


#if 0
/* Old code */
List_T
Path_solve_from_diagonals (bool *foundp, int *found_score, List_T *complete_path, List_T hits,
			   List_T middle_path, List_T right_diagonals, List_T left_diagonals,
			   char *queryptr, Compress_T query_compress, int querylength,
			   Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
			   bool plusp, int genestrand, int nmismatches_allowed,
			   bool paired_end_p, Method_T method, int level) {
  List_T large_kmer_path;
  Univdiag_T diagonal;
  int querystart, queryend;

  List_T local_left_diagonals, local_right_diagonals;
  Univcoord_T low, high;
  Univcoord_T *diagonals;
  int ndiagonals, i;

  Univcoord_T right_univdiagonal_sense, right_univdiagonal_antisense,
    left_univdiagonal_sense, left_univdiagonal_antisense;
  int right_splicepos_sense, right_splicepos_antisense,
    left_splicepos_sense, left_splicepos_antisense;
  int right_nmismatches_sense, right_nmismatches_antisense,
    left_nmismatches_sense, left_nmismatches_antisense;
  Junction_T right_junction_sense, right_junction_antisense,
    left_junction_sense, left_junction_antisense;
  Substring_T right_alts_sense, right_alts_antisense, left_alts_sense, left_alts_antisense;

#ifdef DEBUG13
  List_T p;
#endif


  large_kmer_path = resolve_end_diagonals(&right_univdiagonal_sense,&right_splicepos_sense,
					  &right_nmismatches_sense,&right_junction_sense,&right_alts_sense,
					  &right_univdiagonal_antisense,&right_splicepos_antisense,
					  &right_nmismatches_antisense,&right_junction_antisense,&right_alts_antisense,
					  &left_univdiagonal_sense,&left_splicepos_sense,
					  &left_nmismatches_sense,&left_junction_sense,&left_alts_sense,
					  &left_univdiagonal_antisense,&left_splicepos_antisense,
					  &left_nmismatches_antisense,&left_junction_antisense,&left_alts_antisense,
					  
					  middle_path,right_diagonals,left_diagonals,
					  querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
					  plusp,genestrand,nmismatches_allowed);

  if (localdb == NULL) {
    local_left_diagonals = local_right_diagonals = (List_T) NULL;

  } else {
    if (left_alts_sense != NULL || left_alts_antisense != NULL) {
      /* Do not extend an alternative substring */
      local_left_diagonals = (List_T) NULL;

    } else if (left_univdiagonal_sense > 0 || left_univdiagonal_antisense > 0) {
      /* Do not extend a sense/antisense splice */
      local_left_diagonals = (List_T) NULL;

    } else {
      diagonal = (Univdiag_T) List_head(large_kmer_path);
      debug13(printf("Overall querystart is %d\n",diagonal->querystart));
      if ((querystart = diagonal->querystart) < local1part) {
	/* Localdb not possible */
	local_left_diagonals = (List_T) NULL;
      } else if (querystart >= index1part + index1interval - 1) {
	/* Should have been found by indexdb */
	local_left_diagonals = (List_T) NULL;
      } else {
	low = subtract_bounded(diagonal->univdiagonal,overall_end_distance_genome,chroffset);
	high = add_bounded(diagonal->univdiagonal,max_insertionlen,chrhigh);
	debug13(printf("(1) Calling localdb_get_diagonals at %d..%d at genomic %u..%u (%u..%u), plusp %d\n",0,querystart,
		       low,high,low-chroffset,high-chroffset,plusp));
	if (plusp) {
	  diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,
					    /*pos5*/0,/*pos3*/querystart,low,high,
					    /*plusp*/true,genestrand);
	} else {
	  diagonals = Localdb_get_diagonals(&ndiagonals,localdb2,queryptr,
					    /*pos5*/0,/*pos3*/querystart,low,high,
					    /*plusp*/false,genestrand);
	}
	debug13(printf("Got %d localdb diagonals\n",ndiagonals));

	local_left_diagonals = (List_T) NULL;
	for (i = ndiagonals - 1; i >= 0; i--) {
	  local_left_diagonals = List_push(local_left_diagonals,Univdiag_new(0,querystart,diagonals[i]));
	}
#ifdef LARGE_GENOMES
	FREE(diagonals);
#else
	FREE_ALIGN(diagonals);	/* Aligned from Localdb_get_diagonals */
#endif
      }
    }

    if (right_alts_sense != NULL || right_alts_antisense != NULL) {
      /* Do not extend an alternative substring */
      local_right_diagonals = (List_T) NULL;

    } else if (right_univdiagonal_sense > 0 || right_univdiagonal_antisense > 0) {
      /* Do not extend a sense/antisense splice */
      local_right_diagonals = (List_T) NULL;

    } else {
      diagonal = (Univdiag_T) List_last_value(large_kmer_path);
      debug13(printf("Overall queryend is %d\n",diagonal->queryend));
      if (querylength - (queryend = diagonal->queryend) < local1part) {
	/* Localdb not possible */
	local_right_diagonals = (List_T) NULL;
      } else if (querylength - queryend >= index1part + index1interval - 1) {
	/* Should have been found by indexdb */
	local_right_diagonals = (List_T) NULL;
      } else {
	low = subtract_bounded(diagonal->univdiagonal,max_insertionlen,chroffset);
	high = add_bounded(diagonal->univdiagonal,overall_end_distance_genome,chrhigh);
	debug13(printf("(2) Calling localdb_get_diagonals at %d..%d at genomic %u..%u (%u..%u), plusp %d\n",
		       queryend,querylength,low,high,low-chroffset,high-chroffset,plusp));
	if (plusp) {
	  diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,
					    /*pos5*/queryend,/*pos5*/querylength,low,high,
					    /*plusp*/true,genestrand);
	} else {
	  diagonals = Localdb_get_diagonals(&ndiagonals,localdb2,queryptr,
					    /*pos5*/queryend,/*pos5*/querylength,low,high,
					    /*plusp*/false,genestrand);
	}
	debug13(printf("Got %d localdb diagonals\n",ndiagonals));

	local_right_diagonals = (List_T) NULL;
	for (i = ndiagonals - 1; i >= 0; i--) {
	  local_right_diagonals = List_push(local_right_diagonals,Univdiag_new(queryend,querylength,diagonals[i]));
	}
#ifdef LARGE_GENOMES
	FREE(diagonals);
#else
	FREE_ALIGN(diagonals);	/* Aligned from Localdb_get_diagonals */
#endif
      }
    }
  }
    
  if (local_left_diagonals == NULL && local_right_diagonals == NULL) {
    *complete_path = large_kmer_path;

  } else if (large_kmer_path == NULL) {
    *complete_path = (List_T) NULL;

  } else {
    *complete_path = resolve_end_diagonals(&right_univdiagonal_sense,&right_splicepos_sense,
					   &right_nmismatches_sense,&right_junction_sense,&right_alts_sense,
					   &right_univdiagonal_antisense,&right_splicepos_antisense,
					   &right_nmismatches_antisense,&right_junction_antisense,&right_alts_antisense,
					   &left_univdiagonal_sense,&left_splicepos_sense,
					   &left_nmismatches_sense,&left_junction_sense,&left_alts_sense,
					   &left_univdiagonal_antisense,&left_splicepos_antisense,
					   &left_nmismatches_antisense,&left_junction_antisense,&left_alts_antisense,
					  
					   large_kmer_path,local_right_diagonals,local_left_diagonals,
					   querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
					   plusp,genestrand,nmismatches_allowed);
  }

  if (*complete_path != NULL) {
    /* TODO: Incorporate localdb into solve_via_segments_genome, to solve each sense/antisense possibility separately */
    hits = solve_via_segments_genome(&(*foundp),&(*found_score),hits,*complete_path,
				     right_univdiagonal_sense,right_splicepos_sense,
				     right_nmismatches_sense,right_junction_sense,right_alts_sense,
				     right_univdiagonal_antisense,right_splicepos_antisense,
				     right_nmismatches_antisense,right_junction_antisense,right_alts_antisense,
				     left_univdiagonal_sense,left_splicepos_sense,
				     left_nmismatches_sense,left_junction_sense,left_alts_sense,
				     left_univdiagonal_antisense,left_splicepos_antisense,
				     left_nmismatches_antisense,left_junction_antisense,left_alts_antisense,
				     chrnum,chroffset,chrhigh,chrlength,querylength,query_compress,
				     plusp,genestrand,nmismatches_allowed,paired_end_p,method,level);

    if (*foundp == true) {
      Univdiag_gc(&(*complete_path));
    } else {
      fix_query_ends(*complete_path);
#ifdef DEBUG13
      printf("COMPLETE_PATH\n");
      for (p = *complete_path; p != NULL; p = List_next(p)) {
	diagonal = (Univdiag_T) List_head(p);
	printf("%d..%d\n",diagonal->querystart,diagonal->queryend);
      }
#endif
    }
  }

  return hits;
}

#elif defined(USE_LOCALDB)

List_T
Path_solve_from_diagonals (bool *foundp, int *found_score, List_T *complete_path, List_T hits,
			   Univdiag_T middle_diagonal, List_T right_diagonals, List_T left_diagonals,
			   char *queryptr, int querylength, Compress_T query_compress, 
			   Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
			   bool plusp, int genestrand, int nmismatches_allowed, bool paired_end_p,
			   Method_T method, int level) {
  List_T best_right_path = NULL, best_left_path = NULL, p;
  Univdiag_T middle_diagonal_5, middle_diagonal_3, diagonal;
  int querystart, queryend;

  List_T local_left_diagonals, local_right_diagonals;
  Univcoord_T low, high;
  Univcoord_T *diagonals;
  int ndiagonals, i;

  Univcoord_T right_univdiagonal_sense, right_univdiagonal_antisense,
    left_univdiagonal_sense, left_univdiagonal_antisense;
  int right_splicepos_sense, right_splicepos_antisense,
    left_splicepos_sense, left_splicepos_antisense;
  int right_nmismatches_sense, right_nmismatches_antisense,
    left_nmismatches_sense, left_nmismatches_antisense;
  Junction_T right_junction_sense, right_junction_antisense,
    left_junction_sense, left_junction_antisense;
  Substring_T right_alts_sense, right_alts_antisense, left_alts_sense, left_alts_antisense;
  Substring_T substring;


  if ((substring = Substring_new(/*orig_nmismatches*/-1,middle_diagonal->univdiagonal,
				 middle_diagonal->querystart,middle_diagonal->queryend,querylength,
				 plusp,genestrand,query_compress,/*start_endtype*/END,
				 /*end_endtype*/END,chrnum,chroffset,chrhigh,chrlength,
				 /*trim_left_action*/COMPUTE_TRIM,/*trim_right_action*/COMPUTE_TRIM,
				 /*minlength*/0,/*orig_senseir*/SENSE_NULL,/*transcriptomep*/false,
				 /*chrnum_fixed_p*/false)) != NULL) {
    chrnum = Substring_chrnum(substring);
    chroffset = Substring_chroffset(substring);
    chrhigh = Substring_chrhigh(substring);
    chrlength = Substring_chrlength(substring);
    Substring_free(&substring);
  }


  if (right_diagonals == NULL) {
    middle_diagonal_3 = middle_diagonal;
    queryend = middle_diagonal_3->queryend;

    right_univdiagonal_sense = right_univdiagonal_antisense = (Univcoord_T) 0;
    right_splicepos_sense = right_splicepos_antisense = 0;
    right_nmismatches_sense = right_nmismatches_antisense = -1; /* Need to recalculate */
    right_alts_sense = right_alts_antisense = (Substring_T) NULL;
    /* Continue below for possible localdb extension */

  } else {
    /* best_right_path is independent of sense/antisense */
    best_right_path = resolve_right_diagonals(&right_univdiagonal_sense,&right_splicepos_sense,
					      &right_nmismatches_sense,
#ifdef PRECOMPUTE_JUNCTIONS
					      &right_junction_sense,
#endif
					      &right_alts_sense,
					      &right_univdiagonal_antisense,&right_splicepos_antisense,
					      &right_nmismatches_antisense,
#ifdef PRECOMPUTE_JUNCTIONS
					      &right_junction_antisense,
#endif
					      &right_alts_antisense,

					      best_right_path,middle_diagonal,right_diagonals,
					      querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
					      plusp,genestrand,nmismatches_allowed);
    if (right_alts_sense != NULL || right_alts_antisense != NULL) {
      /* Not possible to add localdb because of the ambiguity */
      queryend = querylength;
    } else if (right_univdiagonal_sense > 0 || right_univdiagonal_antisense > 0) {
      /* Not possible to add localdb until we decide on sense vs antisense */
      queryend = querylength;
    } else if (best_right_path == NULL) {
      middle_diagonal_3 = middle_diagonal;
      queryend = middle_diagonal_3->queryend;
    } else {
      middle_diagonal_3 = (Univdiag_T) List_head(best_right_path);
      queryend = middle_diagonal_3->queryend;
    }
  }
      
  debug13(printf("Overall queryend is %d\n",queryend));
  if (localdb == NULL) {
    /* Localdb not possible */
    local_right_diagonals = (List_T) NULL;
  } else if (querylength - queryend < local1part) {
    /* Localdb not possible */
    local_right_diagonals = (List_T) NULL;
  } else if (querylength - queryend >= index1part + index1interval - 1) {
    /* Should have been found by indexdb */
    local_right_diagonals = (List_T) NULL;
  } else {
    if (best_right_path != NULL) {
      middle_diagonal_3 = (Univdiag_T) List_head(best_right_path);
    } else {
      middle_diagonal_3 = middle_diagonal;
    }

    low = subtract_bounded(middle_diagonal_3->univdiagonal,max_insertionlen,chroffset);
    high = add_bounded(middle_diagonal_3->univdiagonal,overall_end_distance_genome,chrhigh);

    debug13(printf("(3) Calling localdb_get_diagonals at %d..%d at genomic %u..%u (%u..%u), plusp %d\n",
		   queryend,querylength,low,high,low-chroffset,high-chroffset,plusp));
    if (plusp) {
      diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,
					/*pos5*/queryend,/*pos5*/querylength,low,high,
					/*plusp*/true,genestrand);
    } else {
      diagonals = Localdb_get_diagonals(&ndiagonals,localdb2,queryptr,
					/*pos5*/queryend,/*pos5*/querylength,low,high,
					/*plusp*/false,genestrand);
    }
    debug13(printf("Got %d localdb diagonals\n",ndiagonals));
    
    if (ndiagonals > 0) {
      /* Consider localdb diagonals */
      local_right_diagonals = (List_T) NULL;
      for (i = ndiagonals - 1; i >= 0; i--) {
	local_right_diagonals = List_push(local_right_diagonals,Univdiag_new(queryend,querylength,diagonals[i]));
      }
#ifdef LARGE_GENOMES
      FREE(diagonals);
#else
      FREE_ALIGN(diagonals);  /* Aligned from Localdb_get_diagonals */
#endif

      best_right_path = resolve_right_diagonals(&right_univdiagonal_sense,&right_splicepos_sense,
						&right_nmismatches_sense,
#ifdef PRECOMPUTE_JUNCTIONS
						&right_junction_sense,
#endif
						&right_alts_sense,
						&right_univdiagonal_antisense,&right_splicepos_antisense,
						&right_nmismatches_antisense,
#ifdef PRECOMPUTE_JUNCTIONS
						&right_junction_antisense,
#endif
						&right_alts_antisense,

						best_right_path,middle_diagonal,local_right_diagonals,
						querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
						plusp,genestrand,nmismatches_allowed);
      Univdiag_gc(&local_right_diagonals);
    }
  }

  if (left_diagonals == NULL) {
    middle_diagonal_5 = middle_diagonal;
    querystart = middle_diagonal_5->querystart;

    left_univdiagonal_sense = left_univdiagonal_antisense = (Univcoord_T) 0;
    left_splicepos_sense = left_splicepos_antisense = 0;
    left_nmismatches_sense = left_nmismatches_antisense = -1; /* Need to recalculate */
    left_alts_sense = left_alts_antisense = (Substring_T) NULL;
    /* Continue below for possible localdb extension */

  } else {
    /* best_left_path is independent of sense/antisense */
    best_left_path = resolve_left_diagonals(&left_univdiagonal_sense,&left_splicepos_sense,
					    &left_nmismatches_sense,
#ifdef PRECOMPUTE_JUNCTIONS
					    &left_junction_sense,
#endif
					    &left_alts_sense,
					    &left_univdiagonal_antisense,&left_splicepos_antisense,
					    &left_nmismatches_antisense,
#ifdef PRECOMPUTE_JUNCTIONS
					    &left_junction_antisense,
#endif
					    &left_alts_antisense,
					  
					    best_left_path,middle_diagonal,left_diagonals,
					    querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
					    plusp,genestrand,nmismatches_allowed);
    if (left_alts_sense != NULL || left_alts_antisense != NULL) {
      /* Not possible to add localdb because of the ambiguity */
      querystart = 0;
    } else if (left_univdiagonal_sense > 0 || left_univdiagonal_antisense > 0) {
      /* Not possible to add localdb until we decide on sense vs antisense */
      querystart = 0;
    } else if (best_left_path == NULL) {
      middle_diagonal_5 = middle_diagonal;
      querystart = middle_diagonal_5->querystart;
    } else {
      middle_diagonal_5 = (Univdiag_T) List_head(best_left_path);
      querystart = middle_diagonal_5->querystart;
    }
  }

  debug13(printf("Overall querystart is %d\n",querystart));
  if (localdb == NULL) {
    /* Localdb not possible */
    local_left_diagonals = (List_T) NULL;
  } else if (querystart < local1part) {
    /* Localdb not possible */
    local_left_diagonals = (List_T) NULL;
  } else if (querystart >= index1part + index1interval - 1) {
    /* Should have been found by indexdb */
    local_left_diagonals = (List_T) NULL;
  } else {
    if (best_left_path != NULL) {
      middle_diagonal_5 = (Univdiag_T) List_head(best_left_path);
    } else {
      middle_diagonal_5 = middle_diagonal;
    }

    low = subtract_bounded(middle_diagonal_5->univdiagonal,overall_end_distance_genome,chroffset);
    high = add_bounded(middle_diagonal_5->univdiagonal,max_insertionlen,chrhigh);
    debug13(printf("Adding max_insertionlen %d to middle_diagonal %u\n",
		   max_insertionlen,middle_diagonal_5->univdiagonal - chroffset));
    debug13(printf("(4) Calling localdb_get_diagonals at %d..%d at genomic %u..%u (%u..%u), plusp %d\n",0,querystart,
		   low,high,low-chroffset,high-chroffset,plusp));
    if (plusp) {
      diagonals = Localdb_get_diagonals(&ndiagonals,localdb,queryptr,
					/*pos5*/0,/*pos3*/querystart,low,high,
					/*plusp*/true,genestrand);
    } else {
      diagonals = Localdb_get_diagonals(&ndiagonals,localdb2,queryptr,
					/*pos5*/0,/*pos3*/querystart,low,high,
					/*plusp*/false,genestrand);
    }
    debug13(printf("Got %d localdb diagonals\n",ndiagonals));

    if (ndiagonals > 0) {
      /* Consider localdb diagonals */
      local_left_diagonals = (List_T) NULL;
      for (i = ndiagonals - 1; i >= 0; i--) {
	local_left_diagonals = List_push(local_left_diagonals,Univdiag_new(0,querystart,diagonals[i]));
      }
#ifdef LARGE_GENOMES
      FREE(diagonals);
#else      
      FREE_ALIGN(diagonals);	/* Aligned from Localdb_get_diagonals */
#endif

      best_left_path = resolve_left_diagonals(&left_univdiagonal_sense,&left_splicepos_sense,
					      &left_nmismatches_sense,
#ifdef PRECOMPUTE_JUNCTIONS
					      &left_junction_sense,
#endif
					      &left_alts_sense,
					      &left_univdiagonal_antisense,&left_splicepos_antisense,
					      &left_nmismatches_antisense,
#ifdef PRECOMPUTE_JUNCTIONS
					      &left_junction_antisense,
#endif
					      &left_alts_antisense,
					      
					      best_left_path,middle_diagonal,local_left_diagonals,
					      querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
					      plusp,genestrand,nmismatches_allowed);
      Univdiag_gc(&local_left_diagonals);
    }
  }

  /* Form complete_path by extending middle_path with best_right_path and best_left_path */
  *complete_path = (List_T) NULL;

  /* A. Process best_right_path (already in reverse order) */
  for (p = best_right_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    debug13(printf("Pushing right diagonal onto complete path: query %d..%d, diagonal %u\n",
		   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal - chroffset));
    *complete_path = List_push(*complete_path,(void *) diagonal);
  }
  List_free(&best_right_path);

  /* B. Process original middle path */
  *complete_path = List_push(*complete_path,(void *) Univdiag_copy(middle_diagonal));

  /* A. Process best_left_path */
  best_left_path = List_reverse(best_left_path);
  for (p = best_left_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    debug13(printf("Pushing left diagonal onto complete path: query %d..%d, diagonal %u\n",
		   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal - chroffset));
    *complete_path = List_push(*complete_path,(void *) diagonal);
  }
  List_free(&best_left_path);

#ifdef DEBUG13
    printf("COMPLETE_PATH\n");
    for (p = *complete_path; p != NULL; p = List_next(p)) {
      diagonal = (Univdiag_T) List_head(p);
      printf("%d..%d\n",diagonal->querystart,diagonal->queryend);
    }
#endif

    hits = solve_via_segments_genome(&(*foundp),&(*found_score),hits,*complete_path,
				     right_univdiagonal_sense,right_splicepos_sense,
				     right_nmismatches_sense,right_junction_sense,right_alts_sense,
				     right_univdiagonal_antisense,right_splicepos_antisense,
				     right_nmismatches_antisense,right_junction_antisense,right_alts_antisense,
				     left_univdiagonal_sense,left_splicepos_sense,
				     left_nmismatches_sense,left_junction_sense,left_alts_sense,
				     left_univdiagonal_antisense,left_splicepos_antisense,
				     left_nmismatches_antisense,left_junction_antisense,left_alts_antisense,
				     chrnum,chroffset,chrhigh,chrlength,querylength,query_compress,
				     plusp,genestrand,nmismatches_allowed,paired_end_p,method,level);
  
  /* TODO: Return super_path instead of complete_path */

  if (*foundp == true) {
    Univdiag_gc(&(*complete_path));
  } else {
    fix_query_ends(*complete_path);
#ifdef DEBUG13
    printf("COMPLETE_PATH\n");
    for (p = *complete_path; p != NULL; p = List_next(p)) {
      diagonal = (Univdiag_T) List_head(p);
      printf("%d..%d\n",diagonal->querystart,diagonal->queryend);
    }
#endif
  }

  return hits;
}


#else

/* Does not use localdb */
List_T
Path_solve_from_diagonals (bool *foundp, int *found_score, List_T *complete_path, List_T hits,
			   Univdiag_T middle_diagonal, List_T right_diagonals, List_T left_diagonals,
			   char *queryptr, char *queryuc_ptr, int querylength, Compress_T query_compress, 
			   Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
			   bool plusp, int genestrand, int nmismatches_allowed, bool paired_end_p,
			   Method_T method, int level) {

  Univcoord_T right_univdiagonal_sense, right_univdiagonal_antisense,
    left_univdiagonal_sense, left_univdiagonal_antisense;
  int right_splicepos_sense, right_splicepos_antisense,
    left_splicepos_sense, left_splicepos_antisense;
  int right_nmismatches_sense, right_nmismatches_antisense,
    left_nmismatches_sense, left_nmismatches_antisense;
  Junction_T right_junction_sense, right_junction_antisense,
    left_junction_sense, left_junction_antisense;
  Substring_T right_alts_sense, right_alts_antisense, left_alts_sense, left_alts_antisense;
  Substring_T substring;

#ifdef DEBUG13
  List_T p;
  Univdiag_T diagonal;
#endif

  if ((substring = Substring_new(/*orig_nmismatches*/-1,middle_diagonal->univdiagonal,
				 middle_diagonal->querystart,middle_diagonal->queryend,querylength,
				 plusp,genestrand,query_compress,/*start_endtype*/END,
				 /*end_endtype*/END,chrnum,chroffset,chrhigh,chrlength,
				 /*trim_left_action*/COMPUTE_TRIM,/*trim_right_action*/COMPUTE_TRIM,
				 /*minlength*/0,/*orig_senseir*/SENSE_NULL,/*transcriptomep*/false,
				 /*chrnum_fixed_p*/false)) != NULL) {
    chrnum = Substring_chrnum(substring);
    chroffset = Substring_chroffset(substring);
    chrhigh = Substring_chrhigh(substring);
    chrlength = Substring_chrlength(substring);
    Substring_free(&substring);
  }


  if ((*complete_path = resolve_end_diagonals(&right_univdiagonal_sense,&right_splicepos_sense,
					      &right_nmismatches_sense,&right_junction_sense,&right_alts_sense,
					      &right_univdiagonal_antisense,&right_splicepos_antisense,
					      &right_nmismatches_antisense,&right_junction_antisense,&right_alts_antisense,
					      &left_univdiagonal_sense,&left_splicepos_sense,
					      &left_nmismatches_sense,&left_junction_sense,&left_alts_sense,
					      &left_univdiagonal_antisense,&left_splicepos_antisense,
					      &left_nmismatches_antisense,&left_junction_antisense,&left_alts_antisense,

					      middle_diagonal,right_diagonals,left_diagonals,
					      querylength,query_compress,chrnum,chroffset,chrhigh,chrlength,
					      plusp,genestrand,nmismatches_allowed)) != NULL) {

    hits = solve_via_segments_genome(&(*foundp),&(*found_score),hits,*complete_path,
				     right_univdiagonal_sense,right_splicepos_sense,
				     right_nmismatches_sense,right_junction_sense,right_alts_sense,
				     right_univdiagonal_antisense,right_splicepos_antisense,
				     right_nmismatches_antisense,right_junction_antisense,right_alts_antisense,
				     left_univdiagonal_sense,left_splicepos_sense,
				     left_nmismatches_sense,left_junction_sense,left_alts_sense,
				     left_univdiagonal_antisense,left_splicepos_antisense,
				     left_nmismatches_antisense,left_junction_antisense,left_alts_antisense,
				     chrnum,chroffset,chrhigh,chrlength,querylength,query_compress,
				     plusp,genestrand,nmismatches_allowed,paired_end_p,method,level);

    if (*foundp == true) {
      Univdiag_gc(&(*complete_path));
    } else {
      fix_query_ends(*complete_path);
#ifdef DEBUG13
      printf("COMPLETE_PATH\n");
      for (p = *complete_path; p != NULL; p = List_next(p)) {
	diagonal = (Univdiag_T) List_head(p);
	printf("%d..%d\n",diagonal->querystart,diagonal->queryend);
      }
#endif
    }
  }

  return hits;
}
#endif



#if 0
List_T
Path_solve_from_path (bool *foundp, int *found_score, List_T hits, List_T complete_path,
		      Compress_T query_compress, int querylength, Genome_T genomebits,
		      Chrnum_T chrnum, Univcoord_T chroffset, Univcoord_T chrhigh, Chrpos_T chrlength,
		      bool plusp, int genestrand, int max_mismatches_allowed, Method_T method, int level) {

  Stage3end_T hit;
  int sensedir, sense_sensedir, antisense_sensedir;

  List_T p;
  Univdiag_T diagonal, prev_diagonal;
  Chrpos_T splice_distance;
  int prev_nmismatches;
#if 0
  bool abort_sense_p = false, abort_antisense_p = false;
#endif

  int indel_pos;
  int nindels;
  Univcoord_T deletionpos;

  int splice_pos;
  double donor_prob, acceptor_prob;
  int introntype;

  bool sense_acceptable_p, antisense_acceptable_p, sense_intronp, antisense_intronp;
  Univcoord_T left, prev_left;
  Univcoordlist_T sense_lefts = NULL, antisense_lefts = NULL;
  Intlist_T sense_nmismatches = NULL, antisense_nmismatches = NULL;
  Intlist_T sense_endpoints = NULL, antisense_endpoints = NULL;
  List_T sense_junctions = NULL, antisense_junctions = NULL;

  int best_knowni_i, best_knowni_j, best_nmismatches_i, best_nmismatches_j;
  double best_prob_i, best_prob_j;

  Spliceinfo_T spliceinfo;
  int j;

  spliceinfo = Spliceinfo_new(querylength);

#ifdef DEBUG13
  printf("\n");
  printf("Original diagonals:\n");
  for (p = complete_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d at %u [%u (%u - %u)]\n",diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,diagonal->univdiagonal - chroffset,diagonal->univdiagonal,chroffset);
  }
  printf("\n");
#endif

#ifdef DEBUG13
  printf("\n");
  printf("Diagonals on chrnum %d:\n",chrnum);
  for (p = complete_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d at %u [%u] with %d mismatches\n",
	   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,
	   diagonal->univdiagonal - chroffset,diagonal->intscore);
  }
  printf("\n");
#endif


  /*  Step 1:  Fix query ends */
  fix_query_ends(complete_path);

#ifdef DEBUG13
  printf("\n");
  printf("After fixing query ends:\n");
  for (p = complete_path; p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    printf("%d..%d at %u [%u] with %d mismatches\n",
	   diagonal->querystart,diagonal->queryend,diagonal->univdiagonal,
	   diagonal->univdiagonal - chroffset,diagonal->intscore);
  }
  printf("\n");
#endif


  /*  Step 2:  Handle indels and splices */

  p = complete_path;
  prev_diagonal = (Univdiag_T) List_head(p);
  prev_left = prev_diagonal->univdiagonal;

  debug13(printf("Diagonal %d..%d for left %u [%u]\n",
		 prev_diagonal->querystart,prev_diagonal->queryend,prev_left,prev_left - chroffset));

  sense_endpoints = Intlist_push(NULL,prev_diagonal->querystart);
  antisense_endpoints = Intlist_push(NULL,prev_diagonal->querystart);

  /* Previously pushed prev_diagonal->intscore, but that is not
     correct (unless there are no mismatches).  Pushing -1 indicates
     that we need to compute the value */
  if (prev_diagonal->intscore == 0) {
    sense_nmismatches = Intlist_push(NULL,0);
    antisense_nmismatches = Intlist_push(NULL,0);
  } else {
    sense_nmismatches = Intlist_push(NULL,-1);
    antisense_nmismatches = Intlist_push(NULL,-1);
  }

  for (p = List_next(p); p != NULL; p = List_next(p)) {
    diagonal = (Univdiag_T) List_head(p);
    left = diagonal->univdiagonal;
    assert(left != prev_left); /* Because we already handled mismatches above */

    debug13(printf("Diagonal %d..%d at leftpos %u [%u], diff %d\n",
		   diagonal->querystart,diagonal->queryend,left,left - chroffset,left - prev_left));

    if (left < prev_left) {
      /* Insertion */
      nindels = prev_left - left;
#if 0
      max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
      debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
      if (prev_diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
      if (diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
#endif
      if ((indel_pos = Indel_resolve_middle_insertion(&best_nmismatches_i,&best_nmismatches_j,
						      /*left*/prev_left,/*indels*/+nindels,
						      /*mismatch_positions_left*/NULL,/*nmismatches_left*/0,
						      /*mismatch_positions_right*/NULL,/*nmismatches_right*/0,
						      /*ome*/genomebits,/*ome_alt*/genomebits_alt,query_compress,
						      prev_diagonal->querystart,diagonal->queryend,querylength,
						      max_mismatches_allowed,/*plusp:true*/true,genestrand,
						      /*want_lowest_coordinate_p*/true)) <= 0) {
	sense_junctions = List_push(sense_junctions,NULL);
	antisense_junctions = List_push(antisense_junctions,NULL);
      } else {
	sense_junctions = List_push(sense_junctions,Junction_new_insertion(nindels));
	antisense_junctions = List_push(antisense_junctions,Junction_new_insertion(nindels));
	diagonal->querystart += nindels; /* Needed for subsequent indel computation */
      }

#if 0
      if (indel_pos <= Intlist_head(sense_endpoints) + 1) {
	abort_sense_p = true;
      }
      if (indel_pos <= Intlist_head(antisense_endpoints) + 1) {
	abort_antisense_p = true;
      }
#endif

      if ((prev_nmismatches = Intlist_head(sense_nmismatches)) < 0) {
	/* Still need to compute */
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(sense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(sense_nmismatches,-1);
	sense_nmismatches = Intlist_push(sense_nmismatches,-1);
      } else {
	Intlist_head_set(sense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      }

      if ((prev_nmismatches = Intlist_head(antisense_nmismatches)) < 0) {
	/* Still need to compute */
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(antisense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(antisense_nmismatches,-1);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,-1);
      } else {
	Intlist_head_set(antisense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      }

      sense_lefts = Univcoordlist_push(sense_lefts,prev_left);
      antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);

      sense_endpoints = Intlist_push(sense_endpoints,indel_pos);
      antisense_endpoints = Intlist_push(antisense_endpoints,indel_pos);
      debug13(printf("insertion pos in range %d..%d is %d with nmismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,indel_pos,best_nmismatches_i,best_nmismatches_j));

    } else if (left <= prev_left + max_deletionlen) {
      /* Deletion (or short intron) */
      nindels = left - prev_left;
#if 0
      max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
      debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
      if (prev_diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
      if (diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
#endif
      if ((indel_pos = Indel_resolve_middle_deletion_or_splice(&introntype,&best_nmismatches_i,&best_nmismatches_j,
							       /*left*/prev_left,/*indels*/-nindels,
							       /*ome*/genomebits,/*ome_alt*/genomebits_alt,query_compress,
							       prev_diagonal->querystart,diagonal->queryend,querylength,
							       max_mismatches_allowed,/*plusp:true*/true,genestrand,
							       min_intronlength,/*want_lowest_coordinate_p*/true)) <= 0) {
	sense_junctions = List_push(sense_junctions,NULL);
	antisense_junctions = List_push(antisense_junctions,NULL);

      } else if (nindels < min_intronlength || (novelsplicingp == false && splicesites == NULL) || circularp[chrnum] == true) {
	/* Cannot be an intron, so must be a deletion */
	deletionpos = prev_left + indel_pos;
	sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));
	antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

#if 0
	if (indel_pos <= Intlist_head(sense_endpoints) + 1) {
	  abort_sense_p = true;
	}
	if (indel_pos <= Intlist_head(antisense_endpoints) + 1) {
	  abort_antisense_p = true;
	}
#endif

      } else if ((sensedir = Intron_canonical_sensedir(introntype)) == SENSE_NULL) {
	/* No intron dinucleotids found, so must be a deletion */
	deletionpos = prev_left + indel_pos;
	sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));
	antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

#if 0
	if (indel_pos <= Intlist_head(sense_endpoints) + 1) {
	  abort_sense_p = true;
	}
	if (indel_pos <= Intlist_head(antisense_endpoints) + 1) {
	  abort_antisense_p = true;
	}
#endif

      } else {
	deletionpos = prev_left + indel_pos;
	if (plusp == true) {
	  if (sensedir == SENSE_FORWARD) {
	    donor_prob = Maxent_hr_donor_prob(deletionpos,chroffset);
	    acceptor_prob = Maxent_hr_acceptor_prob(deletionpos+nindels,chroffset);
	    sense_junctions = List_push(sense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_FORWARD,
									    donor_prob,acceptor_prob));
	    antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

	  } else {
	    donor_prob = Maxent_hr_antidonor_prob(deletionpos+nindels,chroffset);
	    acceptor_prob = Maxent_hr_antiacceptor_prob(deletionpos,chroffset);
	    antisense_junctions = List_push(antisense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_ANTI,
										    donor_prob,acceptor_prob));
	    sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));

	  }

	} else {
	  sensedir = (sensedir == SENSE_FORWARD) ? SENSE_ANTI : SENSE_FORWARD;
	  if (sensedir == SENSE_FORWARD) {
	    /* check */
	    donor_prob = Maxent_hr_antidonor_prob(deletionpos+nindels,chroffset);
	    acceptor_prob = Maxent_hr_antiacceptor_prob(deletionpos,chroffset);
	    sense_junctions = List_push(sense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_FORWARD,
									    donor_prob,acceptor_prob));
	    antisense_junctions = List_push(antisense_junctions,Junction_new_deletion(nindels,deletionpos));

	  } else {
	    /* check */
	    donor_prob = Maxent_hr_donor_prob(deletionpos,chroffset);
	    acceptor_prob = Maxent_hr_acceptor_prob(deletionpos+nindels,chroffset);
	    antisense_junctions = List_push(antisense_junctions,Junction_new_splice(/*splice_distance*/nindels,SENSE_ANTI,
										    donor_prob,acceptor_prob));
	    sense_junctions = List_push(sense_junctions,Junction_new_deletion(nindels,deletionpos));
	  }
	}
      }

      if ((prev_nmismatches = Intlist_head(sense_nmismatches)) < 0) {
	/* Still need to compute */
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(sense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(sense_nmismatches,-1);
	sense_nmismatches = Intlist_push(sense_nmismatches,-1);
      } else {
	Intlist_head_set(sense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      }

      if ((prev_nmismatches = Intlist_head(antisense_nmismatches)) < 0) {
	/* Still need to compute */
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(antisense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(antisense_nmismatches,-1);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,-1);
      } else {
	Intlist_head_set(antisense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      }

      sense_lefts = Univcoordlist_push(sense_lefts,prev_left);
      antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);

      sense_endpoints = Intlist_push(sense_endpoints,indel_pos);
      antisense_endpoints = Intlist_push(antisense_endpoints,indel_pos);
      debug13(printf("deletion pos in range %d..%d is %d with nmismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,indel_pos,best_nmismatches_i,best_nmismatches_j));

    } else {
      /* Splice */
      spliceinfo->segmenti_donor_nknown = spliceinfo->segmenti_antiacceptor_nknown = 0;
      if (nsplicesites > 0 &&
	  Splicetrie_splicesite_p(prev_left,/*pos5*/1,/*pos3*/querylength) == true) {
	j = binary_search_univcoord(0,nsplicesites,splicesites,prev_left);
	while (j < nsplicesites && splicesites[j] < prev_left + querylength) {
	  if (splicetypes[j] == DONOR) {
	    debug4s(printf("Setting known donor %d for segmenti at %u\n",j,splicesites[j]));
	    spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = splicesites[j] - prev_left;
	    spliceinfo->segmenti_donor_knowni[spliceinfo->segmenti_donor_nknown++] = j;
	  } else if (splicetypes[j] == ANTIACCEPTOR) {
	    debug4s(printf("Setting known antiacceptor %d for segmenti at %u\n",j,splicesites[j]));
	    spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = splicesites[j] - prev_left;
	    spliceinfo->segmenti_antiacceptor_knowni[spliceinfo->segmenti_antiacceptor_nknown++] = j;
	  }
	  j++;
	}
      }
      spliceinfo->segmenti_donor_knownpos[spliceinfo->segmenti_donor_nknown] = querylength + 100;
      spliceinfo->segmenti_antiacceptor_knownpos[spliceinfo->segmenti_antiacceptor_nknown] = querylength + 100;

      spliceinfo->segmentj_acceptor_nknown = spliceinfo->segmentj_antidonor_nknown = 0;
      if (nsplicesites > 0 &&
	  Splicetrie_splicesite_p(left,/*pos5*/1,/*pos3*/querylength) == true) {
	j = binary_search_univcoord(0,nsplicesites,splicesites,left);
	while (j < nsplicesites && splicesites[j] < left + querylength) {
	  if (splicetypes[j] == ACCEPTOR) {
	    debug4s(printf("Setting known acceptor %d for segmentj at %u\n",j,splicesites[j]));
	    spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = splicesites[j] - left;
	    spliceinfo->segmentj_acceptor_knowni[spliceinfo->segmentj_acceptor_nknown++] = j;
	  } else if (splicetypes[j] == ANTIDONOR) {
	    debug4s(printf("Setting known antidonor %d for segmentj at %u\n",j,splicesites[j]));
	    spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = splicesites[j] - left;
	    spliceinfo->segmentj_antidonor_knowni[spliceinfo->segmentj_antidonor_nknown++] = j;
	  }
	  j++;
	}
      }
      spliceinfo->segmentj_acceptor_knownpos[spliceinfo->segmentj_acceptor_nknown] = querylength + 100;
      spliceinfo->segmentj_antidonor_knownpos[spliceinfo->segmentj_antidonor_nknown] = querylength + 100;

      splice_distance = left - prev_left;
#if 0
      max_mismatches_allowed = (diagonal->querystart - prev_diagonal->queryend - 1);
      debug13(printf("max_mismatches %d = %d - %d - 1\n",max_mismatches_allowed,diagonal->querystart,prev_diagonal->queryend));
      if (prev_diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
      if (diagonal->intscore > 0) {
	max_mismatches_allowed += 1;
      }
#endif

      if ((splice_pos = Splice_resolve_sense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
					     &best_prob_i,&best_prob_j,
					     /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
					     prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
					     spliceinfo,max_mismatches_allowed,plusp,genestrand)) <= 0) {
	sense_endpoints = Intlist_push(sense_endpoints,-1); /* Mark as invalid */
	sense_junctions = List_push(sense_junctions,NULL);
      } else if (plusp == true) {
	sense_endpoints = Intlist_push(sense_endpoints,splice_pos);
	sense_junctions = List_push(sense_junctions,Junction_new_splice(splice_distance,SENSE_FORWARD,
									/*donor_prob*/best_prob_i,/*acceptor_prob*/best_prob_j));
      } else {
	sense_endpoints = Intlist_push(sense_endpoints,splice_pos);
	sense_junctions = List_push(sense_junctions,Junction_new_splice(splice_distance,SENSE_FORWARD,
									/*donor_prob*/best_prob_j,/*acceptor_prob*/best_prob_i));
      }
      debug13(printf("sense splice_pos in range %d..%d is %d with mismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,splice_pos,best_nmismatches_i,best_nmismatches_j));
      if ((prev_nmismatches = Intlist_head(sense_nmismatches)) < 0) {
	/* Still need to compute */
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(sense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(sense_nmismatches,-1);
	sense_nmismatches = Intlist_push(sense_nmismatches,-1);
      } else {
	Intlist_head_set(sense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	sense_nmismatches = Intlist_push(sense_nmismatches,best_nmismatches_j);
      }
      sense_lefts = Univcoordlist_push(sense_lefts,prev_left);

      if ((splice_pos = Splice_resolve_antisense(&best_knowni_i,&best_knowni_j,&best_nmismatches_i,&best_nmismatches_j,
						 &best_prob_i,&best_prob_j,
						 /*segmenti_left*/prev_left,/*segmentj_left*/left,chroffset,chroffset,
						 prev_diagonal->querystart,diagonal->queryend,querylength,query_compress,
						 spliceinfo,max_mismatches_allowed,plusp,genestrand)) <= 0) {
	antisense_endpoints = Intlist_push(antisense_endpoints,-1); /* Mark as invalid */
	antisense_junctions = List_push(antisense_junctions,NULL);
      } else if (plusp == true) {
	antisense_endpoints = Intlist_push(antisense_endpoints,splice_pos);
	antisense_junctions = List_push(antisense_junctions,Junction_new_splice(splice_distance,SENSE_ANTI,
										/*donor_prob*/best_prob_j,/*acceptor_prob*/best_prob_i));
      } else {
	antisense_endpoints = Intlist_push(antisense_endpoints,splice_pos);
	antisense_junctions = List_push(antisense_junctions,Junction_new_splice(splice_distance,SENSE_ANTI,
										/*donor_prob*/best_prob_i,/*acceptor_prob*/best_prob_j));
      }
      debug13(printf("antisense splice_pos in range %d..%d is %d with nmismatches %d+%d\n",
		     prev_diagonal->querystart,diagonal->queryend,splice_pos,best_nmismatches_i,best_nmismatches_j));
      if ((prev_nmismatches = Intlist_head(antisense_nmismatches)) < 0) {
	/* Still need to compute */
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      } else if (Intlist_head(antisense_endpoints) != prev_diagonal->querystart) {
	/* Endpoints not quite the same, so need to recompute */
	Intlist_head_set(antisense_nmismatches,-1);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,-1);
      } else {
	Intlist_head_set(antisense_nmismatches,best_nmismatches_i /*+ prev_nmismatches*/);
	antisense_nmismatches = Intlist_push(antisense_nmismatches,best_nmismatches_j);
      }
      antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);
    }

    /* Handle previous segment (for prev_left) */
    prev_left = left;
    prev_diagonal = diagonal;
  }

  /* Finish up lists */
  sense_lefts = Univcoordlist_push(sense_lefts,prev_left);
  antisense_lefts = Univcoordlist_push(antisense_lefts,prev_left);
  sense_endpoints = Intlist_push(sense_endpoints,prev_diagonal->queryend/*+ 1*/);
  antisense_endpoints = Intlist_push(antisense_endpoints,prev_diagonal->queryend/*+ 1*/);


  debug13(printf("After step 2 (indels and splices)\n"));
  debug13(printf("sense (wrt plus): %s\n",Intlist_to_string(sense_endpoints)));
  debug13(printf("antisense (wrt plus): %s\n",Intlist_to_string(antisense_endpoints)));
  debug13(printf("sense nmismatches: %s\n",Intlist_to_string(sense_nmismatches)));
  debug13(printf("antisense nmismatches: %s\n",Intlist_to_string(antisense_nmismatches)));

#if 0
  /*  Step 3:  Handle alternative ends on right */
#endif

  /*  Step 4:  Reverse sense and antisense alignments */
  sense_nmismatches = Intlist_reverse(sense_nmismatches);
  antisense_nmismatches = Intlist_reverse(antisense_nmismatches);
  sense_lefts = Univcoordlist_reverse(sense_lefts);
  antisense_lefts = Univcoordlist_reverse(antisense_lefts);
  sense_endpoints = Intlist_reverse(sense_endpoints);
  antisense_endpoints = Intlist_reverse(antisense_endpoints);
  sense_junctions = List_reverse(sense_junctions);
  antisense_junctions = List_reverse(antisense_junctions);

  debug13(printf("After step 4 (reverse alignments)\n"));
  debug13(printf("sense (wrt plus): %s\n",Intlist_to_string(sense_endpoints)));
  debug13(printf("antisense (wrt plus): %s\n",Intlist_to_string(antisense_endpoints)));
  debug13(printf("sense nmismatches: %s\n",Intlist_to_string(sense_nmismatches)));
  debug13(printf("antisense nmismatches: %s\n",Intlist_to_string(antisense_nmismatches)));

#if 0
  /*  Step 5:  Handle alternative ends on left */
#endif

#ifdef DEBUG13
  printf("Sense junctions\n");
  for (p = sense_junctions; p != NULL; p = List_next(p)) {
    Junction_print(List_head(p));
  }
  printf("\n");
  printf("Antisense junctions\n");
  for (p = antisense_junctions; p != NULL; p = List_next(p)) {
    Junction_print(List_head(p));
  }
  printf("\n");

  printf("Sense lefts: %s\n",Univcoordlist_to_string(sense_lefts));
  printf("Antisense lefts: %s\n",Univcoordlist_to_string(antisense_lefts));
#endif


  /* Need to rely on probability filtering in splice.c to get correct
     results for sense and antisense */
  sense_acceptable_p = endpoints_acceptable_p(&sense_intronp,sense_junctions,sense_endpoints);
  antisense_acceptable_p = endpoints_acceptable_p(&antisense_intronp,antisense_junctions,
						  antisense_endpoints);
  debug13(printf("sense_acceptable_p %d, antisense_acceptable_p %d\n",
		 sense_acceptable_p,antisense_acceptable_p));
  debug13(printf("sense_intronp %d, antisense_intronp %d\n",sense_intronp,antisense_intronp));

  if (sense_acceptable_p == true && antisense_acceptable_p == true) {
    if (sense_intronp == true) {
      sense_sensedir = SENSE_FORWARD;
    } else {
      sense_sensedir = SENSE_NULL;
    }
    if (antisense_intronp == true) {
      antisense_sensedir = SENSE_ANTI;
    } else {
      antisense_sensedir = SENSE_NULL;
    }

    if (sense_sensedir == SENSE_NULL && antisense_sensedir == SENSE_NULL) {
      /* Create just one hit */
      if ((hit = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					  sense_nmismatches,sense_junctions,
					  querylength,query_compress,
					  /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,/*sensedir*/SENSE_NULL,
					  chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					  /*extend_ends_p*/false,method,level)) == NULL) {
	debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
      } else {
	hits = List_push(hits,(void *) hit);
	*foundp = true;
      }

    } else {
      /* Create both sense and antisense hits */
      debug13(printf("Creating both sense and antisense hits\n"));

      if ((hit = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					  sense_nmismatches,sense_junctions,
					  querylength,query_compress,
					  /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sense_sensedir,
					  chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					  /*extend_ends_p*/false,method,level)) == NULL) {
	debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
      } else {
	hits = List_push(hits,(void *) hit);
	*foundp = true;
      }

      if ((hit = Stage3end_new_substrings(&(*found_score),antisense_endpoints,antisense_lefts,
					  antisense_nmismatches,antisense_junctions,
					  querylength,query_compress,
					  /*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,antisense_sensedir,
					  chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					  /*extend_ends_p*/false,method,level)) == NULL) {
	debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
      } else {
	hits = List_push(hits,(void *) hit);
	*foundp = true;
      }
    }

  } else if (sense_acceptable_p == true) {
    if (sense_intronp == true) {
      sensedir = SENSE_FORWARD;
    } else {
      sensedir = SENSE_NULL;
    }
    if ((hit = Stage3end_new_substrings(&(*found_score),sense_endpoints,sense_lefts,
					sense_nmismatches,sense_junctions,
					querylength,query_compress,
					/*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sensedir,
					chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					/*extend_ends_p*/false,method,level)) == NULL) {
      debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
    } else {
      hits = List_push(hits,(void *) hit);
      *foundp = true;
    }

  } else if (antisense_acceptable_p == true) {
    if (antisense_intronp == true) {
      sensedir = SENSE_ANTI;
    } else {
      sensedir = SENSE_NULL;
    }
    if ((hit = Stage3end_new_substrings(&(*found_score),antisense_endpoints,antisense_lefts,
					antisense_nmismatches,antisense_junctions,
					querylength,query_compress,
					/*right_alt*/NULL,/*left_alt*/NULL,plusp,genestrand,sensedir,
					chrnum,chroffset,chrhigh,chrlength,endtrim_allowed,
					/*extend_ends_p*/false,method,level)) == NULL) {
      debug13(printf("ABORTING BECAUSE OF STAGE3END_NEW_SUBSTRINGS\n"));
    } else {
      hits = List_push(hits,(void *) hit);
      *foundp = true;
    }

  } else {
    /* Neither set of junctions/endpoints works */
    debug13(printf("ABORTING BECAUSE ENDPOINTS ARE UNACCEPTABLE (HAVE BREAKS)\n"));
  }


  Intlist_free(&sense_nmismatches);
  Intlist_free(&antisense_nmismatches);
  Univcoordlist_free(&sense_lefts);
  Univcoordlist_free(&antisense_lefts);
  Intlist_free(&sense_endpoints);
  Intlist_free(&antisense_endpoints);
  Junction_gc(&sense_junctions);
  Junction_gc(&antisense_junctions);

  Spliceinfo_free(&spliceinfo);

  return hits;
}
#endif



/* Modified from Stage3end_substrings_run_gmap_plus to use path (list of Univdiag_T) */
/* Stage3end_substrings_run_gmap_plus was modified from run_gmap_plus in sarray-search.c */
/* extend_ends_p is an expensive operation, slowing down speed by 5x */
/* sensedir not used by single_read */
static Stage3end_T
solve_via_gmap_plus (Stage3end_T *hit2, int *sensedir, int *found_score, List_T path,
		     char *queryuc_ptr, int querylength, Compress_T query_compress_fwd, int genestrand,
		     Pairpool_T pairpool, Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR,
		     Oligoindex_array_T oligoindices_minor, Diagpool_T diagpool, Cellpool_T cellpool,
		     bool pick_one_sensedir_p, Method_T method, int level) {
  Stage3end_T hit1;
  List_T stage2pairs, all_stage2_starts, all_stage2_ends;
  List_T p;

  int i;
  Univcoord_T left, alignstart;
  Chrnum_T chrnum;
  Univcoord_T chroffset, chrhigh;
  Chrpos_T chrlength;

  struct Pair_T *pairarray1, *pairarray2;
  List_T pairs1, pairs2;
  Univdiag_T univdiag;
  int querypos, seglength;
  Chrpos_T genomepos;
  char c, g, g_alt, comp;

  int cdna_direction;
  int npairs1, goodness1, matches1, nmatches_posttrim_1,
    max_match_length_1, ambig_end_length_5_1, ambig_end_length_3_1,
    unknowns1, mismatches1, qopens1, qindels1, topens1, tindels1,
    ncanonical1, nsemicanonical1, nnoncanonical1;
  int npairs2, goodness2, matches2, nmatches_posttrim_2,
    max_match_length_2, ambig_end_length_5_2, ambig_end_length_3_2,
    unknowns2, mismatches2, qopens2, qindels2, topens2, tindels2,
    ncanonical2, nsemicanonical2, nnoncanonical2;
  double ambig_prob_5_1, ambig_prob_3_1, avg_splice_score_1;
  double ambig_prob_5_2, ambig_prob_3_2, avg_splice_score_2;
  Splicetype_T ambig_splicetype_5_1, ambig_splicetype_3_1;
  Splicetype_T ambig_splicetype_5_2, ambig_splicetype_3_2;
  Univcoord_T knownsplice_limit_low, knownsplice_limit_high;
  Univcoord_T start, end;
  char *gsequence_orig, *gsequence_alt;
  bool equalp;
  Chrpos_T chrstart, chrend;

  *hit2 = (Stage3end_T) NULL;
  debug14(printf("Entered solve_via_gmap_plus with gmap_mode %d\n",gmap_mode));
#ifdef DEBUG14
  for (p = path; p != NULL; p = List_next(p)) {
    univdiag = (Univdiag_T) List_head(p);
    printf("%d..%d at left %u\n",univdiag->querystart,univdiag->queryend,univdiag->univdiagonal);
    printf("\n");
  }
#endif

  univdiag = (Univdiag_T) List_head(path);
  left = univdiag->univdiagonal;
  chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);


#ifdef HAVE_ALLOCA
    gsequence_orig = (char *) MALLOCA((querylength+1) * sizeof(char));
    gsequence_alt = (char *) MALLOCA((querylength+1) * sizeof(char));
#else
    gsequence_orig = (char *) MALLOC((querylength+1) * sizeof(char));
    gsequence_alt = (char *) MALLOC((querylength+1) * sizeof(char));
#endif


  /* D.  Make all_stage2_starts (paths) */
  if ((gmap_mode & GMAP_ENDS) == 0) {
    all_stage2_starts = (List_T) NULL;

  } else {
    /* Middle path goes from low to high genomic coords */
    univdiag = (Univdiag_T) List_head(path);
    if ((querypos = univdiag->querystart) < 8) {
      debug14(printf("No starts, since querypos %d < 8\n",querypos));
      all_stage2_starts = (List_T) NULL;
    } else {
      left = univdiag->univdiagonal;
      chrend = (left + querypos) - chroffset;
      if (chrend < overall_end_distance_genome) {
	chrstart = 0;
      } else {
	chrstart = chrend - overall_end_distance_genome;
      }
      all_stage2_starts = Stage2_compute_starts(
#ifdef PMAP
						&(queryaaseq_ptr[0]),&(queryaaseq_ptr[0]),
						/*querylength*/querypos,/*query_offset*/0*3,
#else
						&(queryuc_ptr[0]),&(queryuc_ptr[0]),
						/*querylength*/querypos,/*query_offset*/0,
#endif
						chrstart,chrend,chroffset,chrhigh,/*plusp*/true,genestrand,
					    
						oligoindices_minor,pairpool,diagpool,cellpool,
						/*localp should be false*/true,/*skip_repetitive_p*/false,
						/*favor_right_p*/false,/*max_nalignments*/5,/*debug_graphic_p*/false);
      debug14(printf("*Got %d starts\n",List_length(all_stage2_starts)));
    }
  }


  /* E.  Make all_stage2_ends (pairs) */
  if ((gmap_mode & GMAP_ENDS) == 0) {
    all_stage2_ends = (List_T) NULL;

  } else {
    univdiag = (Univdiag_T) List_last_value(path);
    if ((querypos = univdiag->queryend) > querylength - 8) {
      debug14(printf("No ends, since querypos %d > querylength %d - 8\n",querypos,querylength));
      all_stage2_ends = (List_T) NULL;
    } else {
      left = univdiag->univdiagonal;
      chrstart = (left + querypos) - chroffset;
      if (chrstart + overall_end_distance_genome > (chrhigh - chroffset)) {
	chrend = chrhigh - chroffset;
      } else {
	chrend = chrstart + overall_end_distance_genome;
      }
      all_stage2_ends = Stage2_compute_ends(
#ifdef PMAP
					    &(queryaaseq_ptr[querypos]),&(queryaaseq_ptr[querypos]),
					    /*querylength*/querylength - querypos,/*query_offset*/querypos*3,
#else
					    &(queryuc_ptr[querypos]),&(queryuc_ptr[querypos]),
					    /*querylength*/querylength - querypos,/*query_offset*/querypos,
#endif
					    chrstart,chrend,chroffset,chrhigh,/*plusp*/true,genestrand,
					    
					    oligoindices_minor,pairpool,diagpool,cellpool,
					    /*localp should be false*/true,/*skip_repetitive_p*/false,
					    /*favor_right_p*/false,/*max_nalignments*/5,/*debug_graphic_p*/false);
      debug14(printf("Got %d ends\n",List_length(all_stage2_ends)));
    }
  }


  /* Used to count total mismatches here, but cannot do that with Univdiag_T */

  /* F.  Make stage2pairs */
  debug14(printf("Making pairs from univdiags: "));
  stage2pairs = (List_T) NULL;
  for (p = path; p != NULL; p = List_next(p)) {
    univdiag = (Univdiag_T) List_head(p);
    debug14(printf("Handling univdiag for %d..%d at left %u\n",
		   univdiag->querystart,univdiag->queryend,univdiag->univdiagonal));
    querypos = univdiag->querystart;
    seglength = univdiag->queryend - univdiag->querystart;

    left = univdiag->univdiagonal;
    alignstart = left + univdiag->querystart;
    genomepos = alignstart - chroffset;
    Genome_get_segment_blocks_right(gsequence_orig,gsequence_alt,/*left*/alignstart,
				    seglength,chrhigh,/*revcomp*/false);
	
    for (i = 0; i < seglength; i++) {
      c = queryuc_ptr[querypos];
      g = gsequence_orig[i];
      g_alt = gsequence_alt[i];
      /* Previously checked for (g == c || g_alt == c) */
      /* printf("plus: Checking cdna %c against genome %c => %d\n",c,g,Dynprog_consistent_p(c,g,g_alt,genestrand)); */
      if (Dynprog_consistent_p(c,g,g_alt,genestrand) == true) {
	comp = MATCH_COMP;
      } else {
	comp = MISMATCH_COMP;
      }
      debug14(printf("Pushing %c %c %c at %d,%u\n",c,comp,g,querypos,genomepos));
      stage2pairs = Pairpool_push(stage2pairs,pairpool,querypos,genomepos,
				  /*cdna*/c,comp,/*genome*/g,/*genomealt*/g_alt,
				  /*dynprogindex*/0);
      querypos++;
      genomepos++;
    }
    debug14(Pair_dump_list(stage2pairs,true));
    debug14(printf("\n"));
  }

  if (stage2pairs == NULL) {
    hit1 = (Stage3end_T) NULL;
  } else {
    knownsplice_limit_high = ((Pair_T) stage2pairs->first)->genomepos + chroffset;
    stage2pairs = List_reverse(stage2pairs);
    knownsplice_limit_low = ((Pair_T) stage2pairs->first)->genomepos + chroffset;
    
    if ((pairarray1 = Stage3_compute_one(&cdna_direction,&(*sensedir),&pairs1,&npairs1,&goodness1,
					 &matches1,&nmatches_posttrim_1,&max_match_length_1,
					 &ambig_end_length_5_1,&ambig_end_length_3_1,
					 &ambig_splicetype_5_1,&ambig_splicetype_3_1,
					 &ambig_prob_5_1,&ambig_prob_3_1,
					 &unknowns1,&mismatches1,&qopens1,&qindels1,&topens1,&tindels1,
					 &ncanonical1,&nsemicanonical1,&nnoncanonical1,&avg_splice_score_1,

					 &pairarray2,&pairs2,&npairs2,&goodness2,
					 &matches2,&nmatches_posttrim_2,&max_match_length_2,
					 &ambig_end_length_5_2,&ambig_end_length_3_2,
					 &ambig_splicetype_5_2,&ambig_splicetype_3_2,
					 &ambig_prob_5_2,&ambig_prob_3_2,
					 &unknowns2,&mismatches2,&qopens2,&qindels2,&topens2,&tindels2,
					 &ncanonical2,&nsemicanonical2,&nnoncanonical2,&avg_splice_score_2,

					 stage2pairs,all_stage2_starts,all_stage2_ends,
#ifdef END_KNOWNSPLICING_SHORTCUT
					 cutoff_level,/*queryptr*/watsonp ? queryuc_ptr : queryrc,
					 watsonp ? query_compress_fwd : query_compress_rev,
#endif
					 /*queryseq_ptr*/queryuc_ptr,queryuc_ptr,querylength,/*skiplength*/0,
#ifdef EXTRACT_GENOMICSEG
					 /*query_subseq_offset*/0,
#else
					 /*query_subseq_offset*/0,
#endif
					 chrnum,chroffset,chrhigh,chrlength,
					 knownsplice_limit_low,knownsplice_limit_high,/*plusp*/true,genestrand,
					 /*jump_late_p*/false,maxpeelback,pairpool,dynprogL,dynprogM,dynprogR,
					 /*sense_try*/0,/*sense_filter*/0,
					 oligoindices_minor,diagpool,cellpool)) == NULL) {
      /* No possible hits for either sensedir */
      hit1 = (Stage3end_T) NULL;

    } else if (pairarray2 == NULL) {
      /* Possible hit for pairarray1 only */
      start = subtract_bounded(chroffset + Pair_genomepos(&(pairarray1[0])),
			       /*minusterm*/Pair_querypos(&(pairarray1[0])),chroffset);
      end = add_bounded(chroffset + Pair_genomepos(&(pairarray1[npairs1-1])),
			/*plusterm*/querylength - 1 - Pair_querypos(&(pairarray1[npairs1-1])),chrhigh);

      if ((hit1 = Stage3end_new_gmap(&(*found_score),nmatches_posttrim_1,max_match_length_1,
				     ambig_end_length_5_1,ambig_end_length_3_1,
				     avg_splice_score_1,pairarray1,npairs1,
				     /*left*/start,/*plusp*/true,genestrand,
				     querylength,query_compress_fwd,
				     chrnum,chroffset,chrhigh,chrlength,
				     cdna_direction,*sensedir,method,level)) == NULL) {
	/* FREE_OUT(pairarray1); */
      }

    } else {
      /* Possible hit for both pairarrays */
      start = subtract_bounded(chroffset + Pair_genomepos(&(pairarray1[0])),
			       /*minusterm*/Pair_querypos(&(pairarray1[0])),chroffset);
      end = add_bounded(chroffset + Pair_genomepos(&(pairarray1[npairs1-1])),
			/*plusterm*/querylength - 1 - Pair_querypos(&(pairarray1[npairs1-1])),chrhigh);

      if ((hit1 = Stage3end_new_gmap(&(*found_score),nmatches_posttrim_1,max_match_length_1,
				     ambig_end_length_5_1,ambig_end_length_3_1,
				     avg_splice_score_1,pairarray1,npairs1,
				     /*left*/start,/*plusp*/true,genestrand,
				     querylength,query_compress_fwd,
				     chrnum,chroffset,chrhigh,chrlength,
				     /*cdna_direction*/+1,/*sensedir*/SENSE_FORWARD,method,level)) == NULL) {
	/* FREE_OUT(pairarray1); */
      }

      start = subtract_bounded(chroffset + Pair_genomepos(&(pairarray2[0])),
			       /*minusterm*/Pair_querypos(&(pairarray2[0])),chroffset);
      end = add_bounded(chroffset + Pair_genomepos(&(pairarray2[npairs2-1])),
			/*plusterm*/querylength - 1 - Pair_querypos(&(pairarray2[npairs2-1])),chrhigh);

      if ((*hit2 = Stage3end_new_gmap(&(*found_score),nmatches_posttrim_2,max_match_length_2,
				      ambig_end_length_5_2,ambig_end_length_3_2,
				      avg_splice_score_2,pairarray2,npairs2,
				      /*left*/start,/*plusp*/true,genestrand,
				      querylength,query_compress_fwd,
				      chrnum,chroffset,chrhigh,chrlength,
				      /*cdna_direction*/-1,/*sensedir*/SENSE_ANTI,method,level)) == NULL) {
	/* FREE_OUT(pairarray2); */
      }

      if (hit1 == NULL && *hit2 == NULL) {
	/* Nothing to report */
	debug14(printf("Both hits are NULL.  Nothing to report\n"));

      } else if (hit1 == NULL) {
	/* Need to report hit2 only */
	debug14(printf("hit1 is NULL.  Reporting hit2 as hit1\n"));
	hit1 = *hit2;
	*hit2 = (Stage3end_T) NULL;

      } else if (*hit2 == NULL) {
	/* Need to report hit1 only */
	debug14(printf("hit2 is NULL.  Reporting hit1\n"));

      } else if (pick_one_sensedir_p == false) {
	/* Report both */
	debug14(printf("pick_one_sensedir is false.  Reporting both\n"));

      } else if (Stage3end_hit_goodness_cmp(&equalp,hit1,*hit2,/*finalp*/true) >= 0) {
	/* Report only hit1 as best */
	debug14(printf("hit1 is best.  Freeing hit2 and reporting hit1\n"));
	Stage3end_free(&(*hit2));
	*hit2 = (Stage3end_T) NULL;

      } else {
	/* Report only hit2 as best */
	debug14(printf("hit2 is best.  Freeing hit1 and reporting hit2 as hit1\n"));
	Stage3end_free(&hit1);
	hit1 = *hit2;
	*hit2 = (Stage3end_T) NULL;
      }
    }
  }

  List_free(&all_stage2_ends);
  List_free(&all_stage2_starts);

#ifdef HAVE_ALLOCA
  FREEA(gsequence_alt);
  FREEA(gsequence_orig);
#else
  FREE(gsequence_alt);
  FREE(gsequence_orig);
#endif

  return hit1;
}


/* Modified from Stage3end_substrings_run_gmap_minus to use path (list of Univdiag_T) */
/* Stage3end_substrings_run_gmap_minus was modified from run_gmap_minus in sarray-search.c */
/* extend_ends_p is an expensive operation, slowing down speed by 5x */
static Stage3end_T
solve_via_gmap_minus (Stage3end_T *hit2, int *sensedir, int *found_score, List_T path,
		      char *queryuc_ptr, int querylength, Compress_T query_compress_rev, int genestrand,
		      Pairpool_T pairpool, Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR,
		      Oligoindex_array_T oligoindices_minor, Diagpool_T diagpool, Cellpool_T cellpool,
		      bool pick_one_sensedir_p, Method_T method, int level) {
  Stage3end_T hit1;
  List_T stage2pairs, all_stage2_starts, all_stage2_ends, p;

  int i;
  Univcoord_T left, alignstart, alignend;
  Chrnum_T chrnum;
  Univcoord_T chroffset, chrhigh;
  Chrpos_T chrlength;

  struct Pair_T *pairarray1, *pairarray2;
  List_T pairs1, pairs2;
  Univdiag_T univdiag;
  int querypos, seglength;
  Chrpos_T genomepos;
  char c, g, g_alt, comp;

  int cdna_direction;
  int npairs1, goodness1, matches1, nmatches_posttrim_1,
    max_match_length_1, ambig_end_length_5_1, ambig_end_length_3_1,
    unknowns1, mismatches1, qopens1, qindels1, topens1, tindels1,
    ncanonical1, nsemicanonical1, nnoncanonical1;
  int npairs2, goodness2, matches2, nmatches_posttrim_2,
    max_match_length_2, ambig_end_length_5_2, ambig_end_length_3_2,
    unknowns2, mismatches2, qopens2, qindels2, topens2, tindels2,
    ncanonical2, nsemicanonical2, nnoncanonical2;
  double ambig_prob_5_1, ambig_prob_3_1, avg_splice_score_1;
  double ambig_prob_5_2, ambig_prob_3_2, avg_splice_score_2;
  Splicetype_T ambig_splicetype_5_1, ambig_splicetype_3_1;
  Splicetype_T ambig_splicetype_5_2, ambig_splicetype_3_2;
  Univcoord_T knownsplice_limit_low, knownsplice_limit_high;
  Univcoord_T start, end;
  char *gsequence_orig, *gsequence_alt;
  bool equalp;
  Chrpos_T chrstart, chrend;

  *hit2 = (Stage3end_T) NULL;
  debug14(printf("Entered solve_via_gmap_minus with gmap_mode %d\n",gmap_mode));
#ifdef DEBUG14
  for (p = path; p != NULL; p = List_next(p)) {
    univdiag = (Univdiag_T) List_head(p);
    printf("%d..%d at left %u\n",univdiag->querystart,univdiag->queryend,univdiag->univdiagonal);
    printf("\n");
  }
#endif

  univdiag = (Univdiag_T) List_head(path);
  left = univdiag->univdiagonal;
  chrnum = Univ_IIT_get_one(chromosome_iit,left,left);
  Univ_IIT_interval_bounds(&chroffset,&chrhigh,&chrlength,chromosome_iit,chrnum,circular_typeint);


#ifdef HAVE_ALLOCA
  gsequence_orig = (char *) MALLOCA((querylength+1) * sizeof(char));
  gsequence_alt = (char *) MALLOCA((querylength+1) * sizeof(char));
#else
  gsequence_orig = (char *) MALLOC((querylength+1) * sizeof(char));
  gsequence_alt = (char *) MALLOC((querylength+1) * sizeof(char));
#endif


  /* D.  Make all_stage2_starts (paths) */
  if ((gmap_mode & GMAP_ENDS) == 0) {
    all_stage2_starts = (List_T) NULL;

  } else {
    /* Middle path goes from low to high genomic coords.  May need to revise this code for minus */
    univdiag = (Univdiag_T) List_head(path);
    /* Substring_querystart = querylength - univdiag->queryend */
    if ((querypos = querylength - univdiag->queryend) < 8) {
      debug14(printf("No starts, since querystart %d < 8\n",querypos));
      all_stage2_starts = (List_T) NULL;
    } else {
      left = univdiag->univdiagonal;
      chrstart = (left + univdiag->queryend) - chroffset;
      if (chrstart + overall_end_distance_genome > (chrhigh - chroffset)) {
	chrend = chrhigh - chroffset;
      } else {
	chrend = chrstart + overall_end_distance_genome;
      }
      all_stage2_starts = Stage2_compute_starts(
#ifdef PMAP
						&(queryaaseq_ptr[0]),&(queryaaseq_ptr[0]),
						/*querylength*/querypos,/*query_offset*/0*3,
#else
						&(queryuc_ptr[0]),&(queryuc_ptr[0]),
						/*querylength*/querypos,/*query_offset*/0,
#endif
						chrstart,chrend,chroffset,chrhigh,/*plusp*/false,genestrand,
					    
						oligoindices_minor,pairpool,diagpool,cellpool,
						/*localp should be false*/true,/*skip_repetitive_p*/false,
						/*favor_right_p*/false,/*max_nalignments*/5,/*debug_graphic_p*/false);
      debug14(printf("Got %d starts\n",List_length(all_stage2_starts)));
    }
  }


  /* E.  Make all_stage2_ends (pairs) */
  if ((gmap_mode & GMAP_ENDS) == 0) {
    all_stage2_ends = (List_T) NULL;

  } else {
    /* Middle path goes from low to high genomic coords */
    univdiag = (Univdiag_T) List_last_value(path);
    /* Substring_queryend = querylength - univdiag->querystart */
    if ((querypos = querylength - univdiag->querystart) > querylength - 8) {
      debug14(printf("No starts, since querypos %d > querylength %d - 8\n",querypos,querylength));
      all_stage2_ends = (List_T) NULL;
    } else {
      left = univdiag->univdiagonal;
      chrend = (left + univdiag->querystart) - chroffset;
      if (chrend < overall_end_distance_genome) {
	chrstart = 0;
      } else {
	chrstart = chrend - overall_end_distance_genome;
      }
      all_stage2_ends = Stage2_compute_ends(
#ifdef PMAP
					    &(queryaaseq_ptr[querypos]),&(queryaaseq_ptr[querypos]),
					    /*querylength*/querylength - querypos,/*query_offset*/querypos*3,
#else
					    &(queryuc_ptr[querypos]),&(queryuc_ptr[querypos]),
					    /*querylength*/querylength - querypos,/*query_offset*/querypos,
#endif
					    chrstart,chrend,chroffset,chrhigh,/*plusp*/false,genestrand,
					    
					    oligoindices_minor,pairpool,diagpool,cellpool,
					    /*localp should be false*/true,/*skip_repetitive_p*/false,
					    /*favor_right_p*/false,/*max_nalignments*/5,/*debug_graphic_p*/false);
      debug14(printf("Got %d ends\n",List_length(all_stage2_ends)));
    }
  }

  
  /* Used to count total mismatches here, but cannot do that with Univdiag_T */

  /* F.  Make stage2pairs */
  debug14(printf("Making pairs from univdiags: "));
  stage2pairs = (List_T) NULL;
  path = List_reverse(path);
  for (p = path; p != NULL; p = List_next(p)) {
    univdiag = (Univdiag_T) List_head(p);
    debug14(printf("Handling univdiag for %d..%d at left %u\n",
		   querylength - univdiag->queryend,querylength - univdiag->querystart,univdiag->univdiagonal));
    /* Substring_querystart = querylength - univdiag->queryend */
    querypos = querylength - univdiag->queryend;
    seglength = univdiag->queryend - univdiag->querystart;

    /* Don't understand why it isn't chrhigh - 1, but it
       looks like the minus coordinates for substrings are +1 higher
       than they should be */
    /* genomepos = (chrlength + 1) - Substring_alignstart_trim_chr(substring);  -- fails for circular chromosomes */
    left = univdiag->univdiagonal;
    alignstart = left + univdiag->querystart;
    alignend = left + univdiag->queryend;
    genomepos = (chrhigh + 1) - alignend;
    Genome_get_segment_blocks_right(gsequence_orig,gsequence_alt,/*left*/alignstart,
				    seglength,chrhigh,/*revcomp*/true);

    for (i = 0; i < seglength; i++) {
      c = queryuc_ptr[querypos];
      g = gsequence_orig[i];
      g_alt = gsequence_alt[i];
      /* Previously checked for (g == c || g_alt == c) */
      /* printf("minus: Checking cdna %c against genome %c => %d\n",c,g,Dynprog_consistent_p(c,g,g_alt,genestrand)); */
      if (Dynprog_consistent_p(c,g,g_alt,genestrand) == true) {
	comp = MATCH_COMP;
      } else {
	comp = MISMATCH_COMP;
      }
      debug14(printf("Pushing %c %c %c at %d,%u\n",c,comp,g,querypos,genomepos));
      stage2pairs = Pairpool_push(stage2pairs,pairpool,querypos,genomepos,
				  /*cdna*/c,comp,/*genome*/g,/*genomealt*/g_alt,
				  /*dynprogindex*/0);
      querypos++;
      genomepos++;
    }
    debug14(Pair_dump_list(stage2pairs,true));
    debug14(printf("\n"));
  }
  path = List_reverse(path);

  if (stage2pairs == NULL) {
    hit1 = (Stage3end_T) NULL;
  } else {
    knownsplice_limit_low = ((Pair_T) stage2pairs->first)->genomepos + chroffset;
    stage2pairs = List_reverse(stage2pairs);
    knownsplice_limit_high = ((Pair_T) stage2pairs->first)->genomepos + chroffset;

    if ((pairarray1 = Stage3_compute_one(&cdna_direction,&(*sensedir),&pairs1,&npairs1,&goodness1,
					 &matches1,&nmatches_posttrim_1,&max_match_length_1,
					 &ambig_end_length_5_1,&ambig_end_length_3_1,
					 &ambig_splicetype_5_1,&ambig_splicetype_3_1,
					 &ambig_prob_5_1,&ambig_prob_3_1,
					 &unknowns1,&mismatches1,&qopens1,&qindels1,&topens1,&tindels1,
					 &ncanonical1,&nsemicanonical1,&nnoncanonical1,&avg_splice_score_1,

					 &pairarray2,&pairs2,&npairs2,&goodness2,
					 &matches2,&nmatches_posttrim_2,&max_match_length_2,
					 &ambig_end_length_5_2,&ambig_end_length_3_2,
					 &ambig_splicetype_5_2,&ambig_splicetype_3_2,
					 &ambig_prob_5_2,&ambig_prob_3_2,
					 &unknowns2,&mismatches2,&qopens2,&qindels2,&topens2,&tindels2,
					 &ncanonical2,&nsemicanonical2,&nnoncanonical2,&avg_splice_score_2,

					 stage2pairs,all_stage2_starts,all_stage2_ends,
#ifdef END_KNOWNSPLICING_SHORTCUT
					 cutoff_level,/*queryptr*/watsonp ? queryuc_ptr : queryrc,
					 watsonp ? query_compress_fwd : query_compress_rev,
#endif
					 /*queryseq_ptr*/queryuc_ptr,queryuc_ptr,querylength,/*skiplength*/0,
#ifdef EXTRACT_GENOMICSEG
					 /*query_subseq_offset*/0,
#else
					 /*query_subseq_offset*/0,
#endif
					 chrnum,chroffset,chrhigh,chrlength,
					 knownsplice_limit_low,knownsplice_limit_high,/*plusp*/false,genestrand,
					 /*jump_late_p*/true,maxpeelback,pairpool,dynprogL,dynprogM,dynprogR,
					 /*sense_try*/0,/*sense_filter*/0,
					 oligoindices_minor,diagpool,cellpool)) == NULL) {
      /* No possible hits for either sensedir */
      hit1 = (Stage3end_T) NULL;

    } else if (pairarray2 == NULL) {
      /* Possible hit for pairarray1 only */
      start = add_bounded(chroffset + Pair_genomepos(&(pairarray1[0])),
			  /*plusterm*/Pair_querypos(&(pairarray1[0])),chrhigh);
      end = subtract_bounded(chroffset + Pair_genomepos(&(pairarray1[npairs1-1])),
			     /*minusterm*/querylength - 1 - Pair_querypos(&(pairarray1[npairs1-1])),chroffset);

      if ((hit1 = Stage3end_new_gmap(&(*found_score),nmatches_posttrim_1,max_match_length_1,
				     ambig_end_length_5_1,ambig_end_length_3_1,
				     avg_splice_score_1,pairarray1,npairs1,
				     /*left*/end,/*plusp*/false,genestrand,
				     querylength,query_compress_rev,
				     chrnum,chroffset,chrhigh,chrlength,
				     cdna_direction,*sensedir,method,level)) == NULL) {
	/* FREE_OUT(pairarray1); */
      }

    } else {
      /* Possible hit for both pairarrays */
      start = add_bounded(chroffset + Pair_genomepos(&(pairarray1[0])),
			  /*plusterm*/Pair_querypos(&(pairarray1[0])),chrhigh);
      end = subtract_bounded(chroffset + Pair_genomepos(&(pairarray1[npairs1-1])),
			     /*minusterm*/querylength - 1 - Pair_querypos(&(pairarray1[npairs1-1])),chroffset);

      if ((hit1 = Stage3end_new_gmap(&(*found_score),nmatches_posttrim_1,max_match_length_1,
				     ambig_end_length_5_1,ambig_end_length_3_1,
				     avg_splice_score_1,pairarray1,npairs1,
				     /*left*/end,/*plusp*/false,genestrand,
				     querylength,query_compress_rev,
				     chrnum,chroffset,chrhigh,chrlength,
				     /*cdna_direction*/+1,/*sensedir*/SENSE_FORWARD,method,level)) == NULL) {
	/* FREE_OUT(pairarray1); */
      }

      start = add_bounded(chroffset + Pair_genomepos(&(pairarray2[0])),
			  /*plusterm*/Pair_querypos(&(pairarray2[0])),chrhigh);
      end = subtract_bounded(chroffset + Pair_genomepos(&(pairarray2[npairs2-1])),
			     /*minusterm*/querylength - 1 - Pair_querypos(&(pairarray2[npairs2-1])),chroffset);

      if ((*hit2 = Stage3end_new_gmap(&(*found_score),nmatches_posttrim_2,max_match_length_2,
				      ambig_end_length_5_2,ambig_end_length_3_2,
				      avg_splice_score_2,pairarray2,npairs2,
				      /*left*/end,/*plusp*/false,genestrand,
				      querylength,query_compress_rev,
				      chrnum,chroffset,chrhigh,chrlength,
				      /*cdna_direction*/-1,/*sensedir*/SENSE_ANTI,method,level)) == NULL) {
	/* FREE_OUT(pairarray2); */
      }

      if (hit1 == NULL && *hit2 == NULL) {
	/* Nothing to report */
	debug14(printf("Both hits are NULL.  Nothing to report\n"));

      } else if (hit1 == NULL) {
	/* Need to report hit2 only */
	debug14(printf("hit1 is NULL.  Reporting hit2 as hit1\n"));
	hit1 = *hit2;
	*hit2 = (Stage3end_T) NULL;

      } else if (*hit2 == NULL) {
	/* Need to report hit1 only */
	debug14(printf("hit2 is NULL.  Reporting hit1\n"));

      } else if (pick_one_sensedir_p == false) {
	/* Report both */
	debug14(printf("pick_one_sensedir is false.  Reporting both\n"));

      } else if (Stage3end_hit_goodness_cmp(&equalp,hit1,*hit2,/*finalp*/true) >= 0) {
	/* Report only hit1 as best */
	debug14(printf("hit1 is best.  Freeing hit2 and reporting hit1\n"));
	Stage3end_free(&(*hit2));
	*hit2 = (Stage3end_T) NULL;

      } else {
	/* Report only hit2 as best */
	debug14(printf("hit2 is best.  Freeing hit1 and reporting hit2 as hit1\n"));
	Stage3end_free(&hit1);
	hit1 = *hit2;
	*hit2 = (Stage3end_T) NULL;
      }
    }
  }

  List_free(&all_stage2_ends);
  List_free(&all_stage2_starts);

#ifdef HAVE_ALLOCA
  FREEA(gsequence_alt);
  FREEA(gsequence_orig);
#else
  FREE(gsequence_alt);
  FREE(gsequence_orig);
#endif

  return hit1;
}



void
Path_solve_setup (Genome_T genomebits_in, Genome_T genomebits_alt_in, bool *circularp_in,
		  Localdb_T localdb_in, Localdb_T localdb2_in,
		  Chrpos_T shortsplicedist_novelend, int min_intronlength_in, int max_deletionlength,
		  int max_insertionlen_in, 
		  bool novelsplicingp_in, Univcoord_T *splicesites_in, Splicetype_T *splicetypes_in,
		  Chrpos_T *splicedists_in, int nsplicesites_in, 
		  Univ_IIT_T chromosome_iit_in, int circular_typeint_in,
		  int index1part_in, int index1interval_in, int local1part_in, int gmap_mode_in,
		  int maxpeelback_in) {

  genomebits = genomebits_in;
  genomebits_alt = genomebits_alt_in;
  circularp = circularp_in;

  localdb = localdb_in;
  localdb2 = localdb2_in;

  min_intronlength = min_intronlength_in;
  max_deletionlen = max_deletionlength;

  novelsplicingp = novelsplicingp_in;
  splicesites = splicesites_in;
  splicetypes = splicetypes_in;
  splicedists = splicedists_in;
  nsplicesites = nsplicesites_in;

  chromosome_iit = chromosome_iit_in;
  circular_typeint = circular_typeint_in;

  /* For localdb */
  max_insertionlen = max_insertionlen_in;
  if (shortsplicedist_novelend > max_deletionlen) {
    overall_end_distance_genome = shortsplicedist_novelend;
  } else {
    overall_end_distance_genome = max_deletionlen;
  }

  index1part = index1part_in;
  index1interval = index1interval_in;
  local1part = local1part_in;

  gmap_mode = gmap_mode_in;
  maxpeelback = maxpeelback_in;

  return;
}


/* Since GMAP keeps the query in order and inverts the genome, we need only queryuc_ptr */
void
Path_solve_via_gmap (int *found_score, List_T *hits_gplus, List_T *hits_gminus,
		     List_T plus_complete_paths, List_T minus_complete_paths, int genestrand,
		     char *queryuc_ptr, int querylength, Compress_T query_compress_fwd, Compress_T query_compress_rev,
		     Pairpool_T pairpool, Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR,
		     Oligoindex_array_T oligoindices_minor, Diagpool_T diagpool, Cellpool_T cellpool,
		     bool paired_end_p, Method_T method, int level) {

  Stage3end_T hit, gmap2;
  int sensedir;
  List_T *paths, complete_path;

  int n, i;
  int *order, *lengths;

  /* printf("Path_solve_via_gmap: Entered with queryuc_ptr %s\n",queryuc_ptr); */

  if (plus_complete_paths != NULL) {
    /* Sort and filter */
    paths = (List_T *) List_to_array_n(&n,plus_complete_paths);

    lengths = (int *) MALLOC((n+1)*sizeof(int));
    for (i = 0; i < n; i++) {
      complete_path = paths[i];
      lengths[i] = Univdiag_list_length(complete_path);
    }
    order = Sedgesort_order_int(lengths,n);

    if (paired_end_p == false) {
      /* For single-end reads, generate one sensedir */
      for (i = 0; i < n && i < MAX_GMAP_PATHS; i++) {
	/* printf("%d: plus length %d\n",i,lengths[order[n-1-i]]); */
	complete_path = paths[order[n-1-i]];
	if ((hit = solve_via_gmap_plus(&gmap2,&sensedir,&(*found_score),complete_path,
				       queryuc_ptr,querylength,query_compress_fwd,genestrand,
				       pairpool,dynprogL,dynprogM,dynprogR,
				       oligoindices_minor,diagpool,cellpool,
				       /*pick_one_sensedir_p*/true,method,level)) != NULL) {
	  *hits_gplus = List_push(*hits_gplus,(void *) hit);
	}
      }
    } else {
      /* For paired-end reads, generate both sensedirs */
      for (i = 0; i < n && i < MAX_GMAP_PATHS; i++) {
	/* printf("%d: plus length %d\n",i,lengths[order[n-1-i]]); */
	complete_path = paths[order[n-1-i]];
	if ((hit = solve_via_gmap_plus(&gmap2,&sensedir,&(*found_score),complete_path,
				       queryuc_ptr,querylength,query_compress_fwd,genestrand,
				       pairpool,dynprogL,dynprogM,dynprogR,
				       oligoindices_minor,diagpool,cellpool,
				       /*pick_one_sensedir_p*/false,method,level)) != NULL) {
	  *hits_gplus = List_push(*hits_gplus,(void *) hit);
	  if (gmap2 != NULL) {
	    *hits_gplus = List_push(*hits_gplus,(void *) gmap2);
	  }
	}
      }
    }

    
    FREE(order);
    FREE(lengths);
    FREE(paths);
  }


  if (minus_complete_paths != NULL) {
    /* Sort and filter */
    paths = (List_T *) List_to_array_n(&n,minus_complete_paths);

    lengths = (int *) MALLOC((n+1)*sizeof(int));
    for (i = 0; i < n; i++) {
      complete_path = paths[i];
      lengths[i] = Univdiag_list_length(complete_path);
    }
    order = Sedgesort_order_int(lengths,n);

    if (paired_end_p == false) {
      /* For single-end reads, generate one sensedir */
      for (i = 0; i < n && i < MAX_GMAP_PATHS; i++) {
	/* printf("%d: minus length %d\n",i,lengths[order[n-1-i]]); */
	complete_path = paths[order[n-1-i]];
	if ((hit = solve_via_gmap_minus(&gmap2,&sensedir,&(*found_score),complete_path,
					queryuc_ptr,querylength,query_compress_rev,genestrand,
					pairpool,dynprogL,dynprogM,dynprogR,
					oligoindices_minor,diagpool,cellpool,
					/*pick_one_sensedir_p*/true,method,level)) != NULL) {
	  *hits_gminus = List_push(*hits_gminus,(void *) hit);
	}
      }
    } else {
      /* For paired-end reads, generate both sensedirs */
      for (i = 0; i < n && i < MAX_GMAP_PATHS; i++) {
	/* printf("%d: minus length %d\n",i,lengths[order[n-1-i]]); */
	complete_path = paths[order[n-1-i]];
	if ((hit = solve_via_gmap_minus(&gmap2,&sensedir,&(*found_score),complete_path,
					queryuc_ptr,querylength,query_compress_rev,genestrand,
					pairpool,dynprogL,dynprogM,dynprogR,
					oligoindices_minor,diagpool,cellpool,
					/*pick_one_sensedir_p*/false,method,level)) != NULL) {
	  *hits_gminus = List_push(*hits_gminus,(void *) hit);
	  if (gmap2 != NULL) {
	    *hits_gminus = List_push(*hits_gminus,(void *) gmap2);
	  }
	}
      }
    }


    FREE(order);
    FREE(lengths);
    FREE(paths);
  }

  return;
}
